#! /usr/bin/perl -w
################################################################################
# Copyright (C) 2001-2002 Peter J Jones (pjones@pmade.org)
# All Rights Reserved
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 3. Neither the name of the Author nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
################################################################################
#
# configure.pl (bootstrap the sord library)
# Peter J Jones (pjones@pmade.org)
#
################################################################################
#
# Includes
#
################################################################################
use strict;
use Getopt::Long;
use Cwd qw(chdir cwd);
################################################################################
#
# Constants
#
################################################################################
use constant DATE		=> 'Wed Oct 24 13:46:58 2001';
use constant ID			=> '$Id: configure.pl,v 1.1.1.1 2003/12/22 10:33:01 pschojer Exp $';
################################################################################
#
# Global Variables
#
################################################################################
my $lib_name = "sord";
my $source_files;

my $cwd = cwd();
my $dirsep = '/';

if ($^O =~ /mswin/i) {
    $dirsep = '\\';
}

my $mkmf = "'$cwd${dirsep}tools${dirsep}mkmf'";
my $cxxflags = "'$cwd${dirsep}tools${dirsep}cxxflags'";
my $genconfig = "'$cwd${dirsep}tools${dirsep}genconfig'";

my $include_dir = "$cwd${dirsep}include";
my $src_dir = "$cwd${dirsep}src";
my $test_dir = "$cwd${dirsep}tests";
my $build_test_dir = "$test_dir${dirsep}build";
my $test_out = "$build_test_dir${dirsep}test.out";
my $install_spec = "docs${dirsep}/install.spec";
my $version_file = "docs${dirsep}/VERSION";
my $config_script = "sord-config";

my $build_test = "$^X build_test.pl";
my %build_settings;
my $config = "";
my $using_readdir_r = 0;
my $mkmf_flags = "--cxxflags $cxxflags --quiet";
my $wrapper_flags = "";
my $major_number = 1;
my $version_number = 1;
my %clo;

my @common_files = qw(
    directory.cxx
    directory_iterator.cxx
    entry.cxx
);

my @unix_files = qw(
    directory_unix.cxx
    entry_unix.cxx
);

my @windows_files = qw(
    directory_win32.cxx
    entry_win32.cxx
);

my @test_directories = qw(
    entry
    directory
);

my @wrap_directories  = qw(
    src
);
################################################################################
#
# Code Start
#
################################################################################
$|++;

GetOptions(
    \%clo,

    'help|h!',
    'developer!',
    'force-readdir!',
    'enable-tests!',
    'disable-shared!',

    'prefix=s',
    'libdir=s',
    'bindir=s',
    'incdir=s',
) || usage();
$clo{'help'} && usage();

sub usage {
    print "Usage: $0 [options]\n", <<EOT;
  --force-readdir  Don't use readdir_r even if your system supports it
  --enable-tests   Enable building of the sord test harness
  --developer      Configure for use by a sord maintainer
  --disable-shared Don't build a shared library

  --prefix path    Set the installation prefix to path [default /usr/local]
  --incdir path    Set the include install dir. [default prefix/include]
  --bindir path    Set the bin install dir. [default prefix/bin]
  --libdir path    Set the lib install dir. [default prefix/lib]
EOT
    exit 1;
}

if (not exists $ENV{'CXX'}) {
    if ($^O =~ /mswin/i) {
	$ENV{'CXX'} = 'cl.exe';
    } else {
	$ENV{'CXX'} = "c++";
    }

    print "$0: Your CXX environment variable is not set.\n";
    print "$0: Using $ENV{CXX} to compile C++ files.\n\n";
}

$ENV{'CXXFLAGS'}    ||= '';
$clo{'prefix'}	    ||= "${dirsep}usr${dirsep}local";
$clo{'bindir'}	    ||= "$clo{'prefix'}${dirsep}bin";
$clo{'incdir'}	    ||= "$clo{'prefix'}${dirsep}include";
$clo{'libdir'}	    ||= "$clo{'prefix'}${dirsep}lib";

if ($^O =~ /mswin/i) {
    $ENV{'CXXFLAGS'} .= " -DSORD_WIN32";
    $source_files = join(' ', @common_files, @windows_files);
} else {
    $ENV{'CXXFLAGS'} .= " -DSORD_UNIX";
    $source_files = join(' ', @common_files, @unix_files);
}

unless (chdir($build_test_dir)) {
    print STDERR "$0: can't chdir into $build_test_dir: $!\n";
    exit 1;
}

if (system($build_test) != 0) {
    print STDERR "$0: build test failed. Please check it out\n";
}

chdir($cwd);

if (open(CF, $test_out)) {
    while (<CF>) {
	clean($_);

	my ($key, $value) = split(/\s*=\s*/, $_, 2);
	$build_settings{$key} = $value;
    }

    close CF;
} else {
    print STDERR "$0: error opening build test output: $!\n";
}

clean($ENV{CXX});
clean($ENV{CXXFLAGS});

$config .= "sord configuration:\n";
$config .= "        Compiler: $ENV{CXX}\n";
$config .= "  Compiler Flags: $ENV{CXXFLAGS}\n";

if (exists $build_settings{'readdir_r'}
    && $build_settings{'readdir_r'} eq 'yes'
    && not $clo{'force-readdir'})
{
    $config .= " Using readdir_r: Yes\n";
    $ENV{'CXXFLAGS'} .= " -DSORD_USE_READDIR_R";
    $using_readdir_r = 1;
} else {
    $config .= " Using readdir_r: No\n";
    $using_readdir_r = 0;
}

if ($clo{'developer'}) {
    $mkmf_flags .= " --developer";
    $config .= "  Developer Mode: On\n";
    $config .= "    Source Files: $source_files\n";
}

$config .= " bin Install Dir: $clo{'bindir'}\n";
$config .= " inc Install Dir: $clo{'incdir'}\n";
$config .= " lib Install Dir: $clo{'libdir'}\n";

if ($using_readdir_r) {
    $mkmf_flags .= " --mt";
}

parse_version_file();
gen_install_spec();
gen_config_script();

if ($clo{'enable-tests'}) {
    push @wrap_directories, "tests";
    $wrapper_flags .= " --with-test tests";
    gen_test_makefiles();
}

unless (chdir($src_dir)) {
    print STDERR "$0: can't chdir into src: $!\n";
    exit 1;
}

my $src_flags = '';
unless ($clo{'disable-shared'}) {
    $src_flags = "--shared-lib $lib_name --major $major_number";
}

print "Creating Makefile in src...\n";
system("$^X $mkmf $mkmf_flags --include '$include_dir' $src_flags --static-lib $lib_name $source_files");
chdir($cwd);

print "Creating Makefile...\n";
system("$^X $mkmf $mkmf_flags --install $install_spec $wrapper_flags --wrapper @wrap_directories");

print "\n$config\n";

################################################################################
sub gen_test_makefiles {
    if (!chdir($test_dir)) {
	print STDERR "$0: can't chdir to $test_dir: $!\n";
	exit 1;
    }

    print "Creating Makefile in tests...\n";
    my $test_dirs = join(',', @test_directories);
    system("$^X $mkmf $mkmf_flags --with-test '$test_dirs' --wrapper @test_directories");

    foreach my $dir (@test_directories) {
	if (!chdir("$test_dir${dirsep}$dir")) {
	    print STDERR "$0: can't chdir to $test_dir${dirsep}$dir: $!\n";
	    exit 1;
	}

	my @scripts = glob("*.pl");
	my $test_cmds = join(',', map{"$^X -I'$test_dir${dirsep}utility' $_"} @scripts);

	print "Creating Makefile in tests${dirsep}$dir...\n";
	system("$^X $mkmf $mkmf_flags --include '$include_dir' --many-exec --slinkwith '$src_dir,$lib_name' --test-cmds \"$test_cmds\" *.cxx");
    }

    chdir($cwd);
}
################################################################################
sub gen_install_spec {
    unless (open(IS, ">$install_spec")) {
	print STDERR "$0: can't create $install_spec: $!\n";
	exit 1;
    }

    print IS "BINDIR=$clo{'bindir'}\n";
    print IS "INCLUDEDIR=$clo{'incdir'}\n";
    print IS "LIBDIR=$clo{'libdir'}\n";
    print IS "include-dir include${dirsep}sord sord\n";
    print IS "static-lib src $lib_name\n";
    print IS "shared-lib src $lib_name $major_number\n" unless $clo{'disable-shared'};
    print IS "binary $config_script\n";

    close IS;
}
################################################################################
sub parse_version_file {
    unless (open(VF, $version_file)) {
	print STDERR "$0: can't open $version_file: $!\n";
	exit 1;
    }

    my $first_line = <VF>;
    close VF;

    if ($first_line =~ /^([\d.]+)\s+(\d+)/) {
	$version_number = $1;
	$major_number = $2;
    } else {
	print STDERR "$0: warning, $version_file is malformed\n";
    }
}
################################################################################
sub gen_config_script {
    my $libs = "-L$clo{'libdir'} -l$lib_name";
    my $flags = "-I$clo{'incdir'}";

    s/\b-/^/g foreach ($libs, $flags);
    system("$^X $genconfig -o $config_script -v $version_number --libs '$libs' --cxxflags '$flags' --name $lib_name");
}
################################################################################
sub clean {
    chomp($_[0]);
    $_[0] =~ s/^\s+//;
    $_[0] =~ s/\s+$//;
}
