/*
 * Copyright (C) 2001-2002 Peter J Jones (pjones@pmade.org)
 * All Rights Reserved
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/** @file
 * This file contains the definition of the sord::directory class.
**/

#ifndef __sord_directory_h__
#define __sord_directory_h__

#include <sord/entry.h>
#include <iterator>

namespace sord {
    class dimpl; // forward declaration

/**
 * The sord::directory class is used to iterate over the entires
 * in a directory.
**/
class directory {
public:
    //####################################################################
    /** 
     * sord::directory constructor
     *
     * @param path The path to the directory to open
     * @author Peter Jones
    **/
    //####################################################################
    explicit directory (const char *path);

    //####################################################################
    /** 
     * sord::directory constructor
     *
     * @param e A sord::entry that refers to a directory to open
     * @author Peter Jones
    **/
    //####################################################################
    explicit directory (const entry &e);

    //####################################################################
    /** 
     * sord::directory destructor
     *
     * @author Peter Jones
    **/
    //####################################################################
    ~directory (void);

    /// directory iterator
    class const_iterator {
    public:
	typedef entry value_type;
	typedef ptrdiff_t difference_type;
	typedef value_type* pointer;
	typedef value_type& reference;
	typedef std::forward_iterator_tag iterator_category;


	const reference operator* (void) const;
	const pointer operator-> (void) const;

	/**
	 * Only prefix increment is supported so that you cannot make a copy
	 * of a directory iterator.
	**/
	const_iterator& operator++ (void);

	friend bool operator== (const const_iterator &lhs, const const_iterator &rhs);
	friend bool operator!= (const const_iterator &lhs, const const_iterator &rhs);
    private:
	explicit const_iterator (dimpl *parent);
	void set_parent (dimpl *parent);
	mutable dimpl *parent_; 

	friend class directory;
	friend class dimpl;

	// don't allow copy construction or assignment
	const_iterator (const const_iterator &other);
	const_iterator& operator= (const const_iterator &other);
	const_iterator operator++ (int);
    }; // end sord::directory::const_iterator class

    //####################################################################
    /** 
     * Get an iterator to the begining of the directory listing.
     *
     * @return A const_iterator that points to the start of the directory
     * @author Peter Jones
    **/
    //####################################################################
    const_iterator& begin (void) const;

    //####################################################################
    /** 
     * Get an end iterator for the directory listing.
     *
     * @return A const_iterator that points one element past the end.
     * @author Peter Jones
    **/
    //####################################################################
    const_iterator& end (void) const;

    //####################################################################
    /** 
     * Get the name of this directory. This may or may not be the same as
     * the name you gave in the constructor.
     *
     * @return The name of this directory.
     * @author Peter Jones
    **/
    //####################################################################
    const char* get_name (void) const;

    //####################################################################
    /** 
     * Find out if the directory opened was vaild.
     *
     * @return True if the directory was NOT VAILD; false if it was vaild.
     * @author Peter Jones
    **/
    //####################################################################
    bool operator! (void) const;
private:
    dimpl *pimpl_; // private implementation

    // don't allow copy construction or assignment
    directory (const directory&);
    directory& operator= (const directory&);
}; // end sord::directory class

} // end sord namespace
#endif
