/*
 * Copyright (C) 2001-2002 Peter J Jones (pjones@pmade.org)
 * All Rights Reserved
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/** @file
 * This file contains the definition of the sord::entry class.
**/

#ifndef __sord_entry_h__
#define __sord_entry_h__

// system includes
#include <ctime>

namespace sord {
    class dimpl; // forward declaration

/**
 * The sord::entry class is used to hold one directory
 * entry and to retreive information about that entry.
**/
class entry {
public:
    /// type for the entry mode
    typedef unsigned int mode_type;

    /// type for the entry size
    typedef unsigned long size_type;

    /// type for the entry times
    typedef time_t time_type;

    /// all possible types that an entry might be
    enum type {
	type_pipe,	///< FIFO Pipe
	type_character,	///< Character Special
	type_directory,	///< Directory
	type_block,	///< Block Special
	type_regular,	///< Normal File
	type_symlink,	///< Symbolic Link
	type_socket	///< Domain Socket
    };

    /// enum to hold the mode bits
    enum mode_bits {
	mode_user_read	    = 0x000100,	///< readable by owner
	mode_user_write	    = 0x000080,	///< writeable by owner
	mode_user_execute   = 0x000040,	///< executable by owner
	mode_group_read	    = 0x000020,	///< readable by group
	mode_group_write    = 0x000010,	///< writeable by group
	mode_group_execute  = 0x000008,	///< executable by group
	mode_other_read	    = 0x000004,	///< readable by everyone else
	mode_other_write    = 0x000002,	///< writeable by everyone else
	mode_other_execute  = 0x000001	///< executable by everyone else
    };

    //####################################################################
    /** 
     * sord::entry class constructor. Given the name/path of a directory
     * entry, the constructor will load information about that entry.
     *
     * @param name The name of a directory entry to open.
     * @author Peter Jones
    **/
    //####################################################################
    explicit entry (const char *name);

    //####################################################################
    /** 
     * Copy information about an entry from an existing entry.
     *
     * @param other The entry to copy information from.
     * @author Peter Jones
    **/
    //####################################################################
    entry (const entry &other);

    //####################################################################
    /** 
     * Assignment operator. Reset this entries information to that of the
     * other entry.
     *
     * @param other The entry to copy information from.
     * @return description
     * @author Peter Jones
    **/
    //####################################################################
    entry& operator= (const entry &other);

    //####################################################################
    /** 
     * Swap this entry with another one.
     *
     * @param other The other entry to swap information with.
     * @author Peter Jones
    **/
    //####################################################################
    void swap (entry &other);

    //####################################################################
    /** 
     * sord::entry class destructor
     *
     * @author Peter Jones
    **/
    //####################################################################
    ~entry (void);

    //####################################################################
    /** 
     * Get the name of this directory entry. The returned string contains
     * only the name of the entry with no path information.
     *
     * @return The name of the entry.
     * @author Peter Jones
    **/
    //####################################################################
    const char* get_name (void) const;

    //####################################################################
    /** 
     * Get the path and name for this entry. The returned string contains
     * the name of the entry with absolute or relative path information.
     *
     * @param The path and name of the entry.
     * @return description
     * @author Peter Jones
    **/
    //####################################################################
    const char* get_path (void) const;

    //####################################################################
    /** 
     * Get the entry type.
     *
     * @return The entry type.
     * @author Peter Jones
    **/
    //####################################################################
    type get_type (void) const;

    //####################################################################
    /** 
     * Get the entry mode.
     *
     * @return The entry mode.
     * @author Peter Jones
    **/
    //####################################################################
    mode_type get_mode (void) const;

    //####################################################################
    /** 
     * Get the time that the entry was last modified.
     *
     * @return The mtime.
     * @author Peter Jones
    **/
    //####################################################################
    time_type get_mtime (void) const;

    //####################################################################
    /** 
     * Get the time that the entry was last accessed.
     *
     * @return The atime
     * @author Peter Jones
    **/
    //####################################################################
    time_type get_atime (void) const;

    //####################################################################
    /** 
     * Get the size of the entry in bytes.
     *
     * @return The size of the entry.
     * @author Peter Jones
    **/
    //####################################################################
    size_type get_size (void) const;

    //####################################################################
    /** 
     * Check to see if the entry is valid.
     *
     * @return True if the entry is INVALID; false if the entry is valid.
     * @author Peter Jones
    **/
    //####################################################################
    bool operator! (void) const;

private:
    struct pimpl; mutable pimpl *pimpl_;

    entry (void);
    void set (const void *data, const char *path);
    friend class dimpl;
}; // end entry class

} // end sord namespace
#endif
