/*
 * Copyright (C) 2001-2002 Peter J Jones (pjones@pmade.org)
 * All Rights Reserved
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/** @file
 * This file contains the implementation of the win32 version of the
 * directory impl class.
**/

#ifndef _MSC_VER
#   error This code requires Microsoft Visual C++ to build.
#endif

#if _MSC_VER <= 1300
#   pragma warning(disable:4181)
#endif

// sord includes
#include "directory_win32.h"
#include "sord/directory.h"
#include "sord/entry.h"

// standard includes
#include <string>
#include <vector>
#include <cstring>

// system includes
#include <winbase.h>

namespace {
    const char const_directory_sep	    = '\\';
    const char const_current_directory[]    = ".";
    const char const_parent_directory[]	    = "..";
}

//####################################################################
sord::dimpl::dimpl (const char *path)
    : path_(path), vector_iterator_ready_(false),
      dir_pointer_(0), dir_iterator_(this), end_iterator_(0)
{
    is_vaild_ = reset();
}
//####################################################################
sord::dimpl::~dimpl (void) {
    close();
}
//####################################################################
sord::directory::const_iterator& sord::dimpl::begin (void) {
    if (moved_past_begin_ && !reset()) return end_iterator_;
    return dir_iterator_;
}
//####################################################################
sord::directory::const_iterator& sord::dimpl::end (void) {
    return end_iterator_;
}
//####################################################################
const char* sord::dimpl::get_name (void) const {
    return path_.c_str();
}
//####################################################################
bool sord::dimpl::is_vaild (void) const {
    return is_vaild_;
}
//####################################################################
bool sord::dimpl::next (void) {
    moved_past_begin_ = true;

    while (!last_entry_read_) {
	if (!dir_pointer_) {
	    std::string filename(path_);
	    filename += const_directory_sep;
	    filename += "*";

	    if ( (dir_pointer_ = FindFirstFile(filename.c_str(), &finddata_)) == INVALID_HANDLE_VALUE) {
		last_entry_read_ = true;
		break;
	    }
	} else if (!FindNextFile(dir_pointer_, &finddata_)) {
	    last_entry_read_ = true;
	    close();
	    break;
	}

	std::string entry_name(path_);
	entry_name += const_directory_sep;
	entry_name += finddata_.cFileName;
	entry_.set(&finddata_, entry_name.c_str());

	if (entry_.get_type() == entry::type_directory) {
	    if (strcmp(finddata_.cFileName, const_current_directory) == 0
		    || strcmp(finddata_.cFileName, const_parent_directory) == 0)
	    {
		continue;
	    }

	    dirs_.push_back(entry_);
	    continue;
	}

	return true;
    }

    if (dirs_.empty()) return false;

    if (!vector_iterator_ready_) {
	vector_iterator_ = dirs_.begin();
	vector_iterator_ready_ = true;
    } else {
	++vector_iterator_;
    }

    if (vector_iterator_ != dirs_.end()) {
	entry_ = *vector_iterator_;
	return true;
    } else {
	dirs_.clear();
	return false;
    }
}
//####################################################################
sord::entry& sord::dimpl::get_entry (void) {
    return entry_;
}
//####################################################################
bool sord::dimpl::reset (void) {
    close();
    memset(&finddata_, 0, sizeof(finddata_));

    dirs_.clear();
    vector_iterator_ready_ = false;
    last_entry_read_ = false;
    moved_past_begin_ = false;

    if (!next()) {
	moved_past_begin_ = true;
	return false;
    }

    return true;
}
//####################################################################
void sord::dimpl::close (void) {
    if (dir_pointer_) {
	FindClose(dir_pointer_);
	dir_pointer_ = 0;
    }
}
