/*
 * Copyright (C) 2001-2002 Peter J Jones (pjones@pmade.org)
 * All Rights Reserved
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/** @file
 * This file contains the implementation of the sord::entry class.
**/

// sord includes
#include "sord/entry.h"
#include "entry_unix.h"

// standard includes
#include <string>
#include <algorithm>
#include <cstring>

// system includes
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

//####################################################################
sord::entry::pimpl::pimpl (const char *name) {
    clear();

    if (name) {
	common_.name_ = name;
	common_.path_ = name;
    }
}
//####################################################################
void sord::entry::pimpl::clear (void) {
    std::memset(&stat_, 0, sizeof(struct stat));

    common_.type_	= sord::entry::type_regular;
    common_.is_valid_	= false;
    type_valid_		= false;
    did_stat_		= false;
}
//####################################################################
bool sord::entry::pimpl::prepare (const void *member) {
    if (member == &(common_.type_) && type_valid_) 
	return true;

    if (member == &(common_.is_valid_) && common_.is_valid_)
	return true;

    if (!did_stat_)
	do_stat();

    return common_.is_valid_;
}
//####################################################################
void sord::entry::pimpl::set (const void *data, const char *path) {
    clear();

    if (path) common_.path_ = path;
    common_.name_  = static_cast<const dirent*>(data)->d_name;
    common_.is_valid_ = true;

#if defined(__FreeBSD__)
    // we have access to the entry type from stat
    stat_.st_mode = DTTOIF(static_cast<const dirent*>(data)->d_type);
    parse_type(); // pull the type out of the mode
    type_valid_ = true;
    stat_.st_mode = 0; // reset mode
#endif
}
//####################################################################
void sord::entry::pimpl::do_stat (void) {
    did_stat_ = true;
    common_.is_valid_ = false;

    if (stat(common_.path_.c_str(), &stat_) == 0) {
	common_.is_valid_ = type_valid_ = true;
	parse_type();
	parse_mode();

	common_.mtime_	= stat_.st_mtime;
	common_.atime_	= stat_.st_atime;
	common_.size_	= stat_.st_size;
    }
}
//####################################################################
void sord::entry::pimpl::parse_type (void) {
    switch (stat_.st_mode & S_IFMT) {
	case S_IFIFO:  common_.type_ = sord::entry::type_pipe;	    break;
	case S_IFCHR:  common_.type_ = sord::entry::type_character; break;
	case S_IFDIR:  common_.type_ = sord::entry::type_directory; break;
	case S_IFBLK:  common_.type_ = sord::entry::type_block;     break;
	case S_IFREG:  common_.type_ = sord::entry::type_regular;   break;
	case S_IFLNK:  common_.type_ = sord::entry::type_symlink;   break;
	case S_IFSOCK: common_.type_ = sord::entry::type_socket;    break;
	default:       common_.type_ = sord::entry::type_regular;   break;
    }
}
//####################################################################
void sord::entry::pimpl::parse_mode (void) {
    common_.mode_ = 0;

    if (stat_.st_mode & S_IRWXU & S_IRUSR) common_.mode_ |= mode_user_read;
    if (stat_.st_mode & S_IRWXU & S_IWUSR) common_.mode_ |= mode_user_write;
    if (stat_.st_mode & S_IRWXU & S_IXUSR) common_.mode_ |= mode_user_execute;

    if (stat_.st_mode & S_IRWXG & S_IRGRP) common_.mode_ |= mode_group_read;
    if (stat_.st_mode & S_IRWXG & S_IWGRP) common_.mode_ |= mode_group_write;
    if (stat_.st_mode & S_IRWXG & S_IXGRP) common_.mode_ |= mode_group_execute;

    if (stat_.st_mode & S_IRWXO & S_IROTH) common_.mode_ |= mode_other_read;
    if (stat_.st_mode & S_IRWXO & S_IWOTH) common_.mode_ |= mode_other_write;
    if (stat_.st_mode & S_IRWXO & S_IXOTH) common_.mode_ |= mode_other_execute;
}
