/*
 * Copyright (C) 2001-2002 Peter J Jones (pjones@pmade.org)
 * All Rights Reserved
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/** @file
 * This file contains the win32 implementation of the sord::entry::pimpl class.
**/

#ifndef _MSC_VER
#   error This code requires Microsoft Visual C++ to build.
#endif

#if _MSC_VER <= 1300
#   pragma warning(disable:4181)
#endif

// sord includes
#include "entry_win32.h"
#include "sord/entry.h"

// standard includes
#include <string>
#include <algorithm>
#include <cstring>

// system includes
#include <sys/types.h>
#include <sys/stat.h>
#include <windows.h>

//####################################################################
sord::entry::pimpl::pimpl (const char *name) {
    clear();

    if (name) {
	common_.name_ = name;
	common_.path_ = name;
    }
}
//####################################################################
void sord::entry::pimpl::set (const void *data, const char *path) {
    clear();

    const WIN32_FIND_DATA &fd = *(static_cast<const WIN32_FIND_DATA*>(data));

    common_.name_	= fd.cFileName;
    common_.path_	= path;
    common_.is_valid_	= true;
    common_.size_	= fd.nFileSizeLow;
    common_.atime_	= ftime2time(&fd.ftLastAccessTime);
    common_.mtime_	= ftime2time(&fd.ftLastWriteTime);
    common_.mode_	= tdtoxmode(fd.dwFileAttributes, common_.name_.c_str());

    switch (common_.mode_ & S_IFMT) {
	case S_IFDIR:  common_.type_ = sord::entry::type_directory; break;
	default:       common_.type_ = sord::entry::type_regular;   break;
    }

    common_.is_valid_	= true;
    did_stat_		= true;
}
//####################################################################
bool sord::entry::pimpl::prepare (const void *member) {
    if (!did_stat_) do_stat();
    return common_.is_valid_;
}
//####################################################################
void sord::entry::pimpl::clear (void) {
    memset(&stat_, 0, sizeof(struct stat));

    common_.type_	= sord::entry::type_regular;
    common_.is_valid_	= false;
    did_stat_		= false;
}
//####################################################################
void sord::entry::pimpl::do_stat (void) {
    did_stat_ = true;
    common_.is_valid_ = false;

    if (stat(common_.path_.c_str(), &stat_) == 0) {
	common_.is_valid_ = true;
	common_.mode_ = stat_.st_mode;
	common_.mtime_ = stat_.st_mtime;
	common_.atime_ = stat_.st_atime;
	common_.size_ = stat_.st_size;

	switch (stat_.st_mode & S_IFMT) {
	    case S_IFDIR:  common_.type_ = sord::entry::type_directory; break;
	    default:       common_.type_ = sord::entry::type_regular;   break;
	}
    }
}

/*
 * This is taken from STAT.C in the CRT source (heavily modified to work
 * here).  Forgive me for hardcoding values, but they will not change. -if
 */

//####################################################################
unsigned short sord::entry::pimpl::tdtoxmode (int attr, const char *name) {
    unsigned short uxmode;
    unsigned dosmode;
    const char *p;

    dosmode = attr & 0xff;
    if ((p = name)[1] == ':') p += 2;

    /* 
     * check to see if this is a directory - note we must make a special
     * check for the root, which DOS thinks is not a directory
     */
    uxmode = ((((*p == '\\' || *p == '/') && !p[1]) || (dosmode & 0x10) || !*p)
	    ? _S_IFDIR|_S_IEXEC : _S_IFREG);

    // If attribute byte does not have read-only bit, it is read-write
    uxmode |= (dosmode & 0x1) ? _S_IREAD : (_S_IREAD|_S_IWRITE);

    // see if file appears to be executable - check extension of name
    if (p = strrchr(name, '.')) {
	if (!stricmp(p, ".exe") || !stricmp(p, ".cmd") || !stricmp(p, ".bat") || !stricmp(p, ".com"))
	    uxmode |= _S_IEXEC;
    }

    // propagate user read/write/execute bits to group/other fields
    uxmode |= (uxmode & 0700) >> 3;
    uxmode |= (uxmode & 0700) >> 6;

    return(uxmode);
}
//####################################################################
extern "C" time_t __cdecl __loctotime_t(int, int, int, int, int, int, int );
int sord::entry::pimpl::ftime2time(const struct _FILETIME * t) {
    SYSTEMTIME SystemTime;
    FILETIME LocalFTime;

    FileTimeToLocalFileTime(t, &LocalFTime);
    FileTimeToSystemTime(&LocalFTime, &SystemTime);
    return __loctotime_t(SystemTime.wYear, SystemTime.wMonth,
	    SystemTime.wDay, SystemTime.wHour, SystemTime.wMinute, SystemTime.wSecond, -1);
}

// vim ts=8:sw=4:sts=4
