 /***********************************************************************
 *                                                                      *
 * MuViNo - ViTooKi Video Annotation Tool                               *
 *                                                                      *
 * title:	MP7Doc.cpp                                              *
 *                                                                      *
 * $Id: MP7Doc.cpp,v 1.8 2005/05/10 09:22:17 klschoef Exp $
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

#include <assert.h>

#include <qdir.h>
#include <qdom.h>
#include <qfile.h>
#include <qlistview.h>
#include <qptrlist.h>
#include <qregexp.h>
#include <qstring.h>
#include <qstringlist.h>

#include <debug.hpp>

#include "MP7Doc.hpp"
#include "MP7ListViewItem.hpp"

const QString 
MP7Doc::emptyMP7Doc("<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>"
		    "<Mpeg7 xmlns=\"urn:mpeg:mpeg7:schema:2001\" "
		    "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" "
		    "xmlns:mpeg7=\"urn:mpeg:mpeg7:schema:2001\" "
		    "xmlns:xml=\"http://www.w3.org/XML/1998/namespace\" "
		    "xsi:schemaLocation=\"urn:mpeg:mpeg7:schema:2001 Mpeg7-2001.xsd\"/>");

MP7Doc::MP7Doc(QListView *listView)
{
  docMeta.modified = false;
  this->listView = listView;
  listView->setSorting(-1);	// disable sorting
  doc.setContent(emptyMP7Doc);
  updateTreeWidget();
}

MP7Doc::~MP7Doc()
{
  if (listView) {
    /* removes all MP7ListViewItem objects recursively */
    QListViewItem *item, *item1;
    for (item = listView->firstChild(); item != NULL;
	 item = item1) {
      item1 = item->nextSibling();
      delete item;
    }
  }
}

bool MP7Doc::parse(const QString &filename)
{
  QFile file(filename);
  if (!file.open(IO_ReadOnly)) {
    dprintf_err("MP7Doc::parse() ERROR: cannot open file %s\n", filename.latin1());
    return false;
  }
  if (!doc.setContent(&file)) {
    dprintf_err("MP7Doc::parse() ERROR: XML parsing of file %s failed!\n",
		filename.latin1());
    file.close();
    return false;
  }
  file.close();
  updateTreeWidget();
  docMeta.filepath = filename;
  setModified(false);
  return true;
}

void MP7Doc::setModified(bool modified)
{
  if (docMeta.modified != modified) {
    emit modificationChanged(modified);
    docMeta.modified = modified;
  }
}

void MP7Doc::updateTreeWidget()
{
  QDomElement root = doc.documentElement();
  if (root.isNull())
    return;
  MP7ListViewItem *rootItem = (MP7ListViewItem *) listView->firstChild();
  if (!rootItem) {
    rootItem = new MP7ListViewItem(listView);
    rootItem->setOpen(true);
  }
  updateTreeWidget(root, rootItem);
  listView->triggerUpdate();
}

void MP7Doc::updateTreeWidget(QDomElement &elem, MP7ListViewItem *item)
{
  assert(item);
  if (elem.isNull())
    return;
  item->setDomElement(elem);
  item->setText(0, elem.tagName());
  /* merge QDomText children into a single one, except for
   * QDomCDATASection children.
   */
  elem.normalize();
  /* collect text values of elem */
  QString text;
  QDomNode n;
  for (n = elem.firstChild(); !n.isNull(); n = n.nextSibling()) {
    QDomText t = n.toText();
    if (!t.isNull())
      text += t.data();
  }
  if (!text.isEmpty())
    item->setText(1, text);
  /* descend to children of elem */
  MP7ListViewItem *childItem = (MP7ListViewItem *) item->firstChild();
  MP7ListViewItem *prevItem = NULL;
  for (n = elem.firstChild(); !n.isNull(); n = n.nextSibling()) {
    QDomElement e = n.toElement();
    if (!e.isNull()) {
      if (!childItem) {
	childItem = new MP7ListViewItem(item, prevItem);
	childItem->setOpen(true);
      }
      updateTreeWidget(e, childItem);
      prevItem = childItem;
      childItem = (MP7ListViewItem *) childItem->nextSibling();
    }
  }
}

bool MP7Doc::save(const QString &filename)
{
  QFile file(filename);
  if (!file.open(IO_WriteOnly)) {
    dprintf_err("MP7Doc::save() cannot open file for writing: %s\n",
		filename.latin1());
    return false;
  }
  QTextStream stream(&file);
  doc.save(stream, 4);
  file.close();
  docMeta.filepath = filename;
  setModified(false);
  return true;
}

QString MP7Doc::filePath() const
{
  return docMeta.filepath;
}

QString MP7Doc::dirPath() const
{
  if (docMeta.filepath.isEmpty())
    return QString::null;
  QDir dir(docMeta.filepath);
  /* strip filename from path */
  QString sdir(docMeta.filepath);
  uint pos = sdir.findRev(dir.dirName());
  sdir.remove(pos, sdir.length()-pos);
  return sdir;
}

bool MP7Doc::isModified() const
{
  return docMeta.modified;
}

MP7ListViewItem *MP7Doc::addElement(MP7ListViewItem *parent, MP7ListViewItem *after,
				    const QString &name)
{
  if (!parent)
    return NULL;
  QDomElement newElem = doc.createElement(name);
  MP7ListViewItem *newItem = new MP7ListViewItem(parent, newElem, after);
  newItem->setText(0, name);
  newItem->setOpen(true);
  setModified(true);
  return newItem;
}

MP7ListViewItem *MP7Doc::addElement(MP7ListViewItem *parent, const QString &name, int pos)
{
  if (!parent)
    return NULL;
  QDomElement newElem = doc.createElement(name);
  MP7ListViewItem *newItem = new MP7ListViewItem(parent, pos, newElem);
  newItem->setText(0, name);
  newItem->setOpen(true);
  setModified(true);
  return newItem;
}

int MP7Doc::setElementValue(MP7ListViewItem *item, const QString &value)
{
  if (!item)
    return -1;
  QDomElement elem(item->domElement());
  if (elem.isNull()) {
    dprintf_err("MP7Doc::setElementValue() item = %s, value = %s, domElement is NULL!\n",
		item->text(0).latin1(), value.latin1());
    return -1;
  }
  bool found = false;
  QDomText t;
  /* since all QDomNodes are normalized, we expect only one QDomText child
   * of every QDomElement.
   */
  for (QDomNode n = elem.firstChild(); !n.isNull(); n = n.nextSibling()) {
    t = n.toText();
    if (!t.isNull()) {
      found = true;
      break;
    }
  }
  if (!found || t.nodeValue() != value) {
    if (!found) {
      /* create a text child node */
      t = doc.createTextNode(value);
      elem.appendChild(t);
    } else {
      t.setNodeValue(value);
    }
    item->setText(1, value);
    setModified(true);
  }
  return 0;
}

bool MP7Doc::setAttribute(MP7ListViewItem *item, const QString &attribute,
			  const QString &value)
{
  if (!item || attribute.isEmpty())
    return false;
  item->domElement().setAttribute(attribute, value);
  return true;
}

QString MP7Doc::getAttribute(MP7ListViewItem *item, const QString &attribute)
{
	if (!item || attribute.isEmpty())
    	return NULL;
	
	return item->domElement().attribute(attribute);
	
}

bool MP7Doc::moveElement(MP7ListViewItem *item, MP7ListViewItem *newParent,
			 MP7ListViewItem *newSibling)
{
  if (!item || !newParent)
    return false;
  if (newSibling && newSibling->parent() != newParent) {
    dprintf_err("MP7Doc::moveElement(item = %p (%s), newParent = %p (%s), "
		"newSibling = %p (%s)) newSibling is not a child of newParent!\n",
		item, item->text(0).latin1(),
		newParent, newParent->text(0).latin1(),
		newSibling, newSibling->text(0).latin1());
    return false;
  }
  QDomElement elem(item->domElement());
  QDomElement newParentElem(newParent->domElement());
  if (elem.isNull() || newParentElem.isNull()) {
    dprintf_err("MP7Doc::moveElement(item = %p (%s), newParent = %p (%s)) "
		"QDomElement not set!\n",
		item, item ? item->text(0).latin1() : "",
		newParent, newParent ? newParent->text(0).latin1() : "");
    return false;
  }
  /* move QDomElement node */
  if (newSibling) {
    QDomElement newSiblingElem(newSibling->domElement());
    if (newSiblingElem.isNull()) {
      dprintf_err("MP7Doc::moveElement() newSibling = %p (%s) "
		  "element not found in itemNodeMap!\n",
		  newSibling, newSibling ? newSibling->text(0).latin1() : "");
      return false;
    }
    newParentElem.insertAfter(elem, newSiblingElem); // re-parents <elem>
  } else
    newParentElem.insertBefore(elem, QDomNode()); // inserts as first child
  /* move MP7ListViewItem node */
  item->parent()->takeItem(item);
  newParent->insertItem(item);
  if (newSibling)
    item->moveItem(newSibling);
  setModified(true);
  return true;
}

bool MP7Doc::removeElementDeep(MP7ListViewItem *item)
{
  MP7ListViewItem *next = NULL;
  if (!item)
    return false;
  for (MP7ListViewItem *child = (MP7ListViewItem *) item->firstChild(); child; ) {
    next = (MP7ListViewItem *) child->nextSibling();
    removeElementDeep(child);
    child = next;
  }
  delete item;			// removes corresponding QDomElement from tree
  setModified(true);
  return true;
}

int MP7Doc::removeChildrenDeep(MP7ListViewItem *parent, const QString &name)
{
  int rval = 0;
  MP7ListViewItem *next = NULL;
  if (!parent)
    return 0;
  for (MP7ListViewItem *child = (MP7ListViewItem *) parent->firstChild(); child; ) {
    next = (MP7ListViewItem *) child->nextSibling();
    if (name.isNull() || child->text(0).lower() == name.lower())
      if (removeElementDeep(child)) // sets modification state
	rval++;
    child = next;
  }
  return rval;
}

MP7ListViewItem *MP7Doc::nodeByPath(const QString &path, MP7ListViewItem *root,
				    bool create)
{
  if (!root)
    root = (MP7ListViewItem *) listView->firstChild();
  if (!root)
    return NULL;			// no root element present
  /* follow <path> */
  int field = 0;
  QString::SectionFlags flags = QString::SectionSkipEmpty;
  QString name(path.section('/', field, field, flags));
  MP7ListViewItem *item = NULL;
  QString attr;
  QString attrValue;
  while (!name.isEmpty()) {
    attr = "";
    attrValue = "";
    QRegExp rx("\\[@([:\\w]+)(=\"?(\\w+)\"?)?\\]");
    if (rx.search(name) >= 0) {
      attr = rx.cap(1);
      attrValue = rx.cap(3);
      name.replace(rx, "");
    }
    /*
    dprintf_full("MP7Doc::nodeByPath() name = %s, attr = %s, attrValue = %s\n",
		 name.latin1(), attr.latin1(), attrValue.latin1());
    */
    for (item = (MP7ListViewItem *) root->firstChild(); item;
	 item = (MP7ListViewItem *) item->nextSibling()) {
      if (item->text(0).lower() == name.lower()) {
	if (!attr.isEmpty()) {
	  QString val = item->domElement().attribute(attr);
	  if (val != QString::null) {
	    if (!attrValue.isEmpty()) {
	      if (val == attrValue)
		break;
	    } else
	      break;
	  }
	} else
	  break;
      }
    }
    if (!item) {
      if (create) {
	item = addElement(root, name);
	if (!item)
	  return NULL;
	if (!attr.isEmpty()) {
	  if (!setAttribute(item, attr, attrValue))
	    return NULL;
	}
      } else
	return NULL;		       // path does not exist
    }
    /* descend */
    root = item;
    field++;
    name = path.section('/', field, field, flags);
  }
  return root;
}

QPtrList<MP7ListViewItem> MP7Doc::children(MP7ListViewItem *parent,
					   const QString &name) const
{
  QPtrList<MP7ListViewItem> list;
  if (!parent)
    return list;
  for (MP7ListViewItem *child = (MP7ListViewItem *) parent->firstChild(); child;
       child = (MP7ListViewItem *) child->nextSibling()) {
    if (name.isNull() || child->text(0).lower() == name.lower())
      list.append(child);
  }
  return list;
}

MP7ListViewItem *MP7Doc::nearestElement(MP7ListViewItem *start, const QString &name,
					bool searchUp) const
{
  QStringList namelist(name);
  return nearestElement(start, namelist, searchUp);
}

MP7ListViewItem *MP7Doc::nearestElement(MP7ListViewItem *start, const QStringList &names,
					bool searchUp) const
{
  QStringList namelist(names);
  if (!start)
    return NULL;
  if (searchUp) {
    for (MP7ListViewItem *item = start; item; item = (MP7ListViewItem *) item->parent()) {
      if (isElementOf(item->text(0), namelist))
	return item;
    }
    return NULL;
  }
  /* searchUp == false */
  QPtrList<MP7ListViewItem> list;
  list.append(start);
  MP7ListViewItem *rval = searchBreadthFirst(list, namelist);
  return rval;
}

MP7ListViewItem *
MP7Doc::searchBreadthFirst(QPtrList<MP7ListViewItem> &nodes,
			   QStringList &names) const
{
  if (nodes.isEmpty())
    return NULL;
  QPtrList<MP7ListViewItem> childNodes;
  for (MP7ListViewItem *item = nodes.first(); item; item = nodes.next()) {
    if (isElementOf(item->text(0), names))
      return item;
    for (MP7ListViewItem *child = (MP7ListViewItem *) item->firstChild(); child;
	 child = (MP7ListViewItem *) child->nextSibling()) {
      childNodes.append(child);
    }
  }
  return searchBreadthFirst(childNodes, names);
}

bool MP7Doc::isElementOf(const QString &string, QStringList &list)
{
  QString s(string.lower());
  for (QStringList::Iterator it = list.begin(); it != list.end(); it++) {
    if (s == (*it).lower())
      return true;
  }
  return false;
}
