 /***********************************************************************
 *                                                                      *
 * MuViNo - ViTooKi Video Annotation Tool                               *
 *                                                                      *
 * title:	MP7Doc.hpp                                              *
 *                                                                      *
 * $Id: MP7Doc.hpp,v 1.9 2005/02/28 10:36:08 amueller_itec Exp $
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

#ifndef _MP7DOC_HPP_
#define _MP7DOC_HPP_

/** @file
 * Declaration of the MP7Doc class.
 */

#include <qdom.h>
#include <qobject.h>
#include <qptrdict.h>
#include <qptrlist.h>

class QListView;
class QListViewItem;
class QString;
class MP7ListViewItem;

/**
 * This class represents an MPEG-7 document.
 * It manages a graphical tree representation of the document
 * within a QListView widget.
 *
 * @author Mario Taschwer
 * @version $Id: MP7Doc.hpp,v 1.9 2005/02/28 10:36:08 amueller_itec Exp $
 */
class MP7Doc : public QObject
{
  Q_OBJECT

public:
  /** Constructor. Creates an MPEG-7 document consisting of an empty
   * Mpeg7 descriptor.
   * @param listView the QListView widget where the MPEG-7 XML tree
   *                 will be displayed. The widget should have been
   *                 initialized with two columns; XML element names
   *                 will be displayed in the first column, text values
   *                 will be shown in the second column.
   */
  MP7Doc(QListView *listView);

  /* Destructor.
   * Deletes all QListViewItem objects corresponding to this
   * MPEG-7 document.
   */
  ~MP7Doc();

  /** Parse a valid MPEG-7 file in XML format.
   * The QListView widget is populated with QListViewItem objects
   * corresponding to XML elements.
   * @return <code>true</code> on success, <code>false</code> on error.
   */
  bool parse(const QString &filename);

  /** Save the MPEG-7 document in XML format to the given file.
   * @return <code>true</code> on success, <code>false</code> on error.
   */
  bool save(const QString &filename);

  /** Return the filename (full path name) of the document.
   * If none has been set, <code>QString::null</code> is returned.
   */
  QString filePath() const;

  /** Return the directory path of the document's filename.
   * If no filename has been set, <code>QString::null</code> is returned.
   */
  QString dirPath() const;

  /** Return <code>true</code> if this document has been modified. */
  bool isModified() const;

  /** Add an XML element as child of <code>parent</code>.
   * @param parent    parent element.
   * @param after     the sibling after which the new element should be inserted.
   * @param name      the new element's tag name.
   * @return the new XML element, or <code>NULL</code> on error.
   */
  MP7ListViewItem *addElement(MP7ListViewItem *parent, MP7ListViewItem *after,
			      const QString &name);
  
  /** Add an XML element as child of <code>parent</code>.
   * @param parent    parent element.
   * @param name      the new element's tag name.
   * @param pos       the position of the new element within the list of children.
   *                  -1 means end of the list.
   * @return the new XML element, or <code>NULL</code> on error.
   */
  MP7ListViewItem *addElement(MP7ListViewItem *parent, const QString &name, int pos = -1);
  
  /** Set the text value of the XML element corresponding to the MP7ListViewItem object.
   * If there is no text child node of the XML element, one will be created.
   * @param item    both the MP7ListViewItem object and the corresponding XML
   *                element will be updated.
   * @param value   the new text value of the XML element.
   * @return 0 - success.<br>
   *        -1 - the XML element corresponding to <code>item</code>
   *             could not be found.
   */
  int setElementValue(MP7ListViewItem *item, const QString &value);

  /** Set an attribute of the XML element corresponding to the MP7ListViewItem object.
   * @return <code>false</code> on error.
   */
  bool setAttribute(MP7ListViewItem *item, const QString &attribute,
		    const QString &value);
		    
  /** Get an attribute of the XML element corresponding to the MP7ListViewItem object.
  * @return <code>Attribute or NULL</code> on error.
  */
  QString getAttribute(MP7ListViewItem *item, const QString &attribute);

  /** Move an XML element (with all children) to a new parent node,
   * right <em>after</em> the given new sibling. 
   * @param item       element to be moved.
   * @param newParent  new parent element.
   * @param newSibling new sibling element; if <code>NULL</code>
   *                   the item will be inserted as the first child
   *                   of <code>newParent</code>.
   * @return <code>false</code> on error.
   */
  bool moveElement(MP7ListViewItem *item, MP7ListViewItem *newParent,
		   MP7ListViewItem *newSibling = NULL);

  /** Remove an XML element and all its descendants recursively. 
   * @return <code>true</code> if element has actually been removed.
   */
  bool removeElementDeep(MP7ListViewItem *item);

  /** Remove all XML child elements with the given tag name
   * and its descendants recursively.
   * @param parent the parent element.
   * @param name   the tag name; if <code>null</code>, all child
   *               elements will be removed.
   * @return the number of (direct) child elements that have been removed.
   */
  int removeChildrenDeep(MP7ListViewItem *parent,
			 const QString &name = QString::null);
  
  /** Return the element specified by a simple XML path.
   * Path element names are case-insensitive.
   * @param path     the XML path relative to <code>root</code> to the
   *                 desired XML element; the path may consist of
   *                 <code>/</code> characters, element names, and
   *                 an attribute specification. 
   *                 The latter is of the form <code>[\@attr="value"]</code>
   *                 (The quotes are optional).
   *                 If an element name is not unique, the first element with
   *                 that name is chosen to follow the path.
   * @param root     the root of the subtree where to apply <code>path</code>.
   *                 If <code>NULL</code>, the document root is used.
   * @param create   if <code>true</code>, any non existing path elements
   *                 and attributes will be created.
   * @return <code>NULL</code> if the node does not exist.
   */
  MP7ListViewItem *nodeByPath(const QString &path, MP7ListViewItem *root = NULL,
			      bool create = false);

  /** Return all XML child elements with the given tag name.
   * @param parent the parent element.
   * @param name   the tag name; if <code>null</code>, all child
   *               elements will be returned.
   */
  QPtrList<MP7ListViewItem> children(MP7ListViewItem *parent,
				     const QString &name = QString::null) const;

  /** Find the nearest XML element with the given tag name.
   * @param start      element where to start the search.
   * @param name       XML tag name to search for.
   * @param searchUp   if <code>searchUP == true</code>, the XML tree is searched
   *                   along the path to the document root;
   *                   otherwise, the tree is searched in a breadth-first
   *                   manner descending from <code>start</code>.
   * @return <code>NULL</code> if no suitable element could be found.
   */
  MP7ListViewItem *nearestElement(MP7ListViewItem *start, const QString &name,
				  bool searchUp) const;

  /** Find the nearest XML element equal to one of the given tag names.
   * @param start      element where to start the search.
   * @param names      list of XML tag names to search for.
   * @param searchUp   if <code>searchUP == true</code>, the XML tree is searched
   *                   along the path to the document root;
   *                   otherwise, the tree is searched in a breadth-first
   *                   manner descending from <code>start</code>.
   * @return <code>NULL</code> if no suitable element could be found.
   */
  MP7ListViewItem *nearestElement(MP7ListViewItem *start, const QStringList &names,
				  bool searchUp) const;

signals:
  /** This signal is emitted when the modification status of the
   * MPEG-7 document has changed.
   * @param m   if <code>true</code>, the status has changed to <i>modified</i>.
   */
  void modificationChanged(bool m);

protected:
  /** Traverse the DOM tree recursively and construct a tree of QListViewItem objects
   * corresponding to XML elements.
   */
  void updateTreeWidget();

  /** Set the document modification status and emit
   * a modificationChanged() signal if the status is being changed.
   */
  void setModified(bool modified);

private:
  /** Update the MP7ListViewItem tree recursively at the given root node.
   * This function is used internally by {@link #updateTreeWidget()}.
   */
  void updateTreeWidget(QDomElement &elem, MP7ListViewItem *item);

  /** Recursive helper function called by 
   * {@link nearestElement(MP7ListViewItem *, const QStringList &, bool)}.
   * Search list of <code>nodes</code> for given tag names in
   * breadth-first manner.
   * @param nodes      list of nodes at same depth in tree to be searched.
   * @param names      tag names to search for.
   * @return the first node equal to any of the tag names.
   */
  MP7ListViewItem *searchBreadthFirst(QPtrList<MP7ListViewItem> &nodes,
				      QStringList &names) const;

  /** The XML document. */
  QDomDocument doc;

  /** Meta information about the XML document. */
  struct docMeta {
    QString filepath;
    bool modified;
  } docMeta;

  /** The tree view widget. */
  QListView *listView;

  /** The empty MPEG-7 document created by the constructor in XML format. */
  static const QString emptyMP7Doc;

  /** Return <code>true</code> if <code>string</code> is equal
   * to one of the elements of <code>list</code>.
   * The comparison is case-insensitive.
   */
  static bool isElementOf(const QString &string, QStringList &list);

};

#endif /* _MP7DOC_HPP_ */
