 /***********************************************************************
 *                                                                      *
 * MuViNo - ViTooKi Video Annotation Tool                               *
 *                                                                      *
 * title:	MP7Time.cpp                                             *
 *                                                                      *
 * $Id: MP7Time.cpp,v 1.3 2004/08/17 13:42:05 mtaschwer Exp $
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

#include <qregexp.h>

#include <debug.hpp>

#include "MP7Time.hpp"

#define SECS_PER_DAY  (24 * 3600)

MP7Time::MP7Time()
{
  reset();
}

void MP7Time::resetTimePoint()
{
  tpoint.date = QString();
  tpoint.hasTime = false;
  tpoint.sec = 0;
  tpoint.num = 0;
  tpoint.denom = 1;
  tpoint.tz = QString();
}

void MP7Time::resetDuration()
{
  duration.sign = 1;
  duration.days = 0;
  duration.sec = 0;
  duration.num = 0;
  duration.denom = 1;
  duration.tz = QString();
  duration.units = 0;
}

void MP7Time::reset()
{
  resetTimePoint();
  resetDuration();
}

bool MP7Time::setTimePoint(uint value, uint ticksPerSecond)
{
  resetTimePoint();
  if (ticksPerSecond == 0) {
    dprintf_err("MP7Time::setTimePoint(value = %u, ticksPerSecond = %u) invalid ticksPerSecond value!\n",
		value, ticksPerSecond);
    return false;
  }
  tpoint.sec = value / ticksPerSecond;
  tpoint.num = value % ticksPerSecond;
  tpoint.denom = ticksPerSecond;
  tpoint.hasTime = true;
  return true;
}

bool MP7Time::setTimePoint(const QString &mp7Value)
{
  resetTimePoint();
  QString s(mp7Value);
  int pos = s.find('T');
  if (pos < 0)
    pos = s.length();
  /* date part */
  if (pos > 0) {
    tpoint.date = s.left(pos);
    s.remove(0, pos);		// keep the 'T'
  }
  if (s.length() == 0)
    return true;
  /* time zone */
  if ((pos = s.findRev('+')) >= 0 || (pos = s.findRev('-')) >= 0) {
    tpoint.tz = s.right(s.length() - pos);
    s.remove(pos, s.length() - pos);
  }
  if (s.length() == 0)
    return true;
  /* parse time part */
  QRegExp rx("[T:F]");
  int oldpos = s.length()-1;
  uint val;
  bool ok;
  int unit = 1;		// time unit (in seconds) of next field
  while (oldpos >= 0 && (pos = s.findRev(rx, oldpos)) >= 0) {
    if (pos < oldpos) {
      val = s.mid(pos+1, oldpos - pos).toUInt(&ok);
      if (!ok) {
	dprintf_err("MP7Time::setTimePoint(mp7value = %s) integer conversion FAILED!\n",
		    mp7Value.latin1());
	resetTimePoint();
	return false;
      }
    }
    switch (s.at(pos).latin1()) {
    case 'F':
      if (!val) {
	dprintf_err("MP7Time::setTimePoint(mp7value = %s) 0 denominator!\n",
		    mp7Value.latin1());
	resetTimePoint();
	return false;
      }
      tpoint.denom = val;
      unit = 0;			// means "sub-second"
      break;
    case ':':			// fall through
    case 'T':
      if (!unit) {
	tpoint.num = val;
	unit = 1;
      } else {
	tpoint.sec += val * unit;
	unit *= 60;
      }
      break;
    }
    oldpos = pos-1;
  }
  /* correct tpoint.sec if not all three time fields are present, eg. T03:12 */
  if (tpoint.sec && unit <= 3600) {
    tpoint.sec *= 3600 * 60 / unit;
  }
  tpoint.hasTime = true;
  return true;
}

bool MP7Time::setTimePoint(const MP7Time &otherTime)
{
  tpoint = otherTime.tpoint;
  return true;
}

bool MP7Time::moveTimePoint(int value, uint ticksPerSecond)
{
  bool rval = true;
  uint denom = commonMultiple(tpoint.denom, ticksPerSecond);
  tpoint.num *= denom / tpoint.denom;
  value *= (int) (denom / ticksPerSecond);
  tpoint.denom = denom;
  if (value < 0) {
    int sec = (int) tpoint.sec;
    value = -value;
    tpoint.num += denom;
    tpoint.num -= value % denom;
    if (tpoint.num >= denom)
      tpoint.num -= denom;
    else
      sec--;
    sec -= value / denom;
    if (sec < 0) {
      rval = false;
      resetTimePoint();
    } else
      tpoint.sec = (uint) sec;
  } else {
    tpoint.num += value % denom;
    if (tpoint.num >= denom) {
      tpoint.sec++;
      tpoint.num -= denom;
    }
    tpoint.sec += value / denom;
  }
  tpoint.hasTime = true;
  return rval;
}

bool MP7Time::setDuration(uint value, uint ticks)
{
  resetDuration();
  if (ticks == 0) {
    dprintf_err("MP7Time::setDuration(value = %u, ticks = %u) invalid ticks value!\n",
		value, ticks);
    return false;
  }
  duration.sec = value / ticks;
  duration.num = value % ticks;
  duration.denom = ticks;
  return true;
}

bool MP7Time::setDuration(const QString &mp7Value)
{
  resetDuration();
  QString s(mp7Value);
  int pos;
  /* time zone */
  if ((pos = s.findRev('+')) >= 0 || (pos = s.findRev('-')) >= 0) {
    duration.tz = s.right(s.length() - pos);
    s.remove(pos, s.length() - pos);
  }
  /* parse duration */
  uint val;
  bool ok;
  QRegExp rx("[-PDTHMSNF]");
  int oldpos = 0;
  while ((pos = s.find(rx, oldpos)) >= 0) {
    if (pos > oldpos) {
      val = s.mid(oldpos, pos-oldpos).toUInt(&ok);
      if (!ok) {
	dprintf_err("MP7Time::setDuration(mp7Value = %s) integer conversion FAILED!\n",
		    mp7Value.latin1());
	resetDuration();
	return false;
      }
    }
    switch(s.at(pos).latin1()) {
    case '-':
      duration.sign = -1;
      break;
    case 'P':
      break;
    case 'D':
      duration.days = val;
      break;
    case 'T':
      break;
    case 'H':
      duration.sec += val * 3600;
      break;
    case 'M':
      duration.sec += val * 60;
      break;
    case 'S':
      duration.sec += val;
      break;
    case 'N':
      duration.num = val;
      break;
    case 'F':
      if (val > 0)
	duration.denom = val;
      else {
	dprintf_err("MP7Time::setDuration(mp7Value = %s) invalid fraction!\n",
		    mp7Value.latin1());
	resetDuration();
	return false;
      }
    }
    oldpos = pos + 1;
  }
  return true;
}

bool MP7Time::setDuration(const MP7Time &duration)
{
  this->duration = duration.duration;
  return true;
}

bool MP7Time::setDurationFromEnd(const MP7Time &end)
{
  if (!tpoint.hasTime || !end.tpoint.hasTime)
    return false;
  resetDuration();
  uint denom = commonMultiple(tpoint.denom, end.tpoint.denom);
  duration.denom = denom;
  if (compareTimePoint(end) <= 0) {
    duration.sign = 1;
    duration.num = end.tpoint.num * (denom / end.tpoint.denom);
    duration.sec = end.tpoint.sec;
    uint thisnum = tpoint.num * (denom / tpoint.denom);
    if (thisnum > duration.num) {
      duration.num += denom;
      duration.sec--;
    }
    duration.num -= thisnum;
    duration.sec -= tpoint.sec;
  } else {
    MP7Time end2(end);
    if (!end2.setDurationFromEnd(*this))
      return false;
    duration = end2.duration;
    duration.sign = -1;
  }
  if (duration.sec >= SECS_PER_DAY) {
    duration.days = duration.sec / SECS_PER_DAY;
    duration.sec %= SECS_PER_DAY;
  }
  return true;
}

bool MP7Time::addDuration(const MP7Time &otherTime)
{
  if (duration.sign < 0 || otherTime.duration.sign < 0)
    return false;
  if (otherTime.duration.num > 0) {
    uint newDenom = commonMultiple(duration.denom, otherTime.duration.denom);
    duration.num *= newDenom / duration.denom;
    duration.denom = newDenom;
    duration.num += otherTime.duration.num * newDenom / otherTime.duration.denom;
    if (duration.num >= duration.denom) {
      duration.sec++;
      duration.num -= duration.denom;
    }
  }
  duration.sec += otherTime.duration.sec;
  if (duration.sec >= SECS_PER_DAY) {
    duration.days++;
    duration.sec -= SECS_PER_DAY;
  }
  duration.days += otherTime.duration.days;
  return true;
}

QString MP7Time::timePointPretty(bool forceSubSec) const
{
  QString s;
  if (!tpoint.date.isEmpty()) {
    s.append(tpoint.date);
    if (tpoint.hasTime)
      s.append(' ');
  }
  /* avoid returning a null string for "empty" time point */
  if (tpoint.hasTime || tpoint.date.isEmpty()) {
    QString t;
    t.sprintf("%02u:%02u:%02u", tpoint.sec / 3600, (tpoint.sec % 3600) / 60,
	      tpoint.sec % 60);
    s.append(t);
    if (tpoint.num || forceSubSec) {
      s.append('.');
      t.sprintf("%02u", tpoint.num * 100 / tpoint.denom);
      s.append(t);
    }
  }
  if (!tpoint.tz.isEmpty())
    s.append(tpoint.tz);
  return s;
}

QString MP7Time::durationPretty() const
{
  QString s;
  QString t;
  bool hasTime = (duration.sec != 0 || duration.num != 0);
  if (duration.units) {
    t.sprintf("%d * [", duration.units);
    s.append(t);
  }
  if (duration.days) {
    t.sprintf("%u days", duration.days);
    s.append(t);
    if (hasTime || !duration.tz.isEmpty())
      s.append(' ');
  }
  if (hasTime) {
    t.sprintf("%02u:%02u:%02u", duration.sec / 3600, (duration.sec % 3600) / 60,
	      duration.sec % 60);
    s.append(t);
    if (duration.num) {
      t.sprintf(".%02u", duration.num * 100 / duration.denom);
      s.append(t);
    }
  }
  if (!duration.tz.isEmpty())
    s.append(duration.tz);
  if (duration.units)
    s.append(']');
  return s;
}

MP7Time MP7Time::endTimePoint() const
{
  MP7Time end;
  end.tpoint.hasTime = tpoint.hasTime;
  end.tpoint.sec = tpoint.sec;
  end.tpoint.num = tpoint.num;
  end.tpoint.denom = tpoint.denom;
  if (duration.num > 0) {
    end.tpoint.denom = commonMultiple(tpoint.denom, duration.denom);
  }
  int i = 1;
  int sign = duration.sign;
  if (duration.units > 0) {
    i = duration.units;
  } else if (duration.units < 0) {
    i = -duration.units;
    sign = -sign;
  }
  if (sign > 0) {
    if (duration.days)
      end.tpoint.sec += duration.days * SECS_PER_DAY;
    end.tpoint.sec += duration.sec;
    if (duration.num > 0) {
      end.tpoint.num += duration.num * end.tpoint.denom / duration.denom;
      end.tpoint.sec += end.tpoint.num / end.tpoint.denom;
      end.tpoint.num %= end.tpoint.denom;
    }
  } else {
    if (duration.days)
      end.tpoint.sec -= duration.days * SECS_PER_DAY;
    end.tpoint.sec -= duration.sec;
    if (duration.num > 0) {
      uint diff = duration.num * end.tpoint.denom / duration.denom;
      end.tpoint.sec -= diff / end.tpoint.denom;
      diff %= end.tpoint.denom;
      if (diff > end.tpoint.num) {
	end.tpoint.sec--;
	diff -= end.tpoint.denom;
      }
      end.tpoint.num -= diff;
    }
  }
  return end;
}

QString MP7Time::endTimePointPretty(bool forceSubSec) const
{
  return endTimePoint().timePointPretty(forceSubSec);
}

QString MP7Time::toString() const
{
  QString s(timePointPretty());
  QString dur(durationPretty());
  if (!s.isEmpty() && !dur.isEmpty())
    s.append(" [");
  if (!dur.isEmpty()) {
    s.append(dur);
    s.append(']');
  }
  return s;
}

QString MP7Time::timePointMP7() const
{
  QString s;
  if (!tpoint.date.isEmpty()) {
    s.append(tpoint.date);
    if (tpoint.hasTime)
      s.append(' ');
  }
  if (tpoint.hasTime) {
    QString t;
    s.append('T');
    t.sprintf("%02u:%02u:%02u", tpoint.sec / 3600, (tpoint.sec % 3600) / 60,
	      tpoint.sec % 60);
    s.append(t);
    if (tpoint.num) {
      t.sprintf(":%uF%u", tpoint.num, tpoint.denom);
      s.append(t);
    }
  }
  if (!tpoint.tz.isEmpty())
    s.append(tpoint.tz);
  return s;
}

QString MP7Time::durationMP7() const
{
  QString s;
  QString t;
  bool hasTime = (duration.sec != 0 || duration.num != 0);
  if (duration.sign < 0)
    s.append('-');
  s.append('P');
  if (duration.days) {
    t.sprintf("%uD", duration.days);
    s.append(t);
  }
  if (hasTime) {
    s.append('T');
    t.sprintf("%uH%uM%uS", duration.sec / 3600, (duration.sec % 3600) / 60,
	      duration.sec % 60);
    s.append(t);
    if (duration.num) {
      t.sprintf("%uN%uF", duration.num, duration.denom);
      s.append(t);
    }
  }
  if (!duration.tz.isEmpty())
    s.append(duration.tz);
  return s;
}

double MP7Time::timeToSec() const
{
  if (!tpoint.hasTime)
    return 0.0;
  double sec = (double) tpoint.sec;
  if (tpoint.num)
    sec += (double) tpoint.num / (double) tpoint.denom;
  return sec;
}

int MP7Time::timeToMSec() const
{
  if (!tpoint.hasTime)
    return 0;
  int msec = (int) tpoint.sec * 1000;
  if (tpoint.num)
    msec += (int) (tpoint.num * 1000 / tpoint.denom);
  return msec;
}

int MP7Time::durationToMSec() const
{
  int msec = (int) duration.sec * 1000;
  if (duration.days)
    msec += (int) duration.days * SECS_PER_DAY;
  if (duration.num)
    msec += (int) (duration.num * 1000 / duration.denom);
  if (duration.units)
    msec *= duration.units;
  return msec;
}

double MP7Time::durationToSec() const
{
  double sec = (double) duration.sec;
  if (duration.days)
    sec += (double) (duration.days * SECS_PER_DAY);
  if (duration.num)
    sec += (double) duration.num / (double) duration.denom;
  if (duration.units)
    sec *= (double) duration.units;
  return sec;
}

bool MP7Time::timePointIsValid() const
{
  return !tpoint.date.isEmpty() || tpoint.hasTime;
}

bool MP7Time::durationIsNull() const
{
  return !duration.days && !duration.sec && !duration.num;
}

int MP7Time::compareTimePoint(const MP7Time &other) const
{
  if (!tpoint.hasTime && !other.tpoint.hasTime)
    return 0;
  if (!tpoint.hasTime)
    return -1;
  if (!other.tpoint.hasTime)
    return 1;
  if (tpoint.sec < other.tpoint.sec)
    return -1;
  if (tpoint.sec > other.tpoint.sec)
    return 1;
  uint denom = commonMultiple(tpoint.denom, other.tpoint.denom);
  uint thisnum = tpoint.num * (denom / tpoint.denom);
  uint othernum = other.tpoint.num * (denom / other.tpoint.denom);
  if (thisnum < othernum)
    return -1;
  if (thisnum > othernum)
    return 1;
  return 0;
}

bool MP7Time::contains(const MP7Time &time) const
{
  int cmpStart = compareTimePoint(time);
  int cmpEnd = endTimePoint().compareTimePoint(time);
  if (duration.sign < 0) {
    return cmpStart >= 0 && cmpEnd <= 0;
  }
  return cmpStart <= 0 && cmpEnd >= 0;
}

uint MP7Time::commonMultiple(uint op1, uint op2)
{
    if (op1 % op2 == 0)
      return op1;
    if (op2 % op1 == 0)
      return op2;
    return op1 * op2;
}
