 /***********************************************************************
 *                                                                      *
 * MuViNo - ViTooKi Video Annotation Tool                               *
 *                                                                      *
 * title:	MP7VideoSegment.hpp                                     *
 *                                                                      *
 * $Id: MP7VideoSegment.hpp,v 1.10 2005/02/15 14:12:59 amueller_itec Exp $
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

#ifndef _MP7VIDEOSEGMENT_HPP_
#define _MP7VIDEOSEGMENT_HPP_

/** @file
 * Declaration of the MP7VideoSegment class.
 */

#include <vector>

#include "MP7Doc.hpp"
#include "MP7Time.hpp"
#include "MP7View.hpp"
#include "videosegment.hpp"
#include "creationinformation.hpp"
#include "uri.hpp"
#include "variation.hpp"

class MP7ListViewItem;

/** This class represents a view on a temporal video segment of an MPEG-7 document,
 * as used in the DAHL project.
 * <p>
 * The root of this view is always an MPEG-7 Video or VideoSegment descriptor.
 * </p>
 * @author Mario Taschwer
 * @version $Id: MP7VideoSegment.hpp,v 1.10 2005/02/15 14:12:59 amueller_itec Exp $
 */
class MP7VideoSegment : public MP7View
{
public:
  /** Default constructor. */
  MP7VideoSegment();

  /** Constructor. 
   * The <code>root</code> parameter is copied <em>as-is</em>
   * to this object's state.
   * @see #setRoot(MP7ListViewItem *, bool)
   */
  MP7VideoSegment(MP7Doc *doc, MP7ListViewItem *root);

  /* ------------------- Super class methods ------------------------------- */

  /** Return the root of the parent video segment.
   * @see MP7View::parent()
   */
  virtual MP7ListViewItem *parent(MP7ListViewItem *root = NULL) const;

  /** Set this view's root.
   * @see MP7View::setRoot(MP7ListViewItem *, bool)
   */
  virtual MP7ListViewItem *setRoot(MP7ListViewItem *root, bool searchUp = false);

  /** Create an empty video segment description. 
   * @see MP7View::setEmpty()
   */
  virtual void setEmpty();

  /** Show a modal top-level widget allowing the user to enter annotation data
   * of this video segment.
   * @see MP7View::edit()
   */
  virtual bool edit();
  
  /** Show a modal top-level widget allowing the user to enter creation information
   * of this video/ video segment.
   * @see MP7View::edit()
   */
  virtual bool editCreation();
  
  /** Show a modal top-level widget allowing the user to enter URI Paths
   * for video/ videosegments and an URI Path for the audio variation.
   * @see MP7View::edit()
   */
  virtual bool editUri();
  
  /** Show a modal top-level widget allowing the user to enter Variations
   * for video/ videosegments. 
   * @see MP7View::edit()
   */
  virtual bool editVariation();

  /* ------------ Methods specific to this class ----------------------- */

  /** Return the media time value of this video segment. */
  MP7Time mediaTime() const;

  /** Set the media time value of this video segment. 
   * The time point will be set only if it is valid.
   * The duration will be set only if it is not 0.
   * @see MP7Time::timePointIsValid()
   * @see MP7Time::durationIsNull()
   * @return <code>false</code> on error.
   */
  bool setMediaTime(const MP7Time &time);

  /** Modify the media start time (time point) of this segment.
   * The duration value will be updated such that the 
   * segment end time remains constant.
   * @param startTime   only the time point value will be used.
   * @param recursive   if <code>true</code>, this function will be
   *                    applied to the <em>first</em> child segment recursively.
   * @return <code>false</code> on error.
   */
  bool modifyMediaStartTime(const MP7Time &startTime, bool recursive = false);

  /** Modify the media duration of this segment from the given end time.
   * The media start time (time point) remains constant.
   * @param endTime     only the time point value will be used.
   * @param recursive   if <code>true</code>, this function will be
   *                    applied to the <em>last</em> child segment recursively.
   * @return <code>false</code> on error.
   */
  bool modifyMediaEndTime(const MP7Time &endTime, bool recursive = false);

  /** Return the segment title. */
  QString title() const;

  /** Return the segment level (tree depth) of this segment within the MPEG-7 document.
   * The top-level segment is assigned level 0.
   */
  int level() const;

  /** Determine an ancestor video segment.
   * @param result   where to store the result.
   * @param level    the segment level of the ancestor (0 means nearest to
   *                 the document root); must be less than the level of
   *                 this segment.
   *                 if -1, the parent segment will be returned.
   * @return <code>false</code> if the ancestor could not be found.
   * @see #level()
   */
  bool ancestor(MP7VideoSegment &result, int level = -1) const;

  /** Return the child video segments of this segment.
   * @param result     vector to store the result.
   * @return the number of child segments returned in <code>result</code>.
  */
  int childSegments(std::vector<MP7VideoSegment> &result) const;

  /** Add a child segment of this segment to the MPEG-7 document.
   * The child segment is appended as the last child of this segment.
   * @param result    output parameter: new child segment.
   * @return <code>false</code> on error.
   */
  bool addChildSegment(MP7VideoSegment &result);

  /** Move this segment (and all child segments) to a new parent segment,
   * right <em>after</em> the given new sibling segment.
   * If <code>newSibling</code> is <code>NULL</code> or is invalid, this segment will
   * be inserted as first child of <code>newParent</code>.
   * @return <code>false</code> on error.
   * @see MP7View::isValid()
   */
  bool move(MP7VideoSegment *newParent, MP7VideoSegment *newSibling = NULL);

  /** Split this segment at the given media time point into two
   * sibling segments (this and a new one).
   * This segment will keep its time point, but its
   * duration will be shortened. A root segment cannot be split.
   * @param result    output parameter: the new sibling segment.
   * @param time      split time point.
   * @return <code>false</code> on error.
   */
  bool split(MP7VideoSegment &result, const MP7Time &time);
  
  /** Remove this video segment from the MPEG-7 document.
   * When a segment with a sibling is removed, the sibling's
   * media time is increased to include the removed segment.
   * The top-level segment of an MPEG-7 document cannot be removed
   * (an error status is returned).
   * @param recursive if <code>false</code>, 
   *                  child segments are re-parented to a sibling (if exists)
   *                  or to this segment's parent; if <code>true</code>,
   *                  all child segments will be removed recursively.
   * @return <code>false</code> on error.
   */
  bool remove(bool recursive = false);

  /** Search for the video segment containing the given media time point
   * at the specified maximal segment level. Only the segment tree below this segment
   * is searched.
   * @param result      output parameter: the segment searched for.
   * @param time        the media time point to search for.
   * @param level       the maximal segment level of the result.
   * @return <code>false</code> on error.
   * @see #level()
   */
  bool searchByMediaTime(MP7VideoSegment &result, const MP7Time &time, int level) const;

  /** Find the previous generalized sibling segment of this segment.
   * This is the segment of the same depth with maximal media time
   * less than that of this segment.
   * @param result    where to store the result.
   * @return <code>false</code> if no such sibling exists.
   */
  bool prevGenSibling(MP7VideoSegment &result) const;

  /** Find the next generalized sibling segment of this segment.
   * This is the segment of the same depth with minimal media time
   * greater than that of this segment.
   * @param result    where to store the result.
   * @return <code>false</code> if no such sibling exists.
   */
  bool nextGenSibling(MP7VideoSegment &result) const;

  /** Set the TitleImage descriptor of this segment to
   * the given media time point.
   * @return the <code>MP7ListViewItem</code> corresponding to the
   *         created <code>TitleImage/MediaTimePoint</code>
   *         descriptor, or <code>NULL</code> on error.
   */
  MP7ListViewItem *setTitleImage(const MP7Time &time);

private:
  /** Initialize the VideoSegmentDlg's widgets from the MPEG-7 document. 
   * @return <code>false</code> on error.
   */
  bool initDlgFromMP7Doc(VideoSegment &dlg);
  
  bool initCreationDlgFromMP7Doc(CreationInformation &dlg);
  
    
  bool initUriDlgFromMP7Doc(Uri &dlg);
  
  bool initVariationDlgFromMP7Doc(Variation &dlg);
  
  /** Update MPEG-7 document from the VideoSegmentDlg's contents. 
   * @return <code>false</code> on error.
   */
  bool updateMP7DocFromDlg(const VideoSegment &dlg);
  
  bool updateCreationMP7DocFromDlg(const CreationInformation &dlg);
  
  bool updateUriMP7DocFromDlg(const Uri &dlg);
  
  bool updateVariationMP7DocFromDlg(const Variation &dlg);

};

#endif /* _MP7VIDEOSEGMENT_HPP_ */
