 /***********************************************************************
 *                                                                      *
 * MuViNo - ViTooKi Video Annotation Tool                               *
 *                                                                      *
 * title:	MP7View.hpp                                             *
 *                                                                      *
 * $Id: MP7View.hpp,v 1.5 2004/09/24 13:15:09 mtaschwer Exp $
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

#ifndef _MP7VIEW_HPP_
#define _MP7VIEW_HPP_

#include "MP7Doc.hpp"

/** @file
 * Declaration of the MP7View class.
 */

class MP7ListViewItem;

/** This abstract class defines an interface of a <em>view</em> on an MPEG-7 document.
 * We define a view as a subset of XML tree nodes of an MPEG-7 document. The
 * <em>root</em> of a view is the root of the minimal XML subtree of the MPEG-7 document
 * containing all the XML nodes of the view.
 * <p>
 * An implementation (subclass) of the view interface provides methods for easy creation and
 * manipulation of a certain set of MPEG-7 descriptors within an MPEG-7 document.
 * By defining creation and access methods in the view interface, an application
 * could manage different views on the same MPEG-7 document in a generic manner.
 * Moreover, the integration of additional views into the same application should
 * become more feasible.
 * </p>
 * @author Mario Taschwer
 * @version $Id: MP7View.hpp,v 1.5 2004/09/24 13:15:09 mtaschwer Exp $
 */
class MP7View
{
public:
  /** Destructor. */
  virtual ~MP7View() { }

  /** Return <code>true</code> if this view is valid, i.e. if
   * both the MPEG-7 document and the view root have been set.
   */
  virtual bool isValid() const { return doc != NULL && root != NULL; }

  /** Return <code>true</code> if this view is equal to another view,
   * i.e. if both views refer to the same MPEG-7 document and view root.
   */
  virtual bool equals(const MP7View &view) const 
  { 
    return doc == view.doc && root == view.root; 
  }

  /** Specify the MP7Doc object this view is going to operate on. */
  virtual void setMP7Doc(MP7Doc *doc) { this->doc = doc; }

  /** Return the root of this view. */
  MP7ListViewItem *rootItem() const { return root; }

  /** Return the root of the parent view.
   * @param root   the root of the view whose parent should be determined;
   *               if NULL, this view's root will be used.
   * @return <code>NULL</code> if there is no parent view.
   */
  virtual MP7ListViewItem *parent(MP7ListViewItem *root = NULL) const { return NULL; }

  /* ---------------------- Pure Virtual Methods ------------------------- */

  /** Specify the root of this view, i.e. the root of the minimal
   * XML subtree containing this view within the MPEG-7 document.
   * If the given root element's tag name is not what this view
   * expects it to be, the XML tree is searched for the nearest
   * suitable element in the direction specified by the <code>searchUp</code>
   * parameter.
   * @param root     the root of this view, or an element in the vicinity of it.
   * @param searchUp if the tag name of <code>root</code> is not
   *                 what this view expects it to be, the XML tree is searched
   *                 along the path to the document root 
   *                 (<code>searchUP == true</code>), or in a breadth-first
   *                 manner descending from <code>root</code>
   *                 (<code>searchUP == false</code>), until the first suitable
   *                 element is found. This element will become this view's root.
   * @return the new root element, or
   *         <code>NULL</code> if no root element could be determined.
   */
  virtual MP7ListViewItem *setRoot(MP7ListViewItem *root, bool searchUp = false) = 0;

  /** Create an <i>empty</i> XML element structure corresponding to
   * this view. Existing XML elements will not be deleted, but
   * their values will be cleared. So calling this method twice should 
   * do no harm.
   */
  virtual void setEmpty() = 0;

  /** Show a modal top-level widget allowing the user to enter MPEG-7 data
   * for all XML elements of this view.
   * @return <code>true</code> if the MPEG-7 document has been successfully
   *         modified.
   */
  virtual bool edit() = 0;

protected:
  /** The MP7Doc object this view is going to operate on. */
  MP7Doc *doc;

  /** The root of this view (see the class description). */
  MP7ListViewItem *root;

  /** Default constructor. */
  MP7View() { doc = NULL; root = NULL; }
};

#endif /* _MP7VIEW_HPP_ */
