 /***********************************************************************
 *                                                                      *
 * MuViNo - ViTooKi Video Annotation Tool                               *
 *                                                                      *
 * title:	SegmentSlider.cpp                                       *
 *                                                                      *
 * $Id: SegmentSlider.cpp,v 1.8 2005/02/15 14:12:59 amueller_itec Exp $
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

#include <qbutton.h>
#include <qcolor.h>
#include <qframe.h>
#include <qlabel.h>
#include <qmessagebox.h>
#include <qspinbox.h>
#include <qtooltip.h>
#include <qsizepolicy.h>

#include <debug.hpp>
#include <global.hpp>		// u64 data type

#include "MP7ListViewItem.hpp"
#include "SegmentSlider.hpp"

/* width of marker widget (in pixel) */
#define MARKER_WIDTH_PIX    2

SegmentSlider::SegmentSlider(QWidget *parent, const char *name)
  : QSlider(0, 1000, 0, 0, Qt::Horizontal, parent, name)
{
  startTimeLabel = NULL;
  endTimeLabel = NULL;
  levelSpinBox = NULL;
  prevSiblingButton = NULL;
  nextSiblingButton = NULL;
  segmentZooming = true;
  connect(this, SIGNAL(sliderReleased()), this, SLOT(emitSliderMovedSignal()));
}

SegmentSlider::~SegmentSlider()
{
  /* marker widgets are automatically destroyed by Qt */
}

void SegmentSlider::setSegment(const MP7VideoSegment &newSegment, bool emitSignal)
{
  mediaTime = newSegment.mediaTime(); // cache mediaTime
  if (!newSegment.equals(segment)) {
    segment = newSegment;	// copy
    /*
    dprintf_full("SegmentSlider::setSegment() segment = %s (%s), rootItem = %p, "
		 "emitSignal = %s\n",
		 segment.title().latin1(), segment.mediaTime().toString().latin1(),
		 segment.rootItem(), emitSignal ? "true" : "false");
    */
    updateSegmentDisplay();
    /* enable/disable buddy buttons */
    if (prevSiblingButton) {
      MP7VideoSegment sib;
      prevSiblingButton->setEnabled(segment.prevGenSibling(sib));
    }
    if (nextSiblingButton) {
      MP7VideoSegment sib;
      nextSiblingButton->setEnabled(segment.nextGenSibling(sib));
    }
    if (emitSignal && segment.isValid()) {
      emit segmentChanged(segment.rootItem());
    }
  }
}

void SegmentSlider::setTitleLabel(QLabel *label)
{
  titleLabel = label;
}

void SegmentSlider::setStartTimeLabel(QLabel *label)
{
  startTimeLabel = label;
}

void SegmentSlider::setEndTimeLabel(QLabel *label)
{
  endTimeLabel = label;
}

void SegmentSlider::setLevelSpinBox(QSpinBox *spinbox)
{
  levelSpinBox = spinbox;
  connect(spinbox, SIGNAL(valueChanged(int)), this, SLOT(segmentLevelChange(int)));
}

void SegmentSlider::setPrevSiblingButton(QButton *button)
{
  prevSiblingButton = button;
  connect(button, SIGNAL(clicked()), this, SLOT(gotoPrevSibling()));
}

void SegmentSlider::setNextSiblingButton(QButton *button)
{
  nextSiblingButton = button;
  connect(button, SIGNAL(clicked()), this, SLOT(gotoNextSibling()));
}

void SegmentSlider::setMarker(const MP7VideoSegment &subsegment)
{
  MP7Time time(subsegment.mediaTime());
  int value = timeToValue(time);
  /*
  dprintf_full("SegmentSlider::setMarker() time = %s, value = %d\n",
	       time.toString().latin1(), value);
  */
  /* construct segment tool-tip */
  QString s(subsegment.title());
  if (!s.isEmpty())
    s.append(' ');
  s.append(time.toString());
  /* find position where to insert the new marker.
   * if marker at <value> exists already, remove it. */
  std::list<marker>::iterator it;
  for (it = markers.begin(); it != markers.end(); it++) {
    if ((*it).value == value) {
      it = markers.erase(it);
      break;
    }
    if ((*it).value > value)
      break;
  }
  /* create new marker */
  struct marker m;
  m.value = value;
  m.qframe = new QFrame(this, "marker");
  m.qframe->setFixedWidth(MARKER_WIDTH_PIX);
  m.qframe->resize(MARKER_WIDTH_PIX, sliderRect().height());
  m.qframe->setPaletteBackgroundColor(red);
  m.length = durationToValue(time);
  m.segmentTip = s;
  m.segment = subsegment;
  /* insert new marker at <it> */
  markers.insert(it, m);
  m.qframe->show();
  return;
}

void SegmentSlider::removeMarker(int value)
{
  std::list<marker>::iterator it;
  for (it = markers.begin(); it != markers.end(); it++) {
    if ((*it).value == value) {
      delete (*it).qframe;
      markers.erase(it);
      return;
    }
  }
  return;
}

void SegmentSlider::removeMarker(const MP7Time &time)
{
  removeMarker(timeToValue(time));
}

void SegmentSlider::repositionMarkers()
{
  int range = maxValue() - minValue();
  std::list<marker>::iterator it;
  for (it = markers.begin(); it != markers.end(); it++) {
    int x = (*it).value * width() / range;
    (*it).qframe->move(x, 0);
    if (!(*it).segmentTip.isEmpty()) {
      /* remove old segment tool-tip */
      if ((*it).rect.isValid())
	QToolTip::remove(this, (*it).rect);
      /* add new segment tool-tip */
      int segwidth = (*it).length * width() / range;
      QRect segrect(x, 0, segwidth, height());
      QToolTip::add(this, segrect, (*it).segmentTip);
      (*it).rect = segrect;
    }
  }
}

void SegmentSlider::enableSegmentZooming(bool on)
{
  segmentZooming = on;
}

int SegmentSlider::timeToValue(const MP7Time &time) const
{
  double range = (double) (maxValue() - minValue());
  double rangeSec = mediaTime.durationToSec();
  int val = minValue();
  if (rangeSec)
    val += (int) ((time.timeToSec() - mediaTime.timeToSec()) * range / rangeSec);
  return val;
}

MP7Time SegmentSlider::valueToTime(int value) const
{
  MP7Time time(mediaTime);
  u64 range = maxValue() - minValue();
  int msecs = (int) ((u64)(value - minValue()) * (u64)mediaTime.durationToMSec() / range);
  time.moveTimePoint(msecs, 1000);
  return time;
}

int SegmentSlider::durationToValue(const MP7Time &time) const
{
  double range = (double) (maxValue() - minValue());
  double rangeSec = mediaTime.durationToSec();
  int val = 0;
  if (rangeSec)
    val = (int) (time.durationToSec() * range / rangeSec);
  return val;
}


/* --------------------------- Event Handlers ---------------------------------- */

void SegmentSlider::resizeEvent(QResizeEvent *e)
{
  repositionMarkers();
}

void SegmentSlider::mouseDoubleClickEvent(QMouseEvent *e)
{
  e->accept();
  if (segmentZooming) {
    QPoint pos = e->pos();
    /* search for subsegment corresponding to <pos> */
    std::list<marker>::iterator it;
    for (it = markers.begin(); it != markers.end(); it++) {
      if ((*it).rect.contains(pos))
	break;
    }
    if (it == markers.end()) {
      dprintf_small("SegmentSlider::mouseDoubleClickEvent() subsegment not found!\n");
      return;
    }
    setSegment((*it).segment);
    /* reset slider, since it has been moved by the first mouse click */
    setValue(minValue());
  }
}

/* ----------------------- Slot Implementations -------------------------------- */

void SegmentSlider::editSegment()
{
  segment.edit();
  updateSegmentDisplay();
  emit segmentChanged(segment.rootItem());
}

void SegmentSlider::editCreation()
{
  segment.editCreation();
  updateSegmentDisplay();
  emit segmentChanged(segment.rootItem());
}

void SegmentSlider::editUri()
{
  segment.editUri();
  updateSegmentDisplay();
  emit segmentChanged(segment.rootItem());
}

void SegmentSlider::editVariation()
{
  segment.editVariation();
  updateSegmentDisplay();
  emit segmentChanged(segment.rootItem());
}

void SegmentSlider::removeSegment()
{
  bool removed = false;
  MP7VideoSegment parent;
  if (!segment.ancestor(parent))
    return;
  std::vector<MP7VideoSegment> children;
  if (segment.childSegments(children) > 0) {
    switch (QMessageBox::information(this, "Delete VideoSegment",
				     "The current video segment contains subsegments.\n"
				     "Do you want to delete the current segment only,\n"
				     "or all its subsegments, too?",
				     "Current &Only", "&Delete All", "&Cancel",
				     0, 2)) {
    case 0: /* delete current segment only */
      removed = segment.remove(false);
      break;
    case 1: /* delete current segment and all its subsegments recursively */
      removed = segment.remove(true);
      break;
    case 2: /* cancel operation */
      return;
    }
  } else {
    switch (QMessageBox::information(this, "Delete VideoSegment",
				     "Do you want to delete the current video segment?",
				     QMessageBox::Yes, QMessageBox::No)) {
    case QMessageBox::Yes:
      removed = segment.remove();
      break;
    case QMessageBox::No:
      return;
    }
  }
  if (removed) {
    setSegment(parent);
  } else {
    dprintf_err("SegmentSlider::removeSegment() FAILED!\n");
  }
}

void SegmentSlider::segmentLevelChange(int level)
{
  if (level < 0)
    level = -level;
  if (!segment.isValid())
    return;
  int thislevel = segment.level();
  if (segmentZooming && level < thislevel) {
    MP7VideoSegment p;
    if (!segment.ancestor(p, level)) {
      dprintf_err("SegmentSlider::segmentLevelChange(level = %d) parent segment not found!\n",
		  level);
    } else {
      setSegment(p);
    }
  }
}

void SegmentSlider::setSliderFromMediaTime(const MP7Time &time)
{
  if (!segment.isValid())
    return;
  int value = timeToValue(time);
  if (value < minValue() || value > maxValue()) {
    MP7VideoSegment root;
    if (!segment.ancestor(root, 0)) {
      dprintf_err("SegmentSlider::setSliderFromMediaTime(time = %s) root segment NOT FOUND!\n",
		  time.timePointPretty().latin1());
      return;
    }
    MP7VideoSegment newSegment;
    if (!root.searchByMediaTime(newSegment, time, segment.level())) {
      dprintf_err("SegmentSlider::setSliderFromMediaTime(time = %s) searchByMediaTime() "
		  "returned NULL!\n",
		  time.timePointPretty().latin1());
      return;
    }
    setSegment(newSegment);
    value = timeToValue(time);
  }
  setValue(value);
}

void SegmentSlider::updateSegmentDisplay()
{
  /* title label */
  if (titleLabel) {
    if (segment.isValid())
      titleLabel->setText(segment.title());
    else
      titleLabel->setText("(No title)");
  }
  /* level spin box */
  if (levelSpinBox) {
    if (segment.isValid()) {
      int level = segment.level();
      levelSpinBox->setMinValue(-level);
      levelSpinBox->setValue(-level);
    } else {
      levelSpinBox->setMinValue(0);
      levelSpinBox->setMaxValue(0);
      levelSpinBox->setValue(0);
    }
  }
  /* update segment time labels */
  if (segment.isValid()) {
    mediaTime = segment.mediaTime();
  } else {
    mediaTime = MP7Time();
  }
  if (startTimeLabel) {
    startTimeLabel->setText(mediaTime.timePointPretty());
  }
  if (endTimeLabel) {
    endTimeLabel->setText(mediaTime.endTimePointPretty());
  }
  /* clear all markers */
  for (std::list<marker>::iterator it = markers.begin(); it != markers.end(); it++) {
    delete (*it).qframe;
    if ((*it).rect.isValid())
      QToolTip::remove(this, (*it).rect);
  }
  markers.clear();
  /* set segment markers */
  if (segment.isValid()) {
    std::vector<MP7VideoSegment> children;
    int size = segment.childSegments(children);
    for (int i=0; i < size; i++) {
      setMarker(children.at(i));
    }
  }
  repositionMarkers();
}

void SegmentSlider::gotoPrevSibling()
{
  MP7VideoSegment sibSeg;
  if (segment.prevGenSibling(sibSeg)) {
    setSegment(sibSeg);
  }
}

void SegmentSlider::gotoNextSibling()
{
  MP7VideoSegment sibSeg;
  if (segment.nextGenSibling(sibSeg)) {
    setSegment(sibSeg);
  }
}

void SegmentSlider::emitSliderMovedSignal()
{
  dprintf_full("SegmentSlider::emitSliderMovedSignal() media time = %s\n",
	       valueToTime(value()).toString().latin1());
  emit sliderMoved(valueToTime(value()));
}
