 /***********************************************************************
 *                                                                      *
 * MuViNo - ViTooKi Video Annotation Tool                               *
 *                                                                      *
 * title:	SegmentSlider.hpp                                       *
 *                                                                      *
 * $Id: SegmentSlider.hpp,v 1.7 2005/02/15 14:12:59 amueller_itec Exp $
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

#ifndef _SEGMENT_SLIDER_HPP_
#define _SEGMENT_SLIDER_HPP_

/** @file
 * Declaration of the SegmentSlider class.
 */

#include <list>
#include <qslider.h>

#include "MP7Time.hpp"
#include "MP7VideoSegment.hpp"

class QButton;
class QFrame;
class QSpinBox;

/** This class implements a slider widget for representing a video segment.
 * It allows segment markers to be set in media time units.
 * @author Mario Taschwer
 * @version $Id: SegmentSlider.hpp,v 1.7 2005/02/15 14:12:59 amueller_itec Exp $
 */
class SegmentSlider : public QSlider
{
  Q_OBJECT

public:
  /** Constructor. 
   * Initializes this slider with horizontal orientation and range [0..100].
   */
  SegmentSlider(QWidget *parent, const char *name = NULL);

  /** Destructor. */
  ~SegmentSlider();

  /** Set the video segment represented by this slider
   * and update the display accordingly.
   * If <code>emitSignal</code> is <code>true</code> and
   * <code>segment</code> is not equal to the current segment,
   * a {@link #segmentChanged(MP7ListViewItem *)} signal will
   * be emitted.
   */
  void setSegment(const MP7VideoSegment &segment, bool emitSignal = true);

  /** Return the video segment represented by this slider. */
  MP7VideoSegment getSegment() const { return segment; }

  /** Set the segment title buddy label. */
  void setTitleLabel(QLabel *label);

  /** Set the segment start time buddy label. */
  void setStartTimeLabel(QLabel *label);

  /** Set the segment end time buddy label. */
  void setEndTimeLabel(QLabel *label);

  /** Set the segment level spinbox buddy widget. */
  void setLevelSpinBox(QSpinBox *spinbox);

  /** Set the <i>previous sibling</i> buddy button. */
  void setPrevSiblingButton(QButton *button);

  /** Set the <i>next sibling</i> buddy button. */
  void setNextSiblingButton(QButton *button);

  /** Enable or disable segment zooming.
   * Per default, segment zooming is enabled.
   * @see #mouseDoubleClickEvent(QMouseEvent *)
   */
  void enableSegmentZooming(bool on);

public slots:
  /** Invoke an edit dialog on the video segment.
   * A {@link #segmentChanged(MP7ListViewItem *)} signal
   * is emitted.
   * @see MP7VideoSegment::edit()
   */
  void editSegment();
  
  void editCreation();
  
  void editUri();
  
  void editVariation();

  /** Remove the current video segment.
   * The parent segment becomes the current one.
   * The top-level segment cannot be removed.
   */
  void removeSegment();

  /** The level spin box value changed. 
   * If the new value is less than the current segment level
   * and segment zooming is enabled, this slider will zoom-out to
   * the video segment of the specified level.
   */
  void segmentLevelChange(int level);

  /** Update the slider's position according to the current media time. */
  void setSliderFromMediaTime(const MP7Time &time);

  /** Update display of markers and buddy labels according to
   * the MP7VideoSegment. 
   */
  void updateSegmentDisplay();

  /** Make the previous generalized sibling segment the current one (if available). 
   * Two nodes of a tree are <em>generalized siblings</em> if they have the same
   * distance to the root node.
   */
  void gotoPrevSibling();

  /** Make the next generalized sibling segment the current one (if available).
   * Two nodes of a tree are <em>generalized siblings</em> if they have the same
   * distance to the root node.
   */
  void gotoNextSibling();

signals:
  /** The current segment has changed. 
   * The root element of the new segment is emitted.
   */
  void segmentChanged(MP7ListViewItem *segmentRoot);

  /** The slider was moved by the user.
   * The time represented by the new slider position is emitted.
   */
  void sliderMoved(const MP7Time &time);

protected:
  /** Set a segment marker corresponding to the given subsegment.
   * The slider will show a tool-tip with the segment title
   * in the area corresponding to the segment duration.
   * You need to call {@link #repositionMarkers()} afterwards to
   * display the new marker.
   */
  void setMarker(const MP7VideoSegment &subsegment);

  /** Remove the segment marker at the position given by <code>value</code>. */
  void removeMarker(int value);

  /** Remove the segment marker at the given media time point. */
  void removeMarker(const MP7Time &time);

  /** Re-position segment markers and activate segment tool-tips.
   * The marker values are not modified.
   */
  void repositionMarkers();

  /** Convert time point to slider value. */
  int timeToValue(const MP7Time &time) const;

  /** Convert slider value to time point. */
  MP7Time valueToTime(int value) const;

  /** Convert duration to slider value. */
  int durationToValue(const MP7Time &time) const;

  /** Resize event handler. */
  virtual void resizeEvent(QResizeEvent *e);

  /** Double-click event handler. 
   * If segment zooming is enabled, double-clicking this widget
   * will zoom into the subsegment corresponding to the location
   * of the double-click event.
   * @see #enableSegmentZooming(bool)
   */
  virtual void mouseDoubleClickEvent (QMouseEvent *e);

protected slots:
  /** Emit a <code>sliderMoved(const MP7Time &)</code> signal.
   */
  void emitSliderMovedSignal();

private:
  /** A segment marker represents a subsegment.<br>
   * <code>value</code> is the slider value corresponding to the start
   * of the subsegment.
   * <code>length</code> denotes the subsegment length in slider value units
   * (not the marker's width!).<br>
   * <code>segmentTip</code> is the tool-tip text to be displayed in the
   * slider area corresponding to the subsegment.<br>
   * <code>rect</code> is the region corresponding to the subsegment.
   * <code>segment</code> points to the MP7VideoSegment object corresponding
   * to the subsegment.
   */
  struct marker {
    int value;
    int length;
    QFrame *qframe;
    QString segmentTip;
    QRect rect;
    MP7VideoSegment segment;
  };

  /** Segment markers (sorted by value). */
  std::list<marker> markers;

  /** The video segment represented by this slider. */
  MP7VideoSegment segment;

  /** Cache for segment time and duration.
   * Is equal to <code>segment.mediaTime()</code>.
   */
  MP7Time mediaTime;

  /** The segment title buddy label. */
  QLabel *titleLabel;

  /** The segment start time buddy label. */
  QLabel *startTimeLabel;

  /** The segment end time buddy label. */
  QLabel *endTimeLabel;

  /** The segment level spinbox buddy widget. */
  QSpinBox *levelSpinBox;

  /** The <i>previous sibling</i> buddy button. */
  QButton *prevSiblingButton;

  /** The <i>next sibling</i> buddy button. */
  QButton *nextSiblingButton;

  /** Is segment zooming enabled? */
  bool segmentZooming;
};

#endif /* _SEGMENT_SLIDER_HPP_ */
