 /***********************************************************************
 *                                                                      *
 * MuViNo - ViTooKi Video Annotation Tool                               *
 *                                                                      *
 * title: muvinomain.hpp                                                *
 *                                                                      *
 * $Id: muvinomain.hpp,v 1.12 2005/01/20 09:47:08 mkropfbe Exp $
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

#ifndef MUVINOMAIN_H
#define MUVINOMAIN_H

/** @file
 * Declaration of the MuvinoMain class.
 * This file contains also the Doxygen \@mainpage source comment
 * with the \ref index "Main Page" content.
 */

/** @mainpage Muvino - An MPEG-7 Video Annotation Tool
 *
 * <div align="center"><b>Version 0.12</b></div>
 *
 * @section overview_sec Overview
 *
 * Muvino is a tool for creating MPEG-7 metadata of videos. It consists of
 * an embedded video player based on the 
 * <a href="http://vitooki.sourceforge.net/">ViTooKi</a>
 * library and an XML tree view
 * allowing to navigate and edit an MPEG-7 document. It was designed to
 * facilitate the creation and annotation of MPEG-7 <code>%VideoSegment</code>
 * descriptions, by which a video can be divided temporally into a hierarchy 
 * of semantic sections, like the chapters and sections of a printed book.
 *
 * <code>%VideoSegment</code> descriptors are created by splitting the <em>current
 * segment</em> into subsegments at the current media time point shown in the video
 * player. Special slider widgets facilitate quick navigation within the 
 * <code>%VideoSegment</code> tree. The annotation data belonging to a 
 * <code>%VideoSegment</code> can be edited in a user-friendly dialog, there
 * is no need to manipulate the XML tree directly.
 *
 * Muvino supports the following MPEG-7 descriptors within a
 * <code>%VideoSegment</code>:
 * <ul>
 * <li><code>CreationInformation/Creation/Title</code></li>
 * <li><code>CreationInformation/Creation/TitleMedia/TitleImage</code></li>
 * <li><code>CreationInformation/Creation/Abstract</code></li>
 * <li>semantic place and semantic time within the <code>Semantic</code>
 *     descriptor</li>
 * <li><code>FreeTextAnnotation</code> and <code>KeywordAnnotation</code>
 *     within <code>Abstract</code> (see above) and <code>TextAnnotation</code></li>
 * <li><code>MediaTime</code> and <code>TemporalDecomposition</code> - maintained 
 *     automatically by Muvino</li>
 * </ul>
 * 
 * @section install_sec Installation
 *
 * Muvino is part of the <a href="http://vitooki.sourceforge.net/">ViTooKi</a> tool kit.
 * So all system and installation requirements of ViTooKi apply. There is only
 * one binary called <code>muvino</code>, which gets installed automatically when 
 * <code>make install</code> is invoked in the ViTooKi top-level source directory.
 *
 * However, note that Muvino requires Qt 3.1 or newer, which is not freely available for
 * the Windows platform. This is also the reason why Muvino has not yet been compiled
 * and tested on Windows.
 *
 * @section usage_sec Usage
 *
 * Unfortunately, a separate user documentation is not available. However, most of
 * the GUI elements should have an intuitive meaning. If not, the tool-tips appearing
 * when the mouse is moved over the GUI elements may help.
 *
 * To display help for the command-line options invoke <code>muvino -h</code>.
 *
 * @section problems_sec Known Problems and Missing Features
 *
 * Muvino is not yet ready for a file release. The following problems need to
 * be fixed:
 * <ul>
 * <li>The order of <code>%VideoSegment</code>'s XML child nodes does not
 *     conform to MPEG-7.</li>
 * <li>Only <code>VariationDescriptionType</code> is implemented;
 *     at least, <code>ContentDescriptionType</code> should be supported, too.</li>
 * <li>The video segment edit dialog is modal, i.e. interaction with
 *     the main window is not possible while the edit dialog is open.
 *     Concurrent video play-back control should be allowed.</li>
 * <li>Non-leaf video segments cannot be inserted; their creation by moving
 *     selected sibling segments one level down into the new segment
 *     would be desireable.</li>
 * <li>Video play-back cannot be easily started at the start of
 *     a given video segment.</li>
 * <li>The single video frame specified by the <code>TitleImage</code>
 *     descriptor cannot be displayed.</li>
 * </ul>
 *
 * @section author_sec Authors
 *
 * Mario Taschwer &lt;mt@itec.uni-klu.ac.at&gt; <br>
 * Armin M&uuml;ller &lt;armin@itec.uni-klu.ac.at&gt;
 *
 */

#include <DataChannel.hpp>
#include <ReferenceCounter.hpp>
#include <adaptors/MP4Decoder.hpp>

#include "muvinomaindlg.hpp"
#include "MP7Doc.hpp"
#include "MP7Time.hpp"

class QApplication;
class QTimer;
class GlobalTimer;
class ESSynchronizer;
class SegmentSlider;

/**
 * This class implements the application logic of the main GUI class
 * MuvinoMainDlg. The latter class has been generated using the Qt Designer.
 *
 * @author Mario Taschwer
 */

class MuvinoMain : public MuvinoMainDlg  {
  Q_OBJECT

public: 
  MuvinoMain();
  ~MuvinoMain();

protected:
  /** Set the caption of this top-level widget according to the given filename. 
   * @param file     file name (may be full path name), the directory path
   *                 will not be displayed.
   * @param modified if <code>true</code>, an asterisk will be displayed to
   *                 indicate that the document has been modified.
   */
  void setCaptionFromFile(const QString &file = QString::null, bool modified = false);

  /* ----------------------- Event Handlers ----------------------------------------- */

  /** Window close event handler. */
  virtual void closeEvent(QCloseEvent * e);

  /* ----------------------- Menu Actions --------------------------------------- */

public slots:
  virtual void fileNew();
  virtual void fileOpen();
  virtual void fileOpen(const QString &path);
  virtual void fileLoadVideo();
  virtual void fileLoadVideo(const QString &path);
  virtual void fileExit();
  virtual void fileSave();
  virtual void fileSaveAs();

protected:
  void initMenuBar();

  /* ----------------------- Video Player Members --------------------------------- */

public:
  /**
   * Initialize the adaptor chain and the video widget for play-back of an MP4 encoded
   * video. The resolution (width, height) needs to match the
   * dimensions of #videoWidget.
   *
   * @param filename file name of MP4 encoded video.
   * @param demuxDir local directory where the demuxed elementary streams are
   *                 to be stored.
   * @param colorSpace color space which the video should be converted to; must be one
   *                   of: UncompressedVideoFrame::ColorSpaceGRAY8, UncompressedVideoFrame::ColorSpaceRGB24, 
   *                   UncompressedVideoFrame::ColorSpaceRGB32.
   *
   * @return <code>true</code> on success, <code>false</code> on error.
   * @see QLabelRenderer
   */
  bool loadVideo(const char *filename, const char *demuxDir,
		 UncompressedVideoFrame::ColorSpace colorSpace);

public slots:
  virtual void playerStart();
  virtual void playerStop();
  virtual void playerPlay();
  virtual void playerBack3s();
  virtual void setDisplayRateFromSlider(int value);
  virtual void playerAudioStateChange(int state);

  
  
  /** Display current media time in player and emit the
   * <code>currentMediaTimeEvent(const MP7Time &)</code> signal.
   */
  virtual void displayCurrentMediaTime(bool showSubSec = false);

  /**
   * Close the video loaded by #loadVideo().
   */
  void closeVideo();

  /**
   * Start video play-back at a given absolute media time point.
   *
   * @param time   media time point; an invalid (default) time point
   *               means &quot;continue from current time point&quot;.
   * @return <code>true</code> on success, <code>false</code> on error.
   * @see MP7Time::timePointIsValid()
   */
  bool playVideo(const MP7Time &time = MP7Time());

signals:
  void currentMediaTimeEvent(const MP7Time &time);

 protected:
  /**
   * Reset the video widget to its default settings.
   */
  void resetVideoDisplay();

  /** Stop audio play-back. 
   * When this function returns, the audio DataChannel thread
   * has terminated.
   * Make sure that #globalTimer is not paused before calling this method.
   */
  void stopAudio();

  /** Stop video and audio play-back.
   * When this function returns, the video and audio DataChannel threads
   * have terminated, and the input streams have been closed.
   * Make sure that #globalTimer is not paused before calling this method.
   */
  void stopVideoAudio();

  /** Pause video and audio play-back.
   * The DataChannel threads will eventually terminate, leaving the input stream
   * open. This function does not wait for the DataChannel threads to terminate.
   */
  void pauseVideoAudio();

  /** Return the duration of the loaded video.
   * If no video is loaded, a null time will be returned.
   */
  MP7Time videoDuration() const;

  /**
   * Get current video media time stamp.
   * Returns a default MP7Time object if the time stamp cannot be determined.
   */
  MP7Time currentVideoTime() const;

  /** Set the current media time label to the given value. 
   * @param time         media time point.
   * @param showSubSec   determines whether sub-second fractions will
   *                     be shown.
   */
  void setCurrentVideoTimeLabel(const MP7Time &time, bool showSubSec = false);

  /** Set the display rate slider control to the given rate.
   * @see #setDisplayRateFromSlider(int)
   */
  void setDisplayRateSlider(double rate);

  /* ------------------ MP7Doc/MP7TreeView Methods ----------------------------------- */

public:
  /**
   * Load an MPEG-7 file in XML format.
   * @return <code>true</code> on success, <code>false</code> on error.
   */
  bool loadMP7Doc(const QString &filename);

  /** Check MPEG-7 document with loaded video.
   * If no MPEG-7 document has been loaded, a new one
   * with an appropriate Video descriptor will be created.
   * In case of discrepancy, a warning dialog will be shown to the
   * user which lets him choose between opening a different MPEG-7 document, 
   * creating a new one (which is instantly performed), 
   * or cancelling the operation (which is implemented by the caller of this function).
   * @return 0 if the current MPEG-7 document and the video are compatible.<br>
   *         1 if a new (compatible) MPEG-7 document has been created.<br>
   *         2 if the user chose to open a new MPEG-7 document.<br>
   *         3 if the user chose to cancel the operation.
   */
  int checkMP7DocWithVideo();
  
public slots:
  /** Tree view item was selected. */
  virtual void treeViewSelectItem(QListViewItem *item);

  /** Modification status of the MPEG-7 document has changed. 
   * @param m   <code>true</code> means <i>modified</i>.
   */
  virtual void mp7DocModificationChange(bool m);

  /** Set selected item. */
  virtual void treeViewSetSelected(MP7ListViewItem *item);

  /** Set the current tree view item and cause <code>currentSlider</code> to show the
   * video segment containing the XML element <code>item</code>.
   * @param item    must actually be of type <code>MP7ListViewItem *</code>.
   */
  void treeViewClicked(QListViewItem *item);

protected:
  /** Initialize the MPEG-7 tree view widget. */
  void initMP7TreeView();

  /** Ask the user to save the current MPEG-7 document
   * if it has been modified. Save the document if the user
   * chose to do so. A message box with 3 buttons will be
   * displayed: <em>Save</em>, <em>Discard</em>, and <em>Cancel</em>.
   * @return <code>false</code> if the user chose to cancel
   *         the operation; <code>true</code> otherwise.
   */
  bool warnAndSaveMP7Doc();

  /** Create new MPEG-7 document with one video segment of given duration.
   * The existing MPEG-7 document will be deleted without warning!
   */
  void createMP7Doc(const MP7Time &duration);

  /* ------------------ MP7ElementValue Methods ------------------------------ */

public slots:
  virtual void mp7ElementValueChange();

protected:
  /** Reset the mp7ElementValue widget. */
  void resetMP7ElementValue();

  /** Disable the mp7ElementValue widget. */
  void disableMP7ElementValue(bool disable);

  /* ------------------ Segment Navigation Methods --------------------------- */

public slots:
  /** Split current segment at current media time point.
   * If there are child segments, the appropriate child segment
   * will be splitted.
   */
  void splitCurrentSegment();

  /** Set the MPEG-7 TitleImage descriptor of the current segment
   * to the current video frame.
   */
  void setTitleImage();

protected:
  /** Initialize the segmentNav widgets. */
  void initSegmentNav();

  /** Reset the segmentNav widgets to show default values. */
  void resetSegmentNav();

  /* ------------------ Member Variables ------------------------------------- */

private:
  /** The main window caption. */
  static const QString caption;

  /** The DataChannel belonging to the visual elementary stream of a loaded video. */
  DataChannel *videoDC;

  /** The DataChannel belonging to the audio elementary stream of a loaded video. */
  DataChannel *audioDC;

  /** The global timer controlling the DataChannels. */
  GlobalTimer *globalTimer;

  /** The reference counter for using #globalTimer as a SharedAdaptor. */
  ReferenceCounter *globalTimerRefCnt;

  /** The ESSynchronizer controlling the #videoDC DataChannel. */
  ESSynchronizer *essync;

  /** The timer for updating the current media time label. */
  QTimer *mediaTimeUpdater;

  /** The MPEG-7 document. */
  MP7Doc *doc;

  /** The currently selected QListViewItem object. */
  MP7ListViewItem *currentMP7Item;

  /** Segment slider for current video segment. */
  SegmentSlider *currentSlider;

  /** Segment slider for total video. */
  SegmentSlider *totalSlider;
};

#endif
