/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	Presentation.hpp                                              *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

/***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef _PRESENTATION_HPP_
#define _PRESENTATION_HPP_

#include "ContainerInfo.hpp"
#include "DataChannel.hpp"
#include "DataSink.hpp"
#include "Adaptor.hpp"
#include "AdaptorChain.hpp"
#include "ESInfo.hpp"
#include "ReferenceCounter.hpp"
#include "adaptors/GlobalTimer.hpp"
#include "adaptors/ESSynchronizer.hpp"

#include "VideoDisplayIO.hpp"
#include "io/SDLaudioIO.hpp"

class QObject;
#include <qframe.h>
#include <qstring.h>
#include <qtimer.h>

#ifdef ISOMP4
#include "ISOMovies.h"
#endif
#include <list>
#include "VideoDisplayIO.hpp"

#ifndef TEARDOWN_ALL
#define TEARDOWN_ALL 0xFFFFFFFF
#endif

class IO;
class ContainerInfo;
class DataChannel;
class DataSink;
class Adaptor;
class AdaptorChain;
class ESInfo;
class ReferenceCounter;
class GlobalTimer;
class ESSynchronizer;
class VideoDisplayIO;
class SDLaudioIO;
class Statistics;


/* Information about one Presentation 
   This can be local or remote
   In the local case usedAdaptors has to be used
   In the remote case usedTermCaps and rtxEnabled have to be used

*/

struct PresentationInfo {
  bool localPresentation;
  char * URI;
  u32 numberOfAES;
  u32 numberOfVES;
  list <ESInfo *>  origESInfo;
  list <ESInfo *>  actESInfo;
  list <QString> usedAdaptors;
  TerminalCapabilities * usedTermCaps;
  bool rtxEnabled;
  PresentationInfo() {
    localPresentation = false;
    URI = NULL;
    numberOfAES = 0;
    numberOfVES = 0;
    origESInfo.clear();
    actESInfo.clear();
    usedTermCaps = NULL;
    rtxEnabled = false;
  };
};




/**
 *
 * <Class Description goes here>
 *
 *@author Ronald Sowa
 *
 *@version $Id: Presentation.hpp,v 1.16 2004/12/30 12:48:47 mkropfbe Exp $
 *
 *@short <short description>
 *
 */
class Presentation : public VThread {
public:
  enum PresentationState { PRES_INITIALIZED, PRES_READYTOPLAY, PRES_MUTED,PRES_PLAYING, PRES_PAUSED };


  Presentation( u32 x_inSDL, u32 y_inSDL, ReferenceCounter * gtRef,  GlobalTimer * gt);

  ~Presentation();


  /* initializes the presentation. Receives the ContainerInfo */
  virtual int initalizeURI(const char * URI);

  /* sets up the presentation. For each visualStream in the ESInfo - List it
     generates a VideoDisplayIO
  */
  virtual void  setupURI();


  /* returns the ContainerInfo of the current URI */
  const ContainerInfo * getContainerInfo(); 
	
  /* returns all used DataChannels in that presentation */
  const list <DataChannel *> getDataChannels();

  /* returns current URI (complete path) */
  const char * getURI() const;

  /* returns current presentation name (name of the file) */
  virtual const char * getPresentationName();

  /* returns one DataChannel identified by esId */
  DataChannel * getDataChannel (int esId);

  /* redraws all presentation Displays */
  void redrawPresentation(int presNr);

  /* resizes all presentation Displays to w and h */
  void resizePresentation(int w, int h);

  /* sets new presenation position: xpos and ypos define the upper left
     corner of the presentation
  */
  virtual void setPresentationPosition(u32 xpos, u32 ypos);

  /* moves the presentation to the new position which was set by 
     setPresentationPosition
  */
  virtual void movePresentation();

  /* keeps aspect ration - if existing - according to a given height value 
     returns the width value calulated
  */
  virtual int keepAspectRatio(int height);

  /* hides all presentation displays */
  void hidePresentation();

  /* shows all presentaiton displays */
  void showPresentation();

	
  /* play the whole presentation 
	
  */
  virtual bool play() = 0;

  /* pause the whole presentation 
     ToDo: pause only one DataChannel identified by sessionKey
			 
  */
  virtual bool pause(u32 sessionKey) = 0;

  /* check the GlobalTimer if it is paused */
  virtual bool isPaused() = 0;

  /* check the GlobalTimer if it is prefetching */
  virtual bool isPrefetching() = 0;

  /* teardown (stop) the whole presentation
     ToDo: teardown only one DataChannel identified by sessionKey
			 
  */
  virtual bool tearDown(u32 sessionKey) = 0;

  /* mute the whole presentation 
     visualStreams are not decoded and rendered but synchronizing still works
     ToDo: mute only one DataChannel identified by sessionKey
  */
  virtual bool mute(u32 sessionKey, bool muteAudio, bool muteVideo) = 0;

  /*  unmute the whole presentation
      ToDo: unmute only one DataChannel identified by sessionKey
  */
  virtual bool unmute(u32 sessionKey, bool muteAudio, bool muteVideo) = 0;


  virtual void run() = 0;

  void clearPresentationLists();

  /* returns total width of the presentation */
  u32 getPresentationWidth();

  /* returns total height of the presentation */
  u32 getPresentationHeight();

  /* returns x position of the presentation */
  u32 getPresentationXpos();

  /* returns y position of the presentation */
  u32 getPresentationYpos();

  /* returns the sessionKey - not a useable method because each 
     stream has its own sessionKey
  */
  u32 getSessionKey();

  /* set the whole presentation active or inactive
     active means, that the presentation can be switched between fullscreen 
     and windowed mode or that it can be deleted (only when all presenations are stopped)
  */
  void setActive(bool active);	

  /*returns true if presentation is active */
  bool isActive();

  /* returns true if presentation is local otherwise false (remote) */
  bool isLocal();

  /* returns all original VIDEO ESInfo Objects that belong to the presentation */
  list <ESInfo *> getOrigPresentationInfo();

  /* returns all actual VIDEO ESInfo Objects that belong to the presentation 
     remote: new Terminal Capabilities were set
     local: new Adaptors were set
  */
  list <ESInfo *> getActualPresentationInfo();

  /* returns total presentation duration in millisecs */
  int getPresentationDurationMS();

  /* returns the sum of average BW for all ES (interval should be in secs)*/
  virtual int getTotalAverageBandwidth(u32 interval);

  /* returns for a given interval (should be in secs) 
     a list of average FPS each belonging to one VIDEO ES */
  virtual list <float> getFramesPerSecond(u32 interval);
	/* returns the original framerate for each VIDEO ES in this presentation */
	virtual list <float> getOrigFramesPerSecond();
  /* returns for a given interval (should be in secs) 
     a list of average BW each belonging to one VIDEO ES */
  virtual list <int> getAverageBandwidth(u32 interval);

  /* returns a list of actual playing seconds each belonging to 
     one ES */
  virtual list <float> getActualSec();

  virtual float getActualSecOfLongestES();

  virtual list <int> getBufferFillLevel();

  virtual list <double> getBufferFillLevelInSecs();

  /* returns the actual presentation state */
  PresentationState getState() const { return presState; };

  /* returns true if presentation has ended */
  bool endOfPresentation();

  bool isLoadCritical();

  /** Set presentation start and end time, relative to the media start.
   * Time unit is seconds.
   * This causes the Presentation thread to play only the given time range
   * of the presentation.
   * @param starttime   0.0 means media start.
   * @param endtime     0.0 means media end.
   */
  void setTimeRange(double starttime = 0.0, double endtime = 0.0);

  PresentationInfo presInform;

  bool theOnlyOnePresentation;
	
protected:
  GlobalTimer * globalTimer;
	ReferenceCounter * gtRefCount;

  int lastX;
  int lastY;
  list <AdaptorChain *> aChain;
  bool localPres;
  bool needToMove;
  PresentationState presState;
  void addDecoderAdaptor();
  struct presentationManagement {
    DataChannel * dChannel;
    Statistics *stats;
    DataSink * clientObj;
    VideoDisplayIO * videoOutput;
    SDLaudioIO * audioOutput;
    ESInfo * esInfo;
    ESSynchronizer * esSync;
    AdaptorChain * adaptorChain;
    bool decoderSet;
		
    presentationManagement() {
      dChannel = NULL;
      stats = NULL;
      clientObj = NULL;
      videoOutput = NULL;
      audioOutput = NULL;
      esInfo = NULL;
      esSync = NULL;
      adaptorChain = NULL;
      decoderSet = false;
    };
    ~presentationManagement() {
      if(dChannel)
	dChannel = NULL;
      if(stats)
	stats = NULL;
      if(clientObj)
	clientObj = NULL;
      if(videoOutput)
	videoOutput = NULL;
      if(audioOutput)
	audioOutput = NULL;
      if(esInfo)
	esInfo = NULL;
      if(esSync)
	esSync = NULL;
      if(adaptorChain)
	adaptorChain = NULL;
    };
  };
  ContainerInfo * mp4Stream;

  list <presentationManagement *> presManage;
  presentationManagement *presM;
  u32 x_inSDL;
  u32 y_inSDL;
  u32 pWidth;
  u32 pHeight;
  const char * uri;
  u32 sessionKey;
  bool activePresentation;
  list <ESInfo *> origESInfo;
  list <ESInfo *> actESInfo;
  int duration;			// media duration of longest ES (in ms)
  /* Presentation start/end times relative to media start (in seconds).
   * npt means Normal Playing Time (as in RTSP).
   */
  double nptStart;		// 0.0 means media start
  double nptEnd;		// 0.0 means media end
};
#endif
