 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	RemotePresentation.hpp                                        *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef REMOTEPRESENTATION_HPP
#define REMOTEPRESENTATION_HPP

#include "Presentation.hpp"
#include <qurl.h>
#include <qsocketdevice.h>
#include "net/Protocol.hpp"
#include "cache/CacheManager.hpp"
#include "AdaptorChain.hpp"
#include "ReceiverThread.hpp"
#include "MultiVideoPlayerDialog.hpp"

struct rtx_group;
struct rtx_info;
class ContainerInfo;
class DataChannel;
class DataSink;
class RTSP;
class Rtp;
class ESInfo;
class Statistics;

#ifdef WIN32
#include <winsock2.h>
#else

#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <unistd.h>
#endif
#include "metadata/TerminalCapabilities.hpp"

class ESSynchronizer;

#define DEFAULT_RTSP_PORT 554


// some information on errors
// not used yet

#define RPRES_OK					0
#define RPRES_SOCKET_ERROR			-1
#define RPRES_CONNECTION_ERROR		-2
#define RPRES_SERVER_DNS_ERROR		-3
#define RPRES_PROXY_DNS_ERROR		-4

#define WAIT_MSECS_ON_RECEIVE	1



/**
 *
 * class handling a Presentation which is streamed over the network
 * inherited from class Presentation
 *
 *@author Ronald Sowa
 *
 *@version $Id: RemotePresentation.hpp,v 1.16 2005/04/14 15:53:39 klschoef Exp $
 *
 *@short <short description>
 *
 */
class RemotePresentation : public Presentation {

public:
	/* states of the RemotePresentation */
	enum RemotePresentationState { SESSION_ERR, SESSION_CLOSED, SESSION_CONNECTED,
																 SESSION_INITIALIZED, SESSION_PLAYING, SESSION_PAUSED };

	/* Constructor
	   sets dummy TerminalCapabilities
	*/
	RemotePresentation(u32 x_inSDL, u32 y_inSDL,  ReferenceCounter * gtRef, GlobalTimer *gt,
			const char* proxyName=NULL, class MultiVideoPlayerDialog *mvpd = NULL, int proxyPort=3128) ;


	/* Destructor */
	virtual ~ RemotePresentation();

	/* sets up network connection
	   make an RTSP describe by passing URI to the server
	   receives ContainerInfo object
	   returns 0 on success
	*/
	int initalizeURI(const char * URI );

	/* parses ContainerInfo object, sets up VideoDisplays for each VES
	   ToDo: implement audio part
	*/
	void setupURI();

	const char * getPresentationName();

	/** writes the string to the session control channel */
	bool sendResponse(const char *str, int bytes);

	/* connects to the server */
	bool connect();

	bool play();
	bool pause(u32 sessionKey);
	bool isPaused();
	bool isPrefetching();
	bool mute(u32 sessionKey,bool muteAudio, bool muteVideo);
	bool unmute(u32 sessionKey,bool muteAudio, bool muteVideo);
	bool tearDown(u32 sessionKey);
	void run();


	RemotePresentationState getState() const {
		return state;
	};

	/* enable/disable RTX */
	void setRtxEnabled(bool enabled);

	/* assigns new TerminalCapailities + rtx enabled/disabled
	 * further sets if the video should be desplayed upscaled and/or with correct aspect ratio
	 * @param enableSwitching turns on switching and YUV upscaling support
	 */
	bool setTermCaps(TerminalCapabilities * tc, bool rtxEnabled,
									 bool upscale = false, bool keepAspectRatio = true, bool enableSwitching = true);

	/* returns actual TerminalCapabilities */
	TerminalCapabilities * getTermCaps() {
		return modifiedTermCap;
	}

	/* returns true if rtx is enabled otherwise false */
	bool isRtxEnabled() {
		return rtxEnabled;
	};


	bool isUpscaleSet() {
		return upscale;
	};

	bool isKeepAspectSet() {
		return keepAspectRatio;
	};

	bool isSwitchingEnabled() {
		return enableSwitching;
	};

	/* see Presentation.hpp for further information */
	list <float> getFramesPerSecond(u32 interval);

	/* returns the original framerate for each elemStream in this presentation */
	list <float> getOrigFramesPerSecond();

	/* these values are returned from the rtp-channels */
	list <int> getAverageBandwidth(u32 interval);


	float getActualSecOfLongestES();


	list <int> getBufferFillLevel();

	list <double> getBufferFillLevelInSecs();

	/* this summed up value is returned from the rtp-channels */
	int getTotalAverageBandwidth(u32 interval);

	/* these values are returned from the rtp-channels */
	list <float> getActualSec();

	/** look if a message is pending; when yes, add it to the message-list */
	void receiveMessage();
	/** wait until the specified @name message is received.
	    @param buffer contains a pointer to the received message
	    @param bytesRead contains the number of bytes received
	*/
	bool lookForMessage(char *name, char *buffer, int &bytesRead);
	/** handle a received ANNOUNCE message */
	void handleAnnounce(char *buffer);
	/** handle a received GET_PARAMETER message */
	void handleGetParameter(char *buffer);
	/** handle a received OPTIONS message */
	void handleOptions(char *buffer);
	/** handle a received REDIRECT message */
	void handleRedirect(char *buffer);
	/** handle a received SET_PARAMETER message */
	void handleSetParameter(char *buffer);
	/** handle a received TEARDOWN message */
	void handleTeardown(char *buffer);


private:
	/** socket used for reading and sending messages */
	int socket;
	/** buffer used for reading messages */
	char readBuffer[MSG_BUFFER_SIZE];
	/** vector of received messages */
 	vector<char*> recmessages;
	/** time-value used for select on receive */
	struct timeval readtv;
	/** this thread is responsible for receiving messages */
	class ReceiverThread *receiver;
	/** callback pointer to the dialog object */
	class MultiVideoPlayerDialog *mvpd;

	/** wait until a message is received
	    @param sock specifies the socket to listen
	    @param mybuffer contains a pointer to the received message
	    @param bytesRead contains the number of bytes received
	*/
	void waitForMessage(int sock, int &bytesRead, char *mybuffer);



	char proxyName[MAX_STR_LEN];
	int proxyPort;
	/** creates a rtx_group for the given input params */
	bool  setRtxInfoForES(u32 esId, rtx_info * rtx);
	/** returns an rtx_group for the esId, or NULL */
	rtx_group * getRtxGroup(u32 esId);
	/** deep-copies uri, sets mp4Stream by asking the cacheManager*/
	void setUrl(const char *&uri);
	//void setPort(int port);

	sockaddr_in *getDestInfo() {
		return server;
	};

	/** searches for an ES its DataChannel. returns NULL on failure */
	DataChannel * find(int esId);
	/** returns the known protocol or  PROTO_UNKNOWN */
	KnownProtocols determineProtocol(char *request);
	list <AdaptorChain *> aChain;
	int lastX;
	int lastY;
	int longestESfromRtpId;

	bool upscale;
	bool keepAspectRatio;
	bool enableSwitching;


protected:


	void addDecoderAdaptor(IO * output, ESInfo * esInfo, Statistics *stats, int id,
			       ESSynchronizer *ess = NULL);

	/** the url of the source/destination file */
	char *url;

	/** the current state of the video */
	RemotePresentationState state;
	/** The protocol used to parse requests and generate replies.
			FIXME: we hard-code to RTSP, redesign needed for http */
	RTSP *protocol;
	sockaddr_in *server;
	/** buffers the termCap from the response, and then passes
	 * it to the client which is created during SETUP */
	TerminalCapabilities* termCap;
	TerminalCapabilities* modifiedTermCap;



	/** stores for each ES an optional retransmission wish */
	list < rtx_group* > rtxInfo;
	bool rtxEnabled;

	/* a list of RTP channels, each for one ES */
	list <Rtp *> rtp_channels;
	Rtp * rtpChannelOfLongestES;
};
#endif
