/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	Server.cpp                                                    *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

/***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifdef WIN32
#include <winsock2.h>
#else
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <unistd.h>
#include <arpa/inet.h>
#endif

#include <signal.h>
#include <string.h>
//#include <fcntl.h>

#include "Server.hpp"
#include "net/ServerSession.hpp"
#include "net/RTSP.hpp"
#include "cache/AdmissionControl.hpp"
#include "net/HttpServer.hpp"

static const int SELECT_TIMEOUT = 100000;
static const int MAX_CONCURRENT_CONNECTIONS = 10;
static int finish=0;

	//start webserver
static HttpServer *hs;

// destructing sig or quit
void sighandlersrv(int sig)
{
  switch(sig) {
  case SIGINT:
    dprintf_full("handling SIGINT\n");
    if (finish < 5)
      finish++;
    else {
      printf("Server repeatedly didn't respond to shutdown request. Terminating unfriendly...\n");
      if (hs) delete hs;
      hs = NULL;
      exit(1);
    }
      
    break;
  default:
    dprintf("got unhandled signal...\n");
  }
}

/** creates a server, rootDir is deep-copied */
Server::Server(const ServerConfig& conf, CacheManager * c,
			   AdmissionControl* adm,TerminalCapabilities* tc):srvConf(conf), defaultTC(tc)
{
	assert(adm);
	admControl=adm;
	enableRtx=srvConf.enableRtx;
	printf("Server is initializing\r\n");
	
	signal(SIGINT, sighandlersrv);
	cm = c;
	socket = 0;
	
	if (hs != NULL) delete hs;
	hs = new HttpServer(8080, "./metadata/");
};

Server::~Server()
{
	if (socket>0) {
#ifdef WIN32
		closesocket(socket);
#else
		close(socket);
#endif
		socket = 0;
	}
	// find finished Sessions and remove them from the list
	list < Session * >::iterator li;
	if (!sessionList.empty()) {
		li = sessionList.begin();
		while (!sessionList.empty()) {
	    (*li)->tearDown(TEARDOWN_ALL,true,NULL,NULL);
	    delete(*li);
	    li = sessionList.begin();
		}
	}
	if (cm) {
		delete cm;
		cm = NULL;
	}
	if(admControl)
		delete admControl;
	if(defaultTC)
		delete defaultTC;
		
	if (hs)
		delete hs;
};

bool Server::createSocket()
{
	this->socket =::socket(AF_INET, SOCK_STREAM, 0);
	if (socket < 0) {
		perror("create socket");
		return false;
	}

	const int reuse = 1;
	if (setsockopt(socket, SOL_SOCKET, SO_REUSEADDR,
								 (const char *) &reuse, sizeof(reuse)) < 0) {
		perror("setsockopt");
#ifdef WIN32
		closesocket(socket);
#else
		close(socket);
#endif
		socket = 0;
		return false;
	}

	if (srvConf.port != 0) {
		struct sockaddr_in name;
		name.sin_family = AF_INET;
		name.sin_port = htons(srvConf.port);

		struct hostent *ipAddr = NULL;
		if (0 != strcmp(srvConf.bindToIP, SERVER_BIND_ALLDEV)) {
			ipAddr=gethostbyname(srvConf.bindToIP);
			if (ipAddr == NULL)
				dprintf_err("Server: binding specific IP-addr, though no DNS-hostname resolution possible...\n");
		}

		if (0 == strcmp(srvConf.bindToIP, SERVER_BIND_ALLDEV))
			name.sin_addr.s_addr = htonl(INADDR_ANY);
		else {
			name.sin_addr.s_addr=inet_addr(ipAddr->h_addr);
			memcpy((char *)&name.sin_addr, ipAddr->h_addr, ipAddr->h_length);
		}

		if (bind(socket, (struct sockaddr *) &name, sizeof name) != 0) {
			perror("bind socket");
#ifdef WIN32
	    closesocket(socket);
#else
	    close(socket);
#endif
	    socket = 0;
	    //return false;
			::exit(1);
		}
	} else {
		dprintf_err("No RTSP port set\n");
#ifdef WIN32
		closesocket(socket);
#else
		close(socket);
#endif
		socket = 0;
		return false;
	};


	if (0 != listen(socket, MAX_CONCURRENT_CONNECTIONS)) {
		perror("listen");
#ifdef WIN32
		closesocket(socket);
#else
		close(socket);
#endif
		socket = 0;
		return false;
	};

	return true;
};

void Server::startup()
{
  fd_set rdfs;
  struct timeval tv;

	if (false == createSocket()) {
		cerr << "Could not create socket for port " << srvConf.port << endl;
		return;
	};
	int new_socket = -1;
	SessionCounter* counter = new SessionCounter("active_sessions.dat");
	printf("Server up and ready (host: %s port: %i)\r\n",srvConf.bindToIP, srvConf.port);
	counter->start();
	hs->start();
	while (!finish) {
		// listen for new connections
		int sin_size = sizeof(struct sockaddr_in);
		struct sockaddr_in caller;
		//bzero is deprecated
		// bzero(&caller,sin_size);
		memset(&caller,0,sin_size);

		FD_ZERO(&rdfs);
		FD_SET(socket,&rdfs);
		// test for mem leak detection
		 tv.tv_sec=0; // reset to zero
		//tv.tv_sec=0;
		tv.tv_usec=500000;
		if ( (select(FD_SETSIZE,&rdfs,NULL,NULL,&tv) != -1)
				 && (FD_ISSET(socket,&rdfs)) ) {

			if ( (new_socket = accept(socket, (struct sockaddr *) &caller,
#ifndef WIN32
																(unsigned int *)
#endif
																&sin_size)) <= 0) {
				perror("Server::run accept new client");
				::exit(1);
			}
			TerminalCapabilities* tc=NULL;
			if(defaultTC)
				delete defaultTC;
			defaultTC=Server::loadTermCaps();
			if(defaultTC)
				tc=defaultTC->clone();
			// reload rtxinfo from serverconfig!!! and ONLY rtx
			ServerConfig dummySrvCfg;
			if(Server::loadServerConfig(&dummySrvCfg)) {
				srvConf.enableRtx=dummySrvCfg.enableRtx;
				enableRtx=srvConf.enableRtx;
			}
			Session *se =	new ServerSession(new_socket, (struct sockaddr_in *) &caller, 
				cm,admControl,enableRtx,counter,tc);
			sessionList.push_back(se);
			se->start();
		}

		// find finished Sessions and remove them from the list
		list < Session * >::iterator li;
		if (!sessionList.empty()) {
	    li = sessionList.begin();
	    while (!sessionList.empty() && (li != sessionList.end())) {
				if ((*li)->finished()) {
					delete(*li);
					li = sessionList.erase(li);
	/*				static int finishedRequests=0;
					finishedRequests++;
					if(finishedRequests>1)
						finish=true; //FIXME: just for mem leak detection
*/
				} else
					++li;
	    }
		}
	} // !finish

	// remove all sessions
	dprintf("Shutting down server. Closing open sessions. Please wait...\r\n");
	
	if (hs) delete hs;
	hs = NULL;
	
	// find finished Sessions and remove them from the list
	list < Session * >::iterator li;
	if (!sessionList.empty()) {
		li = sessionList.begin();
		while (!sessionList.empty() && (li != sessionList.end())) {
	    (*li)->tearDown(TEARDOWN_ALL,true,NULL,NULL);
	    if (!(*li)->finished()) {
				(*li)->exit();
	    }
	    delete(*li);
	    li = sessionList.erase(li);
		}
	}
	counter->stop();
	counter->wait();
	delete(counter);
#ifdef WIN32
	closesocket(socket);
#else
	close(socket);
#endif
	socket = 0;

}

ContainerInfo *Server::find(const Url *fileName)
{
	if (cm) {
		return cm->getVideo(fileName);
	}
	return NULL;
}


/** returns false, if no serverConfigFile or homedir was found */
bool Server::loadServerConfig(struct ServerConfig* s) {
	RTSP t;
	bool retVal=false;
	FILE* fp;
	int tmpPort=0;

	s->recurseDirectories=false;
	strcpy(s->homeDir,".");
	strcpy(s->movielistXMLFile, "./metadata/MovieList.xml");
        strcpy(s->movielistHTMLFile, "./metadata/index.html");
	strcpy(s->metadataDir, "./metadata");	
	strcpy(s->switchDir,"./switch");
	s->port=t.getDefaultPort();
	strcpy(s->bindToIP,SERVER_BIND_ALLDEV);
	s->enableRtx=true;
	if( !(fp=fopen(s->configfile,"rb")) ) {
		return retVal;
	}

	char temp[1024];
	temp[1023]=0;
	char* tmp=NULL;
	const char* tmp2=NULL;
	while(fgets(temp,1023,fp)) {
		// fgets read one line
		tmp=strstr(temp,HOMEDIR);
		tmp2=strstr(temp,"#");
		// if we have a # before the tmp ignore the line
		// if we have it after the tmp its okay
		// if tmp2 is NULL its also okay
		if(tmp && (tmp2==NULL || tmp2>tmp)) {
			tmp+=strlen(HOMEDIR);
			tmp=stripWhiteSpace(tmp);
			if(strlen(tmp))
				strcpy(s->homeDir,tmp);
			if(s->homeDir && 
				(s->homeDir[strlen(s->homeDir)-1]=='/' ||
				s->homeDir[strlen(s->homeDir)-1]=='\\'))
				s->homeDir[strlen(s->homeDir)-1]='\0';
			retVal=true;
		}
		tmp=strstr(temp,MOVIELISTXMLFILE);
		tmp2=strstr(temp,"#");
		if(tmp && (tmp2==NULL || tmp2>tmp)) {
			tmp+=strlen(MOVIELISTXMLFILE);
			tmp=stripWhiteSpace(tmp);
			if(strlen(tmp))
				strcpy(s->movielistXMLFile,tmp);
			if(s->movielistXMLFile && 
				(s->movielistXMLFile[strlen(s->movielistXMLFile)-1]=='/' ||
				s->movielistXMLFile[strlen(s->movielistXMLFile)-1]=='\\'))
				s->movielistXMLFile[strlen(s->movielistXMLFile)-1]='\0';
			retVal=true;
		}
		tmp=strstr(temp,MOVIELISTHTMLFILE);
		tmp2=strstr(temp,"#");
		if(tmp && (tmp2==NULL || tmp2>tmp)) {
			tmp+=strlen(MOVIELISTHTMLFILE);
			tmp=stripWhiteSpace(tmp);
			if(strlen(tmp))
				strcpy(s->movielistHTMLFile,tmp);
			if(s->movielistHTMLFile && 
				(s->movielistHTMLFile[strlen(s->movielistHTMLFile)-1]=='/' ||
				s->movielistHTMLFile[strlen(s->movielistHTMLFile)-1]=='\\'))
				s->movielistHTMLFile[strlen(s->movielistHTMLFile)-1]='\0';
			retVal=true;
		}		
		tmp=strstr(temp,METADATADIR);
		tmp2=strstr(temp,"#");
		if(tmp && (tmp2==NULL || tmp2>tmp)) {
			tmp+=strlen(METADATADIR);
			tmp=stripWhiteSpace(tmp);
			if(strlen(tmp))
				strcpy(s->metadataDir,tmp);
			if(s->metadataDir && 
				(s->metadataDir[strlen(s->metadataDir)-1]=='/' ||
				s->metadataDir[strlen(s->metadataDir)-1]=='\\'))
				s->metadataDir[strlen(s->metadataDir)-1]='\0';
			retVal=true;
		}
		tmp=strstr(temp,SWITCHDIR);
		tmp2=strstr(temp,"#");
		if(tmp && (tmp2==NULL || tmp2>tmp)) {
			tmp+=strlen(SWITCHDIR);
			tmp=stripWhiteSpace(tmp);
			if(strlen(tmp))
				strcpy(s->switchDir,tmp);
			if(s->switchDir && 
				(s->switchDir[strlen(s->switchDir)-1]=='/' ||
				s->switchDir[strlen(s->switchDir)-1]=='\\'))
				s->switchDir[strlen(s->switchDir)-1]='\0';
			retVal=true;
		}
		tmp=strstr(temp, RECURSE);
		tmp2=strstr(temp,"#");
		if(tmp && (tmp2==NULL || tmp2>tmp)) {
			tmp+=strlen(RECURSE);
			if(strstr(tmp,YES)) {
				s->recurseDirectories=true;
			}			
		}
		tmp=strstr(temp, PORT);
		tmp2=strstr(temp,"#");
		if(tmp && (tmp2==NULL || tmp2>tmp)) {
			tmp+=strlen(PORT);
			tmpPort=atoi(tmp);
			if(tmpPort>0)
				s->port=tmpPort;
		}
		tmp=strstr(temp,BINDTO);
		tmp2=strstr(temp,"#");
		if(tmp && (tmp2==NULL || tmp2>tmp)) {
			tmp+=strlen(BINDTO);
			tmp=stripWhiteSpace(tmp);
			if(strlen(tmp))
				strcpy(s->bindToIP,tmp);
		}
		tmp=strstr(temp,CONFIG_RTX);
		tmp2=strstr(temp,"#");
		if(tmp && (tmp2==NULL || tmp2>tmp)) {
			tmp+=strlen(CONFIG_RTX);
			if(strstr(tmp,NO)) {
				s->enableRtx=false;
			}
		}
	}
	fclose(fp);
	return retVal;
}

char* Server::stripWhiteSpace(char* url) {
	char* temp=url;
	while(	*temp==' '  || *temp=='\r' ||
			*temp=='\n' || *temp=='\t')
			temp++;
	int pos=strlen(url)-1;
	while(	(url+pos)>temp && pos>=0 &&
			(url[pos]==' ' || url[pos]=='\r' ||
			url[pos]=='\n' || url[pos]=='\t') ){
			url[pos]=0;
			pos--;
	}
	return temp;
};

TerminalCapabilities* Server::loadTermCaps()
{
	char tcFile[MAX_STR_LEN];
#ifdef WIN32
	strcpy(tcFile,getenv("USERPROFILE"));
	strcat(tcFile,"\\");
#else
	strcpy(tcFile,getenv("HOME"));
	strcat(tcFile,"/");
#endif
	strcat(tcFile,SERVERTCFILE);
	TerminalCapabilities* tc=new TerminalCapabilities();
	
	if(tc->initFromFile(tcFile))
		return tc;
	delete tc;
	return NULL;
};

bool Server::saveServerConfig(const struct ServerConfig* s)
{
	bool ret=false;
	FILE* fp=fopen(s->configfile,"wb");
	if(!fp) {
		fprintf(stderr,"MP4Server: Failed to open %s\n",s->configfile);
	}
	else { //write defaults to config file
		char hostname[MAX_STR_LEN];
		RTSP rt;
		gethostname(hostname,MAX_STR_LEN);
		fprintf(fp,"#path where all video files are stored (eg. /var/spool/muViServer/data)\n");
		fprintf(fp,"%s%s\n",HOMEDIR,s->homeDir);
		fprintf(fp,"#file-path were generated movielist (XML format) will be stored\n");
		fprintf(fp,"#this path sould also contain the stylesheet 'MovieList.xsl' !\n");
		fprintf(fp,"%s%s\n",MOVIELISTXMLFILE,s->movielistXMLFile);	
		fprintf(fp,"#file-path were generated movielist (HTML format) will be stored\n");
		fprintf(fp,"%s%s\n",MOVIELISTHTMLFILE,s->movielistHTMLFile);	                	
		fprintf(fp,"#path which contains additional meta-data like images or descriptions\n");
		fprintf(fp,"%s%s\n",METADATADIR,s->metadataDir);
		fprintf(fp,"#search HOMEDIR recursively for videos (%s/%s)?\n",YES,NO);
		fprintf(fp,"%s%s\n",RECURSE,(s->recurseDirectories?YES:NO));
		fprintf(fp,"#path where switchsets for video files are stored (eg. /var/spool/muViServer/switch)\n");
		fprintf(fp,"#  add files like switch/coastguard.mp4/CIFversion.mp4 \n");
		fprintf(fp,"#                 switch/coastguard.mp4/QCIFversion.mp4 \n");
		fprintf(fp,"%s%s\n",SWITCHDIR,s->switchDir);
		fprintf(fp,"#RTSP/TCP port (default: %i)\n",rt.getDefaultPort());
		fprintf(fp,"%s%i\n",PORT,s->port);
		fprintf(fp,"#IP-adress of interface, on which RTSP should listen on...\n");
		fprintf(fp,"#if RTSP should listen on all interfaces, set to \"%s\"\n",SERVER_BIND_ALLDEV);
		fprintf(fp,"%s%s\n",BINDTO,s->bindToIP); //hostname); //allow all devices
		fprintf(fp,"#should the server support retransmission of lost packets (%s/%s)\n",YES,NO);
		fprintf(fp,"%s %s\n",CONFIG_RTX,(s->enableRtx?YES:NO));
		fclose(fp);
		ret=true;
	}
	return ret;	
}
