 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title: SemiProxySession.hpp                                          *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 * Video Session Migration System, 2004                                 *
 * Author: Klaus Schoeffmann                                            *
 ************************************************************************/


#ifndef __SemiProxySession_hpp__
#define __SemiProxySession_hpp__

#ifdef WIN32
#include <winsock2.h>
#else
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <unistd.h>
#include <arpa/inet.h>
#endif

#include <string>
//#include <unistd.h>
#include <sys/types.h>
#include <sstream>

#include "SemiGlobals.hpp"
#include "net/Session.hpp"
#include "net/RTSP.hpp"
#include "global.hpp"

#include "metadata/MP21.hpp"
#include "metadata/TerminalCapabilities.hpp"

#include "SemiProxy.hpp"
#include "SSession.hpp"
#include "URIParameters.hpp"
#include "TimeMeasurement.hpp"
#include "Profile.hpp"
#include "TrackList.hpp"

class SemiProxy;

enum PlayerInformation {PI_UNKOWN, PI_MUVIPLAYER, PI_QUICKTIME, PI_REALPLAYER, PI_PVPLAYER};

/**
 * This class is used to handle a single rtsp-session. It contains
 * the most functionality of the semiproxy application.
 *
 * @author Klaus Schoeffmann
 */

class SemiProxySession : public Session {

  //exceptions...
  class GetHostNameProblem {};
  class ParamEndProblem {};

public:
  /** default constructor */
  SemiProxySession(int fdclient, struct sockaddr_in* clientsock, SemiProxy *sp);

  /** default destructor */
  ~SemiProxySession();

  /** Return the reference to a particular session from the static vector of this class. */
  static SemiProxySession* getSemiProxySession(uint RtspSessionId);

  /** Reads messages from client and server until the end of this session.
      The messages will be parsed and forwarded to the opponent. */
  void run();

  /** This function parses the response from the server.
      Depending on this response, this function either
        - creates a new SSession (on successful response to a play-message)
        - sets the SSession to state 'active' again (when the response
          was the ok for a subsequent play-message) */
  void parseResponseFromServer(char *buffer, int len);

  /** This function forwards a message to the server.
      It also looks for parameters specified in the RTSP-URI and
      removes it. If message-type is DESCRIBE and a profile-ID was specified,
      also the TerminalCapabilites (of the specified profile) will be added.
      Furthermore, this function stores the received CSeq value from the client. */
  void forwardToServer(char *buffer, int len);

  /** Forwards a message to the client. Also replaces the CSeq value, set
      by the server to the stored CSeq value received from the client! */
  void forwardToClient(char *buffer, int len);

  /** Handles the OPTIONS-response from the server */
  void handleOptionsResponse(string &buf, int &bytesRead);

  /** Handles the DESCRIBE-response from the server */
  void handleDescribeResponse(string &buf, int &bytesRead);

  /** Handles the SETUP-response from the server */
  void handleSetupResponse(string &buf, int &bytesRead);

  /** Handles the PLAY-response from the server */
  void handlePlayResponse(string &buf, int &bytesRead);

  /** Handles the PAUSE-response from the server */
  void handlePauseResponse(string &buf, int &bytesRead);

  /** Handles the TEARDOWN-response from the server */
  void handleTeardownResponse(string &buf, int &bytesRead);

  /** Handles a GET_PARAMETER message, sent from the client. */
  bool getOptions(const Url* fileName, const char* remaining);

  /** Handles a SET_PARAMETER message, sent from the client. */
  bool setOptions(const Url* fileName, const char* remaining);

  /** Parses a OPTIONS message and forwards it to the server. */
  bool options(const Url* fileName, const char* remaining);

  /** Parses a DESCRIBE message and forwards it to the server. */
  bool connect(const Url* fileName, const char* remaining);

  /** Parses a SETUP message and forwards it to the server.
      Also inserts a destination parameter, which is set to the address of the
      client that sent this message, into the the message. */
  bool setup(const Url* fileName, const char* remaining);

  /** Parses a PLAY message and forwards it to the server.
      If the URI of this message contains an existing RTSP-Session-ID (which
      means, this session should take over the other session), the Range-
      Parameter of this message will be replaced by the current time
      of the other session.
      This function also extracts the RTSP-SessionId of the message. */
  bool play(const Url* fileName, const char* remaining);

  /** Handles a PAUSE message, sent from the client. */
  bool pause(const Url* fileName, const char* remaining);

  /** Handles a TEARDOWN message, sent from the client. */
  bool tearDown(int sessionKey, bool immediate=false, const Url* fileName=NULL, const char* remaining=NULL);

  /** Look if specified SessionID is contained in any track of this session */
  bool containsSessionID(uint SessionID);

  /** @return the RTSP SessionID of this session. */
  //uint getRtspSessionID();

  /** Retrieve current 'client playout time' of this session from server.  */
  uint getServerTimeMs();

  /** Retrieve current 'client playout time' of this session from client.  */
  uint getClientTimeMs();

  void testClientOptions();

  /** Initiate request of current 'client playout time' of this session
      on next main-loop run from the serrver. */
  uint requestServerTimeMs();

  /** Initiate request of current 'client playout time' of this session
      on next main-loop run from the client. */
  uint requestClientTimeMs();

  /** Send a Rtsp-TEARDOWN to the client */
  void sendTeardownToClient();

  /** Close the current session (by closing the socket to client and server)
      Furthermore, the SSession object is marked as 'closed'. */
  void closeSession();

  /** Close the current session (by closing the socket to client and server)
      Furthermore, the SSession object is removed from the sessionlist. */
  void removeSession();

  /** Empty method which is created for conformance to Session.cpp
      @param uri
      @param makeExactMatch  */
  void setUrl(const Url *uri, bool makeExactMatch = true);

  /** Set the session state */
  void setStateClosed();

  PlayerInformation getPlayerInfo() { return playerInfo; };

private:

  PlayerInformation playerInfo;

  /** initialize object */
  void initSemiProxySession();

  /** reset flags (should be called on further DESCRIBE messages) */
  void resetFlags();

  /** contains all currently active SemiProxySessions*/
  static vector<SemiProxySession *> sessionList;

  /** file-descriptors of client and server */
  int fdserver, fdclient;

  /** callback-reference to semiproxy object */
  class SemiProxy *sp;

  /** socket to the client */
  struct sockaddr_in* clientsock;

  /** buffer used for reading messages */
  char *buffer;

  /** number of bytes read on last message */
  int bytesRead;

  /** last CSeq-value received from client */
  uint lastClientCSeq;

  /** last CSeq-value received from server */
  uint lastServerCSeq;

  /** elapsed msecs of old session (on migration) */
  ulong prevDuration;

  /** true if the message from the client contains no termcaps */
  bool noTermCapsFound;

  // --- variables used for active sessions ---
  /** last message-type sent from client */
  uint lastClientRtspMsg;

  /** parameters of URI, sent from the client */
  URIParameters params;

  /** RTSP-Session-ID (of first track)*/
  uint RtspSessionID;

  /** RTSP-Session-ID of last message received from client */
  uint lastCRtspSessionID;

  /** Track list of this session */
  TrackList tracks;


  /** complete rtsp-uri */
  char *SessionURI;
  /** current active streamin session (with <SessionID> as session-Id) */
  SSession *asess;

  /** object used for time measurement */
  TimeMeasurement *timemsmnt;

  /** hostname of last rtsp-request */
  char *requestedHostname;

  /** received playout position from server */
  uint receivedServerPosition;

  /** received playout position from player */
  uint receivedClientPosition;

  /** flag: should a server position request be send on next loop run? */
  bool sendServerPositionRequest;

  /** flag: should a client position request be send on next loop run? */
  bool sendClientPositionRequest;

  /** flag: true if this session has been paused for at least one time */
  bool sessionHasBeenPaused;


  void createRtspServerSocket(const char *host, int port);

  /** read from specified socket until no more data */
  void readFromSocket(int fdsock, struct timeval &tv, fd_set &rfds);

  /** This function checks if the proxy is already connected to the server.
      If not already connected, it opens a new connection to the server. */
  bool checkServerConnection(const Url* fileName);

  /** Extracts the value of a parameter (like 'id=..' or 'host=..')
      from the RTSP-URI and returns it as a string. */
  char* extractStrParameter(string &strbuf, char *pname, bool removeIt = false);

  /** Extracts the value of a parameter (like 'uid=..' or 'proid=..')
      from the RTSP-URI and returns it as an integer. */
  int extractParameter(string &strbuf, char *pname, bool removeIt = false);

  /** Extract the CSeq value from the message
      and store it in the variable @lastCSeq. */
  void extractCSeq(char *buffer, uint &lastCSeq);

  /** Extract the duration of SDP content (by using Range-To) */
  ulong extractDuration(string &strbuf);

  /** Extract the from-value of range in SDP content */
  ulong extractRangeFrom(string &strbuf);
  /** Extract the prebuffer-time, announced by the client */
  ulong extractPrebuffer(string &strbuf);

  /** Return the Content-Length value of the message. */
  uint getContentLength(string strbuf);

  /** Replace the CSeq value of the message in @str to the value of @newCSeq */
  void replaceCSeq(string &str, uint newCSeq);

  /** Replace the range parameter in the content of the message by
      the value of @newVal. */
  void replaceRange(string &strbuf, char *newVal, bool replaceOnlyFromValue = true);

  //void replace(string &str, char* val, char* newval, bool all = false);
  /** Change the Content-Length value of the message. */
  void changeContentLength(string &strbuf, uint newVal);

  /** Checks if the host and port specified in the URI is not the same as
      the host and port, where the semiproxy is running on.
      If there is a specified parameter "host" or "port", this function replaces
      the original host/port by the values of this parameter! */
  bool checkForValidAddress(char *buffer, int &bytesRead);

  /** @return the position of substring (needle1/needle2) in 'str' which
      is nearer to position 'from'. */
  uint getSmallerStringPos(string str, char *needle1, char *needle2, uint from = 0);

  /** @return a string containing the terminal capabilities in mpeg21-format */
  string createTerminalCapabilities(Profile *profile);

};

#endif
