/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:   Job.cpp                                                     *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/


#include "Job.hpp"


/** constructor */
Job::Job(const char *name, Source* es, const char* outFile, u32 targetBitRate, float startPosition, float endPosition)
{
	dprintf_full("\033[32mJob::construct\033[0m\n");
	u32 id=createJobID();
	this->jobID = id;
	assert(name);
	assert(outFile);
	strncpy(this->name, name, MAX_STR_LEN-1);
	this->name[MAX_STR_LEN-1] = '\0';
	//only pointer to source object, no copy
	this->in_es = es;
	strncpy(this->outFile, outFile, MAX_STR_LEN-1);
	this->outFile[MAX_STR_LEN-1] = '\0';
	this->targetBitRate = targetBitRate; //in bits/sec.
	this->startPosition = startPosition;
	this->endPosition = endPosition;
	this->reference_counter=1;
}

Job::Job()
{
	dprintf_full("\033[32mJob::empty construct\033[0m\n");
	name[0] = '\0';
	in_es = NULL;
	outFile[0] = '\0';
	u32 id = createJobID();
	this->jobID = id;
	this->reference_counter=1;
}
	
/** destructor */
Job::~Job()
{
	dprintf_full("\033[32mJob::destruct\033[0m\n");
	if(this->reference_counter != 1)
		assert(1==0);
	this->in_es = NULL;
}

bool Job::isVideo()
{
	return this->is_video;
}

const char* Job::getName() const
{
	dprintf_full("\033[32mJob::getName\033[0m\n");
	this->checkReferences();
	return this->name;
}

const Source* Job::getSource() const
{
	dprintf_full("\033[32mJob::getSource\033[0m\n");
	this->checkReferences();
	assert(this->in_es);
	return this->in_es;
}

const char* Job::getOutFileName() const
{
	dprintf_full("\033[32mJob::getOutFileName\033[0m\n");
	this->checkReferences();
	return this->outFile;
}

u32 Job::getBitRate() const
{
	dprintf_full("\033[32mJob::getBitRate\033[0m\n");
	this->checkReferences();
	return this->targetBitRate;
}

u32 Job::getID() const
{
	dprintf_full("\033[32mJob::getID\033[0m\n");
	this->checkReferences();
	return this->jobID;
}

float Job::getStartPosition() const
{
	dprintf_full("\033[32mJob::getStartPosition\033[0m\n");
	this->checkReferences();
	return this->startPosition;
}

float Job::getEndPosition() const
{
	dprintf_full("\033[32mJob::getEndPosition\033[0m\n");
	this->checkReferences();
	return this->endPosition;
}

u32 Job::createJobID()
{
	dprintf_full("\033[32mJob::createJobID\033[0m\n");
	static u32 i=0;
	return ++i;
}

void Job::increment()
{
	dprintf_full("\033[32mJob::increment\033[0m\n");	
	this->checkReferences();
	this->reference_counter++;
}

void Job::decrement()
{
	dprintf_full("\033[32mJob::decrement\033[0m\n");
	this->checkReferences();
	this->reference_counter--;
}

void Job::checkReferences() const
{
	dprintf_full("\033[32mJob::checkReferences\033[0m\n");
	if(this->reference_counter <= 0)
	{
		dprintf_err("\033[31mJob::checkReferences : error\033[0m\n");
		assert(1==0);
	}
}

void Job::clean()
{
	dprintf_full("\033[32mJob::clean\033[0m\n");
	this->decrement();
	if(this->reference_counter == 1)
	{
		delete(this);
		dprintf_full("\033[32mJob::clean : deleted Job\033[0m\n");
	}
	else
	dprintf_err("\033[31mJob::clean : job is still referenced\033[0m\n");
}

void Job::setName(const char* name)
{
	dprintf_full("\033[32mJob::setName\033[0m\n");
	assert(name);
	this->checkReferences();
	strncpy(this->name, name, MAX_STR_LEN-1);
}

void Job::setSource(Source* es)
{
	dprintf_full("\033[32mJob::setSource\033[0m\n");
	assert(es);
	this->checkReferences();
	this->in_es = es;
}

void Job::setOutFileName(const char* outFile)
{
	dprintf_full("\033[32mJob::setOutFileName\033[0m\n");
	assert(outFile);
	this->checkReferences();
	strncpy(this->outFile, outFile, MAX_STR_LEN-1);
}

void Job::setBitRate(u32 rate)
{
	dprintf_full("\033[32mJob::setBitRate\033[0m\n");
	this->checkReferences();
	this->targetBitRate = rate;
}

void Job::setID(u32 id)
{
	dprintf_full("\033[32mJob::setID\033[0m\n");
	this->checkReferences();
	this->jobID = id;
}

void Job::setStartPosition(float start)
{
	dprintf_full("\033[32mJob::setStartPosition\033[0m\n");
	this->checkReferences();
	this->startPosition = start;
}

void Job::setEndPosition(float end)
{
        dprintf_full("\033[32mJob::setEndPosition\033[0m\n");
	this->checkReferences();
	this->endPosition = end;
}
		

