 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	mp4toyuv.cpp                                                  *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#include <assert.h>
#include "DataChannel.hpp"
#include "adaptors/MP4Decoder.hpp"
#include "DataSink.hpp"
#include "io/MPGStreamIO.hpp"
#include "io/RawStreamIO.hpp"
#include "io/YUVStreamIO.hpp"
#include "io/DevNull.hpp"
#include "io/ffMP4IO.hpp"
#include "ContainerInfo.hpp"
#include "VideoESInfo.hpp"
#include "VideoAdaptorChain.hpp"
#include "metadata/TerminalCapabilities.hpp"
#include "Statistics.hpp"
#include "adaptors/YUVSpatialReductionAdaptor.hpp"
#include "adaptors/YUVScalingAdaptor.hpp"
#include "adaptors/FrameRotate.hpp"
#include "adaptors/Visualizer.hpp"
#include "adaptors/DataDump.hpp"
#include "adaptors/FrameRateAdjust.hpp"
#include "adaptors/MadMP3Decoder.hpp"
#include "AudioAdaptorChain.hpp"
#include "AudioESInfo.hpp"

#ifdef WIN32
#include <crtdbg.h>
#endif
#include "vitooki.hpp"
#include "container/ContainerFile.hpp"


void printHelp(const char* name) {
  printf("%s [--rgb] [--rotate] [--mirror] [--decode-mp3] [--out-fps NN] <.mp4> <out-prefix> [width] [height]\r\n",name);
  printf("Option --rgb: decode to RGB32 instead of YUV.\r\n");
  printf("Option --rotate: rotates the result by 90 degrees.\r\n");
  printf("Option --mirror: mirrors the result by 180 degrees.\r\n");
  printf("        combining --rotate and --mirror results in a 270 degree rotation\r\n");
  printf("Option --decode-mp3: if not set, audio is bypassed and stored directly, else raw PCM is stored\r\n");
  printf("Option --out-fps NN: sets NN as the new output framerate, where incoming frames are dropped/doubled\r\n");
  printf("if width and height is given, the yuv is scaled (eg. from SD to 352x288 or vice versa)!\r\n");
  printf("if more than one MPEG-4 video stream is in the .mp4 container, all will be extracted\r\n");
  printf("             and numbered according to their stream-IDs\r\n");
  printf("note, that the output filenames will look like:  out-prefix-01.yuv .. out-prefix-05.yuv\r\n");
}



int main(int argc, char* argv[]) {
  char *progname = argv[0];
  int opt_rgb = 0;		// '--rgb' option set?
	int opt_rotate = 0;
	int opt_mirror = 0;
	int opt_decode_mp3 =0;
  char *input_file = NULL;
  char *out_prefix = NULL;
  int do_scale = 0;
  int width = 0;
  int height = 0;
	int out_fps = 0;
  const char yuv_suffix[] = ".yuv";
  const char rgb_suffix[] = ".rgb";
  const char mp3_suffix[] = ".mp3";
  const char raw_audio_suffix[] = ".pcm";
  const char *vid_suffix = yuv_suffix;
  const char *aud_suffix = mp3_suffix;
 
#ifdef WIN32
  int tmpFlag = _CrtSetDbgFlag( _CRTDBG_REPORT_FLAG );
  tmpFlag |= _CRTDBG_LEAK_CHECK_DF;
  _CrtSetDbgFlag( tmpFlag );
#endif

  while (--argc > 0 && (*++argv)[0] == '-') {
    if (strcmp(*argv, "--rgb") == 0) {
      opt_rgb = 1;
      vid_suffix = rgb_suffix;
    } else 
    if (strcmp(*argv, "--rotate") == 0) {
      opt_rotate = 1;
    } else 
    if (strcmp(*argv, "--mirror") == 0) {
      opt_mirror = 1;
    } else {
			if (strcmp(*argv, "--decode-mp3") == 0) {
				opt_decode_mp3 = 1;
				aud_suffix = raw_audio_suffix;
			} else {
				if (strcmp(*argv, "--out-fps") == 0) {
					out_fps = atoi(argv[1]); //get fps value
					--argc;
					++argv;					
				} else {
					printf("Unknown option: %s\n", *argv);
					printHelp(progname);
					exit(1);
				}
			}
		}
	}

  if (argc != 2 && argc != 4) {
    printHelp(progname);
    exit(1);
  }
  input_file = argv[0];
  out_prefix = argv[1];
  if (argc == 4) {
    do_scale = 1;
    width = atoi(argv[2]);
    height = atoi(argv[3]);
  }

  vitooki_init_codecs();

  ContainerInfo* mp4=ContainerFile::loadContainerInfo(input_file);
  if(!mp4) {
    printf("Inputfile %s not found\r\n", input_file);
    exit(1);
  }
	
  list < ESInfo * >* streams = mp4->getESList();

  list < ESInfo * >::const_iterator li = streams->begin();
  while (li != streams->end()) {
    if ((*li)->isVisualStream()) {
      char outfile[255];

      VideoESInfo* es=(VideoESInfo*)(*li);
      sprintf(outfile,"%s-%09llu%s", out_prefix, es->getStreamId(), vid_suffix);

      //	Statistics *stats = new Statistics(es,false); //input stats
      IO* input=NULL;
#ifdef ENABLE_FFMPEG
			input=new ffMP4IO(es, input_file, false);
#else
			dprintf_err("mp4toyuv  compiled without ffmpeg!\n");
			::exit(1);
#endif
      assert(es && input);
      VideoAdaptorChain *achain = new VideoAdaptorChain();

			UncompressedVideoFrame::ColorSpace cs;
			if (opt_rgb)
				cs = UncompressedVideoFrame::ColorSpaceRGB32;
			else
				cs = UncompressedVideoFrame::ColorSpaceYV12;

      MP4Decoder *mp4d = new MP4Decoder(es,false,cs);
      achain->addAdaptor(mp4d);
			
			if (out_fps > 0) {
				achain->addAdaptor(new FrameRateAdjust(es, es->getMediaTimeScale() / out_fps));
				//es->setVOPTimeIncrement(es->getMediaTimeScale() / out_fps);
			}

      if (do_scale) {
				YUVScalingAdaptor *ysr = new YUVScalingAdaptor(es, width, height);
				achain->addAdaptor(ysr);
      }

			if (opt_rotate && opt_mirror) { //270 degrees
				achain->addAdaptor(new FrameRotate(es, 270));
			} else if (opt_rotate) { //90 degree
				achain->addAdaptor(new FrameRotate(es, 90));
			}	else if (opt_mirror) { // 180 degrees
				achain->addAdaptor(new FrameRotate(es, 180));
			}

			achain->addAdaptor(new Visualizer((VideoESInfo*)es, Visualizer::SDL,cs ));
			
      DataChannel* dc=new DataChannel(input,es, achain);
			
      YUVStreamIO* output = new YUVStreamIO(es, outfile, true);
      dc->insert(new DataSink(0, output, "127.0.0.1", 0));
      dc->play(0, 0.0);
      dc->start();
      dc->wait();
      delete dc;
			
    } else
			if ((*li)->isAudioStream()) {
				char outfile[255];

				AudioESInfo* es=(AudioESInfo*)(*li);
				sprintf(outfile,"%s-%09llu%s", out_prefix, es->getStreamId(), aud_suffix);
				
				//	Statistics *stats = new Statistics(es,false); //input stats
      IO* input=NULL;
#ifdef ENABLE_FFMPEG
			input=new ffMP4IO(es, input_file, false);
#else
			dprintf_err("mp4toyuv  compiled without ffmpeg!\n");
			::exit(1);
#endif
				assert(es && input);
				AudioAdaptorChain *achain = new AudioAdaptorChain();

				if (opt_decode_mp3)
					achain->addAdaptor(new MadMP3Decoder(es));

				DataChannel* dc=new DataChannel(input,es, achain);
				
				//note: fileSize per frame is determined by AU->size
				RawStreamIO* output = new RawStreamIO(es, outfile, true);
				dc->insert(new DataSink(0, output, "127.0.0.1", 0));
				dc->play(0, 0.0);
				dc->start();
				dc->wait();
				delete dc;
				
			} else {
				dprintf_err("IGNORING Unknown Elementary Stream!!!\n");
			}
    ++li;
  }
	
  delete mp4;
  return 0;
}


