 /***********************************************************************
 *                                                                      *
 * MP7QtPlayer - a simple video player using the ViTooKi library        *
 *               showing MPEG-7 Data                                    *
 *                                                                      *
 * title: MP7QtPlayer.cpp                                               *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

#include <stdlib.h>
#include <cmath>
#include <qlabel.h>
#include <qtimer.h>
#include <qcheckbox.h>
#include <qslider.h>
#include <qlineedit.h>
#include <qtextedit.h>
#include <qslider.h>
#include <qlistbox.h>
#include <qtabwidget.h>

#include <vitooki.hpp>
#include <DataChannel.hpp>
#include <DataSink.hpp>
#include <ContainerInfo.hpp>
#include <ESInfo.hpp>
#include <AudioAdaptorChain.hpp>
#include <VideoAdaptorChain.hpp>
#include <adaptors/MP4audioDecoder.hpp>
#include <adaptors/MP4Decoder.hpp>
#include <adaptors/ESSynchronizer.hpp>
#include <adaptors/GlobalTimer.hpp>
#include <adaptors/SharedAdaptor.hpp>
#include <container/ContainerDemux.hpp>
#include <io/SDLaudioIO.hpp>
#include <io/IOCreator.hpp>
#include <io/MPGStreamIO.hpp>
#include <io/QLabelRenderer.hpp>

#ifdef WIN32
#include <crtdbg.h>
#endif
#include <container/ContainerFile.hpp>

#ifndef ISOMP4
#include <io/ffMP4IO.hpp>
#else
#include <io/isoMP4IO.hpp>
#endif

#include "MP7QtPlayer.hpp"

MP7QtPlayer::MP7QtPlayer()
{
  videoDC = NULL;
  audioDC = NULL;
  globalTimer = NULL;
  essync = NULL;
  mediaTimeAndSliderUpdater = new QTimer(this, "mediaTimeAndSliderUpdater");
  mp7Once = new QTimer(this, "mp7Once");
  mp7Updater = new QTimer(this, "mp7Updater");
  // Timers will be automatically deleted by QObject destructor
  resetVideoDisplay();
  vitooki_init_codecs();
  // display scene specific meta information first
  mainwidget->setCurrentPage(2);
  metawidget->setCurrentPage(1);
  
  // signals and slots connections
  connect(mediaTimeAndSliderUpdater, SIGNAL(timeout()), this, SLOT(displayCurrentMediaTimeAndUpdateSlider()));
  connect(mp7Once, SIGNAL(timeout()), this, SLOT(displayPermanentMP7Information()));
  connect(mp7Updater, SIGNAL(timeout()), this, SLOT(displayChangingMP7Information()));
  
}

MP7QtPlayer::~MP7QtPlayer()
{
  closeVideo();
}

bool MP7QtPlayer::loadMP7(const char *filename)
{	
  FILE *filehandle = MP7Parser::MP7Open(filename);
  if (filehandle == NULL) {
    return false;
  }
 
  vardescrdata = MP7Parser::MP7GetVarDescrData(filehandle);
  if (vardescrdata.errorcode != 0) {
    return false;
  }
     
  bool closed = MP7Parser::MP7Close(filehandle);
  if (!closed) {
    return false;
  }
  
  return true;
}

QString prettyPrintTime(double sec)
{
	QString label;
	uint t = (uint) sec;
	uint h, m, s;
	s = t % 60;
	h = t / 60;
	m = h % 60;
	h /= 60;
	label.sprintf("%02u:%02u:%02u", h, m, s);
	return label;
}

bool MP7QtPlayer::loadVideo(const char *filename, const char *demuxDir,
														UncompressedVideoFrame::ColorSpace colorSpace)
{
  closeVideo();
  ContainerInfo* mp4=ContainerFile::loadContainerInfo(filename);
  if(!mp4) {
    dprintf_err("MP7QtPlayer::loadVideo(): %s not found\n", filename);
    return false;
  }
  if (demuxDir)
    dprintf_full("MP7QtPlayer::loadVideo(): demuxDir = %s\n", demuxDir);
  ContainerDemux::demultiplexAndUpdateIO(mp4, demuxDir);
  list < ESInfo * >* streams = mp4->getESList();
  /* setup video stream first */
  list < ESInfo * >::const_iterator li = streams->begin();
  while (li != streams->end()) {
    ESInfo* es=(*li);
    if (!videoDC && es->isVisualStream()) {
      dprintf_full("MuvinoMain::loadVideo(): ES %llu, visual = %d, input = %s\n",
		   es->getStreamId(), es->isVisualStream(), es->getInput());
      VideoESInfo *esVideo = (VideoESInfo *) es;
      QLabelRenderer *renderer = new QLabelRenderer(esVideo, colorSpace,
						    videoWidget, false);
      if (!renderer) {
	dprintf_err("MuvinoMain::loadVideo(): QLabelRenderer constructor FAILED!\n");
	delete mp4;
	return false;
      }
      IO *input = IOCreator::createInput(es);
      if (!input) {
	dprintf_err("MuvinoMain::loadVideo(): IOCreator::createInput(esVideo) FAILED!\n");
	delete renderer;
	delete mp4;
	return false;
      }
      MP4Decoder *decoder = new MP4Decoder((VideoESInfo *) es, true,
					   colorSpace);
      if (!decoder) {
	dprintf_err("MuvinoMain::loadVideo(): MP4Decoder constructor FAILED!\n");
	delete input;
	delete renderer;
	delete mp4;
	return false;
      }
      if (!globalTimer) {
	globalTimer = new GlobalTimer();
	globalTimerRefCnt = new ReferenceCounter();
      }
      if (!essync)
	essync = new ESSynchronizer(es, globalTimer);
      VideoAdaptorChain *chain = new VideoAdaptorChain();
      chain->setESInfo(es);
      Adaptor* tmp=(Adaptor *) globalTimer; // necessary for VC .NET
      chain->addAdaptor(new SharedAdaptor(globalTimerRefCnt, tmp));
      chain->addAdaptor(essync);
      chain->addAdaptor(decoder);
      videoDC = new DataChannel(input, es, chain);
      if (!videoDC) {
	dprintf_err("MuvinoMain::loadVideo(): DataChannel constructor FAILED!\n");
	delete chain;
	delete input;
	delete renderer;
	delete mp4;
	delete globalTimer;
	globalTimer = NULL;
	return false;
      }
      videoDC->insert(new DataSink(0, renderer, NULL, 0));
      videoDC->play(0, 0.0);
      break;
    }
    ++li;
  }
  /* setup audio stream after video stream */
  li = streams->begin();
  while (li != streams->end()) {
    ESInfo* es=(*li);
    if (!audioDC && es->isAudioStream()) {
      dprintf_full("MuvinoMain::loadVideo(): ES %llu, visual = %d, input = %s\n",
		   es->getStreamId(), es->isVisualStream(), es->getInput());
      AudioESInfo *esAudio = (AudioESInfo *) es;
      IO *input = IOCreator::createInput(es);
      if (!input) {
	dprintf_err("MuvinoMain::loadVideo(): IOCreator::createInput(esAudio) FAILED!\n");
	delete mp4;
	return false;
      }
      if (!globalTimer) {
	globalTimer = new GlobalTimer();
	globalTimerRefCnt = new ReferenceCounter();
      }
      MP4audioDecoder *decoder = new MP4audioDecoder(esAudio, true);
      SDLaudioIO *audio_io = new SDLaudioIO(esAudio, essync);
      AudioAdaptorChain *chain = new AudioAdaptorChain();
      chain->setESInfo(es);
      chain->addAdaptor(decoder);
      Adaptor* tmp=(Adaptor *) globalTimer; // for VC .NET
      chain->addAdaptor(new SharedAdaptor(globalTimerRefCnt, tmp));
      audioDC = new DataChannel(input, es, chain);
      audioDC->insert(new DataSink(0, audio_io, NULL, 0));
      audioDC->play(0, 0.0);
      break;
    }
    ++li;
  }
  /* initialize player controls */
  setCurrentVideoTimeLabel(0.0);
  // set slider range to every 10sec
  slider_range = int(durationOfVideo() / 10);
  mainslider->setRange(0, slider_range-1);

  return true;
}

void MP7QtPlayer::closeVideo()
{
  if (!videoDC)
    return;
  mediaTimeAndSliderUpdater->stop();
  mp7Updater->stop();
  globalTimer->setPaused(false); // de-block before closing
  stopVideoAudio();
  if (videoDC)
    delete videoDC;		// should delete all Adaptors and DataSinks as well
  videoDC = NULL;
  if (audioDC)
    delete audioDC;
  audioDC = NULL;
  if (essync)
    delete essync;
  essync = NULL;
  if (globalTimer)
    delete globalTimer;
  globalTimer = NULL;
}

void MP7QtPlayer::stopAudio()
{
  if (!audioDC)
    return;
  if (audioDC->running()) {
    audioDC->pause(0);
    audioDC->initiateThreadStop(true);
  }
}

void MP7QtPlayer::stopVideoAudio()
{
  if (videoDC && videoDC->running()) {
      videoDC->pause(0);
  }
  if (audioDC && audioDC->running()) {
      audioDC->pause(0);
  }
  if (videoDC && videoDC->running()) {
    videoDC->initiateThreadStop(true);
  }
  if (audioDC && audioDC->running()) {
    audioDC->initiateThreadStop(true);
  }
}

void MP7QtPlayer::pauseVideoAudio()
{
  if (videoDC && videoDC->running()) {
      videoDC->pause(0);
  }
  if (audioDC && audioDC->running()) {
      audioDC->pause(0);
  }
}

void MP7QtPlayer::resetVideoDisplay()
{
  QPixmap pixmap(videoWidget->width(), videoWidget->height());
  pixmap.fill(Qt::black);
  videoWidget->setPixmap(pixmap);
  mediaTimeAndSliderUpdater->stop();
  setCurrentVideoTimeLabel(0.0);
}

bool MP7QtPlayer::playVideo(double sec)
{
  double dur;
  u32 frameNo;
  long iframeNo;
  const ESInfo *es, *audio_es;
  bool enableAudio = false;
  if (!videoDC)			// no video loaded
    return false;
  if ((audioDC != NULL) && (essync->getOutputRateFactor() == 1.0))
    enableAudio = true;
  if (sec == -1.0) {
    /* avoid re-positioning audioDC if DC threads already running */
    if (enableAudio && videoDC->running() && audioDC->running())
      return true;
    sec = currentVideoTime();	// needed for re-positioning audioDC
  } else {
    if (sec < 0.0) {
      dprintf_small("MP7QtPlayer::playVideo(sec = %7.2f) WARNING: sec is set to 0.0\n",
		    sec);
      sec = 0.0;
    }
    es = videoDC->getESInfo();
    if (!es->getMediaTimeScale()) {
      dprintf_err("MP7QtPlayer::playVideo(sec = %7.2f) ERROR: es->getMediaTimeScale() is ZERO!\n",
		  sec);
      return false;
    }
    dur = (double) es->getDuration() / (double) es->getMediaTimeScale();
    if (sec > dur) {
      dprintf_small("MP7QtPlayer::playVideo(sec = %7.2f) WARNING: sec is set to duration = %7.2f\n",
		    sec, dur);
      sec = dur;
    }

    if (videoDC->running())
      playerStop();
    /* clear adaptor buffers */
    globalTimer->setPaused(false); // de-block before closing
    videoDC->getAdaptor()->close();
    /* re-position videoDC */
    frameNo = es->getFrameNumber(sec);
    iframeNo = videoDC->getInput()->setToClosestIFrame(frameNo);
    if (iframeNo < 0) {
      dprintf_err("MP7QtPlayer::playVideo(sec = %7.2f) ERROR: setToClosestIFrame(%u) FAILED!\n",
		  sec, frameNo);
      return false;
    }
    frameNo = (u32) iframeNo;
    sec = es->getMediaTime(frameNo);
    dprintf_full("MP7QtPlayer::playVideo(): adjusting to I-Frame at %7.2f sec (frameNo %u)\n",
		 sec, frameNo);
    essync->adjust(sec);
  } // if (sec != -1.0)
  /* start DataChannel threads */
  if (!globalTimer->isPaused())
    globalTimer->setPaused(true);  // block frame processing in DataChannels
  if (enableAudio) {
    /* clear audioDC buffers */
    if (audioDC->getDataSink(0))
      ((SDLaudioIO *) audioDC->getDataSink(0)->getOutput())->clearBuffer();
    audioDC->getAdaptor()->close();
    /* re-position audioDC  and start audioDC thread */
    audio_es = audioDC->getESInfo();
    frameNo = audio_es->getFrameNumber(sec);
    audioDC->getInput()->setToFrameNumber(frameNo);
    audioDC->play(0, 0.0);
    if (!audioDC->running())
      audioDC->start();
  }
  videoDC->play(0, 0.0);
  if (!videoDC->running())
    videoDC->start();		   // start videoDC thread

  mediaTimeAndSliderUpdater->start(1000);
  mp7Updater->start(500);
  displayCurrentMediaTimeAndUpdateSlider();
  durOfVideo->setText(prettyPrintTime(durationOfVideo()));
  globalTimer->setPaused(false);   // de-block DataChannels
  return true;
}


double MP7QtPlayer::currentVideoTime() const
{
  const ESInfo *es;
  u32 frameNo;
  double t;
  if (!videoDC)
    return 0.0;
  es = videoDC->getESInfo();
  frameNo = videoDC->getInput()->getCurrentFrameNumber();
  t = es->getMediaTime(frameNo);
  dprintf_full("MP7QtPlayer::currentVideoTime(): frameNo = %u, returning %7.2f seconds\n",
	       frameNo, t);
  return t;
}

double MP7QtPlayer::durationOfVideo() const
{
	const ESInfo *es;
	double t;
	
	if (!videoDC)
		return 0.0;
	es = videoDC->getESInfo();
	t = (double)es->getDurationInMs() / (double)1000;
	return t;
}

void MP7QtPlayer::setCurrentVideoTimeLabel(double sec, bool showSubSec)
{
  QString label;
  uint t = (uint) sec;
  uint h, m, s, ss;
  s = t % 60;
  h = t / 60;
  m = h % 60;
  h /= 60;
  if (showSubSec) {
    ss = (uint) (sec * 100.0) - t * 100;
    label.sprintf("%02u:%02u:%02u.%02u", h, m, s, ss);
  } else
    label.sprintf("%02u:%02u:%02u   ", h, m, s);
  currentMediaTimeLabel->setText(label);
}

/* -------- Slot Implementations -------------------------------------------------- */

/* slot: "Start" button clicked */
void MP7QtPlayer::playerStart()
{
  dprintf_full("MP7QtPlayer::playerStart(): entered\n");
  mp7Once->start(500, true);
  playVideo(0.0);
}

/* slot: Play button was clicked */
void MP7QtPlayer::playerPlay()
{
  dprintf_full("MP7QtPlayer::playerPlay(): entered\n");
  mp7Once->start(500, true);
  playVideo();
}

/* slot: Stop button was clicked */
void MP7QtPlayer::playerStop()
{
  dprintf_full("MP7QtPlayer::playerStop(): entered\n");
  if (!videoDC)
    return;
  mediaTimeAndSliderUpdater->stop();
  pauseVideoAudio();
  /* avoid double pausing */
  if (!globalTimer->isPaused())
    globalTimer->setPaused(true);
  displayCurrentMediaTimeAndUpdateSlider();
}


/* slot: update current media time label and slider */
void MP7QtPlayer::displayCurrentMediaTimeAndUpdateSlider()
{  
	int newvalue = int((currentVideoTime() / (durationOfVideo() / (double)slider_range))+0.5);
	if (mainslider->value() != newvalue) {
		mainslider->setValue(newvalue);
	}
	
  	setCurrentVideoTimeLabel(currentVideoTime(), false);
  	
}

void MP7QtPlayer::mainslider_sliderPressed()
{
	mediaTimeAndSliderUpdater->stop();
}

void MP7QtPlayer::mainslider_sliderReleased()
{
	playVideo(ceil((durationOfVideo() / (double)slider_range)*((double)mainslider->value())));
	mediaTimeAndSliderUpdater->start(1000);
}

void MP7QtPlayer::displayPermanentMP7Information()
{
  QString label, tmpstr;
  MP7Parser::MP7_VarSetList *vslist = vardescrdata.vsl_pointer;
         
  // fill source data
  if (vslist != NULL) {
	QString label;
	QString tmp;
	label.append("<font size=\"-1\">");
	if (!vslist->sdata.id.empty()) {
		label.append("<b>ID</b>: ");
		label.append(vslist->sdata.id.c_str());
		label.append("<br>");		
	}
	if (!vslist->sdata.video_codec.empty()) {
		label.append("<b>Video Codec</b>: ");
		label.append(vslist->sdata.video_codec.c_str());
		label.append("<br>");		
	}	
	if (vslist->sdata.video_width != 0) {
		label.append("<b>Video Width</b>: ");
		tmp.sprintf("%d", vslist->sdata.video_width);
		label.append(tmp);
		label.append("<br>");		
	}
	if (vslist->sdata.video_height != 0) {
		label.append("<b>Video Height</b>: ");
		tmp.sprintf("%d", vslist->sdata.video_height);
		label.append(tmp);
		label.append("<br>");		
	}
	if (vslist->sdata.video_rate != 0.0) {
		label.append("<b>Video Rate</b>: ");
		tmp.sprintf("%01.2f", vslist->sdata.video_rate);
		label.append(tmp);
		label.append("<br>");		
	}
	if (vslist->sdata.has_audio) {
		label.append("<b>Audio</b>: present");
		label.append("<br>");
	}
	else {
		label.append("<b>Audio</b>: not present");
		label.append("<br>");
	}
	if (!vslist->sdata.audio_codec.empty()) {
		label.append("<b>Audio Codec</b>: ");
		label.append(vslist->sdata.audio_codec.c_str());
		label.append("<br>");		
	}
	if (!vslist->sdata.path.empty()) {
		label.append("<b>Path</b>: ");
		label.append(vslist->sdata.path.c_str());
		label.append("<br>");		
	}
	if (!vslist->sdata.fileformat.empty()) {
		label.append("<b>Fileformat</b>: ");
		label.append(vslist->sdata.fileformat.c_str());
		label.append("<br>");		
	}
	if (vslist->sdata.filesize != 0) {
		label.append("<b>Filesize</b>: ");
		tmp.sprintf("%ld", vslist->sdata.filesize);
		label.append(tmp);
		label.append("<br>");	
	}
	if (vslist->sdata.bitrate != 0) {
		label.append("<b>Bitrate</b>: ");
		tmp.sprintf("%ld", vslist->sdata.bitrate);
		label.append(tmp);
		label.append("<br>");		
	}		
	label.append("<br>");
	label.append("</font>");
	source->setText(label);
  }
  
  // fill variations selector
  varlist->clear();
  MP7Parser::MP7_VarList *tmp_vlist = vardescrdata.vsl_pointer->vdata.var_list;
  for (int i=1;i<=vslist->nr_variations;i++) {
	  varlist->insertItem(tmp_vlist->vdata.variation_id.c_str(), i);
	  tmp_vlist = tmp_vlist->next;
  }
  varlist->setCurrentItem(0);  
  
  if (vslist != NULL) {
	  double until = vslist->mtime.tpAsDouble() + vslist->mtime.durAsDouble();
	  QString label;
	  label.append("<font size=\"-1\">");
	  if (!vslist->mdata.title.empty()) {
		  label.append("<b>Title</b>: <u>");
		  label.append(vslist->mdata.title.c_str());
		  label.append("</u><br>");
	  }
	  label.append("<b>Total Length: </b>");
	  label.append(prettyPrintTime(until));
	  label.append("<br>");
	  if (!vslist->mdata.free_text_annotations.empty()) {
		  label.append("<b>Free Text Annotations</b>: ");
		  label.append(vslist->mdata.free_text_annotations.c_str());
		  label.append("<br>");
	  }
	  if (!vslist->mdata.keywords.empty()) {
		  label.append("<b>Keywords</b>: ");
		  label.append(vslist->mdata.keywords.c_str());
		  label.append("<br>");
	  }
	  if (!vslist->mdata.semantic_place.empty()) {
		  label.append("<b>Semantic Place</b>: ");
		  label.append(vslist->mdata.semantic_place.c_str());
		  label.append("<br>");
	  }
	  if (!vslist->mdata.semantic_region.empty()) {
		  label.append("<b>Semantic Region</b>: ");
		  label.append(vslist->mdata.semantic_region.c_str());
		  label.append("<br>");
	  }
	  if (!vslist->mdata.semantic_time.empty()) {
		  label.append("<b>Semantic Time</b>: ");
		  label.append(vslist->mdata.semantic_time.c_str());
		  label.append("<br>");
	  }
	  if (!vslist->mdata.creator_name.empty()) {
		  label.append("<b>Creator Name</b>: ");
		  label.append(vslist->mdata.creator_name.c_str());
		  label.append("<br>");
	  }
	  if (!vslist->mdata.location_name.empty()) {
		  label.append("<b>Creator Location</b>: ");
		  label.append(vslist->mdata.location_name.c_str());
		  label.append("<br>");
	  }
	  if (!vslist->mdata.location_region.empty()) {
		  label.append("<b>Creator Region</b>: ");
		  label.append(vslist->mdata.location_region.c_str());
		  label.append("<br>");
	  }
	  if (!vslist->mdata.date.empty()) {
		  label.append("<b>Creation Date</b>: ");
		  label.append(vslist->mdata.date.c_str());
		  label.append("<br>");
	  }
	  if (!vslist->mdata.copyright.empty()) {
		  label.append("<b>Creation Copyright</b>: ");
		  label.append(vslist->mdata.copyright.c_str());
		  label.append("<br>");
	  }	  
	  label.append("<br>");
	  label.append("</font>");
	  mdata->setText(label);
  }
  
}

	
void displayVideoSegmentData (MP7Parser::MP7_VideoSegmentList *list, double cvt, QString *label)
{
	while (list != NULL) {
		double from = list->vseg_data.mediatime.tpAsDouble();
		double until = list->vseg_data.mediatime.tpAsDouble() + list->vseg_data.mediatime.durAsDouble();
		if ((cvt >= from) && (cvt < until)) {
			label->append("<font size=\"-1\">");
			if (!list->vseg_data.title.empty()) {
				label->append("<b>Title</b>: <u>");
				label->append(list->vseg_data.title.c_str());
				label->append("</u><br>");
			}
			label->append("<b>From </b>");
			label->append(prettyPrintTime(from));
			label->append("<b> to </b>");
			label->append(prettyPrintTime(until));
			label->append("<br>");
			label->append("<b>Remaining </b>");
			label->append(prettyPrintTime(until-cvt));
			label->append("<br>");
			if (!list->vseg_data.free_text_annotations.empty()) {
				label->append("<b>Free Text Annotations</b>: ");
				label->append(list->vseg_data.free_text_annotations.c_str());
				label->append("<br>");
			}
			if (!list->vseg_data.keywords.empty()) {
				label->append("<b>Keywords</b>: ");
				label->append(list->vseg_data.keywords.c_str());
				label->append("<br>");
			}
			if (!list->vseg_data.semantic_place.empty()) {
				label->append("<b>Place</b>: ");
				label->append(list->vseg_data.semantic_place.c_str());
				label->append("<br>");
			}
			if (!list->vseg_data.semantic_region.empty()) {
				label->append("<b>Region</b>: ");
				label->append(list->vseg_data.semantic_region.c_str());
				label->append("<br>");
			}
			if (!list->vseg_data.semantic_time.empty()) {
				label->append("<b>Time</b>: ");
				label->append(list->vseg_data.semantic_time.c_str());
				label->append("<br>");
			}			
			label->append("<br>");
			label->append("</font>");
			displayVideoSegmentData(list->children, cvt, label);
		}
		list = list->next;
	}
}	

void MP7QtPlayer::displayChangingMP7Information()
{
 	MP7Parser::MP7_VideoSegmentList *vsglist = vardescrdata.vsl_pointer->tmpdec.vseg_list;
	double cvt = currentVideoTime();
	QString label;
	
	displayVideoSegmentData(vsglist, cvt, &label);
 	scene->setText(label);
}

void MP7QtPlayer::varlist_highlighted_or_selected(int pos) 
{
  QString label, tmp;
  MP7Parser::MP7_VarList *vlist = vardescrdata.vsl_pointer->vdata.var_list;
  
  // go to highlighted variation
  for(int i=1;i<pos+1;i++) {
    if (vlist->next != NULL)
      vlist = vlist->next;
  }
  
  label.append("<font size=\"-1\">");
  if (!vlist->vdata.id.empty()) {
	  label.append("<b>ID</b>: ");
	  label.append(vlist->vdata.id.c_str());
	  label.append("<br>");		
  }
  if (vlist->vdata.fidelity != 0.0) {
	  label.append("<b>Fidelity</b>: ");
	  tmp.sprintf("%01.2f", vlist->vdata.fidelity);
	  label.append(tmp);
	  label.append("<br>");		
  }
  if (vlist->vdata.priority != 0) {
	  label.append("<b>Priority</b>: ");
	  tmp.sprintf("%d", vlist->vdata.priority);
	  label.append(tmp);
	  label.append("<br>");		
  }  
  if (!vlist->vdata.video_codec.empty()) {
	  label.append("<b>Video Codec</b>: ");
	  label.append(vlist->vdata.video_codec.c_str());
	  label.append("<br>");		
  }	
  if (vlist->vdata.video_width != 0) {
	  label.append("<b>Video Width</b>: ");
	  tmp.sprintf("%d", vlist->vdata.video_width);
	  label.append(tmp);
	  label.append("<br>");		
  }
  if (vlist->vdata.video_height != 0) {
	  label.append("<b>Video Height</b>: ");
	  tmp.sprintf("%d", vlist->vdata.video_height);
	  label.append(tmp);
	  label.append("<br>");		
  }
  if (vlist->vdata.video_rate != 0.0) {
	  label.append("<b>Video Rate</b>: ");
	  tmp.sprintf("%01.2f", vlist->vdata.video_rate);
	  label.append(tmp);
	  label.append("<br>");		
  }
  if (vlist->vdata.has_audio) {
	  label.append("<b>Audio</b>: present");
	  label.append("<br>");
  }
  else {
	  label.append("<b>Audio</b>: not present");
	  label.append("<br>");
  }
  if (!vlist->vdata.audio_codec.empty()) {
	  label.append("<b>Audio Codec</b>: ");
	  label.append(vlist->vdata.audio_codec.c_str());
	  label.append("<br>");		
  }
  if (!vlist->vdata.path.empty()) {
	  label.append("<b>Path</b>: ");
	  label.append(vlist->vdata.path.c_str());
	  label.append("<br>");		
  }
  if (!vlist->vdata.fileformat.empty()) {
	  label.append("<b>Fileformat</b>: ");
	  label.append(vlist->vdata.fileformat.c_str());
	  label.append("<br>");		
  }
  if (vlist->vdata.filesize != 0) {
	  label.append("<b>Filesize</b>: ");
	  tmp.sprintf("%ld", vlist->vdata.filesize);
	  label.append(tmp);
	  label.append("<br>");	
  }
  if (vlist->vdata.bitrate != 0) {
	  label.append("<b>Bitrate</b>: ");
	  tmp.sprintf("%ld", vlist->vdata.bitrate);
	  label.append(tmp);
	  label.append("<br>");		
  }
  if (!vlist->vdata.path.empty()) {
	  label.append("<b>Variation Relationship</b>: ");
	  label.append(vlist->vdata.variation_relationship.c_str());
	  label.append("<br>");		
  }  
  label.append("<br>");
  label.append("</font>");
  variations->setText(label);
  
}
