/***********************************************************************
 *                                                                      *
 * MP7QtPlayer - a simple video player using the ViTooKi library        *
 *               showing MPEG-7 Data 					*
 *                                                                      *
 * title:	MP7QtPlayer.hpp                                         *                                                                       
 *									*
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

#ifndef MUVINOMAIN_H
#define MUVINOMAIN_H

#include <DataChannel.hpp>
#include <adaptors/MP4Decoder.hpp>
#include <metadata/MP7Parser.hpp>

#include "MP7QtPlayerDlg.hpp"

class QApplication;
class QTimer;
class GlobalTimer;
class ESSynchronizer;



class MP7QtPlayer : public MP7QtPlayerDlg  {
  Q_OBJECT

 public: 
  MP7QtPlayer();
  ~MP7QtPlayer();

  /**
   * Initialize the adaptor chain and the video widget for play-back of an MP4 encoded
   * video. The resolution (width, height) needs to match the
   * dimensions of #videoWidget.
   *
   * @param filename file name of MP4 encoded video.
   * @param demuxDir local directory where the demuxed elementary streams are
   *                 to be stored.
   * @param colorSpace color space which the video should be converted to; must be one
   *                   of: MP4Decoder::ColorSpaceGRAY8, MP4Decoder::ColorSpaceRGB24, 
   *                   MP4Decoder::ColorSpaceRGB32.
   *
   * @return <code>true</code> on success, <code>false</code> on
   *         error.
   * @see QLabelRenderer
   */
  bool loadVideo(const char *filename, const char *demuxDir,
		 UncompressedVideoFrame::ColorSpace colorSpace);
		 
  bool loadMP7(const char *filename);

public slots:
  virtual void playerStart();
  virtual void playerStop();
  virtual void playerPlay();
  virtual void displayCurrentMediaTimeAndUpdateSlider();
  virtual void displayPermanentMP7Information();
  virtual void displayChangingMP7Information();
  virtual void varlist_highlighted_or_selected(int pos);
  virtual void mainslider_sliderReleased();
  virtual void mainslider_sliderPressed();
  

 protected:
  /**
   * Reset the video widget to its default settings.
   */
  void resetVideoDisplay();

  /**
   * Close the video loaded by #loadVideo().
   */
  void closeVideo();

  /** Stop audio play-back. 
   * Make sure that #globalTimer is not paused before calling this method.
   */
  void stopAudio();

  /** Stop video and audio play-back.
   * Make sure that #globalTimer is not paused before calling this method.
   */
  void stopVideoAudio();

  /** Pause video and audio play-back.
   * The DataChannel threads will eventually terminate, leaving the input stream
   * open. This function does not wait for the DataChannel threads to terminate.
   */
  void pauseVideoAudio();

  /**
   * Start video play-back at a given absolute media time point.
   *
   * @param sec    media time point in seconds; 0.0 means start of video,
   *               -1.0 means &quot;continue from current time point&quot;.
   * @return <code>true</code> on success, <code>false</code> on error.
   */
  bool playVideo(double sec = -1.0);

  /**
   * Get current video media time stamp in seconds.
   * Returns <code>0.0</code> if the time stamp cannot be determined.
   */
  double currentVideoTime() const;
	     
  double durationOfVideo() const;
  
  /** Set the current media time label to the given value. 
   * @param sec          media time in seconds; must be non-negative.
   * @param showSubSec   determines whether sub-second fractions will
   *                     be shown.
   */
  void setCurrentVideoTimeLabel(double sec, bool showSubSec = false);

  /** Set the display rate slider control to the given rate.
   * @see #setDisplayRateFromSlider(int)
   */
  void setDisplayRateSlider(double rate);

 private:
 
  /** Data structure returned by the MPEG-7 Parser **/
  MP7Parser::MP7_VarDescrData vardescrdata;
     
  /** The DataChannel belonging to the visual elementary stream of a loaded video. */
  DataChannel *videoDC;

  /** The DataChannel belonging to the audio elementary stream of a loaded video. */
  DataChannel *audioDC;

  /** The global timer controlling the DataChannels. */
  GlobalTimer *globalTimer;

  /** The reference counter for using #globalTimer as a SharedAdaptor. */
  ReferenceCounter *globalTimerRefCnt;

  /** The ESSynchronizer controlling the #videoDC DataChannel. */
  ESSynchronizer *essync;

  /** The timer for updating the current media time label. */
  QTimer *mediaTimeAndSliderUpdater;
  
  /** timer for changing MPEG-7 information */
  QTimer *mp7Updater;
  
  /** timer for permanent MPEG-7 information */
  QTimer *mp7Once;

  /** range of slider */
  int slider_range;
};

#endif
