 /***********************************************************************
 *                                                                      *
 * QtPlayer - a simple video player using the ViTooKi library           *
 *                                                                      *
 * title:  QtPlayer.cpp                                                 *
 *                                                                      *
 * $Id: QtPlayer.cpp,v 1.11 2005/11/17 16:55:41 mkropfbe Exp $
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

#include <stdlib.h>
#include <cmath>
#include <qlabel.h>
#include <qtimer.h>
#include <qcheckbox.h>
#include <qslider.h>

#include <vitooki.hpp>
#include <DataChannel.hpp>
#include <DataSink.hpp>
#include <ContainerInfo.hpp>
#include <ESInfo.hpp>
#include <AudioAdaptorChain.hpp>
#include <VideoAdaptorChain.hpp>
#include <adaptors/MP4audioDecoder.hpp>
#include <adaptors/MP4Decoder.hpp>
#include <adaptors/ESSynchronizer.hpp>
#include <adaptors/GlobalTimer.hpp>
#include <adaptors/SharedAdaptor.hpp>
#include <container/ContainerDemux.hpp>
#include <io/SDLaudioIO.hpp>
#include <io/IOCreator.hpp>
#include <io/MPGStreamIO.hpp>
#include <io/QLabelRenderer.hpp>

#ifdef WIN32
#include <crtdbg.h>
#endif
#include <container/ContainerFile.hpp>

#ifndef ISOMP4
#include <io/ffMP4IO.hpp>
#else
#include <io/isoMP4IO.hpp>
#endif

#include "QtPlayer.hpp"

QtPlayer::QtPlayer()
{
  videoDC = NULL;
  audioDC = NULL;
  globalTimer = NULL;
  essync = NULL;
  mediaTimeUpdater = new QTimer(this, "mediaTimeUpdater");
  // mediaTimeUpdater will be automatically deleted by QObject destructor
  resetVideoDisplay();
  vitooki_init_codecs();

  // signals and slots connections
  connect(mediaTimeUpdater, SIGNAL(timeout()), this, SLOT(displayCurrentMediaTime()));
}

QtPlayer::~QtPlayer()
{
  closeVideo();
}

bool QtPlayer::loadVideo(const char *filename, const char *demuxDir,
			   UncompressedVideoFrame::ColorSpace colorSpace)
{
  closeVideo();
  ContainerInfo* mp4=ContainerFile::loadContainerInfo(filename);
  if(!mp4) {
    dprintf_err("QtPlayer::loadVideo(): %s not found\n", filename);
    return false;
  }
  if (demuxDir)
    dprintf_full("QtPlayer::loadVideo(): demuxDir = %s\n", demuxDir);
	// SEEMS TO BE NOT NECCESSARY!
	//  ContainerDemux::demultiplexAndUpdateIO(mp4, demuxDir);

  list < ESInfo * >* streams = mp4->getESList();
  /* setup video stream first */
  list < ESInfo * >::const_iterator li = streams->begin();
  while (li != streams->end()) {
    ESInfo* es=(*li);
    if (!videoDC && es->isVisualStream()) {
      dprintf_full("QtPlayer::loadVideo(): ES %llu, visual = %d, input = %s\n",
									 es->getStreamId(), es->isVisualStream(), es->getInput());
      VideoESInfo *esVideo = (VideoESInfo *) es;
      QLabelRenderer *renderer = new QLabelRenderer(esVideo, colorSpace,
																										videoWidget, false);
      if (!renderer) {
				dprintf_err("QtPlayer::loadVideo(): QLabelRenderer constructor FAILED!\n");
				delete mp4;
				return false;
      }
      IO *input = IOCreator::createInput(es);
      if (!input) {
				dprintf_err("QtPlayer::loadVideo(): IOCreator::createInput(esVideo) FAILED!\n");
				delete renderer;
				delete mp4;
				return false;
      }
      MP4Decoder *decoder = new MP4Decoder((VideoESInfo *) es, true,
																					 colorSpace);
      if (!decoder) {
				dprintf_err("QtPlayer::loadVideo(): MP4Decoder constructor FAILED!\n");
				delete input;
				delete renderer;
				delete mp4;
				return false;
      }
      if (!globalTimer) {
				globalTimer = new GlobalTimer();
				globalTimerRefCnt = new ReferenceCounter();
      }
      if (!essync)
				essync = new ESSynchronizer(es, globalTimer);
      VideoAdaptorChain *chain = new VideoAdaptorChain();
      chain->setESInfo(es);
      Adaptor* tmp=(Adaptor *) globalTimer; // necessary for VC .NET
      chain->addAdaptor(new SharedAdaptor(globalTimerRefCnt, tmp));
      chain->addAdaptor(essync);
      chain->addAdaptor(decoder);
      videoDC = new DataChannel(input, es, chain);
      if (!videoDC) {
				dprintf_err("QtPlayer::loadVideo(): DataChannel constructor FAILED!\n");
				delete chain;
				delete input;
				delete renderer;
				delete mp4;
				delete globalTimer;
				globalTimer = NULL;
				return false;
      }
      videoDC->insert(new DataSink(0, renderer, NULL, 0));
      videoDC->play(0, 0.0);
      break;
    }
    ++li;
  }
  /* setup audio stream after video stream */
  li = streams->begin();
  while (li != streams->end()) {
    ESInfo* es=(*li);
    if (!audioDC && es->isAudioStream()) {
      dprintf_full("QtPlayer::loadVideo(): ES %llu, visual = %d, input = %s\n",
									 es->getStreamId(), es->isVisualStream(), es->getInput());
      AudioESInfo *esAudio = (AudioESInfo *) es;
      IO *input = IOCreator::createInput(es);
      if (!input) {
				dprintf_err("QtPlayer::loadVideo(): IOCreator::createInput(esAudio) FAILED!\n");
				delete mp4;
				return false;
      }
      if (!globalTimer) {
				globalTimer = new GlobalTimer();
				globalTimerRefCnt = new ReferenceCounter();
      }
      MP4audioDecoder *decoder = new MP4audioDecoder(esAudio, true);
      SDLaudioIO *audio_io = new SDLaudioIO(esAudio, essync);
      AudioAdaptorChain *chain = new AudioAdaptorChain();
      chain->setESInfo(es);
      chain->addAdaptor(decoder);
      Adaptor* tmp=(Adaptor *) globalTimer; // for VC .NET
      chain->addAdaptor(new SharedAdaptor(globalTimerRefCnt, tmp));
      audioDC = new DataChannel(input, es, chain);
      audioDC->insert(new DataSink(0, audio_io, NULL, 0));
      audioDC->play(0, 0.0);
      break;
    }
    ++li;
  }
  /* initialize player controls */
  setDisplayRateSlider(1.0);
  setCurrentVideoTimeLabel(0.0);
  audioCheckBox->setChecked(audioDC != NULL);
  audioCheckBox->setDisabled(audioDC == NULL);
  audioCheckBoxLabel->setDisabled(audioDC == NULL);
  return true;
}

void QtPlayer::closeVideo()
{
  if (!videoDC)
    return;
  mediaTimeUpdater->stop();
  globalTimer->setPaused(false); // de-block before closing
  stopVideoAudio();
  if (videoDC)
    delete videoDC;		// should delete all Adaptors and DataSinks as well
  videoDC = NULL;
  if (audioDC)
    delete audioDC;
  audioDC = NULL;
  if (essync)
    delete essync;
  essync = NULL;
  if (globalTimer)
    delete globalTimer;
  globalTimer = NULL;
}

void QtPlayer::stopAudio()
{
  if (!audioDC)
    return;
  if (audioDC->running()) {
    audioDC->pause(0);
    audioDC->initiateThreadStop(true);
  }
}

void QtPlayer::stopVideoAudio()
{
  if (videoDC && videoDC->running()) {
      videoDC->pause(0);
  }
  if (audioDC && audioDC->running()) {
      audioDC->pause(0);
  }
  if (videoDC && videoDC->running()) {
    videoDC->initiateThreadStop(true);
  }
  if (audioDC && audioDC->running()) {
    audioDC->initiateThreadStop(true);
  }
}

void QtPlayer::pauseVideoAudio()
{
  if (videoDC && videoDC->running()) {
      videoDC->pause(0);
  }
  if (audioDC && audioDC->running()) {
      audioDC->pause(0);
  }
}

void QtPlayer::resetVideoDisplay()
{
  QPixmap pixmap(videoWidget->width(), videoWidget->height());
  pixmap.fill(Qt::black);
  videoWidget->setPixmap(pixmap);
  mediaTimeUpdater->stop();
  setCurrentVideoTimeLabel(0.0);
}

bool QtPlayer::playVideo(double sec)
{
  double dur;
  u32 frameNo;
  long iframeNo;
  const ESInfo *es, *audio_es;
  bool enableAudio;
  if (!videoDC)			// no video loaded
    return false;
  enableAudio = audioCheckBox->isChecked() &&
    audioDC != NULL && essync->getOutputRateFactor() == 1.0;
  if (sec == -1.0) {
    /* avoid re-positioning audioDC if DC threads already running */
    if (enableAudio && videoDC->running() && audioDC->running())
      return true;
    sec = currentVideoTime();	// needed for re-positioning audioDC
  } else {
    if (sec < 0.0) {
      dprintf_small("QtPlayer::playVideo(sec = %7.2f) WARNING: sec is set to 0.0\n",
		    sec);
      sec = 0.0;
    }
    es = videoDC->getESInfo();
    if (!es->getMediaTimeScale()) {
      dprintf_err("QtPlayer::playVideo(sec = %7.2f) ERROR: es->getMediaTimeScale() is ZERO!\n",
		  sec);
      return false;
    }
    dur = (double) es->getDuration() / (double) es->getMediaTimeScale();
    if (sec > dur) {
      dprintf_small("QtPlayer::playVideo(sec = %7.2f) WARNING: sec is set to duration = %7.2f\n",
		    sec, dur);
      sec = dur;
    }

    if (videoDC->running())
      playerStop();
    /* clear adaptor buffers */
    globalTimer->setPaused(false); // de-block before closing
    videoDC->getAdaptor()->close();
    /* re-position videoDC */
    frameNo = es->getFrameNumber(sec);
    iframeNo = videoDC->getInput()->setToClosestIFrame(frameNo);
    if (iframeNo < 0) {
      dprintf_err("QtPlayer::playVideo(sec = %7.2f) ERROR: setToClosestIFrame(%u) FAILED!\n",
		  sec, frameNo);
      return false;
    }
    frameNo = (u32) iframeNo;
    sec = es->getMediaTime(frameNo);
    dprintf_full("QtPlayer::playVideo(): adjusting to I-Frame at %7.2f sec (frameNo %u)\n",
		 sec, frameNo);
    essync->adjust(sec);
  } // if (sec != -1.0)
  /* start DataChannel threads */
  if (!globalTimer->isPaused())
    globalTimer->setPaused(true);  // block frame processing in DataChannels
  if (enableAudio) {
    /* clear audioDC buffers */
    if (audioDC->getDataSink(0))
      ((SDLaudioIO *) audioDC->getDataSink(0)->getOutput())->clearBuffer();
    audioDC->getAdaptor()->close();
    /* re-position audioDC  and start audioDC thread */
    audio_es = audioDC->getESInfo();
    frameNo = audio_es->getFrameNumber(sec);
    audioDC->getInput()->setToFrameNumber(frameNo);
    audioDC->play(0, 0.0);
    if (!audioDC->running())
      audioDC->start();
  }
  videoDC->play(0, 0.0);
  if (!videoDC->running())
    videoDC->start();		   // start videoDC thread

  mediaTimeUpdater->start(500);    // intervall timer: 500 ms
  displayCurrentMediaTime();	   // remove sub-seconds from displayed media time
  globalTimer->setPaused(false);   // de-block DataChannels
  return true;
}

double QtPlayer::currentVideoTime() const
{
  const ESInfo *es;
  u32 frameNo;
  double t;
  if (!videoDC)
    return 0.0;
  es = videoDC->getESInfo();
  frameNo = videoDC->getInput()->getCurrentFrameNumber();
  t = es->getMediaTime(frameNo);
  dprintf_full("QtPlayer::currentVideoTime(): frameNo = %u, returning %7.2f seconds\n",
	       frameNo, t);
  return t;
}

void QtPlayer::setCurrentVideoTimeLabel(double sec, bool showSubSec)
{
  QString label;
  uint t = (uint) sec;
  uint h, m, s, ss;
  s = t % 60;
  h = t / 60;
  m = h % 60;
  h /= 60;
  if (showSubSec) {
    ss = (uint) (sec * 100.0) - t * 100;
    label.sprintf("%02u:%02u:%02u.%02u", h, m, s, ss);
  } else
    label.sprintf("%02u:%02u:%02u   ", h, m, s);
  currentMediaTimeLabel->setText(label);
}

/* value = round(2 + log2(rate)) */
void QtPlayer::setDisplayRateSlider(double rate)
{
  if (rate < 0.25 || rate > 4.0) {
    dprintf_err("QtPlayer::setDisplayRateSlider(rate = %5.2f): rate out of range!\n",
		rate);
    return;
  }
  displayRateSlider->setValue((int)(2.5 + (log10(rate)/log10(2.0))));
}

/* -------- Slot Implementations -------------------------------------------------- */

/* slot: "Start" button clicked */
void QtPlayer::playerStart()
{
  dprintf_full("QtPlayer::playerStart(): entered\n");
  playVideo(0.0);
}

/* slot: Play button was clicked */
void QtPlayer::playerPlay()
{
  dprintf_full("QtPlayer::playerPlay(): entered\n");
  playVideo();
}

/* slot: Stop button was clicked */
void QtPlayer::playerStop()
{
  dprintf_full("QtPlayer::playerStop(): entered\n");
  if (!videoDC)
    return;
  mediaTimeUpdater->stop();
  pauseVideoAudio();
  /* avoid double pausing */
  if (!globalTimer->isPaused())
    globalTimer->setPaused(true);
  displayCurrentMediaTime(true);
}

/* slot: "-3s" button clicked */
void QtPlayer::playerBack3s()
{
  dprintf_full("QtPlayer::playerBack3s(): entered\n");
  playVideo(currentVideoTime() - 3.0);
}

/* slot: update current media time label */
void QtPlayer::displayCurrentMediaTime(bool showSubSec)
{
  setCurrentVideoTimeLabel(currentVideoTime(), showSubSec);
}

/* slot: set display frame rate factor
 * meaning of value passed from displayRateSlider:
 * factor = 2 ^ (value - 2)
 */
void QtPlayer::setDisplayRateFromSlider(int value)
{
  double factor = pow(2.0,(double) (value - 2));
  if (!essync)
    return;
  dprintf_full("QtPlayer::setDisplayRateFromSlider() value = %d, factor = %4.2f\n",
	       value, factor);
  if (audioDC && factor != 1.0)
    stopAudio();
  essync->setOutputRateFactor(factor);
  audioCheckBox->setDisabled(!audioDC || factor != 1.0);
  audioCheckBoxLabel->setDisabled(!audioDC || factor != 1.0);
  // may cause playerAudioStateChange() to be called
  audioCheckBox->setChecked(audioDC && factor == 1.0);
}

/* slot: audioCheckBox changed its state */
void QtPlayer::playerAudioStateChange(int state)
{
  dprintf_full("QtPlayer::playerAudioStateChange(): audioCheckBox->isChecked() == %d\n",
	       audioCheckBox->isChecked());
  if (audioCheckBox->isChecked()) {
    playVideo();
  } else {
    stopAudio();
  }
}
