 /***********************************************************************
 *                                                                      *
 * QtPlayer - a simple video player using the ViTooKi library           *
 *                                                                      *
 * title: QtPlayer.hpp                                                  *
 *                                                                      *
 * $Id: QtPlayer.hpp,v 1.5 2005/01/20 09:47:51 mkropfbe Exp $
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

#ifndef MUVINOMAIN_H
#define MUVINOMAIN_H

#include <DataChannel.hpp>
#include <adaptors/MP4Decoder.hpp>

#include "QtPlayerDlg.hpp"

class QApplication;
class QTimer;
class GlobalTimer;
class ESSynchronizer;

/**
 * This class implements the application logic of the main GUI class
 * MuvinoMainDlg. The latter class has been generated using the Qt Designer.
 *
 * @author Mario Taschwer
 */

class QtPlayer : public QtPlayerDlg  {
  Q_OBJECT

 public: 
  QtPlayer();
  ~QtPlayer();

  /**
   * Initialize the adaptor chain and the video widget for play-back of an MP4 encoded
   * video. The resolution (width, height) needs to match the
   * dimensions of #videoWidget.
   *
   * @param filename file name of MP4 encoded video.
   * @param demuxDir local directory where the demuxed elementary streams are
   *                 to be stored.
   * @param colorSpace color space which the video should be converted to; must be one
   *                   of: UncompressedVideoFrame::ColorSpaceGRAY8, UncompressedVideoFrame::ColorSpaceRGB24, 
   *                   UncompressedVideoFrame::ColorSpaceRGB32.
   *
   * @return <code>true</code> on success, <code>false</code> on
   *         error.
   * @see QLabelRenderer
   */
  bool loadVideo(const char *filename, const char *demuxDir,
		 UncompressedVideoFrame::ColorSpace colorSpace);

public slots:
  virtual void playerStart();
  virtual void playerStop();
  virtual void playerPlay();
  virtual void playerBack3s();
  virtual void displayCurrentMediaTime(bool showSubSec = false);
  virtual void setDisplayRateFromSlider(int value);
  virtual void playerAudioStateChange(int state);

 protected:
  /**
   * Reset the video widget to its default settings.
   */
  void resetVideoDisplay();

  /**
   * Close the video loaded by #loadVideo().
   */
  void closeVideo();

  /** Stop audio play-back. 
   * Make sure that #globalTimer is not paused before calling this method.
   */
  void stopAudio();

  /** Stop video and audio play-back.
   * Make sure that #globalTimer is not paused before calling this method.
   */
  void stopVideoAudio();

  /** Pause video and audio play-back.
   * The DataChannel threads will eventually terminate, leaving the input stream
   * open. This function does not wait for the DataChannel threads to terminate.
   */
  void pauseVideoAudio();

  /**
   * Start video play-back at a given absolute media time point.
   *
   * @param sec    media time point in seconds; 0.0 means start of video,
   *               -1.0 means &quot;continue from current time point&quot;.
   * @return <code>true</code> on success, <code>false</code> on error.
   */
  bool playVideo(double sec = -1.0);

  /**
   * Get current video media time stamp in seconds.
   * Returns <code>0.0</code> if the time stamp cannot be determined.
   */
  double currentVideoTime() const;

  /** Set the current media time label to the given value. 
   * @param sec          media time in seconds; must be non-negative.
   * @param showSubSec   determines whether sub-second fractions will
   *                     be shown.
   */
  void setCurrentVideoTimeLabel(double sec, bool showSubSec = false);

  /** Set the display rate slider control to the given rate.
   * @see #setDisplayRateFromSlider(int)
   */
  void setDisplayRateSlider(double rate);

 private:
  /** The DataChannel belonging to the visual elementary stream of a loaded video. */
  DataChannel *videoDC;

  /** The DataChannel belonging to the audio elementary stream of a loaded video. */
  DataChannel *audioDC;

  /** The global timer controlling the DataChannels. */
  GlobalTimer *globalTimer;

  /** The reference counter for using #globalTimer as a SharedAdaptor. */
  ReferenceCounter *globalTimerRefCnt;

  /** The ESSynchronizer controlling the #videoDC DataChannel. */
  ESSynchronizer *essync;

  /** The timer for updating the current media time label. */
  QTimer *mediaTimeUpdater;
};

#endif
