
 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	transcode.cpp                                                 *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#include <assert.h>
#include "DataChannel.hpp"
#include "adaptors/MP4Decoder.hpp"
#include "DataSink.hpp"
#include "io/MPGStreamIO.hpp"
#include "io/RawStreamIO.hpp"
#include "io/YUVStreamIO.hpp"
#include "io/DevNull.hpp"
#include "io/IOCreator.hpp"
#include "io/UnbufferedFileByteStream.hpp"
#include "io/BufferedIO.hpp"
#include "io/VITMuxDemuxIO.hpp"
#include "io/OGGMuxDemuxIO.hpp"
#include "ContainerInfo.hpp"
#include "VideoESInfo.hpp"
#include "VideoAdaptorChain.hpp"
#include "metadata/TerminalCapabilities.hpp"
#include "Statistics.hpp"
#include "adaptors/YUVSpatialReductionAdaptor.hpp"
#include "adaptors/YUVScalingAdaptor.hpp"
#include "adaptors/YUVinYUVoverlay.hpp"
#include "adaptors/FrameRotate.hpp"
#include "adaptors/Visualizer.hpp"
#include "adaptors/DataDump.hpp"
#include "adaptors/FrameRateAdjust.hpp"
#include "adaptors/MadMP3Decoder.hpp"
#include "adaptors/MP3Encoder.hpp"
#include "adaptors/MP4Encoder.hpp"
#include "ReferenceCounter.hpp"
#include "adaptors/SharedAdaptor.hpp"
#include "adaptors/ESSynchronizer.hpp"
#include "adaptors/GlobalTimer.hpp"
#include "adaptors/TheoraDecoder.hpp"
#include "adaptors/TheoraEncoder.hpp"
#include "adaptors/VorbisDecoder.hpp"
#include "adaptors/VorbisEncoder.hpp"
#include "AudioAdaptorChain.hpp"
#include "AudioESInfo.hpp"

#ifdef WIN32
#include <crtdbg.h>
#endif
#include "vitooki.hpp"
#include "container/ContainerFile.hpp"
#include "container/Video4LinuxContainerFile.hpp"


void printHelp(const char* name) {
  printf("%s [--v4l] [--add-quit-button FILE.yuv width height] [--rgb] [--rotate] [--mirror] [--decode-video] [--decode-audio] [--out-fps NN] [--encode-video KBPS] [--no-video] [--encode-audio KBPS] [--no-audio] [--write-vit-container] [--write-ogg-container] [--no-visualize] <input.mp4/.avi/.vit> <out-prefix> [width] [height]\r\n",name);
  printf("Option --v4l: use Video 4 Linux card as input.\r\n");	
  printf("Option --add-quit-button: adds any YUV stream (eg. an X-button) in the upper right corner)\r\n");
  printf("Option --rgb: decode to RGB32 instead of YUV.\r\n");
  printf("Option --rotate: rotates the result by 90 degrees.\r\n");
  printf("Option --mirror: mirrors the result by 180 degrees.\r\n");
  printf("        combining --rotate and --mirror results in a 270 degree rotation\r\n");
  printf("Option --decode-video: if not set, video is bypassed and stored directly, else raw YUV is stored\r\n");
  printf("Option --decode-audio: if not set, audio is bypassed and stored directly, else raw PCM is stored\r\n");
  printf("Option --out-fps NN: sets NN as the new output framerate, where incoming frames are dropped/doubled\r\n");
	printf("Option --encode-video KBPS: encodes the decoded YUV to MPEG-4 with KBPS kilobit/sec\n");
	printf("Option --no-video: ignores the video in the input file and only writes (the transcoded) audio\n");
	printf("Option --encode-audio KBPS: encodes the decoded MP3 (or raw PCM input) to MP3 with KBPS kilobit/sec\n");
	printf("Option --no-audio: ignores the audio in the input file and only writes (the transcoded) video\n");
	printf("Option --write-vit-container: writes an interleaved .vit container, uses the out-prefix and postfixes a .vit\n");
	printf("Option --write-ogg-container: writes an interleaved .ogg container, uses the out-prefix and postfixes a .ogg\n");
	printf("Option --no-visualize: will be quiet and will not open a SDL window showing the video\n");
  printf("if width and height is given, the yuv is scaled (eg. from SD to 352x288 or vice versa)!\r\n");
  printf("if more than one MPEG-4 video stream is in the .mp4 container, all will be extracted\r\n");
  printf("             and numbered according to their stream-IDs\r\n");
  printf("note, that the output filenames will look like:  out-prefix-01.yuv .. out-prefix-05.yuv\r\n");
}



int main(int argc, char* argv[]) {
  char *progname = argv[0];
  int opt_rgb = 0;		// '--rgb' option set?
	bool enableVideo4Linux = 0;  
	int opt_rotate = 0;
	int opt_quit_button = 0;
	int quit_button_width, quit_button_height;
	char *quit_button_file;
	int opt_change_rate = 0;
	int opt_mirror = 0;
	int opt_no_audio = 0;
	int opt_no_video = 0;
	int opt_decode_video =0;
	int opt_decode_audio =0;
	int opt_encode_video_kbps = 0;
	int opt_encode_audio_kbps = 0;
	int opt_visualize = 1;
  char *input_file = NULL;
  char *out_prefix = NULL;
	int opt_vit = 0;
	int opt_ogg = 0;
  int do_scale = 0;
  int width = 0;
  int height = 0;
	int out_fps = 0;
  const char yuv_suffix[] = ".yuv";
  const char m4v_suffix[] = ".m4v";
  const char rgb_suffix[] = ".rgb";

  const char mp3_suffix[] = ".mp3";
  const char pcm_suffix[] = ".pcm";

  const char vit_suffix[] = ".vit";
  const char mkv_suffix[] = ".mkv";
  const char ogg_suffix[] = ".ogg";

  const char *video_suffix = m4v_suffix;
  const char *audio_suffix = mp3_suffix;
	const char *container_suffix = NULL;
	DataChannel* vdc = NULL;
	DataChannel* adc = NULL;
	ReferenceCounter *ref; //used for video+audio stream
	GlobalTimer *gt; //used for video+audio stream
	ESSynchronizer *esync;
	Adaptor *sharedAdapt = NULL, *sharedAdapt2 = NULL;
	MuxDemuxIO *muxio;
 
#ifdef WIN32
  int tmpFlag = _CrtSetDbgFlag( _CRTDBG_REPORT_FLAG );
  tmpFlag |= _CRTDBG_LEAK_CHECK_DF;
  _CrtSetDbgFlag( tmpFlag );
#endif

  while (--argc > 0 && (*++argv)[0] == '-') {
    if (strcmp(*argv, "--v4l") == 0) {
      enableVideo4Linux = 1;
    } else
    if (strcmp(*argv, "--rgb") == 0) {
			if (opt_quit_button) {
				fprintf(stderr,"Options --rgb and --add-quit-button dont fit together, the button would use YUV!\n");
				exit(1);
			}
      opt_rgb = 1;
      video_suffix = rgb_suffix;
    } else 
    if (strcmp(*argv, "--add-quit-button") == 0) {
			if (opt_rgb) {
				fprintf(stderr,"Options --rgb and --add-quit-button dont fit together, the button would use YUV!\n");
				exit(1);
			}
      opt_quit_button = 1;
			quit_button_file = argv[1];
			--argc;
			++argv;
			quit_button_width = atoi(argv[1]);
			--argc;
			++argv;
			quit_button_height = atoi(argv[1]);
			--argc;
			++argv;
    } else 
    if (strcmp(*argv, "--rotate") == 0) {
      opt_rotate = 1;
    } else 
    if (strcmp(*argv, "--mirror") == 0) {
      opt_mirror = 1;
    } else 
			if (strcmp(*argv, "--no-video") == 0) {
				opt_no_video = 1;
    } else 
			if (strcmp(*argv, "--no-audio") == 0) {
				opt_no_audio = 1;
			} else 
			if (strcmp(*argv, "--decode-video") == 0) {
				opt_decode_video = 1;
			} else 
			if (strcmp(*argv, "--decode-audio") == 0) {
				opt_decode_audio = 1;
			} else 
				if (strcmp(*argv, "--out-fps") == 0) {
					opt_change_rate = 1;
					out_fps = atoi(argv[1]); //get fps value
					--argc;
					++argv;
				} else 
					if (strcmp(*argv, "--encode-video") == 0) {
						opt_encode_video_kbps = atoi(argv[1]) * 1024; //get KBPS value
						video_suffix = m4v_suffix;
						--argc;
						++argv;					
					} else
						if (strcmp(*argv, "--encode-audio") == 0) {
							opt_encode_audio_kbps = atoi(argv[1]) * 1000; //get KBPS value
							audio_suffix = mp3_suffix;
							--argc;
							++argv;					
						} else
							if (strcmp(*argv, "--write-vit-container") == 0) {
								opt_vit = 1;
								container_suffix = vit_suffix;
							} else
								if (strcmp(*argv, "--write-ogg-container") == 0) {
									opt_ogg = 1;
									container_suffix = ogg_suffix;
								} else
									if (strcmp(*argv, "--no-visualize") == 0) {
										opt_visualize = 0;
									} else {
										printf("Unknown option: %s\n", *argv);
										printHelp(progname);
										exit(1);
									}
	}

	if (opt_decode_video && (opt_encode_video_kbps == 0)) //result is not encoded again...
		video_suffix = yuv_suffix;
	if (opt_decode_audio && (opt_encode_audio_kbps == 0)) //result is not encoded again...
		audio_suffix = pcm_suffix;

  if (argc != 2 && argc != 4) {
    printHelp(progname);
    exit(1);
  }

	//either .vit or .ogg. not both!
	assert((!opt_vit && !opt_ogg) || (opt_vit && !opt_ogg) || (!opt_vit && opt_ogg));


  input_file = argv[0];
  out_prefix = argv[1];
  if (argc == 4) {
    do_scale = 1;
    width = atoi(argv[2]);
    height = atoi(argv[3]);
  }

  vitooki_init_codecs();
   printf("Inputfile: %s\n", input_file);
  ContainerInfo* contInfo = NULL;

  if (!enableVideo4Linux)
  	contInfo=ContainerFile::loadContainerInfo(input_file);
  else
		{	input_file = "";
			contInfo=Video4LinuxContainerFile::loadContainerInfo();
		}
  if(!contInfo) {
    printf("Inputfile %s not found\r\n", input_file);
    exit(1);
  }
	

	//	ref = new ReferenceCounter(); //used for video+audio stream
	//gt = new GlobalTimer(); //used for video+audio stream

	char container_outfile[255];
	if (container_suffix == vit_suffix) {  //pointer comparison!
		sprintf(container_outfile,"%s%s", out_prefix, container_suffix);
		muxio = new VITMuxDemuxIO(new UnbufferedFileByteStream(container_outfile,false));
	} 
#ifdef ENABLE_OGG
	else
		if (container_suffix == ogg_suffix) {  //pointer comparison!
			sprintf(container_outfile,"%s%s", out_prefix, container_suffix);
			muxio = new OGGMuxDemuxIO(new UnbufferedFileByteStream(container_outfile,false));
		}
#endif

  list < ESInfo * >* streams = contInfo->getESList();
  list < ESInfo * >::const_iterator li = streams->begin();
  while (li != streams->end()) {
		dprintf_full("simpleadapt working on STREAM id %llu  of %i\n",(*li)->getStreamId(),streams->size());

		/**************** VIDEO ******************/
    if ((*li)->isVisualStream() && !opt_no_video) {
      char outfile[255];
			IO *output;
		 
      VideoESInfo* in_ves=(VideoESInfo*)(*li);
      VideoESInfo* out_ves=(VideoESInfo*)in_ves->clone(contInfo);

      sprintf(outfile,"%s-%09llu%s", out_prefix, in_ves->getStreamId(), video_suffix);
       dprintf_full("Outputfile: %s\n", outfile);      

      //	Statistics *stats = new Statistics(in_ves,false); //input stats
       IO* input=NULL;
       input=IOCreator::createInput(in_ves, false);
       assert(in_ves && input);
       VideoAdaptorChain *achain = new VideoAdaptorChain();
       
			UncompressedVideoFrame::ColorSpace cs;
			if (opt_rgb)
				cs = UncompressedVideoFrame::ColorSpaceRGB32;
			else
				if(enableVideo4Linux)
					cs = UncompressedVideoFrame::ColorSpaceI420;
				else
					cs = UncompressedVideoFrame::ColorSpaceYV12;


			//be intelligent and decode first, if needed....
      if ((in_ves->getCodecID() != CODEC_ID_RAWVIDEO) && (opt_encode_video_kbps > 0))
				opt_decode_video = 1;
					

      if ((in_ves->getCodecID() != CODEC_ID_RAWVIDEO) && (opt_decode_video)) {	
	/* by tkuglits:
	 * check which Decoder to use!
	 * added TheoraDecoder
	 */
	if(in_ves->getCodecID() == CODEC_ID_THEORA){
	  TheoraDecoder *theod = new TheoraDecoder(in_ves);
	  achain->addAdaptor(theod);
	}else{
	  MP4Decoder *mp4d = new MP4Decoder(in_ves,false,cs);
	  achain->addAdaptor(mp4d);
	}
				out_ves->setCodecID(CODEC_ID_RAWVIDEO);
      }


			if (opt_quit_button > 0) {
				int yuvX = quit_button_width;
				int yuvY = quit_button_height;
				VideoESInfo *watermarkES=new VideoESInfo(99,0,(ContainerInfo*)NULL,3600,(u8*)NULL,0,0,0,0,90000,0,(int)(yuvX*yuvY*1.5),yuvX,yuvY);
				YUVinYUVoverlay *yuvinyuv = new YUVinYUVoverlay(in_ves,new YUVStreamIO(watermarkES, quit_button_file, false), 
																												//put watermarkX in upper right corner
																												in_ves->getWidth()-yuvX, 0,
																												true);
				achain->addAdaptor(yuvinyuv);
			}

			if (opt_change_rate > 0) {
				achain->addAdaptor(new FrameRateAdjust(in_ves, in_ves->getMediaTimeScale() / out_fps));
				out_ves->setFPS(out_fps);
			} else
				out_fps = (int)ceil((in_ves)->getFPS()); //eg. 29.33 --> 30


      if (do_scale) {
				YUVScalingAdaptor *ysr = new YUVScalingAdaptor(in_ves, width, height);
				out_ves->setWidth(width);
				out_ves->setHeight(height);
				out_ves->setAspectRatio(out_ves->getWidth()/out_ves->getHeight());
				achain->addAdaptor(ysr);
      }

			/*
				if (opt_rotate) { //check, if width/height is not swapped...
					if ( ((in_ves->getWidth() > in_ves->getHeight()) && (out_ves->getWidth() > out_ves->getHeight())) ||
							 ((in_ves->getWidth() < in_ves->getHeight()) && (width < height)) ) {
						u32 tmp = out_ves->getHeight();
						out_ves->setHeight(out_ves->getWidth());
						out_ves->setWidth(tmp);
						dprintf_full("simpleadapt: Param width/height given in wrong order, since the video is also rotated.... SWAPPING to %ix%i!\n",width,height);
					}
				}
			*/

			if (opt_rotate && opt_mirror) { //270 degrees
				achain->addAdaptor(new FrameRotate(out_ves, 270));
				u32 tmp = out_ves->getHeight();
				out_ves->setHeight(out_ves->getWidth());
				out_ves->setWidth(tmp);
			} else if (opt_rotate) { //90 degree
				achain->addAdaptor(new FrameRotate(out_ves, 90));
				u32 tmp = out_ves->getHeight();
				out_ves->setHeight(out_ves->getWidth());
				out_ves->setWidth(tmp);
				out_ves->setAspectRatio(out_ves->getWidth()/out_ves->getHeight());
			}	else if (opt_mirror) { // 180 degrees
				achain->addAdaptor(new FrameRotate(out_ves, 180));
			}

			/*
			sharedAdapt = new SharedAdaptor(ref, (Adaptor*&)gt);
			achain->addAdaptor(sharedAdapt);
			esync = new ESSynchronizer(in_ves, gt);
			achain->addAdaptor(esync);
			*/

			if (out_ves->getCodecID() != CODEC_ID_RAWVIDEO) {
				dprintf_err("WARNING: Cannot visualize encoded video stream, use option --decode-video or transcode it!\n");
				opt_visualize = 0;
			}
			if (opt_visualize)
				achain->addAdaptor(new Visualizer(out_ves, Visualizer::SDL, cs ));

			if(opt_encode_video_kbps > 0 && opt_ogg){ // if encode && ogg-container, use TheoraEncoder
			  out_ves->setAvgBandwidth(opt_encode_video_kbps);
			  out_ves->setCodecID(CODEC_ID_THEORA);
			  achain->addAdaptor(new TheoraEncoder(out_ves) );
			}
			else
			if (opt_encode_video_kbps > 0) { //encode, so add MP4Encoder adaptor and use MPGStreamIO output
			  int gop_size = out_fps * 5; //an I_VOP every 5 seconds
			  int num_bvops = 0; // numBVOPs between P
			  achain->addAdaptor(new MP4Encoder(out_ves, opt_encode_video_kbps, out_fps,
							    gop_size /* GOP size */,  num_bvops /* numBVOPs */, 
							    cs,
							    MP4Decoder::FFMPEG, CODEC_ID_MPEG4));
			  out_ves->setAvgBandwidth(opt_encode_video_kbps);
			  out_ves->setCodecID(CODEC_ID_MPEG4);
			}
			
			if (opt_vit || opt_ogg) { //write .vit/.ogg container
			  output = new BufferedIO(out_ves, muxio, true);
			} else
			  if (opt_encode_video_kbps == 0) //write raw YUV
			    output = new YUVStreamIO(out_ves, outfile, true);
			  else 
			    output = new MPGStreamIO(out_ves, outfile, true);
			
			
			if (vdc == NULL) {
				vdc=new DataChannel(input, in_ves, achain);
				vdc->insert(new DataSink(0, output, "127.0.0.1", 0));
				vdc->play(0, 0.0);
				vdc->start();
			} else {
				fprintf(stderr,"simpleAdapt can only handle ONE videostream per container!\n");
				::exit(1);
			}
			

			/**************** AUDIO ******************/
    } else
			if ((*li)->isAudioStream() && !opt_no_audio) {
			  dprintf_full("simpleAdapt: found AudioStream\n");
				char outfile[255];
				IO *output;
				AudioESInfo* in_aes=(AudioESInfo*)(*li);
				AudioESInfo* out_aes=(AudioESInfo*)in_aes->clone(contInfo);
				sprintf(outfile,"%s-%09llu%s", out_prefix, in_aes->getStreamId(), audio_suffix);
				
				//	Statistics *stats = new Statistics(es,false); //input stats
				IO* input=NULL;
				input=IOCreator::createInput(in_aes, false);
				assert(in_aes && input);
				AudioAdaptorChain *achain = new AudioAdaptorChain();

				dprintf_full("simpleAdapt: opt_enc_audio = %d\n", opt_encode_audio_kbps);
				dprintf_full("simpleAdapt: MP2(%i), MP3(%i), VORBIS(%i) == %i\n",
					CODEC_ID_MP2, CODEC_ID_MP3, CODEC_ID_VORBIS, in_aes->getCodecID());
				if (opt_decode_audio || 
				    (
				     ( in_aes->getCodecID() == CODEC_ID_MP2 || 
				       in_aes->getCodecID() == CODEC_ID_MP3 ||
				       in_aes->getCodecID() == CODEC_ID_VORBIS
				       )
				     && (opt_encode_audio_kbps > 0) 
				     )
				    ) { // transcode from mp2/3 or vorbis
				  dprintf_full("simpleAdapt: Decoding Audio\n");

				  if(in_aes->getCodecID() == CODEC_ID_MP2 ||
				     in_aes->getCodecID() == CODEC_ID_MP3){
				    dprintf_full("simpleAdapt: adding MadMP3Decoder\n");
				    achain->addAdaptor(new MadMP3Decoder(in_aes));
				  }
				  else if(in_aes->getCodecID() == CODEC_ID_VORBIS){
				    dprintf_full("simpleAdapt: adding VorbisDecoder\n");
				    achain->addAdaptor(new VorbisDecoder(in_aes));
				  }
				  out_aes->setCodecID(CODEC_ID_PCM_S16BE); //FIXME: big/little endian: or is it S16LE ?
				}

				/*				
				sharedAdapt2 = new SharedAdaptor(ref, (Adaptor*&)gt);
				achain->addAdaptor(sharedAdapt2);
				*/

				if(opt_encode_audio_kbps > 0 && opt_ogg){ //encode ogg-vorbis
				  out_aes->setAvgBandwidth(opt_encode_audio_kbps);
				  out_aes->setCodecID(CODEC_ID_VORBIS);

				  achain->addAdaptor(new VorbisEncoder(out_aes, opt_encode_audio_kbps));
				}
				else if (opt_encode_audio_kbps > 0) { //encode mp3
					achain->addAdaptor(new MP3Encoder(in_aes, opt_encode_audio_kbps));
					out_aes->setAvgBandwidth(opt_encode_audio_kbps);
					out_aes->setCodecID(CODEC_ID_MP3);
				}

				if (opt_vit || opt_ogg) {
					output = new BufferedIO(out_aes, muxio, true);
				} else
					if (opt_encode_audio_kbps == 0) //write raw pcm
						//note: fileSize per frame is determined by AU->size
						output = new RawStreamIO(out_aes, outfile, true);
					else { //encode mp3, use MPGStreamIO output
						output = new MPGStreamIO(out_aes, outfile, true);
					}
				
				if (adc == NULL) {
					adc=new DataChannel(input, in_aes, achain);
					adc->insert(new DataSink(0, output, "127.0.0.1", 0));
					adc->play(0, 0.0);
					adc->start();
				} else {
					fprintf(stderr,"simpleAdapt can only handle ONE audiostream per container!\n");
					::exit(1);
				}
				
			} else {
				dprintf_err("simpleadapt IGNORING Unknown Elementary Stream!!!\n");
			}
    ++li;
  }
	
	if (adc) {
		adc->wait();
		delete adc;
	}

	if (vdc) {
		vdc->wait();
		delete vdc;
	}
	
  delete contInfo;
  return 0;
}


