 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	yuvtompeg.cpp                                                 *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/


#include "vitooki.hpp"
#include "global.hpp"
#include "ContainerInfo.hpp"
#include "AdaptorChain.hpp"
#include "VideoESInfo.hpp"
#include "VideoAdaptorChain.hpp"
#include "Statistics.hpp"
#include "container/ContainerFile.hpp"
#include "adaptors/MP4Decoder.hpp"
#include "adaptors/MP4Encoder.hpp"
#include "io/IOCreator.hpp"
#include "io/MPGStreamIO.hpp"
#include "io/MKIO.hpp"
#include "io/DevNull.hpp"
#include "DataChannel.hpp"
#include "DataSink.hpp"
#include "adaptors/Visualizer.hpp"


void printHelp(const char* name) {
		printf("USAGE:   %s YUVFILE  WIDTH  HEIGHT  FPS  targetBitrate(kbit/sec) [--visualize]\n\n",name);
		printf("         --visualize will show the input YUV on the screen\n");
}


int main(int argc, char *argv[]) {
	int width,height,fps;
	char *filename;
	char *progname;
	char outfile[512];
	int duration;
	int filesize;
	int framesize;
	int bitrate;
	FILE *fp;
	int ticks;
	bool visualize = false;
	int loop_duration = 0;

	progname = argv[0];
	if (argc < 6) {
		printHelp(progname);
		exit(1);
	}
	--argc;	++argv;

	filename=argv[0];	--argc;	++argv;
	width=atoi(argv[0]); --argc;	++argv;
	height=atoi(argv[0]); --argc;	++argv;
	fps=atoi(argv[0]); --argc;	++argv;
	bitrate = atoi(argv[0]) * 1024; --argc;	++argv;

	

  while (--argc > 0 && (*++argv)[0] == '-') {
		if (strcmp(*argv, "--visualize") == 0) {
			visualize=true;
		} else
			if (strcmp(*argv, "--loop") == 0) {
				loop_duration = atoi(argv[1]);	
				--argc;
				++argv;
			} else {
				printf("Unknown option: %s\n", *argv);
				printHelp(progname);
				exit(1);
			}
	}

	framesize=(int)(width*height*1.5);
	ticks=90000/fps;
	if (!(fp=fopen(filename,"r"))) {
		perror("opening yuv file");
		exit(1);
	}
	fseek(fp,0,SEEK_END);
	duration=(ftell(fp)/framesize)*ticks;
	filesize=ftell(fp);
	fclose(fp);
	
	vitooki_init_codecs();

	ContainerInfo* containerInfo=new ContainerInfo(NULL, filename, ContainerInfo::MP4);
	//	ContainerInfo* containerInfo=new ContainerInfo(NULL, filename, ContainerInfo::MATROSKA);
	if(!containerInfo) {
		printf("Cannot open inputfile\n");
		return 1;
	}//if

	AdaptorChain* ach=new VideoAdaptorChain();
	VideoESInfo* es=new VideoESInfo(1,0,containerInfo,ticks,NULL,0,0,
					framesize*fps*8,framesize*fps*8,
					90000,duration,filesize,width,height,
					true, true,
					CODEC_ID_MPEG4, 1.0,
					(float)width/(float)height);
	containerInfo->addESInfo(es); // add this ESInfo to the container of ESInfos

	VideoESInfo *in_es = containerInfo->getFirstVisualES();

	if (visualize) {		
		ach->addAdaptor(new Visualizer((VideoESInfo*)in_es, Visualizer::SDL, 
																	 UncompressedVideoFrame::ColorSpaceYV12 ));
	}
	
	ach->addAdaptor(new MP4Encoder(es, bitrate, (int)es->getFPS(),  30 /* GOP size */,  0 /* numBVOPs */, 
		UncompressedVideoFrame::ColorSpaceYV12,
		MP4Decoder::FFMPEG, CODEC_ID_MPEG4));


	IO* input = IOCreator::createInput((ESInfo*)in_es);

	assert(es && input);
	DataChannel* dc=new DataChannel(input, es, ach);
	
	strcpy(outfile, filename);
	strcpy(strstr(outfile,".yuv"), ".m4v");
	IO *output = NULL;
	output = new MPGStreamIO(es, outfile, true);

#ifdef ENABLE_MKV
	//for testing of Matroska
	//	strcpy(strstr(outfile,".m4v"), ".mkv");
	//output = new MKIO(es, outfile, true);
#endif

	dprintf_full("writing out to %s\n",outfile);
	//note: fileSize per frame is determined by AU->size

	dc->insert(new DataSink(0, output, "127.0.0.1", 0));
	dc->play(0, 0.0);
	dc->start();
	dc->wait();
	delete dc;

	delete containerInfo; //ESInfo is deleted with containerInfo....
	return 0;
}
