/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:   AudioESInfo.cpp                                             *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

#include "AudioESInfo.hpp"
#include "adaptors/MP4audioDecoder.hpp"
#include "ReferenceCounter.hpp"
#include "BitField.hpp"
	/** creates an AudioESInfo object
	  * @param encodeddecconf is only shallow copied and
	  * freed in the destructor
	  * @param avgBW and @param maxBW specify the bandwidth of the given ES
	  * and are in Bits/sec!
	*/
AudioESInfo::AudioESInfo(u32 streamId, u32 streamtype, ContainerInfo * vob, u32 timeIncrement,
												 u8 * encodeddecconf, u32 objtype, u32 bufsize,
												 u32 avgBW, u32 maxBW, u32 timeScale,
												 u64 duration, u32 size,  
												 u32 samplerate, u32 numAudioChannels,
												 bool complete, 
												 CodecID codec_id, float quality, u32 BitsPerSample)
	: ESInfo(streamId,
					 streamType,vob,timeIncrement,encodeddecconf,
					 objtype,bufsize,avgBW,maxBW,
					 timeScale,duration,size,complete,codec_id,quality)
{

  //the following are for audio only
  sample_rate = samplerate;
  audio_channels = numAudioChannels;
  /* MP3 specific channelMod is unsed so far */
  channelMod=0;
  bitsPerSample = BitsPerSample;
	generateDescriptionText();
  /* This is used to set up certain information by parsing the header 
   * of first mp3 frame 
#ifndef ISOMP4 
  MP4audioDecoder *mp4audio = new MP4audioDecoder(this,false);
  mp4audio->initialize();
  delete mp4audio;
#endif
*/
}


u32 AudioESInfo::getHandlerType() const {
	return MP4AudioHandlerType;
}


u32 AudioESInfo::getBitsPerSample() const {
	return bitsPerSample;
}


u32 AudioESInfo::getSampleRate() const {
	return sample_rate;
}


void AudioESInfo::setSampleRate(u32 sampleRate)  {
	sample_rate = sampleRate;
	generateDescriptionText();
}


u32 AudioESInfo::getAudioChannels() const {
	return audio_channels;
}


void AudioESInfo::setAudioChannels(u32 channels) {
	audio_channels = channels;
}


bool AudioESInfo::isVisualStream() const {
	return false;
}


bool AudioESInfo::isAudioStream() const {
	return true;
}


bool AudioESInfo::isODStream() const {
	return false;
}


bool AudioESInfo::isSceneDescriptionStream() const {
	return false;
}


/** returns type of audio -> stereo, mono etc */
u32 AudioESInfo::getChannelMod() const {
	return channelMod;
}


void AudioESInfo::setChannelMod(u32 uiChannelMod) {
	channelMod = uiChannelMod;
	generateDescriptionText();
}


/** creates a deep copy of the ES */
ESInfo *AudioESInfo::clone(ContainerInfo * mp4) {
	dprintf_full("AudioESInfo::clone id %llu\n",streamId);
	u8 *decConf = NULL;
	if (encodedDecoderConfig)
	{ 	
		decConf = new u8[strlen((char *) encodedDecoderConfig) + 1];
		strcpy((char *) decConf, (const char *) encodedDecoderConfig);
	}
	ESInfo *es = new AudioESInfo(streamId, streamType, mp4, vopTimeIncrement, decConf, objectType,
															 bufferSize, avgBandwidth, maxBandwidth,
															 mediaTimeScale, duration, mediaSize, sample_rate,
															 audio_channels, complete, codec_id, this->quality,
															 bitsPerSample);
	es->setInput(input);
	if(!existingClones) {
		existingClones=new ReferenceCounter();
		existingClones->increase();
	}
	existingClones->increase();
	es->setCloneCounter(existingClones);
	BitField* fs=getFrameStatistic();
	if(fs) {
		es->setFrameStatistic(fs->createWithNewSize(fs->getSize(),false));
	}
	((AudioESInfo*)es)->setChannelMod(channelMod);
	return es;
}


void AudioESInfo::generateDescriptionText() {
	char *codecName = new char[MAX_STR_LEN];
	strcpy(codecName,"Unknown");
#ifdef ENABLE_FFMPEG
	AVCodec* av=avcodec_find_decoder(this->codec_id);
	if(av)
		strcpy(codecName,av->name);
#endif
	snprintf(description,MAX_STR_LEN,"%10s: %4ukBps; %5uHz; %3uBit; %1uCh.",
					 codecName,avgBandwidth/1024,sample_rate,bitsPerSample,channelMod);

	delete [] codecName;
}
