 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	BitField.hpp                                                  *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef PS_MPEG4ITEC_BITFIELD
#define PS_MPEG4ITEC_BITFIELD
#include"global.hpp"
#ifdef ISOMP4
#include "ISOMovies.h"
#endif

/**
 *
 * <Class Description goes here>
 *
 *@author Michael Kropfberger and Peter Schojer
 *
 *@version $Id: BitField.hpp,v 1.4 2004/05/19 22:19:01 mkropfbe Exp $
 *
 *@short <short description>
 *
 */
class BitField {
private:	
	u32* bits;	
	u32 maxSize;
public:
	BitField(char* data, u32 size);
	BitField(u32 size, bool initValue);
	virtual ~BitField();
	/** sets the bit at position @param pos to 1. if pos 
	points to an invalid bit, the set is ignored */
	void setBit(u32 pos);
	/** sets the bit at position @param pos to 0. if pos 
	points to an invalid bit, the clear is ignored */
	void clearBit(u32 pos);
	/** gets the bit at position @param pos. if pos 
	points to an invalid bit, always false is returned */
	bool getBit(u32 pos) const;	
	
	/** inverts all 1 to 0 and all 0 to 1 */
	void invert();
	
	/** returns @param nrBits from the bitfield, starting with bitPos @param startPos.
	 * if nrBits is an invalid value (0 or greater 32) or if the specified
	 * bit range exceeds the size of the bitfield, @param nrBits is set to -1.
	 * @param littleEndian specifies endianness. When set to true the decimal value for the
	 * requested bitrange is calculated as if the first bit has the lowest value.
	 * e.g.: Bit 0 1000 Bit 3 -> littleEndian=true --> 1 littleEndian=false --> 8
	 * If everything went fine nrBits returns the number of bits used to create
	 * @param return, otherwise @param return is 0 and @param nrBits is -1
	 */
	u32 getBitRange(u32 startPos, int* nrBits, bool littleEndian=true);
	
	/* writes @param nrBits of @param value starting with @param startBit at position @param startPos.
	 * returns false if startBit+nrBits is greater 32, if nrBits is not in range 1..32,
	 * if startBit is not in range 0..32-nrBits, or startPos+nBits is greater than maxSize
	 * @param littleEndian specifies endianness. When set to true the decimal value for the
	 * requested bitrange is calculated as if the first bit has the lowest value.
	 * e.g.: Bit 0 1000 Bit 3 -> littleEndian=true --> 1 littleEndian=false --> 8
	 */
	bool writeBitRange(u32 startPos, int nrBits, u32 value, u32 startBit, bool littleEndian=true);
	/** returns true if the bitfield consists completely of 1s */
	bool isAllSet() const;
	/** returns true if the bitfield consists completely of 0s */
	bool isAllCleared() const;
	/** returns the percentage of 1s inthe bitfield. Expensive!!! */
	float getPercentageSetBits() const;

	/** creates a new BitField with new size @param size.
	  * If the new size is smaller than the current size,
		* the old values are copied from pos 0 to newSize.
		* if the new size is the same than the current size,
		* this method has clone functionality.
		* if the new size is larger than the current size,
		* the current data is cloned and the values afterwards
		* are set to @param initValueForMissing
		*/
	BitField* createWithNewSize(u32 size, bool initValueForMissing);
	u32 getSize() const { return maxSize;}
	/** simply returns a pointer to the internal bitfield.
	  * returns the size which is multiple of sizeof(u32) */
	const char* toBinaryString(u32* sizeInU32) const;
	/** converts the bitfield into a human viewer 
	friendly 01 ASCII presentation. Time expensive but endian save! */
	char* toAsciiString() const;
	/** converts the bitfield into a hex representation of the form
		* size_0123456789ABCDEF...
		* Example: A Bitfield with the size 9, data:(bit 0) 000111110 (bit 8)
		* is in hex: 00000009_8F0
	*/
	char* toHexString() const;
	/** creates a Bitfield from a string of chars '0' and '1'.
		* Time expensive but endian save! */
	static BitField* fromAsciiString(const char* str);
	/** creates a Bitfield from a binary data pointer. 
	  * fast but not endian save */
	static BitField* fromBinaryString(const char* bin, u32 numBits);
	/** same as above but faster because only a shallow-copy is performed */
	static BitField* fromBinaryString(char* bin, u32 numBits);
	static BitField* fromHexString(const char* bin);
	/** automatic testing of BitField functionality, throws assertion at any error found */
	static void doTest();
};
#endif

	
