 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title: ContainerInfo.hpp                                             *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef PS_CONTAINERINFO_HPP
#define PS_CONTAINERINFO_HPP

#ifdef ISOMP4
#include "ISOMovies.h"
#endif
#ifndef WINCE
#include <fcntl.h>
#endif
#include "global.hpp"

#include "Adaptor.hpp"
#include "IO.hpp"
#include "net/Url.hpp"
struct rtx_group;
struct rtx_info;
class MetaObject;
class ESInfo;
class TerminalCapabilities;
class ReferenceCounter;
class VideoESInfo;
class AudioESInfo;

#define MP4CONF_VERSION 7

struct MetaItem {
	/** creates a MetaItem for the given ES, with quality q, size s and
	  * an adaptor used to create the adapted stream
	  */
	MetaItem(u32 es, float q, u32 s, Adaptor * a) {
		quality = q;
		esId = es;
		size = s;
		adapt = a;
	};
	~MetaItem() {
		if (adapt)
			delete adapt;
	}
	Adaptor *getClonedAdaptor() {
		if (adapt)
			return adapt->clone();
		else
			return NULL;
	};
	MetaItem *clone() {
		return new MetaItem(esId, quality, size, getClonedAdaptor());
	};

	float quality;
	u32 size;
	Adaptor *adapt;
	u32 esId;
};

struct QualityInfo {
	/** creates a qualityInfo struct with quality=1.0, the given size, and the maximum number
	 * of quality steps this ContainerInfo might have */
	QualityInfo(u32 size, u8 maxSteps = 5) {
		lastRequestTime = 0;
		currentQAStep = 0;
		maxQASteps = 1;
		stepLimit = maxSteps;
		steps = new MetaItem *[stepLimit];
		// init array to NULL
		memset((char*)steps,0,sizeof(MetaItem*)*stepLimit);
		// the first entry is the original video, no need for a real Adaptor
		steps[currentQAStep] = new MetaItem(0, 1.0, size, NULL);
	};
	/** creates a deep-copy from qi */
	QualityInfo(QualityInfo & qi) {
		lastRequestTime = 0;
		currentQAStep = qi.currentQAStep;
		maxQASteps = qi.maxQASteps;
		stepLimit = qi.stepLimit;
		steps = new MetaItem *[stepLimit];
		for (int i = 0; i < maxQASteps; i++)
	    steps[i] = qi.steps[i]->clone();
	};
	~QualityInfo() {
		for (int i = 0; i < maxQASteps; i++) {
	    if(steps[i])
				delete steps[i];
		}
		delete[] steps;
	};
	/** adds a MetaItem to qinfo, not thread-safe */
	bool addMetaItem(MetaItem * m) {
		if (maxQASteps >= stepLimit)
	    return false;
		steps[maxQASteps] = m;
		maxQASteps++;
		return true;
	}
	u32 lastRequestTime;
	MetaItem **steps;
	u8 currentQAStep;
	u8 maxQASteps;
	u8 stepLimit;
	u32 esInfoId;
};


/**
 *
 * stores pointers to all available ElemStreams (ESInfos) in this container
 *
 *@author Michael Kropfberger and Peter Schojer
 *
 *@version $Id: ContainerInfo.hpp,v 1.21 2006/02/24 08:17:31 mkropfbe Exp $
 *
 *@short <short description>
 *
 */
class ContainerInfo {
public:

	typedef enum { UNKNOWN_CONTAINER, AVI, MP4, MATROSKA, VITOOKI, DVD, SDP, OGG} ContainerType;


  /** creates a video object for the given url, the url will be deep-copied.
	 * Always create ContainerInfo objects with the ISOMP4ContainerFile class or use the
	 * empty constructor and call loadFromCfgFile afterwards to complete initialization
	 * @param type: is eg. MP4 or DVD
	 */
	ContainerInfo(const char *url, const char* locFile, const ContainerType type);
  /** creates an empty info object. Complete initialization with
	 * loadFromCfgFile
	 */
	ContainerInfo();
	
  /** creates an empty info object. with qualityinfo set. Add ESs manually
	 */
	ContainerInfo(QualityInfo & q);
	
	virtual ~ ContainerInfo();
	
  /** returns the list of ESes for this object. Note that the ESes are destroyed
	 * when the videoobject is deleted
	 */
	list < ESInfo * >* getESList() {
		return &esList;
	};
	
  /** returns for a given streamid the ESInfo object. Returns NULL
	 * if the streamId was not found.
	 * Performance is linear time.
	 */
	ESInfo *getES(u32 streamId) const;
	
  /** Returns the first visual ES it finds */
	VideoESInfo *getFirstVisualES();
	
	/** how many visual streams are part of the Container */
	u32 getNumberOfVisualStreams() const;

	/** how many audio streams are part of the Container */
	u32 getNumberOfAudioStreams() const;

  /** Returns the first visual ES it finds */
	AudioESInfo *getFirstAudioES();

	/* @returns the type from the enumeration like MP4, DVD or SDP  */
	ContainerType getContainerType() { return type; };

	/* sets the container type 
	 * @param ctype:  enumeration like MP4, DVD or SDP  */
	void setContainerType(ContainerType ctype) { type=ctype; };

	/** will be used in conjunction with BufferedIOs created by IOCreator */
	IO *getMuxDemuxIO() { return muxDemuxIO; };

	/** will be used in conjunction with BufferedIOs created by IOCreator */
	void setMuxDemuxIO(IO *demuxio) { muxDemuxIO = demuxio; };

  /** serializes the ContainerInfo to a file.
	 * FIXME: update to include qinfo
	 * FIXME: fails when space is in fileName    
	 */
	bool saveToCfgFile(char *indexName);
	
  /** loads an ContainerInfo object from a file. if sth.
	 * goes wrong, all data loaded so far is freed and false
	 * is returned.
	 * FIXME: update to include qinfo
	 * FIXME: fails when space is in fileName
	 */
	bool loadFromCfgFile(char *indexName);
	
  /** returns  the url of the video, can return null for demultiplexed videos */
  const Url *getUrl() const;
	
	/* shallow copies url */
  void setUrl(Url *uri);
	
	/* deep copies loclFile */
	void setLocalFile(const char* loclFile);
	
	const char* getLocalFile() const {
		return localFile;
	}
  	
  /** sets the iod at the MP4Stream, a deep-copy is done!
	 * iodHandleSize must be a positive number
	 */
	void setIODHandle(u8 * iodHandle, u32 iodHandleSize);
	
  /** returns a pointer to the IOD, will update size to contain the iodHandleSize */
	const u8 *getIODHandle(u32 & size) const {
		size = iodHandleSize;
		return iodHandle;
	};
	
  /** adds an ESInfo object to the end of the list of ES's managed by
	 * this object
	 */
	void addESInfo(ESInfo * esobj) {
		esList.push_back(esobj);
	};
	
	inline u32 getActSize() const {
		return qinfo.steps[qinfo.currentQAStep]->size;
	};

	inline float getActQuality() const {
		return qinfo.steps[qinfo.currentQAStep]->quality;
	};
	
	u32 getLastRequestTime() const {
		return lastAccessTime;
	};

	void setLastRequestTime(u32 last) {
		lastAccessTime = last;
	};
	
	/** returns either the filesize of the mp4 file or the sum of the sizes
	  * of all demuxed streams */
	u32 getFileSize() const;

	/** returns for a non-cached object, the estimated filesize according to
	the SDP description */
	u32 getEstimatedNetworkFileSize();

	void setMetaObject(MetaObject* meta);

	MetaObject* getMetaObject();
	
  /** checks if the MP4 object has one audio stream. Returns false if it hasn't */
	bool hasAudio()const;
	
	QualityInfo *getQualityInfo() {
		return &qinfo;
	};
	/** deep-copies the object */ 
	ContainerInfo * clone();
	
	ReferenceCounter* getUsageCounter();

	/** deletes all locally stored files belonging to this object */
	bool destroy();

	rtx_group* getRtxGroup(u32 esId,list<rtx_group*>* rtxInfo) const;


private:
	void clear();
	ESInfo *parseES(char *line);


protected:
	/** the format of this container */
	ContainerType type;

	list < ESInfo * >esList;

	/** will be used in conjunction with BufferedIOs created by IOCreator */
	IO *muxDemuxIO;

	/** the remote url for this object */
  Url *url;

	/** the local cache file. Can be null */
	char *localFile; 

	/** points back to the MetaObject */
	MetaObject* metaObj;

  	/** iod is not a stream!, seperate it from ES */
	u8 *iodHandle;

  	/** the byte-size of the iod */
	u32 iodHandleSize;

  	/** the default timescale of the video in ticks/second */
	u32 timeScale;

  	/** cache specific code */
	u32 lastAccessTime;

	/** usage counter: how many clients are currently accessing this video */
	ReferenceCounter* usageCounter;

	/** information about the current quality and size of the video */
	QualityInfo qinfo;

	/* Name of the video */
	char *Name;
	
	/* Description of the Video's content */
	char *Description;
	
	/* Video's MimeType */
	char *MimeType;
};

#endif
