/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	DataChannel.hpp                                               *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

/***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef _PS_DATACHANNEL_HPP_
#define _PS_DATACHANNEL_HPP_

#include "global.hpp"
#include "VMutex.hpp"
#include "ESInfo.hpp"
#include "Statistics.hpp"

class IO;
class DataSink;
class Adaptor;
class Frame;
class Url;

/**
 *
 * a DataChannel object reads data from an input. Note that
 * you must at least add one DataSink via play before the DataChannel can
 * actually send data.
 * Each DataChannel has an Adaptor, after sending the data through the
 * the Adaptor, the result of the adaptation is distributed to each DataSink.
 * To each DataSink, there has to be attached an IO output
 *
 *@author Michael Kropfberger and Peter Schojer
 *
 *@version $Id: DataChannel.hpp,v 1.23 2006/01/26 09:50:45 mkropfbe Exp $
 *
 *@short <short description>
 *
 */
class DataChannel:public VThread {
public:
  /** creates a DataChannel only a dummy adaptor is set */
	DataChannel(IO * input, ESInfo * es);

  /** creates a DataChannel object that reads data from an input. Note that
   * you must at least add one DataSink via play before the DataChannel can
   * actually send data.
   * Each DataChannel has an Adaptor, after sending the data through the
   * the Adaptor, the result of the adaptation is distributed to each DataSink.
   */
	DataChannel(IO * input, ESInfo * es, Adaptor * adapt);

	~DataChannel();


  /** changes the adaptor to a new one, works even while
   * the thread is working. If null is passed as parameter,
   * a forwarder is set
   */
	void setAdaptor(Adaptor * adapt);

	/** changes the connected ESInfo
	 */
	void setESInfo(ESInfo *new_es);

	/** set the frame number, where the DC-thread should stop streaming.
	 * this is needed for reaching the switchPoint for stream switching
	 */
	void setFinalFrameNumber(long frameNumber);

	/** waits until the DC-thread is stopped
	 * @param forceTearDown: true will immedately stop the DC-thread,
   *        false will block until the DC is done with input or has reached finalFrameNumber
	 */
	void initiateThreadStop(bool forceTearDown);

	/* changes the connected input IO
	 * @param new_es: the new ESInfo, that fits the according IO *new_io
	 */
	void setInput(IO *new_io);

  /** Moves a previously inserted cliented from
	 * the paused list to the active list
   * returns false when the DataChannel has already ended. In this case
   * the client is not added.
	 * Note that for at least one client play must be called PRIOR
	 * starting the DataChannel, otherwise the thread will terminate
	 * immediately. You can only play clients which were inserted previously.
   */
	bool play(u32 clientId,double prefetch);

  /** mutes a client. The client stays on the active list, and received
	 * frames are still passed through the global DC adaptors (eg. statistics),
	 * but it will bypass the client adaptors (eg. decoding). But the frames
	 * will still be handed over to the I/O output classes' writeFrame for sync.
	 * To return to PLAY state, call DataChannel::unmute(clientId)
   */
	bool mute(u32 clientId);
		
  /** unmutes a MUTED client. This returns the client into PLAY state
   */
	bool unmute(u32 clientId);
		
  /** pauses a client. The client is searched in the active list.
   * If it is found there, it is moved to the pause queue. Otherwise
	 * false is returned. Note that when you pause the last client, the
	 * thread will terminate and has to be restarted where it will continue
	 * from the last frame.
   */
	bool pause(u32 clientId);
		
	/** inserts the client into the paused queue. This is the first mandatory step,
		* if you want to start playing a client.
		*/
	bool insert(DataSink* c);

  /** stores the result of the global Adaptor stream to the disk.
	 * @param proxyMode tells how the filename is generated. if set to true the serverName is included
	 * in the local file name which makes it impossible to reuse demuxed files when the serverName changes
	 * @param return returns a pointer to the new outName.
	 * This will create internally a DataSink with ID 1. Use this id to control the DataSink
	 * e.g. pause(1), teardown(1) etc.
	 */
	char *enableCaching(bool omitHeader, const char* outDir, bool proxyMode=true, const Url* url=NULL);

  /** renders the result of the global Adaptor stream to the screen.
		* Creates internally a DataSink with ID 0. Use this id to control the visualizing DataSink 
		*/
	void visualizeCaching();

  /** returns false, when the client was not found in the active or paused queue.
   * If c was found, the client will be deleted! */
	bool teardown(u32 clientId, bool immediate=false);
	
	bool tearDownFlagSet() const { return tearDown; };

	/** returns the number of active clients */
	int getNumberOfActiveDataSinks(); 
	
	IO* getInput() { return input;};

  Adaptor *getAdaptor() const { return adaptor; }

  /** starts the thread. if for no client play was called previously, the thread terminates immediately.
   * If clients were set, the thread runs till the input reaches End-of-stream.
   * When this state is reached, all buffered data is sent, then the active queue
   * is emptied. The clients are not destroyed. 
   */
	void run();

	void setSessionId(int i) {
		sessionId = i;
	};

	uint getSessionId() const {
		return sessionId;
	};

	const ESInfo *getESInfo() const {
		return es;
	};
	
	/** returns the DataSink with id @param clientId. searches only in paused.
	 * You are not allowed to change clients that are currently playing
	 */
	DataSink* getDataSink(u32 clientId);

	const DataSink* getActiveDataSink(u32 clientId);
	
	list <u32> getListOfActiveClientIds();

	list <u32> getListOfPausedClientIds();


	/** this is BIG HACK... by that, we set the stats of the FIRST dataSink output (eg. RTP), 
			since the code isnt able to cope with more than one networked DataSink anyway :(
	*/
	void setOutputStatistics(Statistics *stats);

	/** this is BIG HACK... by that, you get only one output stats from the first datasink out (eg. RTP)
			since the code isnt able to cope with more than one networked DataSink anyway, it should work
	*/
	Statistics *getOutputStatistics();

private:
	void init(IO * &input, ESInfo * &es, Adaptor * &adapt);
	void deleteInput();
	void closeAllClients(bool immediateClose);

protected:
	int sessionId;

	IO *input;

  /** list of receivers */
	list < DataSink * >active;

	/** list of paused receivers */
	list < DataSink * >paused;

  /** local viewer */
	DataSink *localViewer;
	
  /** Semaphore used to sync access to the active clients */
	VMutex lockDataSink;

  /** Semaphore used to sync access to the Input IO */
	VMutex lockInput;

  /** a semaphore to lock the adaptor */
	VMutex lockAdaptor;
	
 
  /** each DataChannel can have an Adaptor, per default a Forwarder is set here */
	Adaptor *adaptor;
    
	list < Frame * >sendQueue;

  /** specifies if new DataSinks can be added */
	bool channelOpen;

	bool tearDown;
	bool immediateClose;
	ESInfo *es;
	Statistics *outstats;

	long currentFrameNumber;
  long finalFrameNumber;
};


#endif
