 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:   ESInfo.hpp                                                  *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef PS_ELEMENTARYSTREAMINFO_HPP
#define PS_ELEMENTARYSTREAMINFO_HPP


#include "global.hpp"
#include <assert.h>

#ifdef ISOMP4
#include "ISOMP4ContainerFile.hpp"
#include "ISOMovies.h"
#include "avformat.h"
#else
//#include "ffMP4ContainerFile.hpp"
#ifdef __cplusplus
extern "C" {
#include "avformat.h"
}
#endif
#endif

class ContainerInfo;
class BitField;
class ReferenceCounter;


/**
 *
 * Info object that describes one elementary stream
 *
 *@author Michael Kropfberger and Peter Schojer
 *
 *@version $Id: ESInfo.hpp,v 1.24 2006/05/15 10:41:39 mkropfbe Exp $
 *
 *@short <short description>
 *
 */
class ESInfo {
public:
	/** creates an ESInfo object
		* @param encodeddecconf is only shallow copied and
		* freed in the destructor
		* @param avgBW and @param maxBW specify the bandwidth of the given ES
		* and are in Bits/sec !
	*/
	ESInfo(u64 streamId, u32 streamtype, ContainerInfo * vob, u32 timeIncrement,
				 u8 * encodeddecconf, u32 objtype, u32 bufsize,
				 u32 avg, u32 maxBand, u32 timeScale,
				 u64 duration, u32 size,  bool complete=true, 
				 CodecID codec_id = CODEC_ID_NONE, float quality = 1.0);
	
	virtual ~ESInfo();
	
  /** returns the size of the header, returns 0 in case of error.
			If *pheaders is != NULL an assertion is thrown! */
	u32 getHeaders(u8 **pHeaders) const;

	u64 getStreamId() const {
		return streamId;
	};

  /** returns a pointer to the decode config */
	const u8 *getEncodedDecoderConfig() const {
		return encodedDecoderConfig;
	};

  /** creates a shallow copy of enc */
	virtual void setEncodedDecoderConfig(u8 * enc);

  /** encodes dec and deep-copies it */
	void setDecoderConfig(const u8 * dec, int size);

	u32 getObjectType() const {
		return objectType;
	};

	u32 getStreamType() const { 
		return streamType;
	};

	u32 getBufferSize() const {
		return bufferSize;
	};

/*	bool openSucceeded() const {
		return handlerType != (u32) - 1;
	}; */
	virtual u32 getHandlerType() const=0;

	virtual bool isVisualStream() const=0;

	virtual bool isAudioStream() const=0;

	virtual bool isODStream() const=0;

	virtual bool isSceneDescriptionStream() const=0;

	virtual void generateDescriptionText() { strcpy(description,"Generic ESInfo");};
	
	const char *getDescriptionText() const { return description;};
	
  /** creates a deep copy of the ES */
	virtual ESInfo *clone(ContainerInfo * mp4)=0;

	bool getCompleteState() const;

	float getStreamQuality() const { return quality;};

	void setCompleteState(bool comp) {
		complete=comp;
	}
	
  /** sets a deep copy of input at this object, input will be freed when the
   * destructor is called 
   */
	void setInput(const char *inp);

	/** @returns the file name of the elementary stream (or container stream) */
	const char *getInput() const {
		return input;
	};

	/* @returns avg BW in Bits/sec */
	u32 getAvgBandwidth() const {
		return avgBandwidth;
	};
  /**
     Needed in Adaptor classes.
	*/
	void setAvgBandwidth(u32 uiBandwidth) {
		avgBandwidth = uiBandwidth;
	};
	u32 getMaxBandwidth() const {
		return maxBandwidth;
	};
	void setMaxBandwidth(u32 uiMaxBandwidth) {
		maxBandwidth = uiMaxBandwidth;
	};
  /** returns the timescale of the stream measured in ticks/second */
	u32 getMediaTimeScale() const {
		return mediaTimeScale;
	};
  /** set the mediaTimeScale */
	void setMediaTimeScale(u32 uiMediaTimeScale) {
		assert(uiMediaTimeScale>0);
		mediaTimeScale = uiMediaTimeScale;
	};

  /** returns the duration of the ElementaryStream expressed in ticks */
	u64 getDuration() const {
		return duration;
	};
        void setDuration(u64 uiDuration);

  /** returns the duration of the media in ms */
	u64 getDurationInMs() const {
		return (duration * 1000) / mediaTimeScale;
	};
	u32 getOrigMediaSize() const {
		return mediaSize;
	};

	void setOrigMediaSize(u32 mediaSize) {
		this->mediaSize = mediaSize;
	};

	/** returns the size of the current media. Will also work for streams that
	  * are currently streamed to the disk
		*/
	u32 getCurrentDemuxedMediaSize() const;

	/** calculates the frame number for the given time stamp which is expressed in seconds
	 * @param timeInSeconds is a double, which will be mapped onto a frame number
	 * @param upround forces ceil() or floor()
	 * @returns the closest (up-rounded) framenumber, so second 2.25 with 25fps will lead to frame 56.25 ==> 57
	 */
	u32 getFrameNumber(double timeInSeconds, bool upround = true) const;
	
  /** Calculates media time in seconds from frame number. 
   * Returns 0.0 on error.
   */
  double getMediaTime(u32 frameNo) const;

  /** updates the delta times between two consecutive frames in ticks*/

	void setVOPTimeIncrement(u32 ticks);

	u32 getVOPTimeIncrement() const {
		return vopTimeIncrement;
	};
	
	/** @returns number of frames this ES consists of */
	u32 getNumberOfMediaSamples() const {
		if(vopTimeIncrement==0)
			return 0;
		return (duration/vopTimeIncrement);
	};
	
	ContainerInfo *getContainerInfo() {
		return vo;
	};
	
	bool getDecConfUpdated() const { return decConfUpdated;};

	void clearDecConfUpdated() { decConfUpdated=false;};
	
	/** creates a new frame statistic object. fails if already one exists */
	bool enableFrameStatistic();

	/* frees an existing frame statistics object */
	void disableFrameStatistic();

	BitField* getFrameStatistic();

	void setFrameStatistic(BitField*);

	ReferenceCounter* getUsageCounter() { return usageCounter;};

	void setCloneCounter( ReferenceCounter* &ref);

	ReferenceCounter* getCloneCounter() { return existingClones;}

	void setCodecID(CodecID id) {
		codec_id = id;
	}
	
	CodecID getCodecID() {
		return codec_id;
	}

	/** deletes all locally stored files belonging to this object */
	bool destroy();

protected:
	u64 streamId;
	u32 streamType;
	ContainerInfo *vo;
  /** input specifies a local file or an url string */
	char *input;
	u8 *encodedDecoderConfig;
	u32 objectType;
	u32 bufferSize;
	u32 avgBandwidth;
	u32 maxBandwidth;
  /** the timescale of the media, in ticks/second */
	u32 mediaTimeScale;
  /** the duration of the stream expressed in the medias timescale */
	u64 duration;
  /** the byteSize of the original ES */
	u32 mediaSize;
	u32 vopTimeIncrement;
	bool decConfUpdated;
	bool complete;
	BitField* frameStatistic;
	ReferenceCounter* usageCounter;
	/** stores the number of existing clones for this ESInfo including the original one */
	ReferenceCounter* existingClones;
	/** This is the codec id for the stream as provided by ffmpeg
	 * Cannot be used with ISOMP4 */
	CodecID codec_id;
	/* the quality of stream as determined by ffmpeg 
	 * Its a value between 1(good) and 31(bad)
	 * Needed for transcoding.
	 */
	float quality;
	char description[MAX_STR_LEN];
	
	/* Language spoken or displayed in either audio or video */
	char *Language;

private:
	bool loadInfo(char *file);
	
};

#endif
