 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	FastPointerHashTable.hpp                                *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/
#ifndef VITOOKI_PS_FASTPOINTERHASHTABLE_HPP
#define VITOOKI_PS_FASTPOINTERHASHTABLE_HPP

#include <assert.h>
/**
 *
 * A HashTable for pointers only. Hashing is done with a simple MODULO
 * operation. If a collision occurs, the element previously stored at
 * that position is DELETED and the new element is inserted.
 *
 *@author Peter Schojer
 *
 *@version $Id: FastPointerHashTable.hpp,v 1.6 2006/01/20 15:37:17 mkropfbe Exp $
 *
 *@short <hash table with NO collision handling>
 *
 */
template<class T> class FastPointerHashTable 
{
	T invalid;
	T* data;
	/** the maximum size of the hashtable */
	unsigned int size;
	unsigned int* keys;

public:
	FastPointerHashTable(unsigned int siz) {
		invalid=NULL;
		size=siz;
		data=new T[size];
		memset(data,0,sizeof(T)*size); // init to NULL
		keys=new unsigned int[size];
		memset(keys,0u,sizeof(unsigned int)*size); // init to 0
		// 0 is an invalid key for all elements but the first (due to the modulo mapping)
		keys[0]=1u;
	};

	~FastPointerHashTable() {
		freeAllElements();
		delete[] data; delete[] keys;
	};

	/** Insert an element which must be NOT EQUAL to NULL. NULL elements will be ignored.
	  * If a collission is detected no overflow handling happens. The old element is deleted
	  * and the new one inserted. */
	void put(T& elem,unsigned int key) {
		unsigned int pos=key%size;
		assert(elem != NULL);
		if(elem==NULL)
			return;
		// if we receive another element that maps to the same pos but has
		// a different key then we delete it!
		if(data[pos]) 
			delete data[pos];
		data[pos]=elem;
		keys[pos]=key;
	};
	
	/** returns a reference to the element stored in the Table. Manipulating this reference,
	  * will thus also update the element in the Table. If you delete the element outside,
	  * you MUST call unlinkElement prior.
	  */
	T& get(unsigned int key) {
		unsigned int pos=key%size;
		if(keys[pos]==key)
			return data[pos];
		return invalid;	
	};
	
	void freeElement(const unsigned int key) {
		unsigned int pos=key%size;
		if(keys[pos]==key) {
			keys[pos]=key+1; // invalid key which can never be found at this position due to modulo operation
			if(data[pos])
				delete data[pos];
			data[pos]=NULL;
		}
	};

	void freeAllElements() {
		for(unsigned int i=0;i<size;i++) {
			if(data[i]) {
				delete data[i];
				data[i]=NULL;
			}
			keys[i]=i+1;
		}
	};

	void unlinkElement(const unsigned int key) {
		unsigned int pos=key%size;
		if(keys[pos]==key) {
			keys[pos]=key+1; // invalid key which can never be found at this position due to modulo operation
			data[pos]=NULL;
		}
	};
};
#endif
