 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	ESSynchronizer.hpp                                            *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef _ESSYNCHRONIZER_HPP
#define _ESSYNCHRONIZER_HPP
#include "Adaptor.hpp"
#include "AdaptorChain.hpp"
#include "GlobalTimer.hpp"
#include "Frame.hpp"
#include "ESInfo.hpp"

#ifdef WIN32
#ifndef WINCE
#include <sys/timeb.h>
#endif
#else
#include <sys/time.h>
#endif

#include "VCondition.hpp"
#include <time.h>
#include "VMutex.hpp"

/* how many msecs for "behind frames" are tolerated? */
#if (defined (BUILD_ARCH_ARM) || defined (WINCE))
#define ESSYNC_GRACE_MSECS 500
#else
#define ESSYNC_GRACE_MSECS 100
#endif

/* threshold for stream discontinuity detection
 * (in milliseconds of media time)
 */
#define ESSYNC_DISCONT_MSECS 1001 //more than one second...

/**
 * Adaptor class Synchronizer
 * handles stream synchronization
 * can be used for VES and AES
 *
 *@author Michael Kropfberger and Peter Schojer
 *
 *@version $Id: ESSynchronizer.hpp,v 1.15 2006/05/12 11:32:49 mkropfbe Exp $
 *
 *@short <short description>
 *
 */
class ESSynchronizer:public Adaptor {
public:

	/* Constructor
	   each ES needs one Synchronizer object 
	   interval = time interval before average of FPS and aBW is built
	   (presented in ms)
	*/
	ESSynchronizer(ESInfo* esi, GlobalTimer * globalTimer);
	
	virtual ~ESSynchronizer();
	
        /** Blocks the calling data channel until the presentation
	 * time stamp (CTS) of the given frame matches the play-out
	 * media time returned by the associated <code>GlobalTimer</code>.
	 * If the frame lags behind the play-out time by more than
	 * <code>ESSYNC_GRACE_MSECS</code> milliseconds, the frame
	 * will be dropped. If the frame hurries ahead the play-out
	 * time by more than <code>ESSYNC_DISCONT_MSECS</code> milliseconds,
	 * a media stream discontinuity will be assumed and the 
	 * <code>GlobalTimer</code> will be adjusted to the frame's CTS.
	 */
	list < Frame * >adapt(Frame * frm);
	
	Adaptor *clone() { return new ESSynchronizer(es,globalTimer);	};

	void initialize();
	
	list < Frame * >close() {
		list < Frame * >tmp;
		return tmp;
	};

        /** Blocks/deblocks the data channel this ESSynchronizer adaptor
	 * belongs to. The associated <code>GlobalTimer</code> is <em>not</em> paused!
	 * As opposed to the corresponding method of <code>GlobalTimer</code>,
	 * the pause state of this class is binary, i.e. the calling sequence
	 * <code>setPaused(true) - setPaused(true) - setPaused(false)</code>
	 * will yield a pause state of <code>false</code> (no reference counting).
	 * @param pause     if <code>true</code>, the data channel will be blocked;
	 *                  if <code>false</code>, it will be deblocked.
	 * @see GlobalTimer::setPaused(bool)
	 */
	void setPaused(bool pause);

	/** Sets a reference to an ESInfo object. Adaptors are
		allowed to change ESInfo objects. If you don't want that,
		pass an es->clone() 
	*/ 
	void setESInfo(ESInfo * esi) {es=esi;};
	ESInfo * getESInfo() { return es;}

        /** Return associated <code>GlobalTimer</code>. */
        GlobalTimer *getGlobalTimer() const { return globalTimer; }

        /** Set output rate factor w.r.t. media frame rate.
	 * E.g. setting this value to 0.5 will cause a 30 Hz media stream
	 * to be forwarded with 15 Hz.
	 * @param factor      must be positive; 1.0 means media frame rate.
	 * @return <code>true</code> on success.
	 */
        bool setOutputRateFactor(double factor);

        /** Returns output rate factor. 
	 * @see #setOutputRateFactor(double)
	 */
        double getOutputRateFactor() const { return outputRate; }

        /** Adjust the associated global timer to the given media time point.
	 * Note that the GlobalTimer::adjust(double) method
	 * adjusts to a <em>play-out</em> time point, which is
	 * different from the media time point if the output rate
	 * factor is not equal to 1.0.
	 * @param sec    media time point in seconds.
	 * @return <code>true</code> on success.
	 */
        bool adjust(double sec);

	/** returns adaptation costs (CPU only) */
				virtual u32 getTranscodingCosts() const {
					// costs for essync do exist but how high are they per second?
					// assume 5000
					return 5000u;
				}

private:
	GlobalTimer * globalTimer;
        VCondition cond_pause; // condition is true <=> data channel is *not* paused
        VMutex cond_pause_mutex;
        bool paused;

	int dropCount;

protected:
	ESInfo* es;
	struct timeval tv;
	long frmCount;

	/** The output rate factor. 1.0 means media frame rate. */
	double outputRate;
};
#endif

