 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title: FramePatternDetector.cpp                                      *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#include "FramePatternDetector.hpp"

/* Constructor: @param updateEsInfoAtClose set to true if you want the adaptor to
 * automatically overwrite the values in @param esi, when the adaptor is closed (and at least one
 * complete GOP was analyzed!)
*/
FramePatternDetector::FramePatternDetector(VideoESInfo* esi, bool updateEsInfoAtClose)
{
	strcpy(name,"FramePatternDetector");
	memset(bFramesBetweenCounter,0,sizeof(int)*FRAMEPATTERNDETECTOR_MAX_BFRAMES_BETWEEN);
	memset(gopSizesDetected,0,sizeof(int)*FRAMEPATTERNDETECTOR_MAX_GOP_ANALYZE);

	updateESInfo = updateEsInfoAtClose;
	es = esi;
	framesSinceIFrame=0;
	bFramesBetween=0;
	currentGOP=-1;
	staticGopSize=false;
	staticGopPattern=false;
	foundBFramesBetween=-1;
	foundGopSize=-1;
	lastCTSIFrame=0;
	stopBFrames=false;
	stopGOPSize=false;
	bFrameSize=0;
	totalVideoSize=0;
};

void FramePatternDetector::initialize() {
}

FramePatternDetector::~FramePatternDetector()
{
}

list < Frame * > FramePatternDetector::adapt(Frame * frm) {
	assert(frm);
	list < Frame * >tmp;
	u32 thisCTS=0;

	totalVideoSize+=frm->getAU()->size;
	if(frm->getAU()->frameType==Frame::B_VOP)
		bFrameSize+=frm->getAU()->size;
	//standard forwarding adaptor behaviour
	tmp.push_front(frm);
	if(currentGOP>=FRAMEPATTERNDETECTOR_MAX_GOP_ANALYZE) {
		// stop analyze
		return tmp;
	}

	thisCTS = frm->getAU()->cts;
	
	if(frm->getType() == Frame::I_VOP && currentGOP<0) {
		framesSinceIFrame=1;
		bFramesBetween=0;
		currentGOP=0;
		foundBFramesBetween=0;
		foundGopSize=0;
		lastCTSIFrame=frm->getAU()->cts;
		stopBFrames=false;
	}
	else {
		handleFrameType(frm->getType(),thisCTS);
		if(currentGOP>=FRAMEPATTERNDETECTOR_MAX_GOP_ANALYZE) 
			detectGopSize();
	}

	return tmp;

}

void FramePatternDetector::handleFrameType(Frame::FrameType current,u32 ctsCur) {
	int expectedFrames=0;
	if(lastFrame==current) {
		if(lastFrame==Frame::B_VOP) {
			bFramesBetween++;
		}
		else if(lastFrame==Frame::P_VOP) {
			//two consecutive P
			bFramesBetweenCounter[0]++;
		}
		else if(lastFrame==Frame::I_VOP) {
			// two consec. I
			expectedFrames=(ctsCur-lastCTSIFrame)/es->getVOPTimeIncrement();
			gopSizesDetected[currentGOP]=expectedFrames;
			if(foundGopSize==-1) {
				//calc cts value
				foundGopSize=expectedFrames;
				if(expectedFrames!=framesSinceIFrame)
					dprintf_small("FramePatternDetector::handleFrameType: expected %u frames per GOP, only got %u\n",
					expectedFrames,framesSinceIFrame);
			}
			framesSinceIFrame=1;
			lastCTSIFrame=ctsCur;
			analyzeBFramePattern();
			currentGOP++;
			memset(bFramesBetweenCounter,0,sizeof(int)*FRAMEPATTERNDETECTOR_MAX_BFRAMES_BETWEEN);
		}
	}
	else if(current==Frame::B_VOP) {
		// last frame is not a B-frame, a new B-frame start!
		bFramesBetween=1;
	}
	else {
		// if we have an I or P frame, this always ends a bframepattern
		// except in the first GOP which has IPBB instead of IBBPBB
		if(ctsCur>es->getVOPTimeIncrement())
			bFramesBetweenCounter[bFramesBetween]++;
		bFramesBetween=0;
		// if we have an I-frame we have to handle GOPsize additionaly
		if(current==Frame::I_VOP) {
			gopSizesDetected[currentGOP]=(ctsCur-lastCTSIFrame)/es->getVOPTimeIncrement();
				if(gopSizesDetected[currentGOP]!=framesSinceIFrame)
					dprintf_small("FramePatternDetector::handleFrameType: expected %u frames per GOP, only got %u\n",
					expectedFrames,framesSinceIFrame);
			framesSinceIFrame=1;
			lastCTSIFrame=ctsCur;
			analyzeBFramePattern();
			currentGOP++;
			memset(bFramesBetweenCounter,0,sizeof(int)*FRAMEPATTERNDETECTOR_MAX_BFRAMES_BETWEEN);
		}
	}
	lastFrame=current;
}

void FramePatternDetector::analyzeBFramePattern()
{
	if(stopBFrames)
		return;
	//search the maximum in the array
	int maxVal=bFramesBetweenCounter[0];
	int sum=0;
	int maxIndex=0;
	for(int i=1;i<FRAMEPATTERNDETECTOR_MAX_BFRAMES_BETWEEN;i++) {
		sum+=bFramesBetweenCounter[i];
		if(maxVal<bFramesBetweenCounter[i]) {
			maxIndex=i;
			maxVal=bFramesBetweenCounter[i];
		}
	}
	if( ((double)maxVal)/sum>FRAMEPATTERNDETECTOR_BFRAME_THRESHOLD) {
		this->foundBFramesBetween=maxIndex;
		staticGopPattern=true;
		stopBFrames=true;
	}
	else
		staticGopPattern=false;
}

void FramePatternDetector::detectGopSize() {
	if(stopGOPSize)
		return;
	if(currentGOP<=0)
		return;
	int* occurrence=new int[currentGOP];
	int i=0,j=0;
	memset(occurrence,0,sizeof(int)*currentGOP);

	// inefficient, who cares...
	for(i=0;i<currentGOP;i++) {
		for(j=0;j<currentGOP;j++){
			if(gopSizesDetected[i]==gopSizesDetected[j])
				occurrence[i]++;
		}
	}

	int max=occurrence[0];
	int maxpos=0;
	// search the maximum in occurrence
	for(i=1;i<currentGOP;i++) {
		if(occurrence[i]>max) {
			maxpos=i;
			max=occurrence[i];
		}
	}
	if(FRAMEPATTERNDETECTOR_GOP_THRESHOLD < ((double)max)/currentGOP) {
		foundGopSize=gopSizesDetected[maxpos];
		staticGopSize=true;
	}
	else
		staticGopSize=false;
	stopGOPSize=true;
	delete occurrence;
}

list < Frame * > FramePatternDetector::close() {
	list < Frame * >tmp;
	if(!stopGOPSize)
		detectGopSize();
	if(updateESInfo) {
		if(stopBFrames)
			es->setAvgBFrameSize( ((double)this->bFrameSize)/totalVideoSize);
		if(staticGopSize && foundGopSize>0 && stopGOPSize)
			es->setGOP_size(foundGopSize);
		if(staticGopPattern && foundBFramesBetween>=0 && stopBFrames)
			es->setNum_B_frames(foundBFramesBetween);
		es->setStaticFramePattern(staticGopPattern);
	}
	return tmp;
};

