 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	FramePatternDetector.hpp                                        *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef _PS_VITOOKI_FRAMEPATTERNDETECTOR_HPP
#define _PS_VITOOKI_FRAMEPATTERNDETECTOR_HPP
#include "Adaptor.hpp"
#include "Frame.hpp"
#include "VideoESInfo.hpp"

#define FRAMEPATTERNDETECTOR_MAX_BFRAMES_BETWEEN 10
#define FRAMEPATTERNDETECTOR_MAX_GOP_ANALYZE 10
#define FRAMEPATTERNDETECTOR_BFRAME_THRESHOLD 0.9
#define FRAMEPATTERNDETECTOR_GOP_THRESHOLD 0.8

/**
 * Adaptor class FramePatternDetector counts by-passing frames by frame type and
 * detects if a video has a constant frame pattern, how many b-frames are between
 * and how large is the GOP-size
 *@author Peter Schojer
 *
 *@version $Id: FramePatternDetector.hpp,v 1.4 2006/01/20 15:37:17 mkropfbe Exp $
 *
 *@short <short description>
 *
 */
class FramePatternDetector : public Adaptor {
public:

	/* Constructor: @param updateEsInfoAtClose set to true if you want the adaptor to
	 * automatically overwrite the values in @param esi, when the adaptor is closed (and at least one
	 * complete GOP was analyzed!)
	*/
	FramePatternDetector(VideoESInfo* esi, bool updateEsInfoAtClose);
	
	virtual ~FramePatternDetector();
	
	list < Frame * >adapt(Frame * frm);
	
	Adaptor *clone() { return new FramePatternDetector(es,updateESInfo);	};

	void initialize();
	
	list < Frame * >close();

	/** Sets a reference to an ESInfo object. Adaptors are
		allowed to change ESInfo objects. If you don't want that,
		pass an es->clone() 
	*/ 
	void setESInfo(ESInfo * esi) {es=(VideoESInfo*)esi;};
	ESInfo * getESInfo() { return (ESInfo*)es;}
	/** returns if the analyzer has already seen enough GOPs
	  * to decide if the GOP pattern is static or not
		*/
	bool hasDecidedOnGopSize() const { return stopGOPSize;};
	/** returns if the analyzer has already seen enough GOPs
	  * to decide on the number of b-frames in between 
	  */
	bool hasDecidedOnBFrames() const { return stopBFrames;};
	
	/** returns if all GOPs consists of the same number of frames (not type!)*/
	bool isStaticGopSize() const {
		return staticGopSize;
	}
	
	int getGopSize() const {
		return foundGopSize;
	}
	
	int getBFramesBetween() const {
		return foundBFramesBetween;
	}
	/** returns the sum of sizes of all b-frames that passed
	  * through this adaptor. Be careful when using with network input (frame loss!)
		*/
	u32 getTotalSizeOfBFrames() const {
		return bFrameSize;
	}

	/** returns the sum of sizes of all frames that passed
	  * through this adaptor. Be careful when using with network input (frame loss!)
		*/
	u32 getTotalSizeOfFrames() const {
		return totalVideoSize;
	};
	/** returns adaptation costs (CPU only) */
	u32 getTranscodingCosts() const{
		return 5000u;
	}
private:
	void handleFrameType(Frame::FrameType current,u32 ctsCur);
	void analyzeBFramePattern();
	void detectGopSize();

protected:
	VideoESInfo* es;
	bool updateESInfo;
	int bFramesBetween;
	int bFramesBetweenCounter[FRAMEPATTERNDETECTOR_MAX_BFRAMES_BETWEEN];
	/** including I-frame */
	int framesSinceIFrame;
	u32 lastCTSIFrame;
	/** pos[i] in the array stores the GOP size from I-frame [i] to (excluding) I-frame[i+1] */
	int gopSizesDetected[FRAMEPATTERNDETECTOR_MAX_GOP_ANALYZE];
	int foundBFramesBetween;
	int foundGopSize;
	bool staticGopSize;
	bool staticGopPattern;
	int currentGOP;
	Frame::FrameType lastFrame;
	bool stopBFrames;
	bool stopGOPSize;
	u32 bFrameSize;
	u32 totalVideoSize;
};
#endif

