 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	GlobalTimer.hpp                                               *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef _SYNCHRONIZER_HPP
#define _SYNCHRONIZER_HPP
#include "Adaptor.hpp"
#include "AdaptorChain.hpp"
#include "Frame.hpp"
#include "ReferenceCounter.hpp"
#include "VCondition.hpp"
#include "VMutex.hpp"

#ifdef WIN32
#ifndef WINCE
#include <sys/timeb.h>
#endif //wince
#else
#include <sys/time.h>
#endif
#include <time.h>


/**
 * This adaptor class provides a means for synchronizing DataChannels
 * (media streams) by providing methods for obtaining the elapsed time
 * since the start of streaming (not including PAUSED states) 
 * and for blocking DataChannels when set to the PREFETCHING or PAUSED state.
 * <p>
 * All DataChannels representing synchronized media streams (usually 1 video 
 * and 1 audio stream) need to include the same GlobalTimer instance as a
 * SharedAdaptor. Note that a GlobalTimer needs to be coupled with an
 * ESSynchronizer to achieve display rate control for video streams.
 * </p>
 * @see SharedAdaptor
 * @see ESSynchronizer
 *
 *@author Michael Kropfberger, Peter Schojer, Mario Taschwer
 *
 *@version $Id: GlobalTimer.hpp,v 1.13 2006/01/20 15:37:17 mkropfbe Exp $
 *
 *@short Adaptor for synchronizing several DataChannels with PREFETCHING or PAUSED states.
 */
class GlobalTimer:public Adaptor {
public:

	/* Constructor
	   each ES needs the same GlobalTimer object 
	*/
	GlobalTimer();
	
	virtual ~GlobalTimer();
	
	list < Frame * >adapt(Frame * frm);
	
	Adaptor *clone() { return new GlobalTimer();	};

	void reset();

	void initialize();
	
	list < Frame * >close() {
		list < Frame * >tmp;
		return tmp;
	};

	/** Return the time elapsed since the start of streaming,
	 * excluding any time periods when the GlobalTimer was
	 * was blocked.
	 */
	double getActualSec();
	
	/* AudioIO calls this to bring A/V in sync */
	void adjustToTS(int ts, int mediaTimeScale);
	
	/** Adjust timer to play-out time point.
	 * @param sec play-out time point in seconds.
	 */
	void adjust(double sec);
	
	void setPaused(bool pause);
	
	bool isPaused() const { return pauseRefCnt.getUsage() > 0; }

	void setPrefetching(bool pause);
	
	bool isPrefetching() { return prefetchRefCnt.getUsage() > 0; }
	
	/** Drop all frames with a CTS less than the given one.
	 * Frames are dropped even if the GlobalTimer is PAUSED or
	 * PREFETCHING. This allows media streams to be sychronized to
	 * a given time stamp, e.g. after a PAUSE/PLAY in a network player
	 * application.
	 * <p>Note that you need to insert a decoder <em>before</em> the
	 * GlobalTimer into the adaptor chain if you want to use this method.</p>
	 */
	void dropFramesUntil(u32 cts);
	
	/** dummy implementations since a GlobalTimer is used by many ESs
	 */ 
	void setESInfo(ESInfo * esi) {es=esi;};
	
	ESInfo * getESInfo() { return es;}
	
	/** returns adaptation costs (CPU only) */
	u32 getTranscodingCosts() const {
		return 10000u;
	}
	
private: 
	ESInfo* es;
	VCondition cond_block;	// condition: _not_ blocked (paused or prefetching)
	VMutex cond_block_mutex; // protects also wasBlocked
	ReferenceCounter pauseRefCnt;
	ReferenceCounter prefetchRefCnt;
	bool wasBlocked;
	u32 dropUntil;
	VMutex time_mutex; // protecting time values below
	struct timeval tv;
	double actualSec;
	double start_dtime;
	double now_dtime;
	bool firstFrame;
};

#endif

