/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	MP3Encoder.cpp                                                *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

/***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#include "MP3Encoder.hpp"
#include "CompressedVideoFrame.hpp"


/***************************************************/
MP3Encoder::MP3Encoder(AudioESInfo * pESInfo, u32 targetBitRate,
											 enum CodecID outFormat, char* encoderConfig) {
	this->mpEncoderHandle = encoderConfig;
	es = pESInfo;
	muiTargetBitrate = targetBitRate;
	initialized = false;
	strcpy(name,"MP3Encoder-FFMPEG");
	outputFormat=outFormat;
}


/***************************************************/
void MP3Encoder::initialize() {
	if(initialized)
		return;
	dprintf_full("MP3Encoder::initialize() Handle: %p BW: %i bps\n",
							 mpEncoderHandle, muiTargetBitrate);
	initialized = true;
	if (FFMPEGaudioEncoderInit(&mpEncoderHandle, muiTargetBitrate, 44100, 2, 16) != ERR_OK) {
		dprintf_err("MP3Encoder::initialize: Error creating encoder instance!\n");
		initialized = false;
	}
	assert(mpEncoderHandle);
	es->setAvgBandwidth(muiTargetBitrate);
	es->setVOPTimeIncrement(2350); // FIXME: 90000 DIV 38 frames, known by analyzing ffmpeg :(
	muiFrameNumber = 0;
	inBufSize=0;
	firstFrame = true;
}


/***************************************************/
MP3Encoder::~MP3Encoder() {
	FFMPEGaudioEncoderClose(&mpEncoderHandle);
}



/***************************************************/
list < Frame * >MP3Encoder::adapt(Frame * frm) {
	CompressedAudioFrame *mpgFrame = NULL;
	list < Frame * >frmList;
	u32 iFramesize = 0;

	if (!initialized)
		initialize();
	if (!initialized)
		return frmList;

#define PCM_FRAME_SIZE (1152*2*2)  //FIXME: should be read from AVCodecContext->frame_size

	if (inBufSize > 0) { // smth was left from last round....
		memcpy(mpInBuf+inBufSize, frm->getAU()->payload, frm->getAU()->size);
		dprintf_full("MP3Encoder::adapt %i bytes were left from last round...stored further %i bytes\n",inBufSize,frm->getAU()->size);
		inBufSize += frm->getAU()->size;

		if (inBufSize >= PCM_FRAME_SIZE) {
			iFramesize = FFMPEGaudioEncodeFrame(mpEncoderHandle, (char *)mpInBuf, mpOutBuf);
			inBufSize -= PCM_FRAME_SIZE;
			assert(inBufSize <= PCM_FRAME_SIZE); //avoid overlapping rest-buffer
			memcpy(mpInBuf,mpInBuf+PCM_FRAME_SIZE,inBufSize); //move rest to the front

			if (iFramesize > 0) {
				AU *pAU = new AU(frm->getAU());
				u8 *payload = new u8[iFramesize];
				memcpy(payload, mpOutBuf, iFramesize);
			
				pAU->size = iFramesize;
				pAU->payload = payload;
				pAU->cts = muiFrameNumber * es->getVOPTimeIncrement();
			
				mpgFrame = new CompressedAudioFrame(Frame::MP3_AUDIO);
				
				//mpgFrame->detectFrameType();
				mpgFrame->setAU(pAU);
				mpgFrame->setMediaTimeScale(es->getMediaTimeScale());
				frmList.push_back(mpgFrame);
				dprintf_full("MP3Encoder::adapt overlapping frame: size %i bytes, %s frameNo %i, CTS %u:\r\n",
										 pAU->size, Frame::VOPTypeToChar(mpgFrame->getType()), muiFrameNumber,
										 mpgFrame->getAU()->cts);
				muiFrameNumber++;
			} else {
				dprintf_full("MP3Encoder::adapt ... filling encoder queue: no return data yet.\n");
			}
		} else {
			dprintf_full("MP3Encoder::adapt ... still harvesting small pcm bits for one full frame...!\n");
		}
	} // smth was left from last round....
	else { //inBufSize==0, so store for next round
		memcpy(mpInBuf, frm->getAU()->payload, frm->getAU()->size);
		inBufSize= frm->getAU()->size;
		dprintf_full("MP3Encoder::adapt %i tail-bytes saved for next round...\n",inBufSize);
	}

	return frmList;
}


/***************************************************/
list < Frame * >MP3Encoder::close() {
	CompressedAudioFrame *mpgFrame = NULL;
	list < Frame * >frmList;
	u32 iFramesize;


	dprintf_full("MP3Encoder::close ... there are %i bytes left in inBuf, draining....\n",inBufSize);
	
	if (inBufSize > 0) {
		//maybe FIXME: for large input frames, maybe more than one PCM_FRAME_SIZE is still in buffer?  --> while() needed...
		assert(inBufSize <= PCM_FRAME_SIZE); //avoid overlapping rest-buffer
		iFramesize = FFMPEGaudioEncodeFrame(mpEncoderHandle, (char *)mpInBuf, mpOutBuf);
	
		do { 
			if (iFramesize > 0) {
				AU *pAU = new AU();
				u8 *payload = new u8[iFramesize];
				memcpy(payload, mpOutBuf, iFramesize);
				
				pAU->size = iFramesize;
				pAU->payload = payload;
				pAU->cts = muiFrameNumber * es->getVOPTimeIncrement();
				
				mpgFrame = new CompressedAudioFrame(Frame::MP3_AUDIO);
				
				//mpgFrame->detectFrameType();
				mpgFrame->setAU(pAU);
				mpgFrame->setMediaTimeScale(es->getMediaTimeScale());
				frmList.push_back(mpgFrame);
				dprintf_full("MP3Encoder::close FINAL frame: size %i bytes, %s frameNo %i, CTS %u:\n",
										 pAU->size, Frame::VOPTypeToChar(mpgFrame->getType()), muiFrameNumber,
										 mpgFrame->getAU()->cts);
				muiFrameNumber++;
			} else {
				dprintf_err("MP3Encoder::close ... FINAL pcm samples didnt lead to encoded frame?!?!\n");
			}
			iFramesize = FFMPEGaudioEncodeFrame(mpEncoderHandle, NULL /*FLUSH!*/, mpOutBuf);
		}	while (iFramesize > 0);

	} else {
		dprintf_full("MP3Encoder::close ... nothing left to decode.\n");
	}
	
	
	return frmList;
}


/***************************************************/
Adaptor *MP3Encoder::clone() {
	char* copyEncoderConfig=NULL;
#ifdef ENABLE_FFMPEG
		AVCodec *codec=NULL;
		/* Make sure that the required encoder exists in the ffmpeg library */
		codec = avcodec_find_encoder(this->outputFormat);
		if(codec) {
			AVCodecContext* srcClone = avcodec_alloc_context();
			avcodec_get_context_defaults(srcClone);
			
			AVCodecContext* src=(AVCodecContext*) mpEncoderHandle;
			// don't do a memcpy --> pointers!
			srcClone->codec=codec;
			srcClone->b_frame_strategy=src->b_frame_strategy;
			srcClone->b_quant_factor=src->b_quant_factor;
			srcClone->b_quant_offset=src->b_quant_offset;
			srcClone->bit_rate=src->bit_rate;
			srcClone->rtp_payload_size=src->rtp_payload_size;
			srcClone->width=src->width;
			srcClone->height=src->height;
			srcClone->gop_size=srcClone->gop_size;
			srcClone->qmin=src->qmin;
			srcClone->qmax=src->qmax;
			srcClone->max_b_frames=src->max_b_frames;
			srcClone->sample_aspect_ratio.den=src->sample_aspect_ratio.den;
			srcClone->sample_aspect_ratio.num=src->sample_aspect_ratio.num;
			srcClone->rc_max_rate=src->rc_max_rate;
			srcClone->rc_min_rate=src->rc_min_rate;
			srcClone->mb_qmax=src->mb_qmax;
			srcClone->mb_qmin=src->mb_qmin;
			srcClone->time_base=src->time_base;
			// open the codec for encoding
			if (avcodec_open(srcClone, codec) < 0) {
				printf("MP3Encoder::clone(): Could not open codec\n");
				av_free(srcClone);
			}
			else
				copyEncoderConfig=(char*)srcClone;
		}
#else
		dprintf_err("MP3Encoder::clone: compiled without ffmpeg support!\n");
		exit(1);
#endif
	
	return new MP3Encoder(es,muiTargetBitrate,outputFormat,copyEncoderConfig);
}


