 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	MP4Decoder.cpp                                                *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

// MP4Decoder.cpp: Implementierung der Klasse MP4Decoder.
//
//////////////////////////////////////////////////////////////////////

#include "MP4Decoder.hpp"

#include "ESInfo.hpp"
#include "UncompressedVideoFrame.hpp"
#include "IO.hpp"

#define readMarkerBit()	readMarkerBitFunc(__LINE__)

// Table 6-3: Start code values
#define	VISUAL_OBJECT_SEQUENCE_START_CODE	0x1B0
#define VISUAL_OBJECT_SEQUENCE_END_CODE		0x1B1
#define USER_DATA_START_CODE				0x1B2
#define GROUP_OF_VOP_START_CODE				0x1B3
#define VIDEO_SESSION_ERROR_CODE			0x1B4
#define VISUAL_OBJECT_START_CODE			0x1B5
#define VOP_START_CODE 						0x1B6

// Table 6-12: pixel aspect ratio
#define PIXEL_ASPECT_RATIO_Forbidden		0
#define PIXEL_ASPECT_RATIO_1_1				1
#define PIXEL_ASPECT_RATIO_12_11			2
#define PIXEL_ASPECT_RATIO_10_11			3
#define PIXEL_ASPECT_RATIO_16_11			4
#define PIXEL_ASPECT_RATIO_40_33			5
#define PIXEL_ASPECT_RATIO_extended_PAR		0xF

// Table 6-14: VOL shape type
#define	VOL_SHAPE_rectangular	0
#define	VOL_SHAPE_binary		1
#define	VOL_SHAPE_binaryonly	2
#define	VOL_SHAPE_grayscale		3

static u32 __inline
log2bin(u32 value)
{
/* Changed by Chenm001 */
// #if !defined(_MSC_VER)
	int n = 0;

	while (value) {
		value >>= 1;
		n++;
	}
	return n;
	/*
#else
	__asm {
		bsr eax, value 
		inc eax
	}
#endif
	*/
}

/***************************************************/
bool MP4Decoder::isValidVisualFrame(const u8* payload) { // 0 0 1 b6
	return (payload && payload[0]==0 && payload[1]==0 
					&& payload[2]==1 && payload[3] > 0xb0); // 0xb6 or 0xb2
}



/***************************************************/
bool MP4Decoder::isValidDecoderConfig(const u8* payload) { // 0 0 1 X != b6
	return (payload && payload[0]==0 && payload[1]==0 
					&& payload[2]==1 && (payload[3] == 0xb0 || payload[3] == 0x01));
}



/** 
 * Constructor creates an uninitialized Mp4Decoder
*/ 
MP4Decoder::MP4Decoder(VideoESInfo * pESInfo, bool enableSwitching, UncompressedVideoFrame::ColorSpace colorSpace, meCoderType coderType) {
	mpDecoder = NULL;
	mpOutBuf=NULL;
	this->enableSwitching=enableSwitching;  
	es = pESInfo;
	mColorSpace = colorSpace;
	mFrameSize = 0;
	decConfig=NULL;
	decConfigSize=0;
	newDecConfig=NULL;
	meCoder = coderType;
	initialized = false;
	if(coderType==XVID) {
		dprintf_full("MP4Decoder::MP4Decoder-XVID\n");
		strcpy(name,"MP4Decoder-XVID");
	}
	else {
		dprintf_full("MP4Decoder::MP4Decoder-FFMPEG\n");
		strcpy(name,"MP4Decoder-FFMPEG");
	}
	auBuffer[0] = NULL;
	auBuffer[1] = NULL;
	muiWidth=0;
	muiHeight=0;

	mpBuffer = new char[1024*1024*1]; //1 MByte

	dprintf_full("MP4Decoder Construct colorSpace %s, bitDepth %i\n",
							 colorSpace==UncompressedVideoFrame::ColorSpaceGRAY8?"RGB8gray":
							 colorSpace==UncompressedVideoFrame::ColorSpaceRGB24?"RGB24":
							 colorSpace==UncompressedVideoFrame::ColorSpaceRGB32?"RGB32":
							 colorSpace==UncompressedVideoFrame::ColorSpaceRGB565?"RGB565":
							 colorSpace==UncompressedVideoFrame::ColorSpaceI420?"I420":
							 colorSpace==UncompressedVideoFrame::ColorSpaceYV12?"YV12":
							 "UNKNOWN",
							 UncompressedVideoFrame::mapColorSpaceToBitDepth(colorSpace));
}


/***************************************************/
MP4Decoder::~MP4Decoder() {
	dprintf_full("MP4Decoder::~MP4Decoder()\n");
  if (initialized && mpDecoder) {
		DecoderClose((char **) (&mpDecoder));
		mpDecoder=NULL;
	}
	if(decConfig)
		delete[] decConfig;
	decConfig=NULL;

	if(newDecConfig)
		delete[] newDecConfig;
	newDecConfig=NULL;
	
	if(mpOutBuf)
		delete[] mpOutBuf;
	mpOutBuf=NULL;

	if(auBuffer[0])
		delete auBuffer[0];	
	if(auBuffer[1])
		delete auBuffer[1];	

	delete [] mpBuffer;
	mpBuffer = NULL;
}


/***************************************************/
void MP4Decoder::initialize() {
	if (!es) {
		initialized = false;
		return;
	}
	if(initialized)
		return;
	
	initializeMP4HeaderStruct();

	if(decConfig) {
		delete[] decConfig;
		decConfig=NULL;
	}
	if(newDecConfig) {
		delete[] newDecConfig;
		newDecConfig=NULL;
	}
	if(mpOutBuf) {
		delete[] mpOutBuf;
		mpOutBuf=NULL;
	}
	if(auBuffer[0]) {
		delete auBuffer[0];	
		auBuffer[0]=NULL;
	}
	if(auBuffer[1]) {
		delete auBuffer[1];		
		auBuffer[1]=NULL;
	}
  if (mpDecoder) {
		DecoderClose((char **) (&mpDecoder));
		mpDecoder=NULL;
	}
	initialized = true;
	//has to keep the old value for switching
	if (!enableSwitching) {
		muiWidth = 0;
		muiHeight = 0;
		muiLastFrameStart = 0;
		muiLastFrameEnd = 0;
	}
	muiFrameNumber = 0;
	
	decConfig=NULL;
	
	decConfigSize=0;
	newDecConfig=NULL;
    
	// muiLastFrameEnd = pESInfo->getHeaders((u8 *) mpBuffer);
	
	muiLastFrameEnd = es->getHeaders(&decConfig);
	decConfigSize=muiLastFrameEnd;
	memcpy(mpBuffer,decConfig,muiLastFrameEnd);
	

	dprintf_full("MP4Decoder::initialize() %s\r\n",es->getEncodedDecoderConfig());

#if VITOOKI_DEBUG_LEVEL >= 2
	dprintf_full("MP4Decoder::initialize mpBuffer:: ");
    for (uint iiii = 0; iiii < muiLastFrameEnd; iiii++)
			printf("%02x ", (u8)mpBuffer[iiii]);
    dprintf("\r\n");    
#endif

	if (muiLastFrameEnd == 0) {
		dprintf_err("MP4Decoder::initialize: failed to get header for ESInfo object\n");
  }
  initializeParser((u8 *) mpBuffer, muiLastFrameEnd);
  if (!parseVOSHeader())
		dprintf_small("No VOS Header found\n");
  if (!parseVOHeader())
		dprintf_small("No VO Header found\n");
  if (!parseUserData())
		dprintf_small("No UserData found\n");
  if (!parseVideoObjectHeader())
		dprintf_small("No VideoObject Header found\n");
  if (!parseVOLHeader())
		dprintf_small("No VOL Header found\n");

	//on switching, ignore the arriving size, but scale to the very first ES
	//	if (!enableSwitching || muiWidth == 0) { // only set for first time
	if (1==1) { 
		if(es->getCodecID() == CODEC_ID_MPEG4) {
			muiWidth = mp4Header.video_object_layer_width;
			muiHeight = mp4Header.video_object_layer_height;

#ifdef WINCE
			XvidSetHeight(muiHeight);
			XvidSetWidth(muiWidth);
#endif

			if(es->getAspectRatio() <= 0.0)
				es->setAspectRatio((float)muiWidth/muiHeight);
		}
		
		if(muiWidth <=0 || muiHeight <= 0) {
			muiWidth = es->getWidth();
			muiHeight = es->getHeight();
		}
	} else {
		dprintf_full("MP4Decoder::initialize: stream switching: upscale %ix%i --> %ix%i\n",
								 es->getWidth(),es->getHeight(),muiWidth, muiHeight);
	}
	
	dprintf_small("MP4Decoder::initialize: DecoderInit to codec %s (id %i) size %ix%i\n",
								es->getCodecID() == CODEC_ID_MPEG4?"MPEG4":"OTHER",es->getCodecID(), muiWidth, muiHeight);
  int ret = DecoderInit((char **) (&mpDecoder), es->getCodecID(), muiWidth, muiHeight);

	if ((ret != ERR_OK) || (mpDecoder == NULL)) {
		dprintf_err("MP4Decoder::initialize: Error creating decoder handle (ret=%d)\n",ret);
		exit(-1);
  } else {
		dprintf_small("MP4Decoder::initialize: ok, Decoder set to size %ix%i\n", muiWidth, muiHeight);
		//FIXME: hmm... should this be done after streamSwitching and upscale?
		es->setHeight(muiHeight);
		es->setWidth(muiWidth);
		dprintf_full("Mp4Decoder::init Aspect Info %i height %i width %i\n",
			this->mp4Header.aspectRatio.aspect_ratio_info,
			this->mp4Header.aspectRatio.par_height,
			this->mp4Header.aspectRatio.par_width);
		if (mp4Header.vop_time_increment_resolution > 0) {
			if (mp4Header.vop_time_increment_resolution < 61) {
				// treat as frame rate
				// most codecs use it as one
				u32 ti = es->getMediaTimeScale() / mp4Header.vop_time_increment_resolution;
				if(ti > 0)
					es->setVOPTimeIncrement(ti);
			}
			else if ((es->getMediaTimeScale() > mp4Header.vop_time_increment_resolution) && 
							 ((es->getMediaTimeScale() / mp4Header.vop_time_increment_resolution) < 61))
				es->setVOPTimeIncrement(mp4Header.vop_time_increment_resolution);
		}

		// Process ONLY colorspaces defined in MP4Decoder.hpp
		switch (mColorSpace) {
		case UncompressedVideoFrame::ColorSpaceGRAY8:
			mFrameSize = sizeof(u8) * muiWidth * muiHeight;
			break;
		case UncompressedVideoFrame::ColorSpaceRGB565:
			mFrameSize = sizeof(u8) * muiWidth * muiHeight * 2;
			break;
		case UncompressedVideoFrame::ColorSpaceRGB24:
			mFrameSize = sizeof(u8) * muiWidth * muiHeight * 3;
			break;
		case UncompressedVideoFrame::ColorSpaceRGB32:
			mFrameSize = sizeof(u8) * muiWidth * muiHeight * 4;
			break;
		case UncompressedVideoFrame::ColorSpaceYV12: // fall through
		case UncompressedVideoFrame::ColorSpaceI420:
			mFrameSize = sizeof(u8) * muiWidth * muiHeight * 3 / 2;
			break;
		default:
			dprintf_err("MP4Decoder::initialize: Unknown colorspace\n");
			exit(-1);
		}

		dprintf_full("MP4Decoder::initialize: Esinfo %llu BufferSize %i W %i H %i\n",
			es->getStreamId(), es->getBufferSize(), muiWidth,	muiHeight);
	}
}


/***************************************************/
list < Frame * >MP4Decoder::adapt(Frame * frm) {	
  dprintf_full("MP4Decoder::adapt frame CTS %i size %i\n",frm->getAU()->cts,frm->getAU()->size);
  if(es && enableSwitching) {
    if(decConfigSize>0) {
      if(newDecConfig) {
				delete[] newDecConfig;
				newDecConfig=NULL;
      }
			
      u32 nSize=es->getHeaders(&newDecConfig);
      if(nSize != 0 && nSize != decConfigSize) {
				// we have a new decConfig
				if(decConfig)
					delete[] decConfig;
				decConfig=newDecConfig;
				newDecConfig=NULL;
				initialized=false;
      }
      else if(nSize == decConfigSize) {
				// same size but might be different
				if(memcmp(decConfig,newDecConfig,decConfigSize)!=0) {
					// is different
					if(decConfig)
						delete[] decConfig;
					decConfig=newDecConfig;
					newDecConfig=NULL;
					initialized=false;
				}
      }
    }
  }
	
	
	if (!initialized)
		initialize();
	
	list < Frame * >frmList;
	if (!initialized)
		return frmList;
	if(mpOutBuf) {
		delete[] mpOutBuf;
		mpOutBuf=NULL;
	}


#ifndef WINCE

	mpOutBuf = new u8[mFrameSize];
	
	// The Decoder expects in case of BFRAMES_DEC enabled the 
	// B-Frame together with the reference frame. Therefore one frame is 
	// always buffered for this case.
	if (muiFrameNumber == 0) {
	//first send the header to the decoder ...
	lenDecoded = DecodeFrame((char *) mpDecoder, mpBuffer, muiLastFrameEnd,(char *) mpOutBuf, mColorSpace);
	if(lenDecoded <= 0) {
			lenDecoded = abs(lenDecoded);
			dprintf_err("MP4Decoder::adapt :  muiFrameNumber==0, no decoder output for header frame!\n");
		}
		
	// ... and then the first frame. 
	memcpy(mpBuffer, (u8 *) frm->getAU()->payload, frm->getAU()->size);
	muiLastFrameEnd = frm->getAU()->size;
		
	//decode but don't return the output... this saves the first frame in the ffmpeg buffer....
	lenDecoded = DecodeFrame((char *) mpDecoder, mpBuffer, muiLastFrameEnd,
								(char *) mpOutBuf, mColorSpace);
		dprintf_full("MP4Decoder::adapt :  muiFrameNumber==0, decoder output1 len %i\n",lenDecoded);
		
	// ... and again first frame: required to cope with random access after stream switching
	memcpy(mpBuffer, (u8 *) frm->getAU()->payload, frm->getAU()->size);
	lenDecoded = DecodeFrame((char *) mpDecoder, mpBuffer, muiLastFrameEnd,
								(char *) mpOutBuf, mColorSpace);
		dprintf_full("MP4Decoder::adapt :  muiFrameNumber==0, decoder output2 len %i\n",lenDecoded);

	lenDecoded = 0;	// cause this frame to be decoded at next adapt() call
	if(auBuffer[1])
		delete auBuffer[1];
	auBuffer[1] = new AU(frm->getAU());
		if(mpOutBuf) {
			delete[] mpOutBuf;
			mpOutBuf=NULL;
		}
  } else { // muiFrameNumber != 0


		//	int oldMuiLastFE = muiLastFrameEnd;

	if (lenDecoded > 0) {
		memmove(mpBuffer, mpBuffer + lenDecoded, muiLastFrameEnd - lenDecoded);
	}

	//copy input frame to mpBuffer (which is furthermore used by the decoder)
	//FIXME: check for BUFFER and frm->size!!!
	memcpy(mpBuffer + (muiLastFrameEnd - lenDecoded),
					 (u8 *) frm->getAU()->payload, frm->getAU()->size);
	muiLastFrameStart = 0;


	//	char *mpBufferToFrame = mpBuffer;
	//	int currentFrameLen;
	muiLastFrameEnd = muiLastFrameEnd - lenDecoded + frm->getAU()->size;

#endif

	if(auBuffer[0])
		delete auBuffer[0];
	auBuffer[0] = auBuffer[1];
	//assert(auBuffer[1]);
	auBuffer[1] = new AU(frm->getAU());


#ifndef WINCE	
    lenDecoded = DecodeFrame((char *) mpDecoder, mpBuffer, muiLastFrameEnd, (char *) mpOutBuf, mColorSpace);
		dprintf_full("MP4Decoder::adapt : later frames, decoder output len %i\n",lenDecoded);

#else

	u8 *y,*u,*v;
	if (muiFrameNumber == 0) {
		lenDecoded = XVIDDecodeFrameCE(mpBuffer, //(char*)frm->getAU()->payload, //mpBuffer + (muiLastFrameEnd - lenDecoded), 
							muiLastFrameEnd, 
							y, 
							u, 
							v);
		muiLastFrameEnd = frm->getAU()->size;
		auBuffer[0] = new AU(frm->getAU());
	} else {
		muiLastFrameEnd = muiLastFrameEnd - lenDecoded + frm->getAU()->size;
	}
	

	lenDecoded = XVIDDecodeFrameCE((char*)frm->getAU()->payload, //mpBuffer + (muiLastFrameEnd - lenDecoded), 
							muiLastFrameEnd, 
							y, 
							u, 
							v);

#endif


	if(lenDecoded <= 0) {
		lenDecoded = abs(lenDecoded);
	dprintf_err("MP4Decoder::adapt Warning: no decoder output for frame %d\n", muiFrameNumber);
	//			dprintf_full("MP4Decoder::adapt %x %x %x %x\r\n",
	//			(char)mpBuffer[0],(char) mpBuffer[1],(char)mpBuffer[2],(char)mpBuffer[3]);
    } else {
	  //copy result (decoded frame) to new access unit...
	AU * pAU = auBuffer[0];
	auBuffer[0]=NULL;
	pAU->size = mFrameSize;
	pAU->payload = mpOutBuf; 
	dprintf_full("MP4Decoder::adapt Decode Frame No %d uncompressed size %d CTS %i\n",
									 muiFrameNumber, pAU->size, pAU->cts);
	Frame::FrameType type;
	switch (mColorSpace) {
		case UncompressedVideoFrame::ColorSpaceGRAY8:
		case UncompressedVideoFrame::ColorSpaceRGB565:
		case UncompressedVideoFrame::ColorSpaceRGB24:
		case UncompressedVideoFrame::ColorSpaceRGB32:
			type = Frame::RGB_VOP;
			break;
		case UncompressedVideoFrame::ColorSpaceYV12: // fall through
		case UncompressedVideoFrame::ColorSpaceI420:
			type = Frame::YUV_VOP;				
			break;
		default:
			dprintf_err("MP4Decoder::adapt: Unknown colorspace\n");
			exit(-1);
	}
	UncompressedVideoFrame * decFrame = new UncompressedVideoFrame(type, muiWidth, muiHeight);
	decFrame->setAU(pAU);
	dprintf_full("MP4Decoder::adapt(): created UncompressedVideoFrame with type=%d, width=%d, height=%d (ausize=%d)\n", type, muiWidth, muiHeight, decFrame->getAU()->size);
	dprintf_full("MP4Decoder::adapt(): frametype = %d , type = %d\n", decFrame->getType(), type);
	
#ifdef WINCE
	  decFrame->setAUy(y);
	  decFrame->setAUu(u);
	  decFrame->setAUv(v);
#endif

	decFrame->setMediaTimeScale(frm->getMediaTimeScale());
	decFrame->setColorSpace(mColorSpace);
	frmList.push_back(decFrame);
	mpOutBuf=NULL;
    }
    
#ifndef WINCE
  }
#endif
	
	muiFrameNumber++;
	return frmList;
}


/***************************************************/
list < Frame * >MP4Decoder::close() {
	list < Frame * >frmList;
	if (!initialized)
		return frmList;
	if(mpOutBuf) {
		delete[] mpOutBuf;
		mpOutBuf=NULL;
	}

	mpOutBuf = new u8[mFrameSize];
#ifndef WINCE
	if (dec != NULL && muiFrameNumber != 0)
#else
	if (muiFrameNumber != 0)
#endif
		{
#ifndef WINCE
			if (DecodeFrame((char *) mpDecoder, mpBuffer + muiLastFrameStart,
											muiLastFrameEnd, (char *) mpOutBuf, mColorSpace) == 0)
				{
					dprintf_err("MP4Decoder::close Error decoding frame %d\n", muiFrameNumber);
					delete[] mpOutBuf;mpOutBuf=NULL;
					if(auBuffer[1]) {
						delete auBuffer[1];
						auBuffer[1]=NULL;
					}
				}
			
			else
				{
					AU * pAU = auBuffer[1];
					pAU->size = mFrameSize;
					pAU->payload = mpOutBuf;
					dprintf_full("MP4Decoder::close Decode Frame No %d uncompressed size %d  CTS %i\n",
											 muiFrameNumber, pAU->size, pAU->cts);
					muiFrameNumber++;
					Frame::FrameType type;
					switch (mColorSpace) {
					case UncompressedVideoFrame::ColorSpaceGRAY8:
					case UncompressedVideoFrame::ColorSpaceRGB565:
					case UncompressedVideoFrame::ColorSpaceRGB24:
					case UncompressedVideoFrame::ColorSpaceRGB32:
						type = Frame::RGB_VOP;
						break;
					case UncompressedVideoFrame::ColorSpaceYV12: // fall through
					case UncompressedVideoFrame::ColorSpaceI420:
						type = Frame::YUV_VOP;				
						break;
					default:
						dprintf_err("MP4Decoder::close: Unknown colorspace\n");
						exit(-1);
					}
					UncompressedVideoFrame * decFrame = new UncompressedVideoFrame(type, muiWidth, muiHeight);
					decFrame->setAU(pAU);
					dprintf_full("MP4Decoder::close(): created UncompressedVideoFrame with type=%d, width=%d, height=%d (ausize=%d)\n", type, muiWidth, muiHeight, decFrame->getAU()->size);
					dprintf_full("MP4Decoder::close(): frametype = %d , type = %d\n", decFrame->getType(), type);

#ifdef WINCE
					decFrame->setAUy(y);
					decFrame->setAUu(u);
					decFrame->setAUv(v);
#endif

					decFrame->setMediaTimeScale(es->getMediaTimeScale());
					decFrame->setColorSpace(mColorSpace);
					decFrame->setColorSpace(mColorSpace);
					frmList.push_back(decFrame);
					mpOutBuf=NULL;auBuffer[1]=NULL;
				}
#endif
		}
	else {
		delete[] mpOutBuf;mpOutBuf=NULL;
		if(auBuffer[1]) {
			delete auBuffer[1];
			auBuffer[1]=NULL;
		}
	}	
	if (initialized) {
		DecoderClose((char **) (&mpDecoder));
		initialized = false;
	}
	
	return frmList;
}


/***************************************************/
Adaptor * MP4Decoder::clone() {
	return new MP4Decoder(es,enableSwitching,mColorSpace,meCoder);
}


/***************************************************/
bool MP4Decoder::parseVOSHeader() {
  if (nextBits(32) == VISUAL_OBJECT_SEQUENCE_START_CODE)
		{
			getBits(32);
			mp4Header.profile_and_level_indication = getBits(8);
		}
	
  else
		return false;
	
  return true;
}


/***************************************************/
bool MP4Decoder::parseVOHeader() {
	if (nextBits(32) == VISUAL_OBJECT_START_CODE)
		{
			getBits(32);
			mp4Header.is_visual_object_identifier = getBit();
			if (mp4Header.is_visual_object_identifier)
				{
					mp4Header.visual_object_verid = getBits(4);
					mp4Header.visual_object_priority = getBits(3);
				}
			mp4Header.visual_object_type = getBits(4);
			if (mp4Header.visual_object_type == 1	/*video ID */
					|| mp4Header.visual_object_type == 2 /*still texture ID */ )
				{
					video_signal_type_func();
				}
			nextStartCode();
		}
	
	else
		return false;
	return true;
}


/***************************************************/
bool MP4Decoder::video_signal_type_func() {
	int video_signal_type = getBit();
	if (video_signal_type)
		{
			
			//no variables needed, but getBits is advancing in memory space!
			//int video_format = 
			getBits(3);
			//int video_range = 
			getBit();
			int colour_description = getBit();
			if (colour_description)
				{
					//int colour_primaries = 
					getBits(8);
					//int transfer_characteristics = 
			getBits(8);
			//int matrix_coefficients = 
			getBits(8);
				}
		}
	
	else
		return false;
	return true;
}


/***************************************************/
bool MP4Decoder::parseUserData() {
	if (nextBits(32) == USER_DATA_START_CODE)
		{
			getBits(32);
			
			// FIXME: parse till next startcode
		}
	
	else
		return false;
	return true;
}


/***************************************************/
bool MP4Decoder::parseVideoObjectHeader() {
	int iStartCode = 0;
	iStartCode = nextBits(32);
	if (iStartCode >= 0x100 && iStartCode <= 0x11F)
		{
			getBits(32);
			mp4Header.video_object_layer_startcode = iStartCode;
		}
	
	else
		return false;
	return true;
}


/***************************************************/
bool MP4Decoder::parseVOLHeader() {
	if (nextBits(32) >= 0x120 && nextBits(32) <= 0x12F)
		{
			mp4Header.short_video_header = 0;
			mp4Header.video_object_layer_start_code = getBits(32);
			mp4Header.video_object_layer_id =
				mp4Header.video_object_layer_start_code & 0x0F;
			mp4Header.random_accessbile_vol = getBit();
			mp4Header.video_object_type_indication = getBits(8);
			mp4Header.is_object_layer_identifier = getBit();
			if (mp4Header.is_object_layer_identifier) {
				mp4Header.video_object_layer_verid = getBits(4);
				mp4Header.video_object_layer_priority = getBits(3);
			}
			mp4Header.aspectRatio.aspect_ratio_info = getBits(4);
			if (mp4Header.aspectRatio.aspect_ratio_info ==
					PIXEL_ASPECT_RATIO_extended_PAR) {
				mp4Header.aspectRatio.par_width = getBits(8);
				mp4Header.aspectRatio.par_height = getBits(8);
			}
			mp4Header.volControlParameters.vol_control_parameters = getBit();
			if (mp4Header.volControlParameters.vol_control_parameters)
				{
					mp4Header.volControlParameters.chroma_format = getBits(2);
					mp4Header.volControlParameters.low_delay = getBit();
					mp4Header.volControlParameters.vbv_parameters = getBit();
					if(mp4Header.volControlParameters.vbv_parameters)	{
						mp4Header.volControlParameters.first_half_bit_rate =
							getBits(15);
						readMarkerBit();
						mp4Header.volControlParameters.latter_half_bit_rate =
							getBits(15);
						readMarkerBit();
						mp4Header.volControlParameters.first_half_vbv_buffer_size =
							getBits(15);
						readMarkerBit();
						mp4Header.volControlParameters.latter_half_vbv_buffer_size =
							getBits(3);
						mp4Header.volControlParameters.first_half_vbv_occupancy =
							getBits(11);
						readMarkerBit();
						mp4Header.volControlParameters.latter_half_vbv_occupancy =
							getBits(15);
						readMarkerBit();
					}
				}
			mp4Header.video_object_layer_shape = getBits(2);
			readMarkerBit();
			// what does this value tell, its not the vop time increment!!!
			// it seems to be the frame rate, tested with 5 videos crated with 3 different codecs
			// that's stupid! so ignore...
			u32 vop_inc = getBits(16);
			dprintf_full("Detected  VOPTimeIncrement of %u\n",vop_inc);
			if (vop_inc < 9000) //sanity test: inc 9000 means 10 fps.... 
				mp4Header.vop_time_increment_resolution = vop_inc;
			readMarkerBit();
			/*
				int i = mp4Header.vop_time_increment_resolution;
				while (i > 0)
				{
				i = i >> 1;
				mp4Header.vop_time_incrememt_bits_needed++;
				}
			*/
			mp4Header.vop_time_incrememt_bits_needed = log2bin(mp4Header.vop_time_increment_resolution-1);
			mp4Header.fixed_vop_rate = getBit();
			if (mp4Header.fixed_vop_rate) {
				mp4Header.fixed_vop_time_increment = getBits(mp4Header.vop_time_incrememt_bits_needed);
				// not working
			}	else
				dprintf_full("no fixed VOP TIME Increment!\n");
			
		if (mp4Header.video_object_layer_shape != VOL_SHAPE_binaryonly)
			{
				if (mp4Header.video_object_layer_shape ==
						VOL_SHAPE_rectangular)
					{
						readMarkerBit();
						mp4Header.video_object_layer_width = getBits(13);
						readMarkerBit();
						mp4Header.video_object_layer_height = getBits(13);
						readMarkerBit();
					}
				mp4Header.interlaced = getBit();
				mp4Header.obmc_disable = getBit();
				mp4Header.sprite_enable = getBit();
			if (mp4Header.sprite_enable)
				return false;
			mp4Header.not_8_bit = getBit();
			if (mp4Header.not_8_bit)
				{
					mp4Header.quant_precision = getBits(4);
					mp4Header.bits_per_pixel = getBits(4);
				}
			if (mp4Header.video_object_layer_shape == VOL_SHAPE_grayscale)
				{
					mp4Header.no_gray_quant_update = getBit();
					mp4Header.composition_method = getBit();
					mp4Header.linear_composition = getBit();
				}
			mp4Header.quant.quant_type = getBit();
			if (mp4Header.quant.quant_type)
				{
					mp4Header.quant.load_intra_quant_mat = getBit();
					if (mp4Header.quant.load_intra_quant_mat)
						loadQuantizerMatrix(mp4Header.quant.
																intra_quant_matrix);
					mp4Header.quant.load_nonintra_quant_mat = getBit();
				if (mp4Header.quant.load_nonintra_quant_mat)
					loadQuantizerMatrix(mp4Header.quant.
															nonintra_quant_matrix);
				if (mp4Header.video_object_layer_shape ==
						VOL_SHAPE_grayscale)
				{
					mp4Header.quant.load_intra_quant_mat_grayscale =
						getBit();
					if (mp4Header.quant.load_intra_quant_mat_grayscale)
						loadQuantizerMatrix(mp4Header.quant.
																intra_quant_matrix_grayscale);
					mp4Header.quant.load_nonintra_quant_mat_grayscale =
						getBit();
					if (mp4Header.quant.load_nonintra_quant_mat_grayscale)
						loadQuantizerMatrix(mp4Header.quant.
																nonintra_quant_matrix_grayscale);
				}
				}
			mp4Header.complexity_estimation_disable = getBit();
			if (!mp4Header.complexity_estimation_disable)
				return false;
			mp4Header.resync_marker_disable = getBit();
			mp4Header.data_partitioned = getBit();
			if (mp4Header.data_partitioned)
				mp4Header.reversible_vlc = getBit();
			mp4Header.scalability = getBit();
			if (mp4Header.scalability)
				return false;
			}
		
		else
			mp4Header.resync_marker_disable = getBit();
		}
	
	else
		{
			return false;
		}
	return true;
}


/***************************************************/
void MP4Decoder::initializeParser(u8 * pFrame, u32 iSize) {
	mcByte = 0;
	if (pFrame != NULL)
		{
			mpcData = pFrame;
			miSize = iSize;
			
			
			/*
				u8 * bla = new u8[miSize*2+1];
				ISOMP4ContainerFile::encodeDecoderConfig(bla, mpcData, miSize);
				dprintf("---Decoder  %s\n", bla);
				fflush(stdout);
				fflush(stderr);
			*/
		}
	
	else
		{
			mpcData = NULL;
			miSize = 0;
		}
	mbEOD = false;
	miCurrentNoBytes = 0;
	miCurrentBit = -1;
}


/***************************************************/
void MP4Decoder::initializeMP4HeaderStruct() {
	unsigned int intra_quant_mat[64] = 
		{ 8, 17, 18, 19, 21, 23, 25, 27, 17, 18, 19, 21, 23, 25, 27, 28,
			20,
			21, 22, 23, 24, 26, 28, 30, 21, 22, 23, 24, 26, 28, 30, 32, 22,
			23, 24,
			26, 28, 30, 32, 35, 23, 24, 26,
			28, 30, 32, 35, 38, 25, 26, 28, 30, 32, 35, 38, 41, 27, 28, 30,
			32, 35,
			38, 41, 45 
    };
	unsigned int nonintra_quant_mat[64] = 
		{ 16, 17, 18, 19, 20, 21, 22, 23, 17, 18, 19, 20, 21, 22, 23, 24,
			18,
			19, 20, 21, 22, 23, 24, 25, 19, 20, 21, 22, 23, 24, 26, 27, 20,
			21, 22,
			23, 25, 26, 27, 28, 21, 22, 23, 24,
			26, 27, 28, 30, 22, 23, 24, 26, 27, 28, 30, 31, 23, 24, 25, 27,
			28, 30,
			31, 33 
    };
	unsigned int intra_quant_mat_grayscale[64] = 
		{ 8, 17, 18, 19, 21, 23, 25, 27, 17, 18, 19, 21, 23, 25, 27, 28,
			20,
			21, 22, 23, 24, 26, 28, 30, 21, 22, 23, 24, 26, 28, 30, 32, 22,
			23, 24,
			26, 28, 30, 32, 35, 23, 24, 26, 28, 30,
			32, 35, 38, 25, 26, 28, 30, 32, 35, 38, 41, 27, 28, 30, 32, 35,
			38, 41,
			45 
    };
	unsigned int nonintra_quant_mat_grayscale[64] = 
		{ 16, 17, 18, 19, 20, 21, 22, 23, 17, 18, 19, 20, 21, 22, 23, 24,
			18,
			19, 20, 21, 22, 23, 24, 25, 19, 20, 21, 22, 23, 24, 26, 27, 20,
			21, 22,
			23, 25, 26, 27, 28, 21, 22, 23, 24, 26, 27,
			28, 30, 22, 23, 24, 26, 27, 28, 30, 31, 23, 24, 25, 27, 28, 30,
			31, 33 
    };
	
	// Visual Object Sequence
	mp4Header.profile_and_level_indication = 0;
  
	// Visual Object
	mp4Header.is_visual_object_identifier = 0;
	mp4Header.visual_object_verid = 0;
	mp4Header.visual_object_priority = 0;
	mp4Header.visual_object_type = 0;
  
	// Video Object
	mp4Header.video_object_layer_startcode = 0;
  
	// Video Object Layer
	mp4Header.video_object_layer_start_code = 0;
	mp4Header.short_video_header = 0;
	mp4Header.video_object_layer_id = 0;
	mp4Header.random_accessbile_vol = 0;
	mp4Header.video_object_type_indication = 0;
	mp4Header.is_object_layer_identifier = 0;
	mp4Header.video_object_layer_verid = 0;
	mp4Header.video_object_layer_priority = 0;
	mp4Header.aspectRatio.aspect_ratio_info = 0;
	mp4Header.aspectRatio.par_width = 0;
	mp4Header.aspectRatio.par_height = 0;
	mp4Header.volControlParameters.vol_control_parameters = 0;
	mp4Header.volControlParameters.chroma_format = 0;
	mp4Header.volControlParameters.low_delay = 0;
	mp4Header.volControlParameters.vbv_parameters = 0;
	mp4Header.volControlParameters.first_half_bit_rate = 0;
	mp4Header.volControlParameters.latter_half_bit_rate = 0;
	mp4Header.volControlParameters.first_half_vbv_buffer_size = 0;
	mp4Header.volControlParameters.latter_half_vbv_buffer_size = 0;
	mp4Header.volControlParameters.first_half_vbv_occupancy = 0;
	mp4Header.volControlParameters.latter_half_vbv_occupancy = 0;
	mp4Header.video_object_layer_shape = 0;
	mp4Header.vop_time_increment_resolution = 0;
	mp4Header.vop_time_incrememt_bits_needed = 0;
	mp4Header.fixed_vop_rate = 0;
	mp4Header.fixed_vop_time_increment = 0;
	mp4Header.video_object_layer_width = 0;
	mp4Header.video_object_layer_height = 0;
	mp4Header.interlaced = 0;
	mp4Header.obmc_disable = 0;
	mp4Header.sprite_enable = 0;
	mp4Header.not_8_bit = 0;
	mp4Header.quant_precision = 0;
	mp4Header.bits_per_pixel = 0;
	mp4Header.no_gray_quant_update = 0;
	mp4Header.composition_method = 0;
	mp4Header.linear_composition = 0;
	mp4Header.quant.load_intra_quant_mat = 0;
	mp4Header.quant.load_nonintra_quant_mat = 0;
	mp4Header.quant.load_intra_quant_mat_grayscale = 0;
	mp4Header.quant.load_nonintra_quant_mat_grayscale = 0;
	memcpy(mp4Header.quant.intra_quant_matrix, intra_quant_mat,
				 sizeof(mp4Header.quant.intra_quant_matrix));
	memcpy(mp4Header.quant.nonintra_quant_matrix, nonintra_quant_mat,
				 sizeof(mp4Header.quant.nonintra_quant_matrix));
	memcpy(mp4Header.quant.intra_quant_matrix_grayscale,
				 intra_quant_mat_grayscale,
				 sizeof(mp4Header.quant.intra_quant_matrix_grayscale));
	memcpy(mp4Header.quant.nonintra_quant_matrix_grayscale,
		nonintra_quant_mat_grayscale,
				 sizeof(mp4Header.quant.nonintra_quant_matrix_grayscale));
	mp4Header.complexity_estimation_disable = 0;
	mp4Header.resync_marker_disable = 0;
	mp4Header.data_partitioned = 0;
	mp4Header.reversible_vlc = 0;
	mp4Header.scalability = 0;
}


/***************************************************/
void MP4Decoder::nextByte() {
	mcByte = mpcData[miCurrentNoBytes];
	miCurrentBit = 7;
	miCurrentNoBytes++;
	if (miCurrentNoBytes > miSize)
		mbEOD = true;
}


/***************************************************/
bool MP4Decoder::bytealigned() {
	if (miCurrentBit == 7)
		return true;
	else
		{
			nextByte();
			return true;
		}
}


/***************************************************/
int MP4Decoder::getBit() {
	int ret = 0;
	if (miCurrentBit == -1)
		nextByte();
	ret = (mcByte >> miCurrentBit) & 0x01;
	miCurrentBit--;
	if (mbEOD == false)
		return ret;
	
	else
		return EOF;
}


/***************************************************/
int MP4Decoder::getBits(int iNumber) {
	int ret = 0;
	int i;
	for (i = 0; i < iNumber; i++)
		ret = (ret << 1) | getBit();
	if (mbEOD == false)
		return ret;
	
	else
		return EOF;
}


/***************************************************/
int MP4Decoder::nextBits(int noBits) {
	int ret = 0;
	bookmark();
	ret = getBits(32);
	restore();
	if (mbEOD == false)
		return ret;
	
	else
		return EOF;
}


/***************************************************/
void MP4Decoder::nextStartCode() {
	while (!bytealigned())
		getBit();
}


/***************************************************/
void MP4Decoder::loadQuantizerMatrix(unsigned int puiMatrix[]) {
	int i;			// row counter
	int j;			// column counter
	int lastValue = 0;
	bool readLastValue = false;
	for (i = 0; i < 8; i++) {
		for (j = 0; j < 8; j++)	{
			if (!readLastValue)	{
				puiMatrix[j + i * 8] = getBits(8);
				if (puiMatrix[j + i * 8] == 0)
					readLastValue = true;
				else
					lastValue = puiMatrix[j + i * 8];
			}
			else
				puiMatrix[j + i * 8] = lastValue;
		}
	}
}


/***************************************************/
void MP4Decoder::readMarkerBitFunc(int iLine) {
	if (getBit() != 1) {
			dprintf_small("Marker bit invalid at line %d\n", iLine);
		}
}


/***************************************************/
void MP4Decoder::bookmark() {
	miBookmarkedCurrentBit = miCurrentBit;
	miBookmarkedByte = mcByte;
	miBookmarkedCurrentNoBytes = miCurrentNoBytes;
}


/***************************************************/
void MP4Decoder::restore() {
	miCurrentBit = miBookmarkedCurrentBit;
	mcByte = miBookmarkedByte;
	miCurrentNoBytes = miBookmarkedCurrentNoBytes;
}


/***************************************************/
void MP4Decoder::setCoderType(MP4Decoder::eCoderType coderType) {
	meCoder = coderType;
}


/***************************************************/
void MP4Decoder::DecoderClose(char **pDecoderHandle) {
	switch (meCoder) {
	case (MP4Decoder::XVID):
		XVIDDecoderClose(pDecoderHandle);
		break;
#ifdef ENABLE_FFMPEG
	case (MP4Decoder::FFMPEG):
		FFMPEGDecoderClose(pDecoderHandle);
		break;
#endif
	default:
		dprintf_err("Invalid Coder selected\n");
		break;
	}
}


/***************************************************/
int MP4Decoder::DecoderInit(char **pDecoderHandle, CodecID codecID, unsigned int uiWidth, unsigned int uiHeight) {
	switch (meCoder) {
	case (MP4Decoder::XVID):
		return XVIDDecoderInit2(pDecoderHandle, muiWidth, muiHeight);
		break;
#ifdef ENABLE_FFMPEG
	case (MP4Decoder::FFMPEG):
		return FFMPEGDecoderInit(pDecoderHandle, codecID, muiWidth, muiHeight);
		break;
#endif
	default:
		dprintf_err("Invalid Coder selected\n");
		break;
	}

	return ERR;
}


/***************************************************/
int MP4Decoder::DecodeFrame(char *pDecoderHandle, char *pCompressedFrame, 
				int iCFSize, char* pYUVFrame, UncompressedVideoFrame::ColorSpace eColorSpace) {
	int ret = 0;
	
	switch (meCoder) {
	case (MP4Decoder::XVID): 
		dprintf_full("Decoding : using XVID\n");
		ret = XVIDDecodeFrame(pDecoderHandle, pCompressedFrame, iCFSize,
													pYUVFrame, (xvidColorSpace) mapToXVIDColorspace(eColorSpace));
		break;
#ifdef ENABLE_FFMPEG
	case (MP4Decoder::FFMPEG):
		dprintf_full("Decoding : using FFMPEG\n");
		ret = FFMPEGDecodeFrame(pDecoderHandle, pCompressedFrame, iCFSize, 
														pYUVFrame, (ffmpegColorSpace) mapToFFMPEGColorspace(eColorSpace));
		break;
#endif
	default:
		dprintf_err("Invalid Coder selected\n");
		break;
	}
	return ret;
}

/***************************************************/
xvidColorSpace MP4Decoder::mapToXVIDColorspace(UncompressedVideoFrame::ColorSpace colorSpace) {
	xvidColorSpace ret = XVIDColorSpaceUnknown;
	
	switch (colorSpace) {
	case UncompressedVideoFrame::ColorSpaceYV12:
		ret = XVIDColorSpaceYV12;
		break;
	case UncompressedVideoFrame::ColorSpaceRGB24:
		ret = XVIDColorSpaceRGB24;
		break;
	case UncompressedVideoFrame::ColorSpaceI420:
		ret = XVIDColorSpaceI420;
		break;
	case UncompressedVideoFrame::ColorSpaceRGB32:
		ret = XVIDColorSpaceRGB32;
		break;
	default:
		ret = XVIDColorSpaceUnknown;
		break;
	}
	
	return ret;
}


/***************************************************/
ffmpegColorSpace MP4Decoder::mapToFFMPEGColorspace(UncompressedVideoFrame::ColorSpace colorSpace) {
	ffmpegColorSpace ret = FFMPEGColorSpaceUnknown;
	
	switch (colorSpace)	{
		case UncompressedVideoFrame::ColorSpaceYV12:
				ret = FFMPEGColorSpaceYV12;
				break;
		case UncompressedVideoFrame::ColorSpaceRGB24:
				ret = FFMPEGColorSpaceRGB24;
				break;
		case UncompressedVideoFrame::ColorSpaceI420:
				ret = FFMPEGColorSpaceI420;
				break;
		case UncompressedVideoFrame::ColorSpaceRGB32:
				ret = FFMPEGColorSpaceRGB32;
				break;
		case UncompressedVideoFrame::ColorSpaceRGB565:
				ret = FFMPEGColorSpaceRGB565;
				break;
		case UncompressedVideoFrame::ColorSpaceGRAY8:
				ret = FFMPEGColorSpaceGRAY8;
				break;
		default:
				ret = FFMPEGColorSpaceUnknown;
				break;
		}
	
	return ret;
}


/***************************************************/
const char* MP4Decoder::codecToString(CodecID codec_id) {
   static const char* txt[]={  
    "NONE", 
    "MPEG1VIDEO",
    "MPEG2VIDEO",
    "MPEG2VIDEO_XVMC",
    "H263",
    "RV10",
    "RV20",
    "MP2",
    "MP3",   
    "VORBIS",
    "AC3",
    "MJPEG",
    "MJPEGB",
    "LJPEG",
    "SP5X",
    "MPEG4",
    "RAWVIDEO",
    "MSMPEG4V1",
    "MSMPEG4V2",
    "MSMPEG4V3",
    "WMV1",
    "WMV2",
    "H263P",
    "H263I",
    "FLV1",
    "SVQ1",
    "SVQ3",
    "DVVIDEO",
    "DVAUDIO",
    "WMAV1",
    "WMAV2",
    "MACE3",
    "MACE6",
    "HUFFYUV",
    "CYUV",
    "H264",
    "INDEO3",
    "VP3",
    "THEORA",
    "AAC",
    "MPEG4AAC",
    "ASV1",
    "ASV2",
    "FFV1",
    "4XM",
    "VCR1",
    "CLJR",
    "MDEC",
    "ROQ",
    "INTERPLAY_VIDEO",
    "XAN_WC3",
    "XAN_WC4",
    "RPZA",
    "CINEPAK",
    "WS_VQA",
    "MSRLE",
    "MSVIDEO1",
    "IDCIN",
    "8BPS",
    "SMC",
    "FLIC",
    "TRUEMOTION1",
    "VMDVIDEO",
    "VMDAUDIO",
    "MSZH",
    "ZLIB",
    "PCM_S16LE",
    "PCM_S16BE",
    "PCM_U16LE",
    "PCM_U16BE",
    "PCM_S8",
    "PCM_U8",
    "PCM_MULAW",
    "PCM_ALAW",
    "ADPCM_IMA_QT",
    "ADPCM_IMA_WAV",
    "ADPCM_IMA_DK3",
    "ADPCM_IMA_DK4",
    "ADPCM_IMA_WS",
    "ADPCM_MS",
    "ADPCM_4XM",
    "ADPCM_XA",
    "ADPCM_ADX",
    "AMR_NB",
    "AMR_WB",
    "RA_144",
    "RA_288",
    "ROQ_DPCM",
    "INTERPLAY_DPCM",
    "XAN_DPCM",    
    "MPEG2TS",
    "UNKNOWN_CODEC"
    };

	if(codec_id>=CODEC_ID_NONE&&codec_id<=CODEC_ID_MPEG2TS) {
		return txt[codec_id-CODEC_ID_NONE];
	}//if
	else{
		return txt[CODEC_ID_MPEG2TS+1];
	}//else	
}//codecToString


/***************************************************/
int MP4Decoder::stringToCodec(const char* codec_string) {
    int i;
    
    static const char* txt[]={  
    "NONE", 
    "MPEG1VIDEO",
    "MPEG2VIDEO",
    "MPEG2VIDEO_XVMC",
    "H263",
    "RV10",
    "RV20",
    "MP2",
    "MP3",   
    "VORBIS",
    "AC3",
    "MJPEG",
    "MJPEGB",
    "LJPEG",
    "SP5X",
    "MPEG4",
    "RAWVIDEO",
    "MSMPEG4V1",
    "MSMPEG4V2",
    "MSMPEG4V3",
    "WMV1",
    "WMV2",
    "H263P",
    "H263I",
    "FLV1",
    "SVQ1",
    "SVQ3",
    "DVVIDEO",
    "DVAUDIO",
    "WMAV1",
    "WMAV2",
    "MACE3",
    "MACE6",
    "HUFFYUV",
    "CYUV",
    "H264",
    "INDEO3",
    "VP3",
    "THEORA",
    "AAC",
    "MPEG4AAC",
    "ASV1",
    "ASV2",
    "FFV1",
    "4XM",
    "VCR1",
    "CLJR",
    "MDEC",
    "ROQ",
    "INTERPLAY_VIDEO",
    "XAN_WC3",
    "XAN_WC4",
    "RPZA",
    "CINEPAK",
    "WS_VQA",
    "MSRLE",
    "MSVIDEO1",
    "IDCIN",
    "8BPS",
    "SMC",
    "FLIC",
    "TRUEMOTION1",
    "VMDVIDEO",
    "VMDAUDIO",
    "MSZH",
    "ZLIB",
    "PCM_S16LE",
    "PCM_S16BE",
    "PCM_U16LE",
    "PCM_U16BE",
    "PCM_S8",
    "PCM_U8",
    "PCM_MULAW",
    "PCM_ALAW",
    "ADPCM_IMA_QT",
    "ADPCM_IMA_WAV",
    "ADPCM_IMA_DK3",
    "ADPCM_IMA_DK4",
    "ADPCM_IMA_WS",
    "ADPCM_MS",
    "ADPCM_4XM",
    "ADPCM_XA",
    "ADPCM_ADX",
    "AMR_NB",
    "AMR_WB",
    "RA_144",
    "RA_288",
    "ROQ_DPCM",
    "INTERPLAY_DPCM",
    "XAN_DPCM",    
    "MPEG2TS",
    "UNKNOWN_CODEC"
    };


	for(i=0;txt[i]!="UNKNOWN_CODEC";i++){
		if(strcmp(txt[i],codec_string)==0){
			return i;
		}//if
	}//for

      return (i-1);

}//stringToCodec


/** returns adaptation costs (CPU only). Decoding costs depend directly on the number 
	  * of out YUV pixels and frames decoded per second (FIXME: codec not supported)
	*/
u32 MP4Decoder::getTranscodingCosts() const
{
	return (u32)(es->getWidth()*es->getHeight()*es->getFPS());
}
