/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	MP4Decoder.hpp                                                *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

/***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef AFX_MP4DECODER_HPP
#define AFX_MP4DECODER_HPP

#if _MSC_VER > 1000
#pragma once
#endif				// _MSC_VER > 1000


#include "ffmpegInterface.h"
#include "xvidInterface.h"

#include "VideoAdaptor.hpp"
#include "UncompressedVideoFrame.hpp"
#ifdef ISOMP4
#include "ISOMovies.h"
#endif
#include "IO.hpp"

class VideoESInfo;
class UncompressedVideoFrame;


/**
 * Implementation of Adaptor that returns YUVFrame objects
 *
 *@author Bernhard Penz
 *
 *@version $Id: MP4Decoder.hpp,v 1.20 2005/12/21 11:00:43 mkropfbe Exp $
 *
 *@short <short description>
 *
 */
class MP4Decoder:public VideoAdaptor {
public:


	/**
		 Codec selection enumeration, currently available are XviD and FFMPEG
	*/
	typedef enum eCoderType{XVID = 1, FFMPEG} meCoderType;

	/**
		 Creates an uninitialized MP4Decoder instance.
		 @param pESInfo Pointer to ESInfo containing VOS/VO/VOL headers
		 @param enableSwitching specifies if the decoder should reinit itself, when
		 a new header is detected. set to false in the server, set to true in the client
		 @param colorSpace xvidColorSpace of output Frame objects. Defaults to ColorSpaceYV12.
		 @param coderType Video codec to use.
		 @see setCoderType
	*/
	MP4Decoder(VideoESInfo * pESInfo, bool enableSwitching = false, 
						 UncompressedVideoFrame::ColorSpace colorSpace = UncompressedVideoFrame::ColorSpaceYV12,
#ifdef ENABLE_FFMPEG   // default codec (if avail) is ffmpeg
						 meCoderType coderType= FFMPEG);
#else
						 meCoderType coderType= XVID);
#endif

	/**
		 Destroys the MP4Decoder
	*/
	virtual ~ MP4Decoder();
		
	static bool isValidVisualFrame(const u8* payload);

	static bool isValidDecoderConfig(const u8* payload);

	/**
		 Compressed frm in - uncompressed YUV12 frame out.

		 @param frm Pointer to compressed Frame
		 @return List of YUVFrame objects
	*/
	list < Frame * >adapt(Frame * frm);

	/**
		 Flush all buffered uncompressed Frame objects (if any).

		 @return List of YUVFrame objects
	*/
	list < Frame * >close();

	/**
		 Not implemented.

		 @return Defaults to NULL.
	*/
	Adaptor *clone();

	/**
		 Initializes the Decoder. Is called when the first frame arrives 
	*/
	void initialize();

	/**
		 Codec selection

		 @param coderType Type of codec to use. Defaults to XviD
		 @see meCoderType
	*/
	void setCoderType(eCoderType coderType = XVID);

	/**
		 Mapping function between MP4Decoders colorspace (defined in UncompressedVideoFrame::ColorSpace)
		 and XviD codec colorspace.

		 @see UncompressedVideoFrame::ColorSpace
		 @see mapToFFMPEGColorspace
	*/
	static xvidColorSpace mapToXVIDColorspace(UncompressedVideoFrame::ColorSpace colorSpace);
	/*
		Maps a CodecID to a String, returns the String UNKNOWN_CODEC if 		@param codec_id is invalid.
	*/
	static const char* codecToString(CodecID codec_id);

	/*
		Maps a String to a CodecID
	*/
	static int stringToCodec(const char* codec_string);
	
	/**
		 Mapping function between MP4Decoders colorspace (defined in UncompressedVideoFrame::ColorSpace)
		 and FFMPEG codec colorspace.

		 @see UncompressedVideoFrame::ColorSpace
		 @see mapToXVIDColorspace
	*/
	static ffmpegColorSpace mapToFFMPEGColorspace(UncompressedVideoFrame::ColorSpace colorSpace);

	/** returns adaptation costs (CPU only). Decoding costs depend directly on the number 
	  * of out YUV pixels and frames decoded per second (FIXME: codec not supported)
	*/
	u32 getTranscodingCosts() const;
private:
	/**
		 Contains the selected codec type

		 @see setCoderType
	*/
	eCoderType meCoder;

	/**
		 Decoder Initialization method. The decoder chosen with setCoderType is initialized.

		 @param pDecoderHandle Handle to decoder instance
		 @param uiWidth Width of stream to decode
		 @param uiHeight Height of stream to decode
		 @return ERR_OK on success, ERR otherwise
	*/
	int DecoderInit(char **pDecoderHandle, CodecID codecID, unsigned int uiWidth, unsigned int uiHeight);

	/**
		 Close Decoder.

		 @param pDecoderHandle Handle to decoder instance created with DecoderInit
		 @see DecoderInit
	*/
	void DecoderClose(char **pDecoderHandle);

	/**
		 Decode a single Frame.

		 @param pDecoderHandle Handle to valid decoder instance
		 @see DecoderInit
		 @param pCompressedFrame Pointer to memory area where a complete compressed frame is stored.
		 @param iCFSize Size of frame stored in pCompressedFrame
		 @param pYUVFrame Pointer to allocated memory area where the uncompressed frame will
		 be written to.
		 @param eColorSpace Colorspace of the uncompressed frame
		 @see meColorSpace
		 @see mapToXVIDColorspace
	*/
	int DecodeFrame(char *pDecoderHandle, char *pCompressedFrame, 
									int iCFSize, char* pYUVFrame, UncompressedVideoFrame::ColorSpace eColorSpace);

	/**
		 Handle to XviD decoder instance.
	*/
	char *mpDecoder;

	/**
		 Colorspace used for decoding.
	*/
	UncompressedVideoFrame::ColorSpace mColorSpace;

	/**
		 Size of decoded payload of Frame object in bytes.
	*/
	u32 mFrameSize;

	/**
		 The decoder will buffer 2 frames. Much more memory than really needed is
		 allocated. 500 kB should be sufficient even for HDTV.

		 @see *frmBuffer
	*/
	char *mpBuffer;

	/**
		 For internal use only
	*/
	int lenDecoded;

	/**
		 For internal use only
	*/
	uint muiLastFrameStart;

	/**
		 For internal use only
	*/
	uint muiLastFrameEnd;

	/**
		 Contains the number of frames decoded.
	*/
	u32 muiFrameNumber;

	/**
		 For internal use only.
	*/
	u8 *mpOutBuf;

	/**
		 The decoder will buffer 2 frames.

		 @see mpBuffer
	*/
	AU *auBuffer[2];

	/**
		 If the decoder was successful created, this variable indicates
		 the decoded picture width.
	*/
	uint muiWidth;

	/**
		 If the decoder was successful created, this variable indicates
		 the decoded picture height.
	*/
	uint muiHeight;

	/** to support switching the decoder must be able
			to detect when a new decConfig was set at the 
			ESInfo object. Thus, buffer the old decCOnfig
			and also provide temporary storage for the new one
	*/
	u8* decConfig;
	u8* newDecConfig;
	u8 decConfigSize;

	/** specifies if the decoder should check if a new ESHeader
			was set. Set to false in the server, set to true in the client */
	bool enableSwitching;


	typedef struct {
		unsigned int aspect_ratio_info;
		unsigned int par_width;
		unsigned int par_height;
	} structAspectRatio;

	typedef struct {
		unsigned int vol_control_parameters;
		unsigned int chroma_format;
		unsigned int low_delay;
		unsigned int vbv_parameters;
		unsigned int first_half_bit_rate;
		unsigned int latter_half_bit_rate;
		unsigned int first_half_vbv_buffer_size;
		unsigned int latter_half_vbv_buffer_size;
		unsigned int first_half_vbv_occupancy;
		unsigned int latter_half_vbv_occupancy;
	} structVOLControlParameters;

	typedef struct {
		unsigned int quant_type;
		unsigned int load_intra_quant_mat;
		unsigned int load_nonintra_quant_mat;
		unsigned int load_intra_quant_mat_grayscale;
		unsigned int load_nonintra_quant_mat_grayscale;

		unsigned int intra_quant_matrix[64];
		unsigned int nonintra_quant_matrix[64];
		unsigned int intra_quant_matrix_grayscale[64];
		unsigned int nonintra_quant_matrix_grayscale[64];
	} structQuant;

	/**
		 DecoderInit2 will store parsed header information in this struct

	*/
	typedef struct {
		// Visual Object Sequence 
		unsigned int profile_and_level_indication;
		// Visual Object
		unsigned int is_visual_object_identifier;
		unsigned int visual_object_verid;
		unsigned int visual_object_priority;
		unsigned int visual_object_type;
		// Video Object Header
		unsigned int video_object_layer_startcode;
		// Video Object Layer
		unsigned int video_object_layer_start_code;
		unsigned int short_video_header;
		unsigned int video_object_layer_id;
		unsigned int random_accessbile_vol;
		unsigned int video_object_type_indication;
		unsigned int is_object_layer_identifier;
		unsigned int video_object_layer_verid;
		unsigned int video_object_layer_priority;

		structAspectRatio aspectRatio;
		structVOLControlParameters volControlParameters;

		unsigned int video_object_layer_shape;
		unsigned int vop_time_increment_resolution;
		unsigned int vop_time_incrememt_bits_needed;
		unsigned int fixed_vop_rate;
		unsigned int fixed_vop_time_increment;

		unsigned int video_object_layer_width;
		unsigned int video_object_layer_height;

		unsigned int interlaced;
		unsigned int obmc_disable;
		unsigned int sprite_enable;
		unsigned int not_8_bit;
		unsigned int quant_precision;
		unsigned int bits_per_pixel;

		unsigned int no_gray_quant_update;
		unsigned int composition_method;
		unsigned int linear_composition;

		structQuant quant;

		unsigned int complexity_estimation_disable;
		unsigned int resync_marker_disable;
		unsigned int data_partitioned;
		unsigned int reversible_vlc;
		unsigned int scalability;
	} structMP4Header;

	/**
		 If the decoder was successfully created, this variable 
		 contains the information stored in VOS/VO/VOL headers.

		 @see DecoderInit2()
	*/
	structMP4Header mp4Header;

	/**
		 For internal use of parseXYZ methods only.

		 This variable contains the bitbuffer for getBit(), getBits() and nextBits() methods.
		 @see getBit()
		 @see getBits(int iNumber)
		 @see nextBits(int iNumber)
	*/
	u8 mcByte;

	/**
		 For internal use of parseXYZ methods only.

		 Byte buffer for nextByte() method.
		 @see nextByte()
		 @see mpcData
	*/
	u8 *mpcData;

	/**
		 For internal use of parseXYZ methods only.

		 Size of mpcData buffer for nextByte() method.
		 @see nextByte()
	*/
	int miSize;

	/**
		 For internal use of parseXYZ methods only.

		 Is set in nextByte() to true when end of mpcData buffer is reached.
		 @see nextByte()
		 @see mpcData
	*/
	bool mbEOD;

	/**
		 For internal use of parseXYZ methods only.

		 Current byte index withing mpcData. Maintained by nextByte().
		 @see nextByte()
	*/
	int miCurrentNoBytes;

	/**
		 For internal use of parseXYZ methods only.

		 Current bit index within CurrentNoBytes byte. 
		 Maintained by getBit and getBits().
		 @see getBit()
		 @see getBits(int iNumber)
		 @see mpcData
	*/
	int miCurrentBit;

	/**
		 For internal use of parseXYZ methods only.

		 Used by bookmark() and restore().
		 @see bookmark()
		 @see restore()
		 @see mcByte
	*/
	int miBookmarkedCurrentBit;

	/**
		 For internal use of parseXYZ methods only.

		 Used by bookmark() and restore().
		 @see bookmark()
		 @see restore()
	*/
	int miBookmarkedByte;

	/**
		 For internal use of parseXYZ methods only.

		 Used by bookmark() and restore().
		 @see bookmark()
		 @see restore()
	*/
	int miBookmarkedCurrentNoBytes;

	/**
		 Initialize internal buffer structures. 

		 Must be called before any call to parseXYZ functions.

		 @see parseVOSHeader()
		 @see parseVOHeader()
		 @see parseUserData()
		 @see parseVideoObjectHeader()
		 @see parseVOLHeader()
		 @see mpcData
	*/
	void initializeParser(u8 * pFramem, u32 iSize);

	/** 
			Initialize a structMP4Header structure.

			Sets everything in a structMP4Header to 0.
			@see structMP4Header
	*/
	void initializeMP4HeaderStruct();

	/**
		 Fetch next byte from mpcData.

		 @see mpcData
		 @see getBit()
		 @see getBits(int iNumber)
		 @see nextBits(int iNumber)
	*/
	void nextByte();

	/**
		 Check if current bit is byte aligned, if not then seek for next byte-aligned position.

		 @see nextStartCode()
		 @return Always true
	*/
	bool bytealigned();

	/**
		 Fetch next bit from bitstream and advance.

		 @see getBits(int iNumber))
		 @see nextBits(int iNumber)
		 @see mcByte
		 @see mpcData
		 @return Value of next bit in bitstream.
	*/
	int getBit();

	/**
		 Fetch next iNumber bits in bitstream and advance.

		 @see getBit()
		 @see nextBits(int iNumber)
		 @see mcByte
		 @see mpcData
		 @return Value of next iNumber bits in bitstream.
	*/
	int getBits(int iNumber);

	/**
		 Fetch next iNumber bits in bitstream. Do not advance.

		 @see getBit()
		 @see getBits(int iNumber)
		 @see mcByte
		 @see mpcData
		 @return Value of next iNumber bits in bitstream.
	*/
	int nextBits(int iNumber);

	/**
		 Locate next startcode.
	*/
	void nextStartCode();

	/**
		 Read quanzizer matrix values from the bitstream.
	*/
	void loadQuantizerMatrix(unsigned int puiMatrix[]);

	/**
		 Check for correct set marker bit.

		 @iLine Codeline where method is called.
	*/
	void readMarkerBitFunc(int iLine);
	/**
		 Save current status of the bitstream.
		 @see restore()
	*/
	void bookmark();
	/**
		 Restore the last bookmark()-ed status of the bitstream.
		 @see bookmark()
	*/
	void restore();

	/**
		 Parse a Visual Object Seqence Header
		 @return True on success, false on error.
	*/
	bool parseVOSHeader();

	/**
		 Parse a Visual Object Header
		 @return True on success, false on error.
	*/
	bool parseVOHeader();

	bool video_signal_type_func();

	/**
		 Parse user data. Does nothing at the moment
		 @return True on success, false on error.
	*/
	bool parseUserData();

	/**
		 Parse a Video Object header.
		 @return True on success, false on error.
	*/
	bool parseVideoObjectHeader();
	/**
		 Parse a Video Object Layer header.
		 @return True on success, false on error.
	*/
	bool parseVOLHeader();
};

#endif
