/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	MP4Encoder.cpp                                                *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

/***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#include "MP4Encoder.hpp"
#include "CompressedVideoFrame.hpp"


/***************************************************/
MP4Encoder::MP4Encoder(VideoESInfo * pESInfo, u32 targetBitRate,  u32 frameRate,
		       u32 uiKeyFrameInterval,  u32 numBFrames, 
		       UncompressedVideoFrame::ColorSpace colorSpace,
		       MP4Decoder::meCoderType coderType, enum CodecID outFormat, char* encoderConfig) {
  this->mpEncoderHandle = encoderConfig;
  es = pESInfo;
  muiTargetBitrate = targetBitRate;
  if (frameRate > 0)
    this->frameRate = frameRate;
  else
    this->frameRate = (int)es->getFPS();
  
  initialized = false;
  muiKeyFrameInterval = uiKeyFrameInterval;
  this->numBFrames = numBFrames;
  mColorSpace = colorSpace;
  meCoder = coderType;
  if(coderType == MP4Decoder::XVID)
    strcpy(name,"MP4Encoder-XVID");
  else
    strcpy(name,"MP4Encoder-FFMPEG");
  outputFormat=outFormat;
  
  if (targetBitRate < 30 * 1024) {
    dprintf_err("MP4Encoder: Constructor: targetBitRate is unreasonably low (%i bits/second?)... bailing out!\n",targetBitRate);
    exit(1);
  }
}


/***************************************************/
void MP4Encoder::initialize() {
	if(initialized)
		return;
	dprintf_full("MP4Encoder::initialize() Handle: %p Width: %i Height: %i BW: %i kbps FPS: %i  numBFrames: %i\n",
							 mpEncoderHandle, es->getWidth(), es->getHeight(),
							 muiTargetBitrate / 1024, frameRate, numBFrames);
	initialized = true;
	if (EncoderInit(&mpEncoderHandle, es->getWidth(), es->getHeight(),
									muiTargetBitrate, frameRate,  numBFrames,  muiKeyFrameInterval, es) != ERR_OK) {
		dprintf_err("MP4Encoder::initialize: Error creating encoder instance!\n");
		initialized = false;
	}
	assert(mpEncoderHandle);
	es->setAvgBandwidth(muiTargetBitrate);
	muiFrameNumber = 0;
}


/***************************************************/
MP4Encoder::~MP4Encoder() {
	EncoderClose(&mpEncoderHandle);
}



/***************************************************/
list < Frame * >MP4Encoder::adapt(Frame * frm) {
	CompressedVideoFrame *mpgFrame = NULL;
	list < Frame * >frmList;
	int iFramesize = 0;
	if (!initialized)
		initialize();
	if (!initialized)
		return frmList;
	
	if ((muiFrameNumber % muiKeyFrameInterval) == 0)
		iFramesize = EncodeFrame(mpEncoderHandle, (char *) frm->getAU()->payload,
														 mpOutBuf, mColorSpace, 1);
	else
		iFramesize = EncodeFrame(mpEncoderHandle, (char *) frm->getAU()->payload,
														 mpOutBuf, mColorSpace, 0);
	
	dprintf_full("MP4Encoder::adapt  Encode Frame No %d compressed size: %d, first bytes: %x %x %x %x\n", 
							 muiFrameNumber, iFramesize,mpOutBuf[0],mpOutBuf[1],mpOutBuf[2],mpOutBuf[3]);
	if (iFramesize > 0) {
		
		AU *pAU = new AU(frm->getAU());
		u8 *payload = new u8[iFramesize];
		
		memcpy(payload, mpOutBuf, iFramesize);
		
		// First frame is tricky:
		//              - obtain headers to set decoderconfig in ESInfo
		//              - remove header from payload, otherwise it will exist two times. 
		//                      FIXME by mike: why should we need this??
		//                     even worse, we need this header, if it wasn't sent via net RTSP/SDP
		
		// Derive headers for ESInfo object from first frame
		if (muiFrameNumber == 0) {
			u32 uiVESHeaderLen = iFramesize;
			int i;
			u32 ch = 0;
			u32 startCode = 0;
			bool foundVOPcode = false;
			
			// Search for first VOP startcode in freshly encoded frame
			foundVOPcode = false;
			i = 0;
			startCode = 0;
			while (!foundVOPcode) {
				if (i >= iFramesize) {
					dprintf_err("MP4Encoder::adapt:FATAL: couldnt find frame MPEG-4 VOP code (00 00 01 B6) in newly coded frame!!!\n");
					exit(1);
				}
				ch = (u32) payload[i];
				startCode = (startCode << 8) | ch;
				if (startCode == 0x1B6) {
					uiVESHeaderLen = i - 4;
					foundVOPcode = true;
				}
				i++;
			}
			
			// Create new header
			es->setDecoderConfig(payload,uiVESHeaderLen); //deep copy first bytes...
			es->clearDecConfUpdated();


			if (!foundVOPcode) { //assume, there is just the header (by xvid)
				//				iFramesize = 0;
				//delete[] payload;
			}

			/* 
			   // remove header from payload
				 //FIXME by mike: why should we need this?? 
				 //even worse, we need this header, if it wasn't sent via net RTSP/SDP
				 memmove(payload, payload + uiVESHeaderLen+1, iFramesize - uiVESHeaderLen-1);
				 iFramesize -= uiVESHeaderLen+1;
			*/
		}

		if (iFramesize != 0) {
			// All the other frames is easy, just memcopy output.
			pAU->size = iFramesize;
			pAU->payload = payload;
		
			/*
			if ((muiFrameNumber % muiKeyFrameInterval) == 0)
				mpgFrame = new CompressedVideoFrame(Frame::I_VOP, es->getWidth(), es->getHeight());
			else
				mpgFrame = new CompressedVideoFrame(Frame::P_VOP, es->getWidth(), es->getHeight());
			*/
			mpgFrame = new CompressedVideoFrame(Frame::UNKNOWN_VOP, es->getWidth(), es->getHeight());

			mpgFrame->detectFrameType();
			mpgFrame->setAU(pAU);
			mpgFrame->setMediaTimeScale(frm->getMediaTimeScale());
			frmList.push_back(mpgFrame);
			dprintf_full("MP4Encoder::adapt: first 4 bytes of %i bytes, %s frameNo %i, CTS %u: %x %x %x %x\r\n",
									 pAU->size, Frame::VOPTypeToChar(mpgFrame->getType()), muiFrameNumber,
									 mpgFrame->getAU()->cts, mpgFrame->getAU()->payload[0],
									 mpgFrame->getAU()->payload[1],
									 mpgFrame->getAU()->payload[2],
									 mpgFrame->getAU()->payload[3]);
			muiFrameNumber++;
		}
	}
	
	return frmList;
}


/***************************************************/
list < Frame * >MP4Encoder::close() {
	list < Frame * >frmList;
	
	return frmList;
}


/***************************************************/
Adaptor *MP4Encoder::clone() {
	char* copyEncoderConfig=NULL;
	if(mpEncoderHandle && meCoder != MP4Decoder::XVID) {
#ifdef ENABLE_FFMPEG
		AVCodec *codec=NULL;
		/* Make sure that the required encoder exists in the ffmpeg library */
		codec = avcodec_find_encoder(this->outputFormat);
		if(codec) {
			AVCodecContext* srcClone = avcodec_alloc_context();
			avcodec_get_context_defaults(srcClone);
			
			AVCodecContext* src=(AVCodecContext*) mpEncoderHandle;
			// don't do a memcpy --> pointers!
			srcClone->codec=codec;
			srcClone->b_frame_strategy=src->b_frame_strategy;
			srcClone->b_quant_factor=src->b_quant_factor;
			srcClone->b_quant_offset=src->b_quant_offset;
			srcClone->bit_rate=src->bit_rate;
			srcClone->rtp_payload_size=src->rtp_payload_size;
			srcClone->width=src->width;
			srcClone->height=src->height;
			srcClone->gop_size=srcClone->gop_size;
			srcClone->qmin=src->qmin;
			srcClone->qmax=src->qmax;
			srcClone->max_b_frames=src->max_b_frames;
			srcClone->sample_aspect_ratio.den=src->sample_aspect_ratio.den;
			srcClone->sample_aspect_ratio.num=src->sample_aspect_ratio.num;
			srcClone->rc_max_rate=src->rc_max_rate;
			srcClone->rc_min_rate=src->rc_min_rate;
			srcClone->mb_qmax=src->mb_qmax;
			srcClone->mb_qmin=src->mb_qmin;
			srcClone->time_base=src->time_base;
			// open the codec for encoding
			if (avcodec_open(srcClone, codec) < 0) {
				printf("MP4Encoder::clone(): Could not open codec\n");
				av_free(srcClone);
			}
			else	
				copyEncoderConfig=(char*)srcClone;
		}
#else
		dprintf_err("MP4Encoder::clone: compiled without ffmpeg support!\n");
		exit(1);
#endif
	}
	
	return new MP4Encoder(es,muiTargetBitrate,frameRate,muiKeyFrameInterval,numBFrames,
												mColorSpace,meCoder,outputFormat,copyEncoderConfig);
}


/***************************************************/
void MP4Encoder::setCoderType(MP4Decoder::eCoderType coderType) {
	meCoder = coderType;
}


/***************************************************/
void MP4Encoder::EncoderClose(char **pEncoderHandle) {
	
	switch (meCoder)
		{
		case (MP4Decoder::XVID):
			XVIDEncoderClose(pEncoderHandle);
			break;
#ifdef ENABLE_FFMPEG
		case (MP4Decoder::FFMPEG):
			FFMPEGEncoderClose(pEncoderHandle);
			break;
#endif
		default:
			dprintf_err("Invalid Encoder selected\n");
			break;
		}
}


/***************************************************/
int MP4Encoder::EncoderInit( char **pEncoderHandle, unsigned int iWidth, unsigned int iHeight, 
														 int iBitrate,  int iFramerate, int numBFrames, int keyFrameInterval, VideoESInfo * es) {
	int ret = 0;
	
	switch (meCoder)
		{
		case (MP4Decoder::XVID):
			ret = XVIDEncoderInit(pEncoderHandle, iWidth, iHeight, iBitrate, iFramerate, numBFrames);
			break;
#ifdef ENABLE_FFMPEG
		case (MP4Decoder::FFMPEG):
			ret = FFMPEGEncoderInit(pEncoderHandle, outputFormat, iWidth, iHeight, iBitrate, iFramerate, numBFrames, keyFrameInterval);
			break;
#endif
		default:
			dprintf_err("Invalid Coder selected\n");
			break;
		}
	
	return ret;
}


/***************************************************/
int MP4Encoder::EncodeFrame(char *pEncoderHandle, char *pYUVFrame, 
														char *pCompressed, UncompressedVideoFrame::ColorSpace eColorSpace,
														int forceIFrame) {
	int ret = 0;
	
	switch (meCoder) {
	case (MP4Decoder::XVID):
		dprintf_full("Encoding using XVID\n");
		ret = XVIDEncodeFrame(pEncoderHandle, pYUVFrame, pCompressed, 
													(xvidColorSpace) eColorSpace, forceIFrame);
		break;
#ifdef ENABLE_FFMPEG
	case (MP4Decoder::FFMPEG):
		dprintf_full("Encoding using FFMPEG\n");
		ret = FFMPEGEncodeFrame(pEncoderHandle, pYUVFrame, pCompressed, 
														(ffmpegColorSpace) mapToFFMPEGColorspace(eColorSpace), 
														forceIFrame);
		break;
		default:
			dprintf_err("Invalid Coder selected\n");
			break;
#endif
	}
	
	return ret;
}


/***************************************************/
ffmpegColorSpace MP4Encoder::mapToFFMPEGColorspace(UncompressedVideoFrame::ColorSpace colorSpace) {
	ffmpegColorSpace ret = FFMPEGColorSpaceUnknown;

	switch (colorSpace)	{
	case UncompressedVideoFrame::ColorSpaceYV12: 
		ret = FFMPEGColorSpaceYV12;
		break;
	case UncompressedVideoFrame::ColorSpaceRGB24:
		ret = FFMPEGColorSpaceRGB24;
		break;
	case UncompressedVideoFrame::ColorSpaceI420:
		ret = FFMPEGColorSpaceI420;
		break;
	case UncompressedVideoFrame::ColorSpaceRGB32:
		ret = FFMPEGColorSpaceRGB32;
		break;
	default:
		ret = FFMPEGColorSpaceUnknown;
		break;
	}
	
	return ret;
}


/** returns adaptation costs (CPU only). Encoding costs depend directly on the number 
  * of in YUV pixels and frames encoded per second (FIXME: codec not supported)
*/
u32 MP4Encoder::getTranscodingCosts() const {
	return (u32)(es->getWidth()*es->getHeight()*es->getFPS()*2);
}


/***************************************************/
u8* MP4Encoder::createNewHeader(VideoESInfo* vid, u32 targetBitRate, u32 width, u32 height,
																u32 uiKeyFrameInterval, int framerate, u32* headerSize) {

	dprintf_full("MP4Encoder::createNewHeader\n");
	assert(headerSize);
	if(!headerSize)
		return NULL;
	UncompressedVideoFrame* frm = new UncompressedVideoFrame(Frame::YUV_VOP, width, height);
	AU* au = new AU();
	au->payload = new u8[(int)ceil(1.5*width*height)];
	au->size = (int)(1.5*width*height);
	frm->setAU(au,false);
	VideoESInfo* pCVid = (VideoESInfo*)vid->clone(vid->getContainerInfo());
	pCVid->setWidth(width);
	pCVid->setHeight(height);
	MP4Encoder* mp4 = new MP4Encoder(pCVid, targetBitRate, framerate, uiKeyFrameInterval);
	mp4->initialize();
	list<Frame*> retFrm = mp4->adapt(frm);
	if(retFrm.empty()) {
		au->cts = au->dts = 3000;
		retFrm = mp4->adapt(frm);
	}
	if(retFrm.empty()) {
		*headerSize = 0;
		delete frm;
		return NULL;
	}
	if (*retFrm.begin() != frm)
		delete frm;
	// otherwise we get a header
	Frame* headerFrame = *retFrm.begin();
	delete headerFrame;
	*headerSize = 0;
	
	u8* newHeader = NULL;
	*headerSize = pCVid->getHeaders(&newHeader);
	
	return newHeader;
}

