/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	MP4Encoder.hpp                                                *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

/***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef AFX_MP4ENCODER_HPP
#define AFX_MP4ENCODER_HPP

#if _MSC_VER > 1000
#pragma once
#endif				// _MSC_VER > 1000

#include "VideoAdaptor.hpp"
#include "UncompressedVideoFrame.hpp"
#include "VideoESInfo.hpp"
#include "MP4Decoder.hpp"

/**
 * Implementation of Adaptor that takes UncompressedVideoFrame as input
 * and returns an CompressedVideoFrame as output.
 *
 *@author Michael Kropfberger and Peter Schojer
 *
 *@version $Id: MP4Encoder.hpp,v 1.16 2005/05/31 06:24:58 mkropfbe Exp $
 *
 *@short <short description>
 *
 */
class MP4Encoder:public VideoAdaptor {
public:
	/**
	 * Constructor.
	 * Creates a MP4Encoder instance. Do set attributes of pESInfo object carefully.
	 * Width, Height, and FPS must match the YUV input!
	 * @param pESInfo Pointer to ESInfo object with proper encoding parameters.
	 * @param targetBitRate is set in bits/second
	 * @param outputFormat which video format should be created as output, will be ignored
	 * when XVid is used for encoding which only allows for CODEC_ID_MPEG4
	 * Used attributes of ESInfo: avgBandwidth, width, height, FPS (via getFPS()).
	 * Width, Height, and FPS must match the YUV input.
	 */
	MP4Encoder(VideoESInfo * pESInfo, 
						 u32 targetBitRate,
						 u32 frameRate,
						 u32 uiKeyFrameInterval = 300,
						 u32 numBFrames = 0, 
						 UncompressedVideoFrame::ColorSpace colorSpace = UncompressedVideoFrame::ColorSpaceI420,
#ifdef ENABLE_FFMPEG
						 MP4Decoder::eCoderType coderType = MP4Decoder::FFMPEG,
#else
						 MP4Decoder::eCoderType coderType = MP4Decoder::XVID,
#endif
						 enum CodecID outputFormat=CODEC_ID_MPEG4,
						 char* mp4EncoderConfig=NULL);

	/** Destructor.	*/
	virtual ~ MP4Encoder();

	/**
	 * YUVFrame in - MPGFrame out.
	 * Input YUVFrame object will be deleted in this method!
	 * Adaptor implementation for YUVFrame objects.
	 * @param *frm Pointer to YUVFrame object.
	 * @return List containing MPGFrame objects. Will contain in general
	 * only one MPGFrame object unless B frame support is activated in
	 * the XviD library.
	 */
	list < Frame * >adapt(Frame * frm);
	
	/**
	 * Flush all buffered MPGFrame objects (if any).
	 *	 @return List of MPGFrame objects
	 */
	list < Frame * >close();

	/** Not implemented.
	 * @return Defaults to NULL.
	 */
	Adaptor *clone();
	
	void initialize();
	
	void setCoderType(MP4Decoder::eCoderType ct);
	
	/**
	 * Mapping function between MP4Decoders colorspace (defined in meColorSpace)
	 * and FFMPEG codec colorspace.
	 *
	 * @see meColorSpace
	 * @see mapToXVIDColorspace
	 */
	static ffmpegColorSpace mapToFFMPEGColorspace(UncompressedVideoFrame::ColorSpace colorSpace);
	
	/** returns adaptation costs (CPU only). Decoding costs depend directly on the number 
	 * of out YUV pixels and frames decoded per second (FIXME: codec not supported)
	 */
	u32 getTranscodingCosts() const;
	
	static u8* createNewHeader(VideoESInfo* vid,u32 targetBitRate,u32 width, u32 height,
														 u32 uiKeyFrameInterval, int framerate, u32* headerSize);
	
private:
	MP4Decoder::eCoderType meCoder;
	
	int EncoderInit( char **pEncoderHandle, unsigned int iWidth, unsigned int iHeight, 
									 int iBitrate,  int iFramerate, int numBFrames, int keyFrameInterval, VideoESInfo * es);
	
	void EncoderClose(char **pEncoderHandle);
	
	int EncodeFrame(char *pEncoderHandle, char *pYUVFrame, 
									char *pCompressed, UncompressedVideoFrame::ColorSpace eColorSpace, int forceIFrame);
	
	/** Handle to XviD encoder instance.*/
	char *mpEncoderHandle;
	
	/** Colorspace used for encoding.	*/
	UncompressedVideoFrame::ColorSpace mColorSpace;
	
	/** Size of decoded payload of Frame object in bytes.	*/
	u32 mFrameSize;
	
	/** Internal output buffer for XviD encoder. */
	char mpOutBuf[MAX_BUFFER_SIZE];

	/** Contains the number of frames encoded.	*/
	u32 muiFrameNumber;


	/** The target bitrate for the encoder */
	u32 muiTargetBitrate;

	/* Key Frame Interval. */
	u32 muiKeyFrameInterval;

	u32 numBFrames;

	u32 frameRate;

	enum CodecID outputFormat;
};

#endif

