 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	Adaptor.cpp                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#include "TimeMeasureAdaptor.hpp"
#include "../ESInfo.hpp"
#include "../Adaptor.hpp"
#include <assert.h>

TimeMeasureAdaptor::TimeMeasureAdaptor(Adaptor* a,FILE* fp):
	adaptor(a),muiTimeSpent(0),muiTimeSpentLastFrame(0),output(fp)
{
	assert(a!=NULL);
	const char* tmp=adaptor->getName();
	strcpy(name,"TimeMeasureAdaptor:");
	strcat(name,tmp);
	initialized=false;
};

TimeMeasureAdaptor::~TimeMeasureAdaptor()
{
	if(adaptor)
		delete adaptor;
};

long TimeMeasureAdaptor::getdifftime(struct timeval *t1, struct timeval *t2)
{
	long ret = 0;

	ret = t2->tv_usec - t1->tv_usec;
	ret += (t2->tv_sec - t1->tv_sec) * 1000000;

	return ret;
};

unsigned int TimeMeasureAdaptor::getTimeSpent()
{
	return muiTimeSpent;
};

unsigned int TimeMeasureAdaptor::getTimeSpentForLastFrame()
{
	return muiTimeSpentLastFrame;
};


void TimeMeasureAdaptor::startTimer()
{
	gettimeofday(&mtStart, NULL);
};

void TimeMeasureAdaptor::endTimer()
{
	gettimeofday(&mtEnd, NULL);
	muiTimeSpentLastFrame = getdifftime(&mtStart, &mtEnd);
	muiTimeSpent+=muiTimeSpentLastFrame;
};

/** 
	Accepts as input a frame, and checks if it can adapt the frame,
	If the frame is modified, a new Frame object is created and inserted into the return list.
	This new frame is returned only, if the size of the payload is
	larger than zero. If the frame was not modified, it is also inserted
	into the list.
	For more complex adaptors, it will happen, that the adaptor caches up one complete GOP
	and returns the full GOP in network order. A caching adaptor has to create
	deep-copies for each frames.
	Never directly modify the payload of the input frm, always create copies!!!

	@param frm object to adapt. An Adaptor is never allowed to free
	the input frame. This has to be done in the caller.
	@return List of adapted Frame objects
*/
list<Frame*> TimeMeasureAdaptor::adapt(Frame * frm)
{
	if(!initialized)
		initialize();
	startTimer();
	list<Frame*> f=adaptor->adapt(frm);
	endTimer();
	return f;
};

/**
	Close and destroy an Adaptor.

	Releases all internally buffered frame objects, and deallocate 
	all allocated memory. The Adaptor must not be used after calling
	the close method!!!

	@return List of adapted Frame objects.
*/
list<Frame*> TimeMeasureAdaptor::close()
{
	startTimer();
	list<Frame*> f=adaptor->close();
	endTimer();
	if(output) {
		fseek(output,0,SEEK_END);
		fprintf(output,"%s(%s):spent %u microsec\n",name,
		(adaptor->getESInfo()?adaptor->getESInfo()->getInput():"Unknown ESInfo"),getTimeSpent());
	}
	return f;
};

/**
	Create a shallow copy of the Adaptor.

	Creates an Adaptor with the same setup (without copying the current status).
	@return An Adaptor object.
*/
Adaptor* TimeMeasureAdaptor::clone()
{
	Adaptor	*a=	adaptor->clone();
	return new TimeMeasureAdaptor(a,output);
};

/** Sets a reference to an ESInfo object. Adaptors are
	allowed to change ESInfo objects. If you don't want that,
	pass an es->clone() 
*/ 
void TimeMeasureAdaptor::setESInfo(ESInfo * esi)
{
	adaptor->setESInfo(esi);
};

ESInfo * TimeMeasureAdaptor::getESInfo()
{
	return adaptor->getESInfo();
};

/**
	Initialize internal data structures.
*/
void TimeMeasureAdaptor::initialize() 
{
	if(initialized) 
		return;
	initialized=true;
	muiTimeSpent=0;
	startTimer();
	adaptor->initialize();
	endTimer();
	if(output) {
		fseek(output,0,SEEK_END);
		fprintf(output,"%s(%s):spent %u microsec in Initialize\n",name,
		(adaptor->getESInfo()?adaptor->getESInfo()->getInput():"Unknown ESInfo"),getTimeSpent());
	}
};
	
