 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	AdmissionControl.cpp                                          *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#include "AdmissionControl.hpp"
#include "VideoESInfo.hpp"
#include "AudioESInfo.hpp"
#include "adaptors/Forwarder.hpp"
#include "adaptors/MP4Encoder.hpp"
#include "adaptors/MP4Decoder.hpp"
#include "adaptors/TemporalAdaptor.hpp"
#include "adaptors/YUVColorReductionAdaptor.hpp"
#include "adaptors/YUVSpatialReductionAdaptor.hpp"
#include "MetaObject.hpp"
#include "VideoAdaptorChain.hpp"
#include "net/ProxySession.hpp"

AdmissionControl::AdmissionControl(const ResourceLimit* lim,const ResourcePrices* p):currentUsage()
{
	assert(lim);
	assert(p);
	limit=new ResourceLimit(lim);
	prices=new ResourcePrices(p);
	activeRequests=0;
	lock.initialize("AdmissionControl-lock");
};
AdmissionControl::~AdmissionControl()
{
	if(limit)
		delete limit;
	if(prices)
		delete prices;
};
/** returns true if it will be possible to insert the resource usage */
bool AdmissionControl::canAdmitRequest(const ResourceUsage* usage)
{
	return (currentUsage.cpu+usage->cpu<1.0) && 
		(currentUsage.network+usage->network<1.0) &&
		(currentUsage.disk+usage->disk<1.0);
}

/** returns true if we could successfully insert the usage */
bool AdmissionControl::insertRequest(const ResourceUsage* usage)
{
	assert(usage);
	bool ret=false;
	dprintf_full("AdmissionControl::insertRequest(N=%f,C=%f,H=%f)\n",usage->network,usage->cpu,usage->disk);
	if(usage->cpu==usage->disk && usage->disk==usage->network && usage->network<0.00000001) {
		// ignore zero req, don't change activeRequests
		return true;
	}
	dprintf_full("AdmissionControl::insertRequest act %i ((N=%f,C=%f,H=%f)\n",
		activeRequests,currentUsage.network,currentUsage.cpu,currentUsage.disk);

	lock.lock();
	if(activeRequests==0) {
		// reset resourceusage to 0 to deal with fp rounding errors
		// which otherwise would accumulate
		currentUsage.cpu=currentUsage.network=currentUsage.disk=0.0;
	}
	if(canAdmitRequest(usage)) {
		currentUsage.add(usage);
		activeRequests++;
		dprintf_full("AdmissionControl::insertRequest: INSERTED\n");
		ret=true;
	}
	else
		prxStat.rejected++;
	lock.release();
	return ret;
}

/** immediately frees the resource usage */
bool AdmissionControl::freeRequest(const ResourceUsage* usage)
{
	assert(usage);
	bool ret=false;
	dprintf_full("AdmissionControl::freeRequest(N=%f,C=%f,H=%f)\n",usage->network,usage->cpu,usage->disk);
	lock.lock();
	if(usage->cpu==usage->disk && usage->disk==usage->network && usage->network<0.00000001) {
		// ignore zero req, don't change activeRequests
		ret=true;
	}
	else if(activeRequests>0) {
		// at least one request is open!
		ret=true;
		currentUsage.subtract(usage);
		dprintf_full("AdmissionControl::freeRequest: SUCCESS \n");
		activeRequests--;
	}
	lock.release();
	return ret;
}

/** takes the properties specified by the user and tries to match these to the
  * properties of the video. If they match, a Forwarder is returned, if not a
  * different Adaptor that will generate a video according to the specified params.
  */
Adaptor* AdmissionControl::makeExactMatch(unsigned int dimX, 
									unsigned int dimY, 
									unsigned int bitRate, 
									bool greyScale, float frameRate,
									ContainerInfo* stream,ResourceUsage* ru,
									bool allowHigherThanSrcBitRate) {
	dprintf_full("AdmissionControl::makeExactMatch IN: dimx %u dimy %u bitRate %u grey %i\r\n",
		dimX, dimY, bitRate, greyScale);
	assert(ru);
	
	if (!stream)
		return NULL;
	// doesn't access variations, no need to lock
	VideoESInfo* es= stream->getFirstVisualES();
	AudioESInfo* es2=stream->getFirstAudioES();
	if(es2)
		bitRate-=es2->getAvgBandwidth();

	if(!es) {
		dprintf_err("AdmissionControl::makeExactMatch: Video with no Visual ES?\r\n");
		ru->cpu=ru->disk=ru->network=0.0;
		return new Forwarder();
	}

	// we have a cache hit! the source is always read from the disk
	Feature src(es);
	const char* fileTmp=es->getInput();
	if(fileTmp) {
		FILE* fpTmp=fopen(fileTmp,"rb");
		if(fpTmp) {
			src.isCached=true;
			fclose(fpTmp);
		}
		else
			src.isCached=false;
	}
	else
		src.isCached=false;
	
	Feature dst(es);
	dst.isCached=false;
	dst.bitrate=bitRate;
	dst.dimX=dimX;
	dst.dimY=dimY;
	dst.color=!greyScale;
	dst.frameRate=frameRate;
	ResourceUsage* ruTmp=CostFunction::calcResourceUsage(&src,&dst,limit);

	ru->network=ruTmp->network;
	ru->disk=ruTmp->disk;
	ru->cpu=ruTmp->cpu;
	delete ruTmp;
	// CPU is calculated on the adaptor!
	VideoAdaptorChain* result=NULL;
	// FIXME:? no upscaling support!!!!
	if(es->getWidth()<dimX)
		dimX=es->getWidth();
	if(es->getHeight()<dimY)
		dimY=es->getHeight();
	bool transcode=false;
	int adaptorCount=0;
	if( (es->getWidth()/8) != (dimX/8) || 

		(es->getHeight()/8) != (dimY/8) || greyScale 

		|| bitRate < es->getAvgBandwidth() || es->getFPS()>frameRate) {

		// we need transcoding!!!

		transcode=true;

		result=new VideoAdaptorChain();

	}
	if(!transcode && allowHigherThanSrcBitRate && bitRate!=es->getAvgBandwidth()) {
		transcode=true;
		result=new VideoAdaptorChain();
	}
	if(es->getFPS()>frameRate)
		result->addAdaptor(new TemporalAdaptor(es));
	if(transcode) {
		//ATTENTION: make sure to use the same dec/enc lib, so either both xvid or both ffmpeg!!
		result->addAdaptor(new MP4Decoder(es, false, UncompressedVideoFrame::ColorSpaceYV12, MP4Decoder::FFMPEG));
		adaptorCount++;
	}

	if( (es->getWidth()/8) != (dimX/8) || 
		(es->getHeight()/8) != (dimY/8) ) {
		// it's up to the user to decide if he wants to keep
		// the aspect ratio, we stick to dimX and dimY
		result->addAdaptor(new YUVSpatialReductionAdaptor(es, &dimX, &dimY));
		adaptorCount++;
	}
	if(greyScale) {
		result->addAdaptor(new YUVColorReductionAdaptor(es));
		adaptorCount++;
	}
	if (bitRate > es->getAvgBandwidth() && !allowHigherThanSrcBitRate) {
		bitRate=es->getAvgBandwidth();
	}
	if(transcode) {
		u32 new_bitrate = dimX * dimY * 6;
		if(!allowHigherThanSrcBitRate) {
			if (new_bitrate > bitRate)
				new_bitrate = bitRate;
		}
		else
			new_bitrate = bitRate;
		result->addAdaptor(new MP4Encoder(es, new_bitrate, (int)es->getFPS(), 300 /*GOP size*/,  0 /*BVOPs*/, 
																			UncompressedVideoFrame::ColorSpaceYV12, MP4Decoder::FFMPEG));
		result->initialize(); // must be called prior cost calc
		// fix cpu to correct value
		ru->cpu= ((double)result->getTranscodingCosts())/limit->cpu;
		adaptorCount++;
	}
	else {
		assert(result==NULL);
		dprintf_small("MetaObject::makeExactMatch created a Forwarder\n");
		ru->cpu=0.0;
		if(result)
			delete result;
		if(!insertRequest(ru)) {
			ru->network=ru->disk=0.0; //rejected
		}
		return new Forwarder();
	}
	dprintf_small("AdmissionControl::makeExactMatch created AdaptorChain of size %i\n",adaptorCount);
	if(!insertRequest(ru)) {
			ru->network=ru->disk=ru->cpu=0.0; //rejected
			delete result;result=NULL;
	}
	return result;
};

/** similar to makeExactMatch but it will also consider costs and will
	* return the version which offers the best cost/quality ratio.
	* @param return is a Forwarder if no adaptation is needed, @param outStream will always be set
	* to the source used for adaptation */
Adaptor* AdmissionControl::makeBestMatch(const UserPreferences* up,
			int delayToServer,
			MetaObject* meta,
			/* out parameters */
			ContainerInfo** ci, ResourceUsage* ru)
{
	assert(up && ci && meta && ru);
	ResourceUsage resUse;
	bool 	originalBest=false;
	static const int bitrates[]={1,2,4,8};
	static const int nrBR=4;
	
	int dimx_min=((up->dimX.min+43)/44)*44;
	int dimx_max=(up->dimX.max/44)*44;
	Video* target=NULL;
	double fr_min=up->frameRate.min;
	double fr_max=up->frameRate.max;
	double result,quality,costs;
	Video* bestTarget=NULL;
	double bestResultCosts=FLT_MAX;
	
	ContainerInfo* bestVideo=NULL;
	// always check the original videos first!
	int realDelay=0;
	ResourceUsage* job=NULL;
	// locks the meta lock!!!!
	std::list<ContainerInfo*>* variations=meta->getVariations();
	std::list<ContainerInfo*>::iterator origVid;
	for(origVid=variations->begin();origVid!=variations->end();++origVid) {
		// all videos stored in variations are hits (more or less :-))
		Feature sourceVideo( (*origVid)->getFirstVisualES());
		const char* fileTmp=(*origVid)->getFirstVisualES()->getInput();
		if(fileTmp) {
			FILE* fpTmp=fopen(fileTmp,"rb");
			if(fpTmp) {
				sourceVideo.isCached=true;
				fclose(fpTmp);
			}
			else
				sourceVideo.isCached=false;
		}
		else
			sourceVideo.isCached=false;
		
		job=CostFunction::calcResourceUsage( &sourceVideo, &sourceVideo, limit);
		costs=CostFunction::calcWeightedMonetaryCosts(&currentUsage,job,prices,
		CostFunction::baseCosts,sourceVideo.contentCosts,realDelay,up->delayInMS.point);
		quality=up->calcQuality(&sourceVideo,realDelay);
		if(quality<0.0001)
				result=FLT_MAX;
		else if(costs+CostFunction::baseCosts+sourceVideo.contentCosts>up->pay)
			result=FLT_MAX;
		else
			result=CostFunction::calcFinalCosts(quality,up->pay,costs);
		// minimize transcoding were possible, otherwise check here for full hits!
		if(quality>0.01) {
			if(result<bestResultCosts) {
				bestResultCosts=result;
				bestVideo=(*origVid);
				ru->cpu=job->cpu;ru->network=job->network;ru->disk=job->disk;
			}
		}
		delete job;job=NULL;		
	}
	if(bestVideo) {
		*ci=bestVideo;
		meta->releaseVariationsLock();
		if(!insertRequest(ru)) {
			ru->cpu=ru->network=ru->disk=0.0;
		}
		return NULL;
	}

	originalBest=false;
	// if we come that far we have to do adaptation
	for(origVid=variations->begin();origVid!=variations->end();++origVid) {
		// all videos stored in variations are hits (more or less :-))
		Feature sourceVideo( (*origVid)->getFirstVisualES());
		const char* fileTmp=(*origVid)->getFirstVisualES()->getInput();
		if(fileTmp) {
			FILE* fpTmp=fopen(fileTmp,"rb");
			if(fpTmp) {
				sourceVideo.isCached=true;
				fclose(fpTmp);
			}
			else
				sourceVideo.isCached=false;
		}
		else
			sourceVideo.isCached=false;
		
		for(int dimx=dimx_min;dimx<=dimx_max;dimx+=44) {
			int br_min_pos=0;
			int nrPixels=(int)(dimx*(((double)dimx)/sourceVideo.getAspectRatio()));

			while(up->bitrate.min>(nrPixels*bitrates[br_min_pos]) && br_min_pos<nrBR)
				br_min_pos++;

			int br_max_pos=nrBR-1;
			while(up->bitrate.max<(nrPixels*bitrates[br_max_pos]) && br_min_pos<br_max_pos)
				br_max_pos--;
			for(int brPos=br_min_pos;brPos<=br_max_pos;brPos++) {
				int br=bitrates[brPos]*nrPixels;
				for(double fr=fr_min;fr<=fr_max+0.001;fr+=5.0) {
					// assume not cached
					// assume that content cost degrade lineary with dimX
					int ccost=(int)(((double)dimx)/sourceVideo.dimX*sourceVideo.contentCosts);
					target=new Video(dimx,(int)(((double)dimx)/sourceVideo.getAspectRatio()),
							up->color.point,br,0,ccost,fr,sourceVideo.durationInSec);
					// assume startup delay of 1 second
		//			result=CostFunction::calcFinalCosts(req,target,r,systemLoad,prices,delayToServer);
					job=CostFunction::calcResourceUsage( &sourceVideo, target, limit);
					costs=CostFunction::calcWeightedMonetaryCosts(&currentUsage,job,prices,
					CostFunction::baseCosts,target->contentCosts,delayToServer,up->delayInMS.point);
					quality=up->calcQuality(target,delayToServer);
					if(quality<0.0001)
						result=FLT_MAX;
					else if(costs+CostFunction::baseCosts+ccost>up->pay)
						result=FLT_MAX;
					else
						result=CostFunction::calcFinalCosts(quality,up->pay,costs);

					if(result<bestResultCosts && job) {
						if(bestTarget)
							delete bestTarget;
						bestTarget=target;
						// frame dropping hack: when dropping B-frames never store the adapted version in the cache
						// store the original one, detect this case by checking cpu usage of the task
						if(job->cpu<0.00001)
							printf("Detected TEMPORAL ADAPTATION\r\n");
						originalBest=(job->cpu<0.00001);
						//*originalBest=false;
						bestResultCosts=result;
						bestVideo=(*origVid);
						target=NULL;
						ru->cpu=job->cpu;ru->network=job->network;ru->disk=job->disk;
					}
					if(target) {
						delete target;
						target=NULL;
					}
					if(job) {
						delete job;
						job=NULL;
					}
				}
			}
		}
	}

	if(target) {
		delete target;
		target=NULL;
	}
	if(job) {
		delete job;
		job=NULL;
	}
	if(bestVideo) {
		// make adaptorchain for bestVideo --> bestTarget
		// makeexact also has already inserted the costs for ru!
		Adaptor* ada=AdmissionControl::makeExactMatch(bestTarget->dimX,bestTarget->dimY,
			bestTarget->bitrate,!bestTarget->color,(float)bestTarget->frameRate,bestVideo,ru,true);
		// calculate costs
		meta->releaseVariationsLock();
		*ci=bestVideo;
		if(bestTarget)
			delete bestTarget;
		return ada;
	}
	meta->releaseVariationsLock();
	*ci=NULL;
	ru->cpu=ru->network=ru->disk=0.0;
	return NULL;
};
