
 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	AdmissionControl.hpp                                          *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef PS_VITOOKI_ADMISSIONCONTROL
#define PS_VITOOKI_ADMISSIONCONTROL
#include "cache/CostFunction.hpp"
#include "global.hpp"
#include "VMutex.hpp"

#define MAX_JOB_ENTRIES 1024

typedef struct Request{
	UserPreferences up;
	Video sourceVideo;
	unsigned long startTime;
	unsigned long endTime; // calculated as video.duration+startTime
	Request(const UserPreferences& u, const Video& vid,unsigned long start,unsigned long end):up(u),sourceVideo(vid)
	{
		startTime=start;
		endTime=end;
	}
	~Request() {
	};
} Request;

class AdmissionControl
{
protected:
	ResourceLimit* limit;
	ResourcePrices* prices;
	ResourceUsage currentUsage;
	VMutex lock;
	int activeRequests;
	
	/** returns true if it will be possible to insert the resource usage */
	bool canAdmitRequest(const ResourceUsage* usage);	
public:
	AdmissionControl(const ResourceLimit* lim,const ResourcePrices* p);
	~AdmissionControl();

	const ResourceLimit* getResourceLimit() const {
		return limit;
	}


	/** returns true if we could successfully insert the usage */
	bool insertRequest(const ResourceUsage* usage);

	/** immediately frees the resource usage */
	bool freeRequest(const ResourceUsage* usage);

	/** takes the properties specified by the user and tries to match these to the
	  * properties of the video. If they match, a Forwarder is returned, if not a
	  * different Adaptor that will generate a video according to the specified params.
	  * If you want to keep the old version of the ContainerInfo object, be sure to pass
	  * a clone for @param stream! The adaptor will modify it!
		* @param ru: pass an ResourceUsage struct here, so
		* that the AdmissionControl will calculate the costs for making an exact match
		* and inserting it into the AdmissionControl. If there are not enough resources to make an
		* exact match a Forwarder will be returned and resource costs will be set to zero
	  * FIXME: support TemporalAdaptation
		* NOTE: You MUST NOT call insertRequest with @param ru. This already happened
		* within the function, but you must call freeRequest as soon as you are finished with the job
		* If @param ru contains usage values of 0.0, the request was rejected.
	  */
	Adaptor* makeExactMatch(unsigned int dimX, unsigned int dimY, 
							unsigned int bitRate, 
							bool greyScale, float frameRate,
							ContainerInfo* stream,ResourceUsage* ru,bool allowHigherThanSrcBitRate=false);

	/** similar to makeExactMatch but it will also consider costs and will
	  * return the version which offers the best cost/quality ratio.
		* @param return is a Forwarder if no adaptation is needed, but then @param outStream will be set
		* otherwise @param return will be set to the visual adaptor needed, and  @param outStream will 
	  * also be set to the source used for adaptation.
		* NOTE: You MUST NOT call insertRequest with @param ru. This already happened
		* within the function, but you must call freeRequest as soon as you are finished with the job
		* If @param ru contains usage values of 0.0, the request was rejected.
	*/
	Adaptor* makeBestMatch(const UserPreferences* up,
			int delayToServer,
			MetaObject* meta,
			/* out parameters */
			ContainerInfo** outStream, ResourceUsage* ru);

};

#endif

