 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	CostFunction.cpp                                              *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#include "CostFunction.hpp"
#include "cache/AdmissionControl.hpp"

int CostFunction::baseCosts=10;

ResourceUsage* CostFunction::calcResourceUsage( const Feature* origVideo, const Feature* targetVideo, const ResourceLimit* r)
{
	assert(origVideo);
	assert(targetVideo);
	assert(r);
	ResourceUsage* u=new ResourceUsage();
	int decoding=cpuCostDecoding(origVideo->dimX,origVideo->dimY,origVideo->frameRate);
	int encoding=cpuCostEncoding(targetVideo->dimX,targetVideo->dimY,targetVideo->frameRate);
	int spatial=cpuCostResizing(targetVideo->dimX,targetVideo->dimY,targetVideo->frameRate);
	int recoding=decoding+encoding;
	//we ignore dimY, only dimX can be specified
	// we also assume that framerate adjustement can be achieved by B-frame dropping
	// we assume only one B-frame between (IPBPBPB) -> (GOP-1)/2 are B-frames
	
	bool transcode=checkTranscoding(origVideo,targetVideo);
	
	if(targetVideo->isCached) {
		u->network=((double)targetVideo->bitrate)/r->network; // just sending to client
		u->disk=((double)targetVideo->bitrate/8.0)/r->disk; // reading from disk
		u->cpu=0.0; // no transcoding
	}
	else {
		// targetvideo not cached	
		if(origVideo->dimX!=targetVideo->dimX)	
			recoding+=spatial;
		
		if(origVideo->isCached) {
			u->network=((double)targetVideo->bitrate)/r->network; // just sending to client
			
			if(!transcode) {
				
				u->disk=((double)(targetVideo->bitrate/8.0))/r->disk; // reading from disk
				u->cpu=0.0; // no transcoding	
			}
			else {
				// transcoding, 
				u->disk=((double)(origVideo->bitrate+targetVideo->bitrate)/8.0)/r->disk; // reading from disk+caching
				u->cpu=((double)(recoding))/r->cpu;
			}
		}
		else {
			// original not cached
			u->network=((double)targetVideo->bitrate+origVideo->bitrate)/r->network; // sending to client+reading orig
			if(transcode)
				u->disk=((double)(origVideo->bitrate+targetVideo->bitrate)/8.0)/r->disk; // writing orig+transcoded to disk	
			else
				u->disk=((double)(origVideo->bitrate)/8.0)/r->disk; // writing orig to disk	
			if(transcode) {
				u->cpu=((double)(recoding))/r->cpu;
			}
			else
				u->cpu=0.0;	
		}
	}
// DEBUG	
//	dprintf_full("(%i,%i,%i,%f) --(%i,CPU=%.2lf,NET=%.2lf,DSK=%.2lf)--> (%i,%i,%i,%f)\n",origVideo->dimX,origVideo->dimY,origVideo->bitrate,origVideo->frameRate, recoding,u->cpu,u->network,u->disk,targetVideo->dimX,targetVideo->dimY,targetVideo->bitrate,targetVideo->frameRate);

	return u;
};		

double CostFunction::calcWeightedMonetaryCosts(
			const ResourceUsage* totalUsed, 
			const ResourceUsage* req,
			const ResourcePrices* prices,
			int baseCosts,
			int contentCosts,
			int startUpDelay, 
			int maximumDelay)
{
	double delay=0.0;
	if( startUpDelay >= 0 && startUpDelay<=maximumDelay) {
		if(maximumDelay>0)
			delay=((double)startUpDelay)/maximumDelay;
		else
			delay=0.0;
	}
	else {
		// return infinite
		return FLT_MAX;
	}
	if( (totalUsed->network+req->network)>0.99999 ||
			(totalUsed->cpu+req->cpu)>0.99999 || 
			(totalUsed->disk+req->cpu)>0.99999) {
		// return infinite
//		printf("RESOURCE OVERFLOW\r\n");
		return FLT_MAX;
	}
	return (req->network/(1-totalUsed->network-req->network))*prices->network+
		(req->cpu/(1-totalUsed->cpu-req->cpu))*prices->cpu+
		(req->disk/(1-totalUsed->disk-req->disk))*prices->disk+delay;
};

double CostFunction::calcFairMonetaryCosts(
		const ResourceUsage* req, 
		const ResourcePrices* priceInCent, 
		int baseCosts, 
		int contentCosts,
		double videoduration)
{
	double result=(double)baseCosts+contentCosts;
	double network=0.0,disk=0.0,cpu=0.0, total=0.0;
	network=(req->network*priceInCent->network);
	network*=videoduration;
	disk=(req->disk*priceInCent->disk);
	disk*=videoduration;
	cpu=req->cpu*priceInCent->cpu;
	cpu*=videoduration;
	total=disk+network+cpu;
	result+=total;
	return result;
};


double CostFunction::calcFinalCosts(double quality, double pay, double weightedcosts)
{
	if(weightedcosts>pay || quality < 0.0001)
		return FLT_MAX;
	// return (quality*pay/totalcosts)+(weightedcosts*totalcosts/pay);
	// return (1-quality)+weightedcosts*(totalcosts/pay); //v1
	// return 1-quality+weightedcosts+totalcosts/pay;	//v2
	// return (1-quality)*weightedcosts*3; // 2ndbest so far
	//	return 10*(1-quality)*(weightedcosts/pay); // best so far
	return (1-quality)*weightedcosts;
};

double CostFunction::calcFinalCosts(
			const Request* origVideo, 
			const Feature* targetVideo, 
			const ResourceLimit* r,
			const ResourceUsage* systemLoad,
			const ResourcePrices* prices,int startUpDelay)
{
		ResourceUsage* job=CostFunction::calcResourceUsage( &origVideo->sourceVideo, targetVideo, r);
		double monetaryCosts=CostFunction::calcWeightedMonetaryCosts(systemLoad,job,prices,
			CostFunction::baseCosts,targetVideo->contentCosts,startUpDelay,origVideo->up.delayInMS.point);
		double quality=origVideo->up.calcQuality(targetVideo,startUpDelay);
//		printf("Quality %lf\r\n",quality);
		delete job;
		if(quality<0.0001)
			return FLT_MAX;
		if(monetaryCosts+CostFunction::baseCosts+targetVideo->contentCosts>origVideo->up.pay)
			return FLT_MAX;
		return CostFunction::calcFinalCosts(quality,origVideo->up.pay,monetaryCosts);
}

// only compares original versions of videos, ignores layers
bool CostFunction::checkTranscoding(const Feature* origVideo, const Feature* targetVideo) {
	// only check dimX, dimY is defined through aspect ratio (a constant!!!) and dimX
	bool transcode=origVideo->dimX!=targetVideo->dimX;
	if(!transcode && origVideo->color) {
		// doesn't make sense to transcode from grey to color :-)
		transcode=origVideo->color!= targetVideo->color;
	}
	if(!transcode) {
		// bitrate too high
		if(origVideo->bitrate>targetVideo->bitrate) {
			double newFR; int newBR;
			// can we adapt by dropping B-frames?
			if(origVideo->canReduceBFramesToBitrate(targetVideo->bitrate,&newFR,&newBR)) {
				if(newFR<targetVideo->frameRate)
					transcode=true;	
			}
			else
				transcode=true;
		}
	}
	return transcode;
}

Feature* CostFunction::findBestVariation(const Request* req,const ResourceUsage* systemLoad,
			const ResourceLimit* r,const ResourcePrices* prices, double* lowestCosts, int delayToServer, bool* originalBest)
{
	// try all possible combinations of features specified in the userpreferences
	assert(req && systemLoad && r && prices && lowestCosts);
	assert(originalBest);
//	static const int bitrates[]={64000,128000,192000,256000,384000,512000,768000,1024000};
		static const int bitrates[]={1,2,4,8};
	static const int nrBR=4;
	*originalBest=false;
	
	int dimx_min=((req->up.dimX.min+43)/44)*44;
	int dimx_max=(req->up.dimX.max/44)*44;
	Video* target=NULL;
	double fr_min=req->up.frameRate.min;
	double fr_max=req->up.frameRate.max;
	double result,quality,costs;
	Video* bestTarget=NULL;
	double bestResultCosts=FLT_MAX;
//	FILE* fp=fopen("findbestvar.txt","wb");

	// always check the original video first!
	int realDelay=0;
	if(!req->sourceVideo.isCached)
		realDelay=delayToServer;
//	bestResultCosts=CostFunction::calcFinalCosts(req,&req->sourceVideo,r,systemLoad,prices,realDelay);
	ResourceUsage* job=CostFunction::calcResourceUsage( &req->sourceVideo, &req->sourceVideo, r);
	costs=CostFunction::calcWeightedMonetaryCosts(systemLoad,job,prices,
	CostFunction::baseCosts,req->sourceVideo.contentCosts,realDelay,req->up.delayInMS.point);
	quality=req->up.calcQuality(&req->sourceVideo,realDelay);
	if(quality<0.0001)
			bestResultCosts=FLT_MAX;
	else if(costs+CostFunction::baseCosts+req->sourceVideo.contentCosts>req->up.pay)
		bestResultCosts=FLT_MAX;
	else
		bestResultCosts=CostFunction::calcFinalCosts(quality,req->up.pay,costs);
	// minimize transcoding were possible, otherwise check here for full hits!
	if(quality>0.01) {
		*originalBest=true;
		*lowestCosts=bestResultCosts;
		delete job;
		return new Video(&req->sourceVideo);
	}
	else {
		bestTarget=NULL;
		bestResultCosts=FLT_MAX;
	}
	delete job;job=NULL;
//	printf("(%i,%i)Dimx(%i,%i)\r\n",req->up.dimX.min,req->up.dimX.max,dimx_min,dimx_max);
//	printf("Framerate(%.2lf,%0.2lf)\r\n",fr_min,fr_max);
//	fprintf(fp,"Bitrate\tdimx\tframerate\tresult\tquality\tcosts\r\n");
	for(int dimx=dimx_min;dimx<=dimx_max;dimx+=44) {
		int br_min_pos=0;
		int nrPixels=(int)(dimx*(((double)dimx)/req->sourceVideo.getAspectRatio()));

		while(req->up.bitrate.min>(nrPixels*bitrates[br_min_pos]) && br_min_pos<nrBR)
			br_min_pos++;

		int br_max_pos=nrBR-1;
		while(req->up.bitrate.max<(nrPixels*bitrates[br_max_pos]) && br_min_pos<br_max_pos)
			br_max_pos--;
		for(int brPos=br_min_pos;brPos<=br_max_pos;brPos++) {
			int br=bitrates[brPos]*nrPixels;
//			printf("Bitrate %i\r\n",br);
			for(double fr=fr_min;fr<=fr_max+0.001;fr+=5.0) {
				// assume not cached
				// assume that content cost degrade lineary with dimX
				int ccost=(int)(((double)dimx)/req->sourceVideo.dimX*req->sourceVideo.contentCosts);
				target=new Video(dimx,(int)(((double)dimx)/req->sourceVideo.getAspectRatio()),
						req->up.color.point,br,0,ccost,fr,req->sourceVideo.durationInSec);
				// assume startup delay of 1 second
	//			result=CostFunction::calcFinalCosts(req,target,r,systemLoad,prices,delayToServer);
				job=CostFunction::calcResourceUsage( &req->sourceVideo, target, r);
				costs=CostFunction::calcWeightedMonetaryCosts(systemLoad,job,prices,
				CostFunction::baseCosts,target->contentCosts,delayToServer,req->up.delayInMS.point);
				quality=req->up.calcQuality(target,delayToServer);
				if(quality<0.0001)
					result=FLT_MAX;
				else if(costs+CostFunction::baseCosts+ccost>req->up.pay)
					result=FLT_MAX;
				else
					result=CostFunction::calcFinalCosts(quality,req->up.pay,costs);
//				fprintf(fp,"%6i %6i %.6lf %.6lf %.6lf %.6lf\r\n",br,dimx,fr,result,quality,costs);
//				printf("Costs: %.3lf Quality: %.3lf Result: %.3lf \r\n",costs,quality,result);
				if(result<bestResultCosts) {
					if(bestTarget)
						delete bestTarget;
					bestTarget=target;
					// frame dropping hack: when dropping B-frames never store the adapted version in the cache
					// store the original one, detect this case by checking cpu usage of the task
					if(job->cpu<0.00001)
						printf("Detected TEMPORAL ADAPTATION\r\n");
					*originalBest=(job->cpu<0.00001);
					//*originalBest=false;
					bestResultCosts=result;
					target=NULL;
				}
				if(target) {
					delete target;
					target=NULL;
				}
				if(job) {
					delete job;
					job=NULL;
				}
			}
		}
	}


	
//	fprintf(fp,"%6i %6i %.6lf %.6lf %.6lf %.6lf\r\n",req->sourceVideo.bitrate,
//		req->sourceVideo.dimX,req->sourceVideo.frameRate,result,quality,costs);

	if(target) {
		delete target;
		target=NULL;
	}
	if(job) {
		delete job;
		job=NULL;
	}

//	fclose(fp);
	*lowestCosts=bestResultCosts;
	return bestTarget;

};

/** eliminates impossible user requirements, like requesting a video
	* with a frame rate higher than the original frame rate */
void CostFunction::cleanUserPreferences(UserPreferences* u, const Video* source)
{
	if(u->frameRate.min>source->frameRate) {
		u->frameRate.min=source->frameRate;
		u->frameRate.max=source->frameRate;
	}
	if(u->frameRate.max>source->frameRate)
		u->frameRate.max=source->frameRate;
	if(u->bitrate.min>source->bitrate)
		u->bitrate.min=source->bitrate-1;
};

