 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	MetaObject.hpp                                                *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef _MPEG4ITEC_METAOBJECT_HPP_
#define _MPEG4ITEC_METAOBJECT_HPP_

#include "ContainerInfo.hpp"
#include "cache/CostFunction.hpp"

#include "VMutex.hpp"
class Adaptor;
class Url;
class UserPreferences;
/**
 *
 * <Class Description goes here>
 *
 *@author Michael Kropfberger and Peter Schojer
 *
 *@version $Id: MetaObject.hpp,v 1.10 2006/01/20 15:37:17 mkropfbe Exp $
 *
 *@short <short description>
 *
 */
class MetaObject {
public:
	/** creates a new meta-object for the given stream.
	  * @param stream is shallow copied and destroyed with the meta-object
	  */
	MetaObject(ContainerInfo* stream);
	
	/** creates a new empty meta-object and tries to
	  * init from a config file
	  */
	MetaObject(const char* configFile);

	/** frees memory for all ContainerInfo objects */
	~MetaObject();

	/** search for a stream with specified properties 
	  * @param lowerDimX specifies the minimum width of the video 
	  * @param upperDimX specifies the maximum width of the video 
	  * @param lowerDimY specifies the minimum height of the video 
	  * @param lowerDimY specifies the maximum height of the video 
	  * @param lowerBitRate specifies the minimum bit-rate in bit/sec
	  * @param upperBitRate specifies the maximum bit-rate in bit/sec
	  * @param greyScale specifies if the video should have color or not
	  *	@param isExactMatch specifies if an exact match was found or not.
	  * An exact match is anything that fits the specified range
	  * If you want to exclude some params from the search, set the lower Value to zero
	  * and the upper value to the largest int.
	  * Note that @param return will only return NULL, if no ContainerInfo object
	  * is stored. If no exact match is found, the
	  * object nearest to the specified parameters is returned and @param isExactMatch
	  * is set to false. If you do require an exact match, take the returned ContainerInfo
	  * and call makeExactMatch.
	  * FIXME: support TemporalAdaptation
	  */
	ContainerInfo* findMp4Stream(u32 lowerDimX, u32 upperDimX, 
								 u32 lowerDimY, u32 upperDimY, 
								 u32 lowerBitRate, u32 upperBitRate, 
								 bool greyScale, float frameRate,
								 bool* isExactMatch);



	/** @param return returns false if a stream with the same properties already
	  * exists, otherwise the video is added to the MetaObject and true is returned
	  */
	bool addMP4Stream(ContainerInfo* stream);
	/** returns true iff the stream is found */
	bool deleteMP4Stream(ContainerInfo* stream);
	const Url* getUrl() const { return url;};
	bool saveToConfigFile(const char* cfg);
	int getNumberOfVariations() const {return variations.size();};
	/** returns a pointer to the internal variations. locks and KEEPS the lock.
	  * You must call releaseVariationsLock(), when you are done with the list
		*/
	std::list<ContainerInfo*>* getVariations();
	void releaseVariationsLock();
	
protected:
	/** if this method is called, it will destroy
	  * all previously stored ContainerInfo objects
	  * and delete the url. Then it will load the cfg
	  * file.
	  * Recommended use is to call it only from a
	  * constructor
	  */
	bool loadFromConfigFile(const char* cfg);
	/** calculates quality inside the range min, max for value val
		* @param min specifies the lower border of the range
		* @param max specifies the upper border of the range
		* @param val specifies a point inside the range
		* @param return will be -1 if @param min > @param max, or will be zero
		* if @param val is outside of the range.otherwise a value between
		* 0 and 1 will be returned, which states how close @param val is to @param max
		* if @param val == @param max then 1.0 will be returned
		*/
	double calcSimpleQuality(u32 min, u32 max, u32 val);
protected:
	list<ContainerInfo*> variations;
	Url* url;
	VMutex lockAccess;
	u32 pendingOutsideUnlock;
};

#endif
