 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	ContainerDemux.cpp                                            *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#include "ContainerDemux.hpp"
#include "ContainerInfo.hpp"
#include "Adaptor.hpp"
#include "DataChannel.hpp"
#include "io/IOCreator.hpp"
#include "io/MPGStreamIO.hpp"
#include "io/DevNull.hpp"
#include "DataSink.hpp"


/** takes the info object and demultiplexes every ES into
 * a separate file. A copy of the info object with updated
 * input descriptors is returned.
 * Only call this method with an .mp4 file as input.
 * at each visualstream a visualadaptor is set 
 */
ContainerInfo *ContainerDemux::demultiplex(ContainerInfo * &mp4Info, const char* outputDir,
																		const Adaptor * visualAdaptor,
																		bool omitHeader, bool deleteMP4,
																		bool visualizeAdaptation) {
	ContainerInfo * mp4New = mp4Info->clone();
	ContainerDemux::demultiplexAndUpdateIO(mp4New, outputDir, visualAdaptor, omitHeader,
																	deleteMP4, visualizeAdaptation);
	return mp4New;
}


/** same as demultiplex(..) but the input descriptors are updated 
* to point to the new input. Per default the MP4FILE is NOT DELETED!
* If an adaptor is set, each visualstream will be adapted with this adaptor    
*/
void ContainerDemux::demultiplexAndUpdateIO(ContainerInfo * &mp4Info,
																						const char* outputDir,
																						const Adaptor * visualAdaptor,
																						bool omitHeader, bool deleteMP4,
																						bool visualizeAdaptation) {

	list < ESInfo * > *es = mp4Info->getESList();
	assert(es && es->size() > 0);
	list < ESInfo * >::iterator li = es->begin();
	DataChannel **dc = new DataChannel *[es->size()];
	for (int i = 0; i < (int) es->size(); i++) {
		// for each ES create a DataChannel, MP4IO will be used for reading
		dc[i] = new DataChannel(IOCreator::createInput((*li), false), (*li));

		char *out = dc[i]->enableCaching(omitHeader,outputDir);

		(*li)->setInput(out);

		delete[] out;

		if ((*li)->isVisualStream() && visualAdaptor != NULL) {
			Adaptor* at=((Adaptor*)visualAdaptor)->clone();
			at->setESInfo(*li);
			dc[i]->setAdaptor(at);
		}
		
		if (visualizeAdaptation && (*li)->isVisualStream()) {
			dc[i]->visualizeCaching();
		}

		++li;
	}
	
	// rewind the iterator
	li = es->begin();
	for (int j = 0; j < (int) es->size(); j++) {
		// start & wait for each thread to finish 
		// -> ISOMP4 lib is not thread safe
		dc[j]->start();
		if(dc[j]->running())
			dc[j]->wait();
		/* the stream may have contained empty frames,
		 * so update the duration to avoid inconsistent
		 * ESInfo::frameStatistics
		 */
		ESInfo *ess = (ESInfo *) dc[j]->getESInfo();
		u32 essFrames = ess->getNumberOfMediaSamples();
		u32 readFrames = dc[j]->getInput()->getCurrentFrameNumber();
		if (readFrames != essFrames) {
		  dprintf_full("ContainerDemux::demultiplexAndUpdateIO() stream %d: read %u frames, correcting ESInfo (was %u)\n",
			       j, readFrames, essFrames);
		  ess->setDuration((u64) readFrames * (u64) ess->getVOPTimeIncrement());
		}
		ess->setCompleteState(true);
		delete dc[j];
	}
	delete[] dc;
	// now delete the mp4file
	if (deleteMP4) {
		dprintf_full("ISOMP4ContainerFile::demultiplexAndUpdateIO: Deleting ISOMP4ContainerFile\r\n");
		FILE *fp = fopen(mp4Info->getLocalFile(), "w");

		if (fp) {
			fclose(fp);
			remove(mp4Info->getLocalFile());
		}
	}
}

	
void ContainerDemux::adaptESInfo(ESInfo* es, const char* outputFile, Adaptor* adaptor, u32 startFrame, u32 endFrame, bool omitHeader, bool visualize) {
	assert(es);
	
	DataChannel* dc = ContainerDemux::createDataChannel(es,outputFile,adaptor,startFrame,endFrame,omitHeader,visualize);
	if(!dc)
	    return;
	    
	dc->start();
	if(dc->running())
	{
		dc->wait();
	}
	delete dc;
	
}


/** adapts single ESInfo, sets outputfilename, start and endframe will only
     work with demultiplexed es! */	
void ContainerDemux::adaptESInfo(ESInfo* es, const char* outputFile, Adaptor* adaptor, 
	double startSecond, double endSecond, bool omitHeader, bool visualize) {
	assert(es);
	u32 startFrame=es->getFrameNumber(startSecond);
	u32 endFrame=4294967295u; //2^32-1
	if(endSecond > 0.0)
		endFrame=es->getFrameNumber(endSecond);
	ContainerDemux::adaptESInfo(es,outputFile,adaptor,(u32)startFrame,(u32)endFrame,omitHeader,visualize);
}


/** creates a DataChannel which is ready for immediate use. Just start the returned DataChannel */
DataChannel* ContainerDemux::createDataChannel(ESInfo* es, const char* outputFile, 
					Adaptor* adaptor, u32 startFrame, u32 endFrame, 
					bool omitHeader, bool visualize) {
	assert(es);
	if(!es)	// for release build where all asserts vanish
		return NULL;

	IO* input = IOCreator::createInput(es, false);

	input->setToClosestIFrame(startFrame);
	input->setEndFrameNumber(endFrame);
	IO* output=NULL;
	if(outputFile)
		output = new MPGStreamIO(es, outputFile, true, omitHeader);
	else
		output = new DevNull();
	
	DataSink* datasink = new DataSink(0, output, "127.0.0.1", 0, NULL);
	DataChannel* dc = new DataChannel(input, es);
	if(visualize)
		dc->visualizeCaching();		
	dc->setAdaptor(adaptor);
	dc->insert(datasink);
	dc->play(0, 0.0);
	return dc;
}
