
/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	ContainerFile.cpp                                    *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#include "net/Url.hpp"

#ifdef ENABLE_DVDREAD
#include "DVDContainerFile.hpp"
#endif

#include "ContainerFile.hpp"
#include "ISOMP4ContainerFile.hpp"
#include "ffMP4ContainerFile.hpp"
#include "MKContainerFile.hpp"
#include "VITContainerFile.hpp"
#include "OGGContainerFile.hpp"
#include "io/HttpByteStream.hpp"


/** init. a videoobject from a pre-stored ContainerFile. 
 * No data is read, just metainfo
 */
ContainerInfo *ContainerFile::loadContainerInfo(const char *fileName, Url* uri, const ParseType type) {
	ContainerInfo * info = NULL;
	HttpByteStream *stream = NULL;
	HttpMimeType mimeType = mime_unknown;

	dprintf_full("ContainerFile::loadContainerInfo: ####HANDLE####\n");
#ifdef ENABLE_DVDREAD	
	dvd_handle_t *dvd_handle = DVDContainerFile::openDVD(fileName);
	
 	if(dvd_handle) {
 		// treat as DVD
 		u32 main_title_no = DVDContainerFile::getMainTitleNo(dvd_handle);
 		/* load the containerinfo for the main video title set */
 		info = DVDContainerFile::loadContainerInfo(dvd_handle,main_title_no);
 		delete dvd_handle;
 	}
#endif


	if(info==NULL) {
		// was not detected by DVD?

		if (!uri)
			uri = new Url(fileName);

		if (strstr(uri->getProtocol(), "http")) {
			stream = new HttpByteStream(uri->toString(), true /* forReading */);
			if (!stream->open()) {
				dprintf_err("ContainerFile::loadContainerInfo: file not found: %s!\n",fileName);
				return NULL;
			}
			mimeType = stream->getMimeTypeEnum();
		}
	

		const char* suffix=strrchr(fileName,'.');
		if(suffix || (mimeType != mime_unknown)) {
			
			// handle VITOOKI file format
			if((mimeType == mime_vit) || strstr(fileName,".vit") || strstr(fileName,".VIT")) {
				dprintf_full("ContainerFile::loadContainerInfo(): VITOOKI container detected!\n");
				info = VITContainerFile::loadContainerInfo(uri,stream);
			}// if .vit

#ifdef ENABLE_OGG
			// handle OGG file format
			else if((mimeType == mime_ogg) || strstr(fileName,".ogg") || strstr(fileName,".OGG")) {
				dprintf_full("ContainerFile::loadContainerInfo(): OGG container detected!\n");
				info = OGGContainerFile::loadContainerInfo(uri,stream);
			}// if .ogg
#endif

#ifdef ENABLE_MKV
			// handle MATROSKA file format
			else if((mimeType == mime_mkv) || strstr(fileName,".mkv") || strstr(fileName,".MKV")) {
				if (stream)
					stream->close(); //direct stream passing not yet supported....
				dprintf_full("ContainerFile::loadContainerInfo(): MATROSKA detected!\n");
				info = MKContainerFile::loadContainerInfo(fileName,uri);
			}// if .mkv
#endif

			else if(strstr(fileName,".mp4") || strstr(fileName,".MP4") || mimeType == mime_mp4) {
				if (stream)
					stream->close(); //direct stream passing not yet supported....
				
				// we can now choose between ISOMP4 or ffmpeg
				// assume that if ISOMP4 is defined this is the preferred one
				dprintf_full("ContainerFile::loadContainerInfo(): MP4 detected!\n");
#ifdef ISOMP4
				info = ISOMP4ContainerFile::loadContainerInfo(fileName,uri);
#endif
#ifdef ENABLE_FFMPEG
				// if this fails try ffmpeg
				if(info==NULL)
					info = ffMP4ContainerFile::loadContainerInfo(fileName,uri,(ffMP4ContainerFile::ParseType)type);
#endif
			}//if .mp4
			else {
				if (stream)
					stream->close(); //direct stream passing not yet supported....
				
#ifdef ENABLE_FFMPEG
				dprintf_full("ContainerFile::loadContainerInfo(): UNDEFINED detected!\n");
				// NOT KNOWN, still give it a try with ffmpeg....
				info = ffMP4ContainerFile::loadContainerInfo(fileName,uri,(ffMP4ContainerFile::ParseType)type);	
#endif
			}// NOT KNOWN extension
		} //has suffix
		else {
			// no suffix and no mimetype????? try ffmpeg and hope
				if (stream)
					stream->close(); //direct stream passing not yet supported....
#ifdef ENABLE_FFMPEG
			info = ffMP4ContainerFile::loadContainerInfo(fileName,uri,(ffMP4ContainerFile::ParseType)type);	
#endif
		}	
	}

	return info;	
}


/** FIXME not implemented: multiplexes a set of ES into a MP4ContainerFile */
bool ContainerFile::saveToContainerFile(const char *fileName, ContainerInfo *ci) {
	dprintf_err("ContainerFile::saveToContainerFile:  FIXME not implemented\n");
	return false;
}
