 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title: DVDContianerFile.hpp                                          *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef WIN32

#include "global.hpp"

class ContainerInfo;
class ESInfo;

/**                          **/
/** dvd specific definitions **/
/**                          **/
	
/** video definitions **/

/* mpeg version */
#define MPEG1 0   // mpeg1 video
#define MPEG2 1   // mpeg2 video

/* video format */
#define NTSC 0   // ntsc format
#define PAL 1    // pal fomrat

/* display aspect ratio*/
#define NORM 0   // 4:3
#define WIDE 1   // 16:9

/* letterboxed */
#define NOTLETTERBOXED 0   // normal
#define LETTERBOXED 1      // already letterboxed

/* bit rate */
#define VBR 0   // variable bit rate
#define CBR 1   // constant bitrate


/** audio definitions **/

/* audio format */
#define AC3 0        // dolby digital
#define MPEG1A 2     // mpeg1 audio
#define MPEG2EXT 3   // extended mpeg2 audio
#define LPCM 4       // lpcm audio
#define DTS_AUDIO 6  // dts audio

/* has multichannel extension */
#define NOMULTI 0   // no multichannel extension
#define MULTI 1     // multichannel extension

/* application mode */
#define NOSPECAM 0   // not specified
#define KARAOKE 1    // karaoke mode
#define SURROUND 2   // surround mode

/* quantization */
#define BIT16 0   // 16 bit
#define BIT20 1   // 20 bit
#define BIT24 2   // 24 bit
#define DRC 3     // drc

/* sample frequency */
#define KHZ48 0       // 48Khz
#define UNKNOWNSF 1   // unknown frequency

/* code extension */
#define NOSPECCE 0   // not specified
#define NORMCAP 1    // normal caption
#define VIMP 2       // audio for visually impaired
#define DIRCOMM1 3   // directors comment 1
#define DIRCOMM2 4   // directors comment 2

/* dolby surround */
#define NODOLBY 0   // no dolby surround
#define DOLBY 1     // dolby surround


/** subpicture definitions **/

/* coding mode */
#define RLE 0     // rle
#define CMEXT 1   // extended coding mode
/* including libdvdread stuff */
#include <dvdread/dvd_reader.h>
#include <dvdread/ifo_read.h>
#include <dvdread/ifo_types.h>


/**                   **/
/** structs and types **/
/**                   **/
/* handles one instance of a dvd */
struct dvd_handle_s {
		char filename[MAX_STR_LEN];
		dvd_reader_t *dvd_reader;
		u32 nr_vts;
		u32 main_title_no;
		char *target;
};

/* opaque type for one dvd instance */
typedef struct dvd_handle_s dvd_handle_t;

/* used internal for one video title set */
typedef struct vts_info_s vts_info_t;


/**                         **/
/** class description       **/
/** and function prototypes **/
/**                         **/

/* Container for DVDs */
class DVDContainerFile {
	public:
		
		/* opens a dvd
		 * @param file (device or image)
		 * @return handle for one instance of a dvd or 0 on fail
		 */
		static dvd_handle_t * openDVD(const char *file);
		
		/* close dvd
		 * @param dvd handle
		 * @return void
		 */
		static void closeDVD(dvd_handle_t *dvd_handle);
		
		/* get number of video title sets
		 * @param dvd handle
		 * @return number of video title sets
		 */
		static u32 getNrOfVTS(dvd_handle_t *dvd_handle);
	
		/* get the filename of the current dvd
		 * @param dvd_handle
		 * @return filename
		 */
		static char * getFileName(dvd_handle_t *dvd_handle);
	
		/* get suggested main title number
		 * @param dvd_handle
		 * @return suggested main title
		 */
		static u32 getMainTitleNo(dvd_handle_t *dvd_handle);
		
		/* set main title to a custom VTS number
		 * @param dvd_handle
		 * @param vts number
		 * @return void_function
		 */
		static void setMainTileNo(dvd_handle_t *dvd_handle, u32 main_title_no);
		
		/* set temporary target for storing the vob
		 * @param path+filename e.g. /home/foo/myvob.vob
		 */
		static void setTargetName(dvd_handle_t *dvd_handle,char *target);
		
		/* get temporary target name
		 * @param dvdhandle
		 * @return path+filename
		 */
		static char * getTargetName(dvd_handle_t *dvd_handle);
				
		/* load container information for one video title set
		 * @param dvd_handle, video title set number
		 * @return container information
		 */
		static ContainerInfo * loadContainerInfo(dvd_handle_t *dvd_handle, u32 vts_nr);
		
		/* rip video title
		 * @param dvd_handle
		 * @param percentage to rip (1-100 in %), 0 means 100%
		 * @return u32 error code
		 */
		static u32 copyMain(dvd_handle_t *dvd_handle, u32 percentage);
				
	private:
		
		/* get basic video title set information
		 * (number of audio and subpicture streams + ifo handle
		 * @param dvd handle, video title set number
		 * @return basic vts info
		 */
		static vts_info_t * getVTSInfo(dvd_handle_t *dvd_handle, u32 vts_nr);
	
		/* get video ESInfo
		 * @param vts info
		 * @return video ESInfo
		 */
		static ESInfo * getVideoInfo(vts_info_t *vts_info, ContainerInfo *container);
	
		/* get audio ESInfo
		 * @param vts info, audio stream number
		 * @return audio ESInfo, 0 on fail
		 */
		static ESInfo * getAudioInfo(vts_info_t *vts_info, u32 stream_nr, ContainerInfo *container);
	
		/* get subpicture ESInfo
		 * @param vts_info, subpicture stream number
		 * @return subpicture ESInfo, 0 on fail
		 */
		static ESInfo * getSubInfo(vts_info_t *vts_info, u32 stream_nr, ContainerInfo *container);
};

#endif /* WIN32 */

