 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title: ISOMP4ContainerFile.cpp                                       *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifdef ISOMP4


#include "ISOMP4ContainerFile.hpp"
#include "ContainerInfo.hpp"
#include "ESInfo.hpp"
#include "DataChannel.hpp"
#include "DataSink.hpp"
#include "MPGStreamIO.hpp"
#include "Adaptor.hpp"
#include "AdaptorChain.hpp"
#include "MP4Decoder.hpp"
#include "Renderer.hpp"
#include "DevNull.hpp"
#include "MPGSDLRenderer.hpp"
#include "Url.hpp"
#include "SDP.hpp"
#include <string.h>

/** init. an object from a pre-stored ISOMP4ContainerFile. No data is read, just meta-
* info
*/
ContainerInfo *ISOMP4ContainerFile::loadContainerInfo(const char *fileName,const Url* uri)
{
	const char* pUri=NULL;
	if(uri)
		pUri=uri->toString();
	
	ContainerInfo * result = new ContainerInfo(pUri,fileName, ContainerInfo::MP4);
	
	dprintf_full("ISOMP4ContainerFile::loadMP4Info START-MARK\r\n");
	
	ISOErr err;
	ISOMovie moov;
	
	char *mp4File = new char[strlen(fileName) + 1];
	strcpy(mp4File, fileName);
	
	err = ISOOpenMovieFile(&moov, mp4File, MP4OpenMovieNormal);
	if (err != MP4NoErr) {
		dprintf_err("ISOOpenMovieFile failed\r\n");
		delete result;
		delete mp4File;
		return NULL;
	}
	dprintf_full("ISOOpenMovieFile is ok\r\n");
	
	u32 timeScale = 0;
	err = MP4GetMovieTimeScale(moov, &timeScale);
	if (err != MP4NoErr) {
		dprintf_err("Failed to get timeScale\r\n");
		timeScale = 0;
	}
	
	ISOHandle iodhandle;
	err = ISONewHandle(0, &iodhandle);
	if (err != MP4NoErr) {
		dprintf_err("ISONewHandle failed\r\n");
		ISODisposeMovie(moov);
		delete result;
		delete mp4File;
		return NULL;
	}
	
	err = MP4GetMovieInitialObjectDescriptor(moov, iodhandle);
	if (err != MP4NoErr) {
		dprintf_err("MP4GetMovieInitialObjectDescriptor failed\r\n");
		ISODisposeMovie(moov);
		delete result;
		delete mp4File;
		return NULL;
	}
	u32 iodHandleSize = 0;
	MP4GetHandleSize(iodhandle, &iodHandleSize);
	result->setIODHandle((u8 *) * iodhandle, iodHandleSize);
	
	// now parse for streams
	u32 streamcnt = 0;
	err = ISOGetMovieTrackCount(moov, &streamcnt);
	if (err != MP4NoErr) {
		dprintf_err("ISOGetMovieTrackCount failed\r\n");
		ISODisposeMovie(moov);
		delete result;
		delete mp4File;
		return NULL;
	}
	ISOTrack trak;
	u32 trackId;
	u32 size = 0;
	for (u32 i = 1; i <= streamcnt; i++) {
		err = MP4GetMovieIndTrack(moov, i, &trak);
		if (err != MP4NoErr) {
	    ISODisposeMovie(moov);
	    delete result;
	    delete mp4File;
	    return NULL;
		}
		err = MP4GetTrackID(trak, &trackId);
		if (err != MP4NoErr) {
	    ISODisposeMovie(moov);
	    delete result;
	    delete mp4File;
	    return NULL;
		}
		ESInfo *es = ISOMP4ContainerFile::loadESInfo(moov, result, trackId);
		if (!es) {
	    delete es;
	    ISODisposeMovie(moov);
	    delete result;
	    delete mp4File;
	    return NULL;
		}
		dprintf_full("loadFromISOMP4ContainerFile creating ES %i\r\n", i);
		es->setInput(mp4File);
		
		result->addESInfo(es);
		size += es->getOrigMediaSize();
	}
	
	ISODisposeMovie(moov);
	ISODisposeHandle(iodhandle);
	
  QualityInfo *qi = result->getQualityInfo();
	qi->steps[0]->size = size;
	return result;
};


ESInfo *ISOMP4ContainerFile::loadESInfo(ISOMovie & moov, ContainerInfo * vob,
														u32 streamId)
{
	ISOTrack trak;
	ISOErr err;
	err = ISOGetMovieTrack(moov, streamId, &trak);
	if (err != MP4NoErr) {
		return NULL;
	}
	
	ISOMedia media;
	err = ISOGetTrackMedia(trak, &media);
	if (err != MP4NoErr) {
		return NULL;
	}
	u32 width = 0, height = 0;
    // FIXME: returns invalid dimension????
	err = MJ2GetTrackDimensions(trak, &width, &height);
	if (err != MP4NoErr) {
		width = 0;
		height = 0;
		
	}
	u32 handlerType;
	err = ISOGetMediaHandlerDescription(media, &handlerType, 0);
	if (err != MP4NoErr) {
		return NULL;
	}
	ISOTrackReader reader;
	err = ISOCreateTrackReader(trak, &reader);
	if (err != MP4NoErr) {
		return NULL;
	}
	
	ISOHandle dcHandle;
	err = ISONewHandle(0, &dcHandle);
	if (err != MP4NoErr) {
		return NULL;
	}
	err = MP4TrackReaderGetCurrentDecoderConfig(reader, dcHandle);
	if (err != MP4NoErr) {
		return NULL;
	}
	
	u32 mediaTimeScale = 0;
	u32 totalSamples = 0;
	ISOGetMediaTimeScale(media, &mediaTimeScale);
	ISOGetMediaSampleCount(media, &totalSamples);
	u64 duration = 0;
	ISOGetMediaDuration(media, &duration);
	u32 ticksDiff=duration/totalSamples;
	dprintf_full("Ticks/sec %i DurationinTicks %i numberofSamples %u\n", mediaTimeScale,
							 int (duration),totalSamples);
	u32 dcsize = 0;
	MP4GetHandleSize(dcHandle, &dcsize);
	
	u8 *dptr = (u8 *) (*dcHandle);
	
	u8 *encodedDecoderConfig = new u8[dcsize * 2 + 1];
	SDP::encodeDecoderConfig(encodedDecoderConfig, dptr, dcsize);
	
	u32 objectType, streamType, bufferSize, upstream, maxbit = 0, avgbit =
		0;
	//MP4GetMediaDecoderType(media, 1, &objectType, &streamType, &bufferSize, 0);
	
	MP4GetMediaDecoderInformation(media, 1, &objectType, &streamType,
																&bufferSize, &upstream, &maxbit, &avgbit,
																0);
	
	// determine the size of the ES
	u32 mediaSize = 0;
	u32 tmpSize = 0;
	u32 dummy;
	u32 dummy2, dummy3;
	while (MP4TrackReaderGetNextAccessUnit
				 (reader, dcHandle, &tmpSize, &dummy, &dummy2,
					&dummy3) == MP4NoErr) {
		mediaSize += tmpSize;
	}
	ISODisposeHandle(dcHandle);
	ISODisposeTrackReader(reader);
	return new ESInfo(streamId, handlerType, vob, ticksDiff, encodedDecoderConfig,
										objectType, streamType, bufferSize, width, height,
										avgbit, maxbit, mediaTimeScale, duration, mediaSize,
										totalSamples);
};


#endif
