/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	VITContainerFile.cpp                                        *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/


#include "VITContainerFile.hpp"
#include "ContainerInfo.hpp"
#include "VideoESInfo.hpp"
#include "AudioESInfo.hpp"
#include "DataChannel.hpp"
#include "io/BufferedFileByteStream.hpp"
#include "io/UnbufferedFileByteStream.hpp"
#include "io/HttpByteStream.hpp"
#include "io/VITMuxDemuxIO.hpp"


ContainerInfo *VITContainerFile::loadContainerInfo(const Url* uri, ByteStream *stream) {
	ContainerInfo *ci = NULL;

	const char* pUri=NULL;
	if(uri)
		pUri=uri->toString();
	else
		return NULL;

	dprintf_full("VITContainerFile: :loadContainerInfo: Init... URI %s\n",pUri);

	if (stream == NULL) { //reopen the byteStream
		if (strstr(uri->getProtocol(), "file"))
			stream = new UnbufferedFileByteStream(uri->getPath(), true /* forReading */);
		else if (strstr(uri->getProtocol(), "http"))
			stream = new HttpByteStream(uri->toString(), true /* forReading */);
	}
	
	if (!stream->open()) {
	  dprintf_err("VITContainerFile::loadContainerInfo: file not found: %s!\n",pUri);
	  return NULL;
	}
		
	ci = new ContainerInfo(pUri, uri->getPath(), ContainerInfo::VITOOKI);
	assert(ci);
 	ci->setMuxDemuxIO(new VITMuxDemuxIO(stream));
	
	//FOURCC
	char fourcc[5];
	stream->read( fourcc,4); //only read first 4 bytes, add \0
	fourcc[4] = '\0';
	if (strcmp(fourcc,"VIT ") != 0) {
		dprintf_err("VITContainerFile::loadContainerInfo: file %s is no ViTooKi VIT container file! (got #%s#)\n",uri->toString(),fourcc);
		delete ci;
		return NULL;
	}
	
	//version number
	u32 version;
	stream->read( &version, sizeof(version));
	if (version != VIT_VERSION_NUMBER) {
		dprintf_err("VITContainerFile::loadContainerInfo: VIT file %s has wrong version number %i (expected %i)!\n",
								pUri,version,VIT_VERSION_NUMBER);
		delete ci;
		return NULL;
	}
	
	//number of following (muxed) tracks
	u32 numTracks;
	stream->read( &numTracks, sizeof(numTracks));
	dprintf_full("VITContainerFile::loadContainerInfo: VIT file contains %i elementary streams\n",numTracks);
	
	
	//get list of tracks
	for (u32 curr=0; curr < numTracks; curr++) {
		
		VITTrackInfo ti;
		stream->read(&ti, sizeof(ti));
		dprintf_full("VITContainerFile::loadContainerInfo: Header with TrackID %i CodecID %i StreamType %i\n", 
								 ti.streamID, ti.codecID, ti.streamType);

		ESInfo *esi;
		
		if (ti.streamType == VIT_VIDEO) {
			esi = new VideoESInfo(ti.streamID, ti.streamType, ci, ti.vopinc /*time increment*/,
														NULL /*encodedDecoderConfig*/, 0, 0, 
														ti.kbps * 1024, ti.kbps * 1024, 90000, 0 /*duration*/, 
														0 /*mediaSize*/, ti.esi.vi.width, ti.esi.vi.height,
														true, true, (CodecID)ti.codecID, 1.0 /*quality*/, ti.esi.vi.aspectRatio);
		} else
			if (ti.streamType == VIT_AUDIO) {
				esi = new AudioESInfo(ti.streamID, ti.streamType, ci, ti.vopinc /*time increment*/,
															NULL /*encodedDecoderConfig*/, 0, 0, 
															ti.kbps * 1024, ti.kbps * 1024, 90000, 0 /*duration*/, 
															0 /*mediaSize*/, ti.esi.ai.sampleRate, ti.esi.ai.channels, true,
															(CodecID)ti.codecID, 1.0 /*quality*/, 16 /* bits_per_sample */);
				char lang[3];
				strncpy(lang,ti.esi.ai.language,2);
				lang[3]='\0';
				((AudioESInfo*)esi)->setLanguage(lang);
				
			} else {
				dprintf_err("VITContainerFile: TrackID %i CodecID %i has UNKNOWN StreamType %i\n", 
										ti.streamID, ti.codecID, ti.streamType);
				delete ci;
				return NULL;
			}
		
		esi->setInput(uri->toString());
		assert(ci);
		ci->addESInfo(esi);
	}
	
	//NOTE: dont close the stream since it is needed in VITMuxDemuxIO for reading the frames !!!
	//stream->close();
	return ci;
}

