 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	ffMP4ContainerFile.cpp                                        *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#include <string.h>

#include "ffMP4ContainerFile.hpp"
#include "ContainerInfo.hpp"
#include "VideoESInfo.hpp"
#include "AudioESInfo.hpp"
#include "DataChannel.hpp"
#include "DataSink.hpp"
#include "Adaptor.hpp"
#include "AdaptorChain.hpp"
#include "MP3Parser.hpp"

#include "../adaptors/MP4Decoder.hpp"

#include "../io/DevNull.hpp"
#include "../io/MPGStreamIO.hpp"

#include "../net/Url.hpp"
#include "../net/SDP.hpp"

/* This structure is there temporarily .... I know it looks ugly.
   will be removed when FFmpeg becomes consistent in providing
   all required parameters
*/
struct stInfo {
	CodecID codec_id;     //eg. CODEC_ID_MPEG4==15,  CODEC_ID_MPEG4AAC=0x15003
	float rate;			    	//frame_rate for Video; sample_rate for audio
	long size;				    //size of Elementary Stream
	u32 packet_size;
	int totalSamples;			//total samples in ES
	long avgBitRate;		
	u32 bits_per_sample;	//bits per sample of uncompressed wave frame (eg. 16)
	u32 channels;         //audio only (eg. 2...stereo)
	u32 packetDuration;
	float duration;				//duration of ES in ticks
	float test_duration;  //used internally for calculation               
	bool done;            //       -do-
};
static stInfo strm[MAX_STREAMS]; 			

//Some default values need to be declared here;
static int frame_width  = 352;
static int frame_height = 288;
//static int frame_rate = 30; /*@ * FRAME_RATE_BASE;*/
//static int frame_rate_base = 1;
static int audio_sample_rate = 44100;
static int audio_channels = 2;

	
/**********************************************************/
/** init. an object from a pre-stored ffMP4ContainerFile. 
 * No data is read, just metainfo
 */
ContainerInfo *ffMP4ContainerFile::loadContainerInfo(const char *fileName, const Url* uri, const enum ParseType type) {
	int streamcnt = 0;
	const char* pUri=NULL;
	if(uri)
		pUri=uri->toString();
	
	ContainerInfo * result = new ContainerInfo(pUri,fileName, ContainerInfo::MP4);

	enum ParseType parseType = type;
	if (strstr(fileName+strlen(fileName)-4,".m4v") || strstr(fileName+strlen(fileName)-4,".M4V") ) { 
		//force run_through for single elem streams
		parseType = RUN_THROUGH;
	}

	dprintf_full("ffMP4ContainerFile::loadContainerInfo %s START-MARK parseType %i\r\n",fileName,parseType);
	
	switch(parseType) {
		case RUN_THROUGH:
			streamcnt = getStreamParameters_run_through(fileName);
			break;
		case ESTIMATE1:
			streamcnt = getStreamParameters_estimate1(fileName);
			break;
		case ESTIMATE2:
			streamcnt = getStreamParameters_estimate2(fileName);
			break;
		default:
			streamcnt = getStreamParameters_run_through(fileName);
	}
	
		
	if(streamcnt < 0) {
		dprintf_full("ffMP4ContainerFile::loadContainerInfo: Error in getting stream parameters\n");
		delete result;
		return NULL;
	}
	
	AVInputFormat *informat = NULL;
  AVFormatContext *ic=NULL;
	AVFormatParameters params, *ap = &params;

	//AVOutputFormat *fmt; //this has been introduced to support multiple file formats.
	/* auto detect the output format from the name. default is mp4. */
	/*fmt = guess_format(NULL, fileName, NULL);
	if (!fmt) {
		printf("Could not deduce output format from file extension: using mp4.\n");
	        fmt = guess_format("mp4", NULL, NULL);
	}
	if (!fmt) {
		fprintf(stderr, "Could not find suitable output format\n");
	        exit(1);
	}*/

	
	
	/* set some default parameters if required */
	memset(ap, 0, sizeof(*ap));
	ap->sample_rate = audio_sample_rate;
	ap->channels = audio_channels;
	//	ap->frame_rate = frame_rate;
	ap->width = frame_width;
	ap->height = frame_height;
 
	int ret;
	char *mp4File = new char[strlen(fileName) + 1];
	strcpy(mp4File, fileName);
	

	//open the input file with the generic libavformat function(avformat.h)
	ret = av_open_input_file(&ic, fileName, informat, 0, ap);
	if(ret < 0)	{
		//HACK FOR DVB-HTTP STREAM: assume&force mpeg
		informat = av_find_input_format("mpeg");
		dprintf_full("informat is %p\n",informat);
		ret = av_open_input_file(&ic, fileName, informat, 0, ap);
		if(ret < 0)	{
			//print_error(fileName, ret);
			dprintf_full("ffMP4ContainerFile::loadContainerInfo : Error in opening file!!\n");
			delete result;
			delete[] mp4File;
			return NULL;
		}
	}
	dprintf_full("ffMP4ContainerFile::loadContainerInfo : av_open_input_file is ok\r\n");

		
	//If not enough info to get the stream parameters, we decode the
	//first frames to get it. (used in cases when file has no headers eg. mpeg1)
	ret = av_find_stream_info(ic);
	if (ret < 0) {
		dprintf_full("ffMP4ContainerFile: %s - could not find codec parameters\n", fileName);
		exit(1);
	}
	
	
	//	unsigned int timeScale = 0;
	// timeScale = ic->pts_den; 			 
	
	
	/*FIXME DUMMY IOD passed 
	  iodhandle not set because ffmpeg-0.4.6 does not provide it 
	  set the iodhandle here when needed
	  NOTE: IOD is not required by most of the players so it may not cause any problem
	  However, the ISO/IEC Reference code player does expect the IOD 
	*/
	u32 iodHandleSize = 30; 	//Arbit size
	u8 iod[] = "11223344556677889900AABBCCDDEEFF11223344556677889900AABBCCDD";
        u8 *iodhandle = iod;
        result->setIODHandle(iodhandle, iodHandleSize);
	

	
	streamcnt = 0;
	streamcnt = ic->nb_streams;

	if (streamcnt == 0) {
		dprintf_err("ffMP4ContainerFile::loadFromffMP4ContainerFile  couldn't find any ElemStreams in container!\n");
		::exit(1);
	}
	
	unsigned int size = 0;	
	for (int streamIndex = 0; streamIndex < streamcnt; streamIndex++) 
	{
		ESInfo *es = ffMP4ContainerFile::loadESInfo(ic, result, streamIndex);
		if (!es) {
			delete es;
			delete result;
			av_close_input_file(ic);	
			delete[] mp4File;
			return NULL;
		}
		
		dprintf_full("ffMP4ContainerFile::loadFromffMP4ContainerFile creating ES %i\r\n", streamIndex);
		es->setInput(mp4File);
		result->addESInfo(es);
		size += es->getOrigMediaSize();
	}
	
	QualityInfo *qi = result->getQualityInfo();
	qi->steps[0]->size = size;
	av_close_input_file(ic);	
	delete[] mp4File;
 	return result;	
};


/**********************************************************/
ESInfo *ffMP4ContainerFile::loadESInfo(AVFormatContext *ic, ContainerInfo *vob, int streamIndex)
{
	//XXX currently some parameters for audio are set in MP4audioDecoder::initialize() 
	//this is becaue ffmpeg fails to deliver the necessary information for audio.	
	AVStream *st;
	st = ic->streams[streamIndex];
	
	unsigned int streamId, handlerType, width, height, mediaTimeScale, totalSamples, mediaSize;
	unsigned int objectType=0, streamType=0, bufferSize=0, avgbit = 0, maxbit = 0; 
	u64 duration = 0;              
	float aspect_ratio = -1;
	
	streamId = st->id;
	width = st->codec->width;   		
	height = st->codec->height; 
	streamType = st->codec->codec_type; 
	
	if(st->codec->sample_aspect_ratio.num == 0)
		aspect_ratio = (float)width / (float)height;
	else if (height != 0)
		aspect_ratio = av_q2d(st->codec->sample_aspect_ratio) * (float)width / (float)height;
	if (aspect_ratio <= 0)
		aspect_ratio = (float)width / (float)height;
	
	handlerType = st->codec->codec_type; 		//video = 0; audio = 1; others = -1	
	if(handlerType == 0)				//FIXME set the #defines in global.hpp for ffmpeg and replace these lines
		handlerType = MP4VisualHandlerType;
	if(handlerType == 1)
		handlerType = MP4AudioHandlerType;
	
	//	mediaTimeScale = ic->pts_den;
	//some videos (divx?) come with 1000000, instead of 90000
	mediaTimeScale = 90000; //FIXME: quickhack for divx
 	//convert time base of duration from secs to 90000 timebase
	duration = (int)ceil(strm[streamIndex].duration*mediaTimeScale);
	avgbit = strm[streamIndex].avgBitRate; 			
	maxbit = 0;  
	mediaSize = strm[streamIndex].size;
	totalSamples = strm[streamIndex].totalSamples;
	if (strm[streamIndex].totalSamples < 0 || strm[streamIndex].totalSamples == 0x7FFFFFFF) {
		dprintf_err("FATAL: ffMP4ContainerFile::loadESInfo could not determine number of totalSamples (was %i)!\n",totalSamples);
		dprintf_err("so assuming arbitrary high values for live-streaming\n");
		totalSamples = 999999;
		duration = totalSamples * (mediaTimeScale/25); //assume 25 fps
		mediaSize = duration/mediaTimeScale * avgbit;
		//::exit(1);

		/*
		dprintf_err("so assuming VOPTimeIncrement of lame-friendly 2351!\n");
		//FIXME: absolute hack alarm!
		// vopInc 2351 seems to be a "lame"-friendly value, this results in 39 packets, take 40, to be on the safe side
		totalSamples = 40 * (int)((float)(duration) / (float)(mediaTimeScale));
		*/
	}

	dprintf_full("ffMP4ContainerFile::loadESInfo Ticks/sec %i Duration: %6.2f secs (ticks %li, samples %u) \n",
							 mediaTimeScale,
							 (float) ((float)(duration) / (float)(mediaTimeScale)),
							 (long)duration, totalSamples );
	int dcsize = 0;
	u8 *dptr = NULL;
	dptr = getDecoderConfiguration(ic, dcsize, streamIndex);
	if(dptr == NULL)
	{	
		dprintf_full("ffMP4ContainerFile : loadESInfo got NULL instead of DecoderConfiguration !! \n");
		dprintf_full("ffMP4ContainerFile : unknown %s codec %i????\n",st->codec->codec_type==CODEC_TYPE_VIDEO?"VIDEO"
								 :st->codec->codec_type==CODEC_TYPE_AUDIO?"AUDIO":"UNKNOWN_STREAM",st->codec->codec_id);
		exit(1);
	}
	u8 *encodedDecoderConfig = new u8[dcsize * 2 + 1];
	SDP::encodeDecoderConfig(encodedDecoderConfig, dptr, dcsize);

	//XXX some of these things are not set before passing
	if(handlerType == MP4VisualHandlerType) {
		dprintf_full("ffMP4ContainerFile::loadESInfo VideoESInfo CodecID %i,  ticks: %u\n",
								 st->codec->codec_id, (unsigned)(duration/totalSamples));
		return new VideoESInfo(streamId, streamType, vob, duration/totalSamples,
													 encodedDecoderConfig,
													 objectType, bufferSize, 
													 avgbit, maxbit, mediaTimeScale, duration, 
													 mediaSize, width, height,
													 true, true,  strm[streamIndex].codec_id, st->quality, aspect_ratio);
	}
	else if(handlerType == MP4AudioHandlerType) {
		u32 channels =  st->codec->channels;
#ifdef BUILD_ARCH_ARM
		if (channels <2 ) {
			dprintf_err("ffMP4ContainerFile::loadESInfo iPAQ hack: simulating stereo audio!\n");
			channels=2; //always set to stereo on ipaq... else --> mickey mouse
		}
#endif
		dprintf_full("ffMP4ContainerFile::loadESInfo AudioESInfo CodecID %i,  %i Hz  Channels: %i, ticks: %u\n",
								 st->codec->codec_id, st->codec->sample_rate, channels, (unsigned)(duration/totalSamples));
		return new AudioESInfo(streamId, streamType, vob, duration/totalSamples,
													 encodedDecoderConfig,
													 objectType, bufferSize,
													 avgbit, maxbit, mediaTimeScale, duration, mediaSize, 
													 st->codec->sample_rate, channels, true,
													 st->codec->codec_id, st->quality, strm[streamIndex].bits_per_sample);
	}
	return NULL;
}


/**********************************************************/
int ffMP4ContainerFile::getStreamParameters_estimate1(const char *fileName)
{
	AVInputFormat *informat = NULL;
	AVFormatContext *temp = NULL;                                
	AVFormatParameters params, *ap = &params;
	AVStream * st;
	structMPEGaudioHeader mp3header; 
	MP3Parser mp3;
	int dcsize = 0;
	u8 *dptr = NULL;
	int ret, totalBitrate;
	
	memset(ap, 0, sizeof(*ap));
	ap->sample_rate = audio_sample_rate;
	ap->channels = audio_channels;
	//	ap->frame_rate = frame_rate;
	//	ap->frame_rate_base = frame_rate_base;
	ap->width = frame_width;
	ap->height = frame_height;
	
	//	if (!strstr(fileName,".")) //no suffix, assume mp3
	//informat = av_find_input_format("mp3");


	
	//open the input file with the generic libavformat function(avformat.h)
	ret = av_open_input_file(&temp, fileName, informat, 0, ap);
	if(ret < 0) {
		//HACK FOR DVB-HTTP STREAM: assume and force mpeg
		informat = av_find_input_format("mpeg");
		dprintf_full("informat is %p\n",informat);

		ret = av_open_input_file(&temp, fileName, informat, 0, ap);
		if(ret < 0) {
			dprintf_full("ffMP4ContainerFile::getStreamParameters_estimate1 Error %i in opening file!!\n",ret);
			return -1;
		}
	}
	
	dprintf_full("ffMP4ContainerFile::getStreamParameters_estimate1 : av_open_input_file is ok\r\n");
	
	//If not enough info to get the stream parameters, we decode the
	//first frames to get it. (used in cases when file has no headers)
	ret = av_find_stream_info(temp);
	if (ret < 0) {
		dprintf_full("ffMP4ContainerFile %s: could not find codec parameters\n", fileName);
		exit(1);
	}

	int noStreams = temp->nb_streams;
	dprintf_small("ffMP4ContainerFile::getStreamParameters_estimate1() streams %d \n",	noStreams);
	int i=0, j;
	totalBitrate = temp->bit_rate;
	for(i = 0; i < noStreams; i++) {
		st = temp->streams[i];
		strm[i].duration = st->duration / st->time_base.den;
		/* HACK (mt): replace duration by value of container
		 * if stream duration is not plausible
		 */
		if (strm[i].duration < 1.0) {
			//		  strm[i].duration = temp->duration / 1000000.0;
			strm[i].duration = 99999999;
		  dprintf_err("ffMP4ContainerFile::getStreamParameters_estimate1() "
				"stream %d contains suspicious duration %6.2f, corrected to %6.2f\n",
				i, st->duration/1000000.0, strm[i].duration);
		}

		strm[i].codec_id = st->codec->codec_id;

		if(st->codec->codec_type == CODEC_TYPE_AUDIO) {
			dptr = getDecoderConfiguration(temp, dcsize, i);
			if(dptr == NULL) {
				dprintf_full("ffMP4ContainerFile : getStreamParameters() got NULL instead of DecoderConfiguration !! \n");
				dprintf_full("ffMP4ContainerFile : (the file is not an mp3, maybe raw PCM ?? ) \n");
				//exit(1);
			}
			
			strm[i].rate = (float)(st->codec->sample_rate);
			strm[i].channels = st->codec->channels;

			switch (st->codec->codec_id) {
			case CODEC_ID_MP3:	
			case CODEC_ID_AAC:
			case CODEC_ID_MPEG4AAC:
				if(dptr && dptr[0] == 0xFF && dptr[1]>>5 == 0x07)
					mp3header = mp3.parseMPEGaudioHeader(dptr, dcsize);
				else
					dprintf_err("ffMP4ContainerFile::getStreamParameters_estimate1: Valid mp3 header not found\n");
				strm[i].avgBitRate = mp3header.bitrate*1000;
				strm[i].bits_per_sample = mp3header.bits_per_sample;
				strm[i].totalSamples = (int) ceil(strm[i].duration /
																					((mp3header.frame_duration / 90000.0) *
																					 (strm[i].packet_size/mp3header.frame_size_in_bytes)));
				break;
			case CODEC_ID_MP2:
				if(dptr && dptr[0] == 0xFF && dptr[1]>>5 == 0x07)
					mp3header = mp3.parseMPEGaudioHeader(dptr, dcsize);
				else
					dprintf_err("ffMP4ContainerFile::getStreamParameters_estimate1: Valid mp2 header not found\n");
				strm[i].avgBitRate = mp3header.bitrate*1000;
				strm[i].bits_per_sample = mp3header.bits_per_sample;
				strm[i].totalSamples = (int) ceil(strm[i].duration /
																					((mp3header.frame_duration / 90000.0) *
																					 (strm[i].packet_size/mp3header.frame_size_in_bytes)));
				break;
			case CODEC_ID_PCM_S8:
			case CODEC_ID_PCM_U8:
			case CODEC_ID_PCM_S16LE:
			case CODEC_ID_PCM_S16BE:
			case CODEC_ID_PCM_U16LE:
			case CODEC_ID_PCM_U16BE:

				if (st->codec->codec_id == CODEC_ID_PCM_S8 || st->codec->codec_id == CODEC_ID_PCM_U8) {
					strm[i].bits_per_sample = 8;
					dprintf_err("ffMP4ContainerFile::getStreamParameters_estimate1: PCM Audio with 8bits not yet supported, eg. in DSPaudioIO\n");
					exit(1);
				} else
					strm[i].bits_per_sample = 16;
				//for pcm, get duration from AVFormatContext, not stream...
				strm[i].duration = temp->duration/1000000.0;
				strm[i].avgBitRate = (int)ceil(strm[i].rate) * strm[i].channels * strm[i].bits_per_sample;
				//				strm[i].totalSamples = (int)ceil(strm[i].duration); // store 1 second of raw pcm in one frame...

				//FIXME: ffmpeg reads 256 bytes for one PCM-raw frame with av_get_frame()....
				strm[i].totalSamples = strm[i].avgBitRate/8/256 * (int)ceil(strm[i].duration);
				break;
			default:
				dprintf_err("ffMP4ContainerFile: unknown audio stream codec ID %i!\n",st->codec->codec_id);
				exit(1);
			}
		}
	}//run thru all streams
	
	for(i = 0; i < noStreams; i++) {
		st = temp->streams[i];
		if(st->codec->codec_type == CODEC_TYPE_VIDEO) {
		  if (st->r_frame_rate.den > 0) 
				strm[i].rate = (float)(st->r_frame_rate.num/st->r_frame_rate.den);
			else //fails on .avi  :(
				strm[i].rate = 25;

			strm[i].totalSamples = (int)strm[i].duration * (int)strm[i].rate;
			strm[i].avgBitRate = totalBitrate;
			for (j = 0; j < noStreams; j++) {
				if(j != i)
					strm[i].avgBitRate -= strm[j].avgBitRate;
			}
		}
		strm[i].size = (int)(strm[i].avgBitRate*strm[i].duration/8.0);
		if (st->codec->codec_type == CODEC_TYPE_VIDEO) {
			dprintf_full("ffMP4ContainerFile::getStreamParameters_estimate1 VIDEO stream %i: codecID: %i fps %2.1f "
									 "duration %6.2f  size %li numFrms: %i avg BW is %ld bit/sec\n",
									 i, strm[i].codec_id, strm[i].rate, strm[i].duration, strm[i].size,strm[i].totalSamples,
									 strm[i].avgBitRate);
		} else if (st->codec->codec_type == CODEC_TYPE_AUDIO) {
			dprintf_full("ffMP4ContainerFile::getStreamParameters_estimate1 AUDIO stream %i: codecID: %i "
									 "rate %2.1f bits: %i "
									 "channels: %i duration %6.2f  size %li numFrms: %i avg BW is %ld bit/sec\n",
									 i, strm[i].codec_id, strm[i].rate, strm[i].bits_per_sample, strm[i].channels, 
									 strm[i].duration, strm[i].size,
									 strm[i].totalSamples, strm[i].avgBitRate);
		} else {
			dprintf_err("ffMP4ContainerFile::getStreamParameters_estimate1 stream %i is neither video nor audio (is: %i)!!!\n",i, st->codec->codec_type);
			exit(1);
		}
	}
	
	av_close_input_file(temp);
	
	return noStreams;
}



/**********************************************************/
int ffMP4ContainerFile::getStreamParameters_estimate2(const char *fileName)
{
	AVInputFormat *informat = NULL;
	AVFormatContext *temp;                                
	AVStream *st;
	AVFormatParameters params, *ap = &params;
	
	structMPEGaudioHeader mp3header; MP3Parser mp3;
	int dcsize = 0;
	u8 *dptr = NULL;
			
	memset(ap, 0, sizeof(*ap));
	ap->sample_rate = audio_sample_rate;
	ap->channels = audio_channels;
	//ap->frame_rate = frame_rate;
	//ap->frame_rate_base = frame_rate_base;
	ap->width = frame_width;
	ap->height = frame_height;

	int ret;
	
    	//open the input file with the generic libavformat function(avformat.h)
	ret = av_open_input_file(&temp, fileName, informat, 0, ap);
	if(ret < 0)
		{
			dprintf_full("ffMP4ContainerFile::getStreamParameters_estimate2 Error in opening file!!\n");
			return -1;
		}
	
	
	dprintf_full("ffMP4ContainerFile::getStreamParameters : av_open_input_file is ok\r\n");
	
	//If not enough info to get the stream parameters, we decode the
	//first frames to get it. (used in cases when file has no headers)
	ret = av_find_stream_info(temp);
	if (ret < 0) {
		dprintf_full("ffMP4ContainerFile %s: could not find codec parameters\n", fileName);
		exit(1);
	}
	
	AVPacket pkt1, *pkt = &pkt1;
	int stIndex=0, noStreams = temp->nb_streams, i=0, done = 0;
        u32 data = 0;
	int startOfHeader = 0;
	bool first = true;
	
	for(i = 0; i < noStreams; i++) {
                st = temp->streams[i];
		strm[i].size = 0; strm[i].totalSamples = 0; strm[i].avgBitRate = 0; 
		strm[i].done = false;
		strm[i].test_duration = MAX_TEST_DURATION;
		//if(st->codec->codec_type == CODEC_TYPE_VIDEO)
		//	strm[i].rate = st->r_frame_rate/st->r_frame_rate_base; /*@/FRAME_RATE_BASE;*/
			
		strm[i].duration = st->duration/1000000.0; //strm[i].totalSamples/strm[i].rate; //The divide by 1000000 is done to convert the 
													    //duration from TIME_BASE 1000000 to secs	
		if(strm[i].duration < strm[i].test_duration)
			strm[i].test_duration = strm[i].duration;
        }
	//here we go ..... read packets(or frames) and perform calculations
	for(;;)	{
		
		ret = av_read_frame(temp, pkt);
		if(ret < 0) {
				dprintf_full("ffMP4ContainerFile : Error in retrieving packet or EOF reached\n");
				break;
		}
		
		if (!pkt->size) {
			dprintf_full("ffMP4ContainerFile : Seems the packet is empty!!\n");
		}
	
		stIndex = pkt->stream_index;
			
		if(temp->streams[stIndex]->codec->codec_type == CODEC_TYPE_AUDIO && first)
                {
                        data = 0;
			startOfHeader = 0;
                        while((data>>21) != 0x000007FF && startOfHeader <= pkt->size)    //this is the start code for the decConf info
                        {
                                data <<= 8;
                                data += pkt->data[startOfHeader];
                                startOfHeader++;
                        }
                        if((data>>21) == 0x000007FF)
                        {
                                dcsize = 4; 
				startOfHeader -= 4;
                                dptr = pkt->data + startOfHeader;
                        }
                        mp3header = mp3.parseMPEGaudioHeader(dptr, dcsize);
			strm[stIndex].bits_per_sample = mp3header.bits_per_sample;
			strm[stIndex].rate = (90000/((float)mp3header.frame_duration*(float)(pkt->size/mp3header.frame_size_in_bytes)));
			if(strm[stIndex].rate <= 0.0)
				strm[stIndex].rate = 1.0;
			if(mp3header.frame_size_in_bytes/pkt->size < 0.9)
				strm[stIndex].packetDuration = mp3header.frame_duration*(pkt->size/mp3header.frame_size_in_bytes);
			else
				strm[stIndex].packetDuration = mp3header.frame_duration;
			
			first = false;
                }
		if(strm[stIndex].totalSamples < (int)(strm[stIndex].rate*strm[stIndex].test_duration)) {
			strm[stIndex].size += pkt->size;
			strm[stIndex].totalSamples++;
		}
		else
			strm[stIndex].done = true;
		av_free_packet(pkt);
		done = 0;
		for(i  = 0; i< noStreams; i++) {
			if(strm[i].done)
				done++;
		}
		if(done == noStreams)
			break;
		
	}
	
	for(i = 0; i < noStreams; i++)
	{
		strm[i].avgBitRate = (long int)((strm[i].size/strm[i].test_duration) * 8.0);
		strm[i].size = (int)((strm[i].avgBitRate*strm[i].duration)/8.0);
		strm[i].totalSamples = (int)((strm[i].totalSamples/strm[i].test_duration)*strm[i].duration);
		dprintf_full("ffMP4ContainerFile::getStreamParameters_estimate2 stream %i: duration %6.2f  size %li numFrms: %i avg BW is %ld bit/sec\n",i,strm[i].duration,strm[i].size,strm[i].totalSamples, strm[i].avgBitRate);
	}

	av_close_input_file(temp);
	
	return noStreams;
}


/**********************************************************/
int ffMP4ContainerFile::getStreamParameters_run_through(const char *fileName)
{
	AVInputFormat *informat = NULL;
	AVFormatContext *temp;                                
	AVFormatParameters params, *ap = &params;
	
	memset(ap, 0, sizeof(*ap));
	ap->sample_rate = audio_sample_rate;
	ap->channels = audio_channels;
  //ap->frame_rate = frame_rate;
	//ap->frame_rate_base = frame_rate_base;
	ap->width = frame_width;
	ap->height = frame_height;

	int ret;
	
    	//open the input file with the generic libavformat function(avformat.h)
	ret = av_open_input_file(&temp, fileName, informat, 0, ap);
	if(ret < 0)
		{
			dprintf_full("ffMP4ContainerFile::getStreamParameters_run_through Error in opening file!!\n");
			return -1;
		}
	
	
	dprintf_full("ffMP4ContainerFile::getStreamParameters_run_through : av_open_input_file is ok\r\n");
	
	//If not enough info to get the stream parameters, we decode the
	//first frames to get it. (used in cases when file has no headers)
	ret = av_find_stream_info(temp);
	if (ret < 0) {
		dprintf_full("ffMP4ContainerFile::getStreamParameters_run_through %s: could not find codec parameters\n", fileName);
		exit(1);
	}
	
	AVPacket pkt1, *pkt = &pkt1;
	int stIndex=0, noStreams = temp->nb_streams;
	int i=0;

	//initialize
	for(i = 0; i< noStreams; i++)	{
		strm[i].size = 0; strm[i].totalSamples = 0; strm[i].avgBitRate = 0; 
		
		//FIXME : so far rate is correct for video only; arbit value set for audio
		//This part requires fixing when upgraded to newer version of ffmpeg.
		//For video: rate = frame rate         audio: rate = sample rate.

		/*
		if(temp->streams[i]->codec->codec_type == CODEC_TYPE_VIDEO)
			strm[i].rate = temp->streams[i]->r_frame_rate/temp->streams[i]->r_frame_rate_base;
		else if(temp->streams[i]->codec->codec_type == CODEC_TYPE_AUDIO)
			strm[i].rate = 50.0; 				
		*/
	}
	
	//here we go ..... read packets(or frames) and perform calculations
	for(;;)	{
		
		ret = av_read_frame(temp, pkt);
		if(ret < 0) {
				dprintf_full("ffMP4ContainerFile : Error in retrieving packet or EOF reached\n");
				break;
		}
		
		if (!pkt->size) {
			dprintf_full("ffMP4ContainerFile : Seems the packet is empty!!\n");
		}
	
		stIndex = pkt->stream_index;
		strm[stIndex].size += pkt->size;
		strm[stIndex].totalSamples++;
		
		av_free_packet(pkt);
	}
	
	for(i = 0; i < noStreams; i++)
	{
		strm[i].duration = temp->streams[i]->duration/1000000; //strm[i].totalSamples/strm[i].rate; //The divide by 1000000 is done to convert the 
													    //duration from TIME_BASE 1000000 to secs	
		/* Duration of raw mpeg files is not obtained from ffmpeg
		 * .... so calculate it
		 */
		if(strstr(fileName, ".m4v") || strstr(fileName, ".M4V"))
			strm[i].duration = strm[i].totalSamples/strm[i].rate;
		
		strm[i].avgBitRate = (long int)((strm[i].size/strm[i].duration) * 8.0);
		dprintf_full("ffMP4ContainerFile::getStreamParameters_run_through stream %i: duration %6.2f  size %li numFrms: %i avg BW is %ld bit/sec\n",i,strm[i].duration,strm[i].size,strm[i].totalSamples, strm[i].avgBitRate);
	}

	av_close_input_file(temp);
	
	return noStreams;
}


/**********************************************************/
u8 *ffMP4ContainerFile::getDecoderConfiguration(AVFormatContext *ic, int& pktSize, int streamIndex)
{
	AVPacket pkt;
	bool parseVideo = false;
	if(ic->streams[streamIndex]->codec->codec_type == CODEC_TYPE_VIDEO) {
		switch(ic->streams[streamIndex]->codec->codec_id) {
		case CODEC_ID_MPEG4 :
		case CODEC_ID_MSMPEG4V1 :
		case CODEC_ID_MSMPEG4V2 :
		case CODEC_ID_MSMPEG4V3 :
			pktSize = ic->streams[streamIndex]->codec->extradata_size;
			if(pktSize != 0 && (u8*)ic->streams[streamIndex]->codec->extradata != NULL)
				return (u8*)ic->streams[streamIndex]->codec->extradata;
			else {
				parseVideo = true; 
				break;
			}
			break;	
		case CODEC_ID_MPEG1VIDEO :
		case CODEC_ID_MPEG2VIDEO :
			pktSize = 20;
			return ((u8*)"dummyDecoderConfiguration"); //XXX Dummy header sent for MPEG1/2 
		default :
			return NULL;
		}
	}	
	

	for(;;)	{				//XXX Remove the infinite loop
		if (av_read_frame(ic, &pkt) < 0) {
			dprintf_full("ffMP4ContainerFile::getDecoderConfiguration - Error in retrieving packet or decConfig not found\n");
			return NULL;
		}
		
		if (!pkt.size) {
			dprintf_full("ffMP4ContainerFile : Seems the packet is empty!!\n");
		}

		/* NOT NEEDED AS INFO HAS BEEN FOUND ABOVE FOR MP4 VIDEO 
		 * BUT if you dont get it above than parse Video to get it
		 */
		if(parseVideo && ic->streams[pkt.stream_index]->codec->codec_type == CODEC_TYPE_VIDEO 
			 && ic->streams[streamIndex]->codec->codec_type == CODEC_TYPE_VIDEO) {	
				pktSize = 0;
				int data = 0;
				while((data>>8) != 0x00000001 && pktSize <= pkt.size) {   //this is the start code for the decConf info
					data <<= 8;
					data += pkt.data[pktSize];
					pktSize++;
				}
				if((data>>8) == 0x00000001) {
					pktSize = 0;
					data = 0; 
					while(data != 0x000001B6 && pktSize <= pkt.size) {  //this is the start code for the VOP header
						//we need only the decoder configuration info
						data <<= 8;
						data += pkt.data[pktSize];
						pktSize++;
					}
					pktSize--;				//since last 32 bits are startcode for the VOP header
					return pkt.data;
				}
		}
		
		//The code below id for decoder configuration for MP3 files ONLY 
		if(ic->streams[pkt.stream_index]->codec->codec_type == CODEC_TYPE_AUDIO 
			 && ic->streams[streamIndex]->codec->codec_type == CODEC_TYPE_AUDIO) {
			pktSize = 0;
			u32 data = 0;
			int startOfHeader = 0;
			while((data>>21) != 0x000007FF && startOfHeader <= pkt.size) {  //this is the start code for the decConf info
				data <<= 8;
				data += pkt.data[startOfHeader];
				startOfHeader++;
			}
			if((data>>21) == 0x000007FF) {
				pktSize = 4; 
				
				startOfHeader -= 4;
				strm[streamIndex].packet_size = pkt.size;
				return (pkt.data + startOfHeader);
			}
		}
	}
	
	return NULL;
}
