/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	FilteredIO.cpp                                                *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#include "FilteredIO.hpp"
#include "BitField.hpp"
#include "Frame.hpp"
#include "ESInfo.hpp"

FilteredIO::FilteredIO(IO* _io, BitField* _filter) 
{
	assert(_io && _filter);
	io=_io;
	filter=_filter;
	assert(filter->isAllSet());
	framesSeen=0;
};


FilteredIO::~FilteredIO()
{
	if(io)
		delete io;

	if(filter)
		delete filter;
};

	/** returns a frame if one complete frame is available, otherwise null 
		is returned. This function is typically blocking. Don't use a NULL
		return value to conclude STREAMEOF, always check with getState()!
	*/
Frame * FilteredIO::getFrame() {
	Frame* frm = NULL;
	bool  stop = false;
	u32 vopTimeIncrement = io->getESInfo()->getVOPTimeIncrement();
	do {
		frm=io->getFrame();
		if(frm) {
			if(filter->getBit(frm->getAU()->cts/vopTimeIncrement) || (frm->getAU()->cts==0))
				stop=true;
			else {
				dprintf_full("FilteredIO::getFrame: throwing away CTS %i VopIncr %i\n",frm->getAU()->cts,vopTimeIncrement);
				delete frm;
				frm=NULL;
			}
		}
	} while(!stop && io->getState()!=IO::STREAMEOF && io->getState()!=IO::CLOSED &&
				io->getState()!=IO::CLOSING && io->getState()!=IO::STREAMERR);
	framesSeen++;

#ifdef _POSIX_PRIORITY_SCHEDULING
	sched_yield();
#else
	// for windows pthread lib!
#ifdef WIN32
#ifndef WINCE
	sched_yield();
#else
	Sleep(0);
#endif 
#endif
#endif
	dprintf_full("FilteredIO::getFrame() CTS %u\n",frm?frm->getAU()->cts:0);
	return frm;
};

/** returns the number of packets sent. Returns 0 on error */
int FilteredIO::writeFrame(Frame * frm, ESInfo *out_es) 
{
	assert(frm);
	framesSeen++;
	if(filter->getBit(frm->getAU()->cts/io->getESInfo()->getVOPTimeIncrement())) {
		return io->writeFrame(frm);
	}

#ifdef _POSIX_PRIORITY_SCHEDULING
	sched_yield(); //this is necessary to give parallel getFrames a chance
#endif

	return 1;
};

/** opens the IO connection. State is set to OPENING. Depending on the
		underlying QIODevice, a network connection or a file connection
		is established. When the connection is ready for use, State is OPEN */
bool FilteredIO::open() 
{
	framesSeen=0;
	return io->open();
};

/** closes the IO class. @param immediate specifies, if (optional)
	* buffered data should be read/sent to the client (==false), or immediately
	* dumped (==true)
	*/
bool FilteredIO::close(bool immediate)
{
	return io->close(immediate);
};

long FilteredIO::setToClosestIFrame(u32 frameNumber, bool backwards)
{
	dprintf_full("FilteredIO::setToClosestIFrame\n");
	return io->setToClosestIFrame(frameNumber, backwards);
};

/** deletes the data source for this IO class. Will only work with
 * local files, remote files can not be deleted and will always return true.
 */
bool FilteredIO::destroy()
{
	return io->destroy();
};

IO::State FilteredIO::getState() const
{
	return io->getState();
};

/** returns the newly set state */
IO::State FilteredIO::setState(IO::State s)
{
	return io->setState(s);
};

/** returns a value  from 0..100 indicating buffer usage */
int FilteredIO::getBufferFillLevel() const
{
	return io->getBufferFillLevel();
};

/** flushes the buffer from starting TimeStamp to end TS
 * @param from_ts: 0 is beginning, everything else is somewhere in the stream
 * @param to_ts: -1 is ending, everything else is somewhere in the stream
 * @returns the number of freed bytes
 */
int FilteredIO::flushBuffer(long from_ts, long to_ts)
{
	return io->flushBuffer(from_ts,to_ts);
};

/** returns a pointer to the local file name or an URL */
const char *FilteredIO::getURL() const
{
	return io->getURL();
};

bool FilteredIO::setToFrameNumber(u32 frameNumber)
{
	dprintf_full("FilteredIO::setToFrameNumber\n");
	return io->setToFrameNumber(frameNumber);
};

void FilteredIO::setResendFrameHeader(bool set)
{
	io->setResendFrameHeader(set);
};

void FilteredIO::run()
{
	if(!io->running())
		io->start();
	io->wait();
};

void FilteredIO::setESInfo(ESInfo * es)
{
	io->setESInfo(es);
};

void FilteredIO::setRtxInfo(const rtx_info* rtx)
{
	io->setRtxInfo(rtx);
};

IO::State FilteredIO::play(double prefetchTime)
{
	return io->play(prefetchTime);
};

IO::State FilteredIO::pause()
{
	return io->pause();
};
IO::State FilteredIO::mute()
{
	return io->mute();
};

/** deletes the old BitField and sets a new one, 
	* returns false if the old and new BitField differ in size
	*/
bool FilteredIO::setBitField(BitField* b)
{
	bool ret=false;
	if(b && b->getSize()>0) {
#ifdef DEBUG
		char* tmp=b->toHexString();
		dprintf_full("FilteredIO::setBitField %s\n",tmp);
		delete tmp;
#endif
		ret=(filter->getSize()==b->getSize());
		if(filter)
			delete filter;
		filter=b;
	}
	return ret;
}

ESInfo* FilteredIO::getESInfo() {
	return io->getESInfo();
}

