 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	FilteredIO.hpp                                                *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef PS_MPEG4ITEC_FILTEREDIO_HPP
#define PS_MPEG4ITEC_FILTEREDIO_HPP
#include "IO.hpp"
class BitField;
class Frame;
class ESInfo;

/**
 *
 * <Class Description goes here>
 *
 *@author Michael Kropfberger and Peter Schojer
 *
 *@version $Id: FilteredIO.hpp,v 1.14 2006/01/26 11:42:00 mkropfbe Exp $
 *
 *@short <short description>
 *
 */
class FilteredIO : public IO
{
public:
	/** Note that the get-/writeFrame() functions of this class will only return the 
	  * frames from the underlying io class
	  * where the corresponding entry in BitField is 1!	
	  */
	FilteredIO(IO* io, BitField* filter);

	virtual ~FilteredIO();

	/** returns a frame if one complete frame is available, otherwise null 
		is returned. This function is typically blocking. Don't use a NULL
		return value to conclude STREAMEOF, always check with getState()!
	*/
	Frame * getFrame();

	/** returns the number of packets sent. Returns 0 on error */
	int writeFrame(Frame * frm, ESInfo *out_es = NULL);

	/** opens the IO connection. State is set to OPENING. Depending on the
			underlying QIODevice, a network connection or a file connection
			is established. When the connection is ready for use, State is OPEN */
	bool open();

	/** closes the IO class. @param immediate specifies, if (optional)
	  * buffered data should be read/sent to the client (==false), or immediately
	  * dumped (==true)
	  */
	bool close(bool immediate=false);

	/** deletes the data source for this IO class. Will only work with
	 * local files, remote files can not be deleted and will always return true.
	 */
	bool destroy();

	State getState() const;

	/** returns the newly set state */
	State setState(State s);

	/** returns a value  from 0..100 indicating buffer usage */
	int getBufferFillLevel() const;

	/** flushes the buffer from starting TimeStamp to end TS
	 * @param from_ts: 0 is beginning, everything else is somewhere in the stream
	 * @param to_ts: -1 is ending, everything else is somewhere in the stream
	 * @returns the number of freed bytes
	 */
	int flushBuffer(long from_ts = 0, long to_ts = -1);

	/** returns a pointer to the local file name or an URL */
	const char *getURL() const;

	/** repositions the IO class to the given frame.
	 * Will return false in the following cases:
	 *  - an illegal frame number (too large) was specified
	 *  - the stream is not seekable, because the underlying 
	 *    device is a network device and the requested frame 
	 *    is already out of range: FIXME: implement
	 */
	bool setToFrameNumber(u32 frameNumber);

  /** Returns the current frame number of the associated IO object. */
  virtual u32 getCurrentFrameNumber() const {
    return io ? io->getCurrentFrameNumber() : 0;
  }

  /** Sets the end frame number of the associated IO object. 
   * @return <code>true</code> on success.
   * @see IO::setEndFrameNumber(u32)
   */
  virtual bool setEndFrameNumber(u32 stopNumber) {
    return io ? io->setEndFrameNumber(stopNumber) : false;
  }

  /** Returns the end frame number of the associated IO object. */
  virtual u32 getEndFrameNumber() const {
    return io ? io->getEndFrameNumber() : 0;
  }

	/** indicates if the next call to getFrame will send the
	 * frame header. */
	void setResendFrameHeader(bool set);

	void run();

	void setESInfo(ESInfo * es);

	ESInfo* getESInfo();

	/** repositions the IO class to @param frameNumber, if this
	  * Frame is an I-frame. Otherwise a frameNumber lower than
		* frameNumber is searched and returned.
		* @param return contains the chosen frameNumber.
		* Will return -1 in case of an error, will only
		* work with IO classes opened for reading.
		*/
	long setToClosestIFrame(u32 frameNumber, bool backwards);
	
	/** deletes the old BitField and sets a new one, 
		* returns false if the old and new BitField differ in size
		*/
	bool setBitField(BitField* b);

	/** returns the number of frames received or sent. 
	 * counter will be resetted when open is called */
	virtual u32 getFramesSeen() const { return framesSeen;};

	/** sets a shallow copy of @param rtx at the ESInfo object */
	void setRtxInfo(const rtx_info* rtx); 

	State play(double prefetchTime=0.0);

	State pause();

	State mute();	

protected:
	IO* io;
	BitField* filter;
};
#endif

