 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title: IOCreator.cpp                                                 *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/
#include "IOCreator.hpp"

#include "../ContainerInfo.hpp"
#include "../BitField.hpp"
#include "../ReferenceCounter.hpp"
#include "../ESInfo.hpp"
#include "../IO.hpp"

#include "ffMP4IO.hpp"
#include "MPGStreamIO.hpp"
#include "YUVStreamIO.hpp"
#include "FilteredIO.hpp"
#include "VITMuxDemuxIO.hpp"
#include "BufferedIO.hpp"

#ifdef ENABLE_MKV
#include "MKIO.hpp"
#endif

#ifdef ENABLE_DVDREAD
#include "DVDIO.hpp"
#endif

IO *IOCreator::createInput(ESInfo* es,  bool wrapInFilteredIO) {
	// different cases:
	// vo set & vo.url is set and  this.input is set
	//       url==input -> create an MP4IO
	//       not equal and this.input exists -> create MPGStreamIO
	
	char* tmp=NULL;
	const char* input;
	ContainerInfo* vo=es->getContainerInfo();
	
	//extract only file part of URL or PATH
	input=strrchr(es->getInput(),'/');

	if (!input || strlen(input)==0) {
		dprintf_full("IOCreator::createInput(): bad input %s, falling back to localFile\n",input);
		input =  vo->getLocalFile();
	}

	if(input) {

			tmp=strrchr(input,'.');
			dprintf_full("IOCreator::createInput(): input %s localFile %s checking on %s codec id %i\n", es->getInput(), vo->getLocalFile(),input, es->getCodecID() );
			
			//check for extensions first
			if (tmp && (strstr(tmp,".yuv") || strstr(tmp,".YUV")) ) {
				dprintf_full("IOCreator::createInput(): YUVStreamIO created\r\n");
				YUVStreamIO* yio=new YUVStreamIO((VideoESInfo*)es,vo->getLocalFile(),false);
				es->setInput(vo->getLocalFile());
				return yio;
			}
			
#ifdef ENABLE_MKV 
			if (tmp && (strstr(tmp,".mkv") || strstr(tmp,".MKV")) ) {
				dprintf_full("IOCreator::createInput(): MKIO created\r\n");
				MKIO* mkio=new MKIO((VideoESInfo*)es,vo->getLocalFile(),false);
				es->setInput(vo->getLocalFile());
				return mkio;
			}
#endif
			
			if (tmp && (strstr(tmp,".vit") || strstr(tmp,".VIT")) ) {
				dprintf_full("IOCreator::createInput(): VIT BufferedIO created\r\n");
				BufferedIO* bufio=new BufferedIO((VideoESInfo*)es, (MuxDemuxIO*)vo->getMuxDemuxIO(), false);
				es->setInput(vo->getLocalFile());
				return bufio;
			}
			
			if (tmp && (strstr(tmp,".ogg") || strstr(tmp,".OGG")) ) {
				dprintf_full("IOCreator::createInput(): OGG BufferedIO created\r\n");
				BufferedIO* bufio=new BufferedIO((VideoESInfo*)es, (MuxDemuxIO*)vo->getMuxDemuxIO(), false);
				es->setInput(vo->getLocalFile());
				return bufio;
			}
			
#ifdef ENABLE_DVDREAD
			if(tmp && (strstr(tmp,"vob") || strstr(tmp,"VOB") || strstr(tmp,"iso") || strstr(tmp,"ISO"))) {
				DVDIO* dio=new DVDIO(es,input);
				if (wrapInFilteredIO) {
					BitField* b=new BitField(es->getNumberOfMediaSamples(),true);
					FilteredIO* fio=new FilteredIO(dio,b);
					return fio;
				} else
					return dio;
			}
#endif		


		/* no fitting extension, so ffmpeg shall try it's best...
		 * else verify for supported container formats  */
		if ( (es->getCodecID() == CODEC_ID_MP3) ||  (es->getCodecID() == CODEC_ID_AAC) || // pass audio to MP4IO
				 (!tmp || IOCreator::verifyContainerExtension(tmp)) ) {
#ifdef ENABLE_FFMPEG
			ffMP4IO *mio = new ffMP4IO(es,vo->getLocalFile(),false);
			dprintf_full("IOCreator::createInput(): no extension detected, fallback: ffMP4IO created\r\n");
			if (wrapInFilteredIO) {
				BitField* b=new BitField(es->getNumberOfMediaSamples(),true);
				FilteredIO* fio=new FilteredIO( mio ,b);
				return fio;
			} else
				return mio;
#else
			dprintf_full("IOCreator::createInput(): ffMP4IO disabled in this build!\r\n");
			exit(1);
#endif
		}
		
		if ( (es->getCodecID() == CODEC_ID_MPEG4AAC) || 
				 //the above is specially mp4-packaged AAC/mp3 (as stored in ./demux)
				 (es->getCodecID() == CODEC_ID_MPEG4) ||
				 (tmp && strstr(tmp,".m4v") || strstr(tmp,".M4V")) // raw mpeg4
				 ) { // in demux directory
			dprintf_full("IOCreator::createInput(): raw MPGStreamIO created\r\n");
			MPGStreamIO *sio = new MPGStreamIO(es, input, false);
			if (wrapInFilteredIO) {
				BitField* b=new BitField(es->getNumberOfMediaSamples(),true);
				FilteredIO* fio = new FilteredIO(sio, b);
				return fio;
			} else
				return sio;
		}
		
			// all other combinations are illegal 
			dprintf_err("IOCreator::createInput(): Unknown filetype (input %s  localFile %s)\r\n", input, vo->getLocalFile());
			return NULL;
		
	}

	dprintf_err("IOCreator::createInput(): fatal file detection (input %s  localFile %s) \r\n", input, vo->getLocalFile());
	return NULL;
}


bool IOCreator::verifyContainerExtension(const char *file) {
	assert(file);
	if (strstr(file,".mp4") || strstr(file,".MP4") 
			|| strstr(file,".mpg") || strstr(file,".MPG") 
			|| strstr(file,".mpeg") || strstr(file,".MPEG")
			|| strstr(file,".vit") || strstr(file,".VIT") //ViTooKis own container format
			|| strstr(file,".mkv") || strstr(file,".MKV") 
			|| strstr(file,".mp3") || strstr(file,".MP3")   // also handle mp3 audio within MP4IO
			|| strstr(file,".avi") || strstr(file,".AVI") )
		return true;
	else
		return false;
}
