 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	IncompleteIO.hpp                                              *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef _PS_MEG4ITEC_INCOMPLETEIO_HPP
#define _PS_MEG4ITEC_INCOMPLETEIO_HPP
#include "IO.hpp"
#include "global.hpp"


class MPGStreamIO;
class SimpleRtp;
class Session;
class ESInfo;
class PacketizationLayer;
class TerminalCapabilities;
class GlobalTimer;
class Statistics;
class Frame;

#include "PointerRingBuffer.hpp"

/**
 *
 * <Class Description goes here>
 *
 *@author Michael Kropfberger and Peter Schojer
 *
 *@version $Id: IncompleteIO.hpp,v 1.12 2006/01/20 15:37:18 mkropfbe Exp $
 *
 *@short <short description>
 *
 */
class IncompleteIO: public IO
{
	/** how many frames can we buffer */
#define INCOMPLETEIO__MAX_FRAMES 64
#define INCOMPLETEIO__MAX_FRAME_LOOK_AHEAD 10
public:
/** creates an RTP object
			@param localFile: the partially cached data for this stream
      @param url: the source/dest for this IO class, deep-copied
      @param remotePrt: the remote port where data is sent to
      @param address: the ip address of the remote host, deep-copied
			@param localPrt: the local port, where we are binding and receiving
      @param es: the ElementaryStream sent, shallow-copied
      @param packetization: specifies the transport mode: can be RFC3016,
                            multiSL or flexmux
      @param writeonly specifies if the RTP connection is for sending
      (writeonly is set to true), of for reading
			@param tc: the TerminalCapabilities, shallow-copied (eg. RTX on/off)
  */
  IncompleteIO(const char* localFile, const char *url, int remotePrt, 
			const char *address, int localPrt, 
			ESInfo * es, PacketizationLayer * packetization, bool writeOnly,
			GlobalTimer * globalTimer, Statistics *stats);
  virtual ~IncompleteIO();
	/** returns a frame if one complete frame is available, otherwise null 
		is returned. This function is typically blocking. Don't use a NULL
		return value to conclude STREAMEOF, always check with getState()!
	*/
	Frame * getFrame();
	/** returns the number of packets sent. Returns 0 on error */
	int writeFrame(Frame * frm, ESInfo *out_es = NULL);
	/** opens the IO connection. State is set to OPENING. Depending on the
			underlying QIODevice, a network connection or a file connection
			is established. When the connection is ready for use, State is OPEN */
	bool open();
	/** closes the IO class. @param immediate specifies, if (optional)
	  * buffered data should be read/sent to the client (==false), or immediately
	  * dumped (==true)
	  */
	bool close(bool immediate=false);
	/** deletes the data source for this IO class. Will only work with
	 * local files, remote files can not be deleted and will always return true.
	 */
	bool destroy();
	State getState() const;
	/** returns the newly set state */
	State setState(State s);
	/** returns a value  from 0..100 indicating buffer usage */
	int getBufferFillLevel() const;
	/** flushes the buffer from starting TimeStamp to end TS
	 * @param from_ts: 0 is beginning, everything else is somewhere in the stream
	 * @param to_ts: -1 is ending, everything else is somewhere in the stream
	 * @returns the number of freed bytes
	 */
	int flushBuffer(long from_ts = 0, long to_ts = -1);
	/** returns a pointer to the local file name or an URL */
	const char *getURL() const;
	/** repositions the IO class to the given frame.
	 * Will return false in the following cases:
	 *  - an illegal frame number (too large) was specified
	 *  - the stream is not seekable, because the underlying 
	 *    device is a network device and the requested frame 
	 *    is already out of range: FIXME: implement
	 */
	bool setToFrameNumber(u32 frameNumber);
	
	long setToClosestIFrame(u32 frameNumber, bool backwards);

	/** indicates if the next call to getFrame will send the
	 * frame header. */
	void setResendFrameHeader(bool set);
	void run();

	/** sets a shallow copy of @param rtx at the ESInfo object */
	void setRtxInfo(const rtx_info* rtx); 
	State play(double prefetchTime=0.0);
	State pause();
	State mute();	
	void setESInfo(ESInfo * es);
	const ESInfo* const getESInfo() const;
protected:
	void flushLocalCache();
	/** returns a pointer and removes the frame from the cache */
	Frame* findAndUnlinkFrame(u32 cts);
	void setFrame(Frame* frm);
protected:
	MPGStreamIO* localInput;
	SimpleRtp* remoteInput;
	/** a backup from the voptimeincrement in the es. needed because rtp overwrites the value with a wrong detected increment */
	u32 vopTimeIncrement;
	/* maps cts to frames, a INCOMPLETEIO__SETS-associative frame cache with pseudo LRU, we delete the one with the oldest CTS*/
	PointerRingBuffer<Frame*> frameCache;	
	/** cacheUsage: how many frames are currently stored in the cache */
	int cacheUsage;
	u32 lastCTS;
	ESInfo* es;
};

#endif

