 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 *  title:   QFrameBuffer.hpp                                           *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *  $Id: QFrameBuffer.hpp,v 1.5 2005/02/22 18:26:11 mkropfbe Exp $
 *                                                                      *
 ***********************************************************************/

#ifndef _QFRAMEBUFFER_HPP
#define _QFRAMEBUFFER_HPP

#include "IO.hpp"
#include "VideoESInfo.hpp"
#include "adaptors/MP4Decoder.hpp"
#include <qevent.h>
#include <qimage.h>
#include <qwidget.h>
#include "VideoFrameBuffer.hpp"

class QLabel;
class QCustomEvent;

/**
 * @short This class serves as a frame buffer and event handler of a QLabel 
 * widget for displaying video frames.
 * <p>
 * It manages a QImage object as a frame buffer whose pixels may be set using
 * the {@link #setPixel()} method. For painting the frame buffer's contents into
 * the QLabel widget's pixmap, a {@link #customEvent()} handler is implemented.
 * </p>
 * <p>
 * Although this class is an extension of QWidget, it is not intended to be used
 * as a replacement of the QLabel instance actually displaying the video frames.
 * Actually, a QFrameBuffer widget will never become visible.
 * </p>
 *
 * @author Mario Taschwer
 * @version $Id: QFrameBuffer.hpp,v 1.5 2005/02/22 18:26:11 mkropfbe Exp $
 */
class QFrameBuffer : public QWidget, public VideoFrameBuffer {
  Q_OBJECT

public:
  /** Constructor. 
   * @param qlabel      the QLabel widget where to display the video frames.
   *                    This object will become an invisible child of <code>qlabel</code>
   *                    in the widget hierarchy.
   *                    The frame buffer will be initialized to the width
   *                    and height of <code>qlabel</code>.
   * @param depth       the color depth of the frame buffer (8, 24, 32).
   * @param numColors   the color table size of the frame buffer
   *                    (if <code>depth == 8</code>).
   * @param gray        if <code>true</code> and <code>numColors > 0</code>, set the
   *                    color table entries to all gray values.
   * @param eventType   the custom event type used by {@link #postPaintEvent()}
   *                    to request displaying the frame buffer on the <code>qlabel</code>
   *                    widget. The value of this parameter should be in the range
   *                    <code>[QEvent::User, QEvent::MaxUser]</code>.
   * @see QImage::QImage(int, int, int, int, QImage::Endian)
   */
  QFrameBuffer(QLabel *qlabel, int depth, int numColors = 0, bool gray = true,
	       QEvent::Type eventType = QEvent::User);

  /** Destructor. */
  virtual ~QFrameBuffer();

  /** Set a pixel in the frame buffer.
   * If this method is called from a thread other than the Qt
   * GUI thread, the frame buffer should be locked before
   * by calling the {@link #lock()} method.
   * @see QImage::setPixel(int, int, uint)
   */
  inline void setPixel(int x, int y, uint index_or_rgb) {
    image->setPixel(x, y, index_or_rgb);
  }

  /** Post a custom event for <code>this</code> object
   * to display the frame buffer on the QLabel widget.
   * This is thread-safe and will cause the GUI thread to invoke
   * {@link #customEvent()}. We need to use custom events, as
   * hidden widgets will not receive paint events.
   */
  void postPaintEvent();

protected:
  /** Event handler for displaying the QImage buffer's contents in the 
   * associated QLabel widget. 
   */
  virtual void customEvent(QCustomEvent *event);

private:
  /** The QLabel widget where to display video frames. */
  QLabel *qlabel;

  /** The frame buffer. */
  QImage *image;

  /** Custom event type used by {@link #postPaintEvent()}. */
  QEvent::Type eventType;

};

#endif /* _QFRAMEBUFFER_HPP */
