 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 *  title:   QLabelRenderer.hpp                                         *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *  $Id: QLabelRenderer.hpp,v 1.18 2006/01/20 15:37:18 mkropfbe Exp $
 *                                                                      *
 ***********************************************************************/

#ifndef _QLABELRENDERER_HPP
#define _QLABELRENDERER_HPP

#include "IO.hpp"
#include "VideoESInfo.hpp"
#include "adaptors/MP4Decoder.hpp"
#include <qevent.h>
#include <qimage.h>
#include <qwidget.h>

#include "QFrameBuffer.hpp"
#include "VideoRenderer.hpp"

class QFrameBuffer;


/**
 * @short This I/O class renders a gray-scale or RGB stream  
 * (via a QFrameBuffer) onto a QLabel widget. 
 *
 * It is intended to be used as a DataSink connected to a DataChannel.
 * <p>
 * Video frames will be (down or up) scaled to fit into the QLabel widget by
 * preserving the aspect ratio given in VideoESInfo.</p>
 * <p>
 * The expected format of the gray-scale stream is: 1 byte per pixel.</p>
 * <p>
 * The expected format of the RGB24 stream is: 3 bytes per pixel,
 * in the order RGB (red, green, blue).</p>
 * <p>
 * The expected format of the RGB32 stream is: 4 bytes per pixel,
 * in the order BGRA (blue, green, red, alpha). This conforms to the
 * PIX_FMT_RGBA32 format defined in ffmpeg/libavcodec/avcodec.h on
 * little endian architectures.</p>
 *
 *@author Mario Taschwer
 *@version $Id: QLabelRenderer.hpp,v 1.18 2006/01/20 15:37:18 mkropfbe Exp $
 */
class QLabelRenderer : public VideoRenderer
{
public:

  /** Constructor. If the parameters are invalid, the #open() method will fail.
   * @param es        elementary stream information about the video stream to
   *                  be rendered.
   * @param colorSpace the color space of the video stream to be rendered;
   *                   must be one of: UncompressedVideoFrame::ColorSpaceGRAY8,
   *                   UncompressedVideoFrame::ColorSpaceRGB24, UncompressedVideoFrame::ColorSpaceRGB565
   *                   UncompressedVideoFrame::ColorSpaceRGB32.
   * @param qlabel    the QLabel object to render frames on; if not already set,
   *                  it will be initialized with a pixmap of the same size.
   * @param allowUpscaleX if <code>true</code>, video frames may be upscaled
   *                      in x-dimension to fit the qlabel, which may
   *                      result in significant quality loss, since a very
   *                      simple upscaling algorithm is used.
   * @param eventType   the custom event type used internally
   *                    to request displaying a video frame on the <code>qlabel</code>
   *                    widget. The value of this parameter should be in the range
   *                    <code>[QEvent::User, QEvent::MaxUser]</code>.
   */
  QLabelRenderer(const VideoESInfo *es, UncompressedVideoFrame::ColorSpace colorSpace,
		 QLabel *qlabel, bool allowUpscaleX = false,
		 QEvent::Type eventType = QEvent::User);

  /** Destructor. */
  virtual ~QLabelRenderer();

  /** Renders a frame to the destination QLabel widget.
   * @return 1 on success, 0 on error.
   */
  virtual int writeFrame(Frame * frm, ESInfo *out_es=NULL);

  /** Creates internal frame buffers and checks whether the stream to be
   * rendered matches the widget dimensions.
   * @return <code>true</code> on success, <code>false</code> on error.
   * @see IO::open()
   */
  virtual bool open();


protected:
  /** Update {@link #frame_info} (except <code>size</code>)
   * and {@link #display_region}.
   * This function is not thread-safe, and is internally
   * called by the {@link #open()} method.
  */
  bool updateDisplayRegion();


private:
  /** QLabel object where frames are to be rendered upon. */
  QLabel *qlabel;

  /** Custom event type used with QFrameBuffer. */
  QEvent::Type eventType;
};


#endif /* _QLABELRENDERER_HPP */
