
/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	Rtp.cpp                                                       *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

/***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#include <vector>
#ifndef WINCE
#include <sched.h> 
#endif

#include "Rtp.hpp"
#include "CompressedVideoFrame.hpp"
#include "CompressedAudioFrame.hpp"
#include "net/SyncLayer.hpp"
#include "net/PacketizationLayer.hpp"
#include "adaptors/MP4Decoder.hpp"
#include "adaptors/MP4audioDecoder.hpp"
#include "Statistics.hpp"
#include "metadata/TerminalCapabilities.hpp"

/* when there is less than eg. 2.0 percent pkt loss, increase net_BW by INC_PACKETs */
#define RTP_BW_INC_LOSS 1.5
#define RTP_BW_INC_PACKETS 6
//#define RTP_BW_INC_FACT 1.15

/* when there is more than eg. 5.0 percent pkt loss, decrease net_BW by eg. 20% (0.80) */
#define RTP_BW_DEC_LOSS 10.0
#define RTP_BW_DEC_FACT 0.85


/* how may secs of video should be added to the standard sec, just to fill buffers and lose quality?
 * Type: SECONDS */
#define RTP_BUF_FILL_FACT 0.30

/* how may secs of video should be removed from the standard sec, just to drain buffers and keep quality?
 * Type: SECONDS */
#define RTP_BUF_DRAIN_FACT 0.30

/* if the buffer is 95% full, we stop adapting, just fully draining buffer... 
 * further, we reduce network bandwidth to a bare minimum of 3 Kbyte/sec
 * we call this "superdrain mode" :)
 * Type: PERCENT */
#define RTP_BUF_SUPERHIGHWATER 90

/* if client buffer is drained from SUPERHIGHWATER, it will return into normal mode when the
 * buffer fill level falls below the RTP_BUF_SUPERHIGHWATER_LOW
 * Type: PERCENT */
#define RTP_BUF_SUPERHIGHWATER_LOW 70


/* if client buffer allows it, dont reduce quality less than eg. 50% 
 * for B-frame dropping, a pattern with 4b  IBBBBPBBBBPBBBB and B-quant28
 *                           gives about 50% scalability
 * Type: PERCENT */
#define RTP_MAX_ADAPT 50

/* what is the lowest acceptable framerate? (eg. fps > 15)
 * Type: FPS */
#define RTP_MIN_PRIO 14 //14 or 5
//#define RTP_MIN_PRIO IO_NETWORK_HIGHEST_PRIORITY

/* if client has less than eg. 0.3 secs of video in buffer, we force prefetching
 * Type: SECONDS */
#define RTP_BUF_PREFETCH  0.1

/* byte size of server-side preQ */
#define RTP_SERVER_PreQ_MAX  10*1024*1024

/* byte size of client-side preQ */
//NOT USED! OVERRIDDEN BY: RTP_CLIENT_PreQ_MAX_SECS * avgBW
#define RTP_CLIENT_PreQ_MAX  4*1024*1024


// for stupid MSVC
#ifdef WIN32
static const int TTL = 16;
static const int MAX_PAYLOAD_SIZE = 65000;
static const int RTP_PAYLOAD_TYPE = 96;
#endif
#define MPEG_AUDIO_HEADER_SIZE 4
#define RTP_SEQ_MOD  0x10000
#define RTP_BW_MINIMUM 3*1024 //dont go below that net_bw value...
#define UCL_TIMEOUT 20

//minimum udp packet inter-sendout time
#define RTP_MIN_INTERARRIVAL_TIME 6

//if act_wait is below this, we do busy waiting, to get better granularity
#define RTP_BUSY_WAIT_MS 5

PortGenerator portGen;

/***************************************************************/
Rtp::Rtp(const char *uri, int remotePrt, const char *addr,
	 int localPrt, ESInfo * e,
	 PacketizationLayer * p, bool w, const TerminalCapabilities *tc,
	 GlobalTimer *gt, Statistics *statistics,bool avtd) : IO() 
{
	assert(e);
	assert(p);
	assert(uri && addr);
	allowVopTimeDetection=avtd;
  url = new char[strlen(uri) + 1];
  rtx=NULL;
  remotePort = remotePrt;
  remoteAddress = new char[strlen(addr) + 1];
  localPort = localPrt;
  globalTimer = gt;
  stats = statistics;

  writeToNetwork = w;
  strcpy(remoteAddress, addr);
  strcpy(url, uri);
  setState(CLOSED);
  session = NULL;
  pLayer = p;
  firstFrame=false; // must use open to set to true
  memset(&sendQ.Q,0,sizeof(sendQ.Q));
  preQ.Q = preQ.Q_tail = NULL;
  sendQ.size = preQ.size = 0;
  resendFrameHeader=false;
  preQ.needsPrefetching=false; //initial prefetching autodetected

  preQ.MUTEX_Q.initialize("RTP-preQ.MUTEX_Q");

  preQ.MUTEX_Q_DATA_AVAIL.initialize("RTP-preQ.MUTEX_Q_DATA_AVAIL");
  preQ.MUTEX_Q_DATA_AVAIL.lock();

  preQ.MUTEX_Q_FULL.initialize("RTP-preQ.MUTEX_Q_FULL");
  preQ.MUTEX_Q_DONE.initialize("RTP-preQ.MUTEX_Q_DONE");
  preQ.MUTEX_Q_DONE.lock();

  sendQ.MUTEX_Q.initialize("RTP-sendQ.MUTEX_Q");

  last_rr_seq_no = -1;
  last_rr_sum_pkts_lost = -1;

  last_rtp_seq=0;
  last_rtx_rtp_seq=0;
  interarrival_time=0;
  //net_bw_before_superdrain=-1;
  real_bw=0;
  first_rtp_seq = -1;
  first_rtx_rtp_seq = -1;
	ts_offset = 0;

  feedbackEnabled = false;
  rtxEnabled = false;
	threadWasPaused=false;

	this->setESInfo(e);
	if(tc)
		stats->setBaseStreamBW(tc->getNetworkCapacityInByte()); //set base, unaffected by switching
	else
		stats->setBaseStreamBW(stats->getStreamBW()); //set base, unaffected by switching


  dprintf_full("Rtp::Rtp (url=%s, remotePort=%i, remoteAddr=%s, localPort=%i, writeOnly=%i, avgBW=%i, TCbw=%i)\r\n",
	       url, remotePort,remoteAddress,localPort,(int)writeToNetwork,stats->getStreamBW()/128, stats->getBaseStreamBW()/128 );

}


/***************************************************************/
void Rtp::setESInfo(ESInfo * new_es) {
	int netBW;

	es = new_es;

	stats->setESInfo(new_es);

  stats->setStreamBW( (int)(es->getAvgBandwidth() / 8 * 1.15)); //add 15 %
#if (defined NETBENCH)
	dprintf_full("Rtp compiled with NETBENCH, so simulated network behaviour!!!\n");
  //override with exact avgBW
  stats->setStreamBW( (int)(es->getAvgBandwidth() / 8 * 1));
#endif
	//stats->setStreamBW(0.8*1024*1024/8);

	dprintf_full("Rtp::setESInfo with new streamBW %i\n",stats->getStreamBW()/128);

	endFrameNumber = 0;

	//set adaption right for new stream
	netBW = stats->getStreamBW();

	if (state != CLOSED) {  //only for stream switching
		calcAdaptPreQValues(netBW);
		adaptPreQ(netBW);
	}
}


/***************************************************************/
Rtp::~Rtp() {
  if (remoteAddress) {
    delete[] remoteAddress;
    remoteAddress = NULL;
  }
  // FIXME: pLayer needs a refCount
  if (pLayer) {
    delete pLayer;
    pLayer = NULL;
  }
  if (state != CLOSING || state != CLOSED) {
    close(true);
  }

	
  if(session)	{
    rtp_send_bye(session);
    //rtp_done(session); 
    session=NULL;
  } // delete happens in rtp_done(..)  
  flushBuffer(0,-1);
  if(rtx) {
    if(rtx->session)	{
      rtp_send_bye(rtx->session);
      //rtp_done(rtx->session); 
      rtx->session=NULL;
    } // delete happens in rtp_done(..)  
    delete rtx; 
    rtx=NULL;
  }	

}

/***************************************************************/
bool Rtp::open() {

  dprintf_full("Rtp::open Rtp is %s, state is %i\n",(writeToNetwork) ? "write" : "read",state);
  if (state == OPEN)
    return true;
  setState(OPENING);

	/*
  setRoundRobinScheduling(1);
	*/

  currentFrameNumber=0;
	endFrameNumber=-0;
  IOFrameStat.numProcessedFrames=0;
  IOFrameStat.seqno  = -1;
  IOFrameStat.act_ts = -1; //impossibly first CTS
  firstFrame=true;
  framesSeen = 0;
  if (remoteAddress && !session) {
    dprintf_full("Rtp::open init RTP channel ports %i <--> %i\n",localPort,remotePort);
		int p1, p2;
		p1=portGen.getSockFromPort(localPort);
		p2=portGen.getSockFromPort(localPort+1);
		dprintf_full("PortGenerator::dome\n"); fflush(stdout);
    session = ucl_rtp_init(remoteAddress,
													 localPort,
													 remotePort,
													 TTL, 100, Rtp::RtpCallback, (uint8_t *) this, 
													 p1,p2);
  }
	dprintf_full("PortGenerator::\n"); fflush(stdout);
  if (session) {
    rtp_set_option(session, RTP_OPT_WEAK_VALIDATION, FALSE);
    rtp_set_option(session, RTP_OPT_PROMISC, TRUE);

#ifndef WITHOUT_RTP_EXT
    if (!writeToNetwork) { // DataSink
      if (rtx) {
				dprintf_full("Rtp::open enabled to recv/send feedback!\n");
				feedbackEnabled = true;
				//should we send feedback? enable it at UCL lib!
				if ((rtx->state == ACK) || (rtx->state == NACKACK)) 
					rtp_set_option(session, RTCP_OPT_FB_ACK, TRUE);
				if ((rtx->state == NACK) || (rtx->state == NACKACK)) 
					rtp_set_option(session, RTCP_OPT_FB_NACK, TRUE);
      }
    } else
      feedbackEnabled = true; //server will always accept extended feedback
#endif

    if (rtx) {
      if (rtx->remotePort ==0) {
				dprintf_full("Rtp::open init RTP RTX channel port is ZERO! RTX disabled! still capable of recv/send feedback!\n");
      } else {
#ifndef WITHOUT_RTP_EXT
				dprintf_full("Rtp::open init RTP RTX channel ports %i <--> %i\n",rtx->localPort,rtx->remotePort);
				rtxEnabled = true;
				rtx->session = ucl_rtp_init(remoteAddress,
																		rtx->localPort,
																		rtx->remotePort,
																		TTL, 100, Rtp::RtpCallback, (uint8_t *) this, 
																		portGen.getSockFromPort(rtx->localPort),
																		portGen.getSockFromPort(rtx->localPort+1));
				if (!rtx->session) {
					setState(STREAMERR);
					return false;
				}
				rtp_set_option(rtx->session, RTP_OPT_WEAK_VALIDATION, FALSE);
				rtp_set_option(rtx->session, RTP_OPT_PROMISC, TRUE);
				if (!writeToNetwork) { // DataSink
					//should we send rtx feedback?
					if ((rtx->state == ACK) || (rtx->state == NACKACK)) 
						rtp_set_option(rtx->session, RTCP_OPT_FB_ACK, TRUE);
					if ((rtx->state == NACK) || (rtx->state == NACKACK)) 
						rtp_set_option(rtx->session, RTCP_OPT_FB_NACK, TRUE);
				}
#endif
      }
    }//if rtx


    setState(OPEN);
    firstFrame=true;
		// FIXME: on server-side this should be recv'd via RTSP/SDP!
    stats->setClientPreQMaxSize((RTP_CLIENT_PreQ_MAX_SECS+1) *  
																(int) ((float)(es->getAvgBandwidth() / 8) * (1+RTP_BUF_FILL_FACT) ) );
		
    if (writeToNetwork) //server has larger Q
      preQ.max_size = RTP_SERVER_PreQ_MAX;
    else 
      preQ.max_size = stats->getClientPreQMaxSize();
		
    preQ.size = preQ.num_elems = preQ.num_full_frames = 0;
    preQ.avg_bw = 0;
    //    preQ.avg_bw = es->getAvgBandwidth();
    preQ.highest_rtp_seq=0;
    preQ.highest_ts=0;

    dprintf_full("Rtp::open (ports %i:%i) is starting the according RTP thread\n",localPort,remotePort);
    this->start();
    return true;
  } else {
    setState(STREAMERR);
    return false;
  }
}

/***************************************************************/
IO::State Rtp::play(double prefetchTime) {
  dprintf_full("Rtp::play with prefetch %f (old state: %i)\n",prefetchTime,state);

  if ((getState() == IO::CLOSED) ||
      (getState() == IO::PAUSED) ||
      (getState()  == IO::MUTED)) {
    if (isInput()) {
      if (!preQ.needsPrefetching) {
				preQ.needsPrefetching=true;
				stats->setStillToPrefetchSecs(	stats->getStillToPrefetchSecs() + prefetchTime);
				dprintf_full("Rtp::play requesting stillToPrefetchSecs %f\n", stats->getStillToPrefetchSecs());
				flushBuffer(0,-1); //delete all
				IOFrameStat.seqno = -1; //getFrame will accept new rtp_seq offset
				IOFrameStat.act_ts = -1; //impossibly first CTS
      }
    } else {
      stats->setPrefetchedSecs(stats->getPrefetchedSecs() + prefetchTime);
		}
    if (getState() != IO::CLOSED) {
      setState(IO::OPEN);
			while (threadWasPaused)
				msleep(10);
		}
  }
	return getState();
}

/***************************************************************/
IO::State Rtp::pause() {
  dprintf_full("Rtp::pause\n");
  if (getState() == IO::OPEN || 
      getState()  == IO::OPENING || 
      getState()  == IO::CLOSING || 
      getState()  == IO::STREAMEOF || 
      getState()  == IO::MUTED) {
    setState(IO::PAUSED);
		//well, we ignore preQ locking... just read...
		if (preQ.Q != NULL) { // popPreQ might hang in preQ.MUTEX_Q_DATA_AVAIL
				while (!threadWasPaused) //wait until RtpThread is really paused!
					msleep(10);
		}
	}

	return getState();
}
 
/***************************************************************/
IO::State Rtp::mute() {
  dprintf_full("Rtp::mute\n");
  if (getState() == IO::OPEN || 
      getState()  == IO::OPENING || 
      getState()  == IO::CLOSING || 
      getState()  == IO::STREAMEOF || 
      getState()  == IO::PAUSED)
    return setState(IO::MUTED);
  else
    return getState();}



/***************************************************************/
char *Rtp::getSSRC() {
  if (state != OPEN)
    return NULL;
  
  char *result = new char[9];
  sprintf(result, "%08x", rtp_my_ssrc(session));
  return result;
}


/*****************************************************************/
bool Rtp::setEndFrameNumber(u32 frameNumber) {
	long incr = es->getVOPTimeIncrement();

	if (endFrameNumber == 0)
		flushBuffer(frameNumber * incr, -1);
	else
		flushBuffer(frameNumber * incr, endFrameNumber * incr);

 	endFrameNumber=frameNumber;
	return true;
}

/*****************************************************************/
bool Rtp::setToFrameNumber(u32 frameNumber) {

  flushBuffer(0, -1);
  currentFrameNumber=frameNumber;

  return true;
}

/*****************************************************************/
int Rtp::flushBuffer(long from_ts, long to_ts) {
  NetPacket_t *p=NULL;
  NetPacket_t *pr=NULL;
  int flushed_bytes=0;


  dprintf_full("Rtp::flushBuffer preQ from TS %li to TS %li\n",from_ts,to_ts);
	assert(to_ts == -1 || from_ts <= to_ts);
  preQ.MUTEX_Q.lock();

	if (to_ts < -1)
		to_ts = -1; //fix wrong inputs
	
	if (isInput())
		stats->setHighestPacketTS(0);
  p=pr=preQ.Q;
  while (p != NULL) {
		//dprintf_full("   checking rtp seq %i TS %i (%i elems left, next %p)\n", p->rtp_seq, p->rtp_ts,preQ.num_elems,p->next);
    if ( (p->rtp_ts >= (unsigned)from_ts) &&
				 ((to_ts == -1) || (p->rtp_ts <= (unsigned)to_ts)) ) {
      preQ.size -= p->pl_size;
      flushed_bytes += p->pl_size;
      preQ.num_elems--;
      if (p->last_frag)
				preQ.num_full_frames--;
			 
      preQ.avg_bw = 0;
      if (preQ.num_elems == 0)
				preQ.Q_tail = NULL;
      else
				if (preQ.num_full_frames > 0)
					preQ.avg_bw = (int)(preQ.size/preQ.num_full_frames * 
															((double)es->getMediaTimeScale()) / 
															((double)es->getVOPTimeIncrement()));

      if (p == preQ.Q)
				preQ.Q = pr = p->next;
      else
				pr->next = p->next;

      if (p == preQ.Q_tail) {
				preQ.Q_tail = pr;
				if (pr != NULL) //very last elem deleted
					pr->next = NULL;
      }
			 
      //free everything!
      assert(p->payload != NULL);
      delete p->payload;
      p->payload=NULL;
      delete p;p=NULL;
			 
      //keep old pr
      if(pr==NULL)
				p=NULL;
      else if (pr == preQ.Q)
				p=preQ.Q;
      else
				p=pr->next;
			
    } else {
      if (isInput() && (stats->getHighestPacketTS() < p->rtp_ts)) {
				dprintf_full("Rtp::flushBuffer: set new highest TS: %i\n",p->rtp_ts);
				stats->setHighestPacketTS(p->rtp_ts);
      }
      pr=p;
      p=p->next;
    }
  }
	 
  preQ.MUTEX_Q_FULL.release();
  if (preQ.Q != NULL) { //still data available
    preQ.MUTEX_Q_DATA_AVAIL.release();
  }
 
  preQ.MUTEX_Q.release();
  return flushed_bytes;
}


/***************************************************************/
/** returns the number of packets sent. Returns 0 on error */
int Rtp::writeFrame(Frame * frm, ESInfo *out_es) {
  bool result = true;
  SLPacketList slpl;
  AU au;
  u8* newPayload=NULL;
  bool last = false;
  int counter = 0;
	
  if (!writeToNetwork || state != OPEN || !pLayer || !frm)
    return 0;

  dprintf_full("Rtp::writeFrame CTS %i prio %i size %i\n",frm->getAU()->cts, frm->getAU()->prio, frm->getAU()->size);
  
	if (endFrameNumber>0 && frm->getAU()->cts / es->getVOPTimeIncrement() > endFrameNumber) {
		//this comes from unflushed DC senqueue, and is out of range, so drop it!
		dprintf_full("Rtp::writeFrame nr %i TS %i exceeds endFrameNumber %i, ignoring!\n",frm->getAU()->cts / es->getVOPTimeIncrement(),frm->getAU()->cts, endFrameNumber );
		return 0;
	}

	//HEADER DATA
  if ( (!firstFrame && resendFrameHeader ) 
       || (firstFrame && !MP4Decoder::isValidDecoderConfig(frm->getAU()->payload)) ) { //HEADER already in first frame
		
    resendFrameHeader=false;
    firstFrame=false;
	
    //send extra header!
    u8* esHeaderBuffer=NULL;
    u32 headerSize=es->getHeaders(&esHeaderBuffer);
    if(headerSize>0) {
      newPayload=new u8[headerSize + frm->getAU()->size];
      memcpy(newPayload, esHeaderBuffer, headerSize);
      
      au.payload=newPayload;
      au.size=headerSize;
      au.cts=frm->getAU()->cts;
      au.dts=frm->getAU()->dts;
      au.duration=0;
      slpl=pLayer->createSLPacketList(&au);
      if(!slpl.empty()) {
				SLPacketList::iterator li = slpl.begin();
				last = false;
				counter = 0;

				dprintf_full("Rtp::writeFrame: HEADER (size %i) first 4 bytes of first Packet, CTS %u: %x %x %x %x\r\n",
										 headerSize, frm->getAU()->cts, (*li)->getPayload()[0],(*li)->getPayload()[1],
										 (*li)->getPayload()[2],(*li)->getPayload()[3]);
			
				while(!slpl.empty()) {
					if(slpl.size()==1) {
						last=true;
					}
					li = slpl.begin();
					result &= insertToPreQ((*li)->getPayload(), (*li)->getPayloadSize(),
																 frm->getType(), au.size, IO_NETWORK_HIGHEST_PRIORITY,
																 frm->getAU()->dts, counter, last, 0,Rtp::Q_TS);
					(*li)->unsetPayload(); //payload will be freed when ACKd
					delete *li; // here, there is no payload to free any more...
					(*li)=NULL;
					slpl.pop_front();
					counter++;
				}
      }
    }
    if(esHeaderBuffer) {
      delete[] esHeaderBuffer;
      esHeaderBuffer=NULL;
    }
  }

  //FRAME DATA
  slpl = pLayer->createSLPacketList(frm->getAU());
  firstFrame=false;

  if (slpl.empty()) {
    dprintf_full("Rtp::writeFrame() SLPacketList is empty! AU: size = %u, cts = %u\n",
		 frm->getAU()->size, frm->getAU()->cts);
    return 1; //fake success, since the SLPacket is buffered and not yet full MTU reached...
  }
  SLPacketList::iterator li = slpl.begin();

  dprintf_full("Rtp::writeFrame: FRAME first 4 bytes of Frame, CTS %u: %x %x %x %x\r\n",
							 frm->getAU()->cts, frm->getAU()->payload[0],
							 frm->getAU()->payload[1],
							 frm->getAU()->payload[2],frm->getAU()->payload[3]);
  last=false;
  //counter stays to old value (might be with header)

  while(!slpl.empty()) {
    if(slpl.size()==1) {
      last=true;
    }
    li = slpl.begin();
    result &= insertToPreQ((*li)->getPayload(), (*li)->getPayloadSize(),
													 frm->getType(), frm->getAU()->size, frm->getAU()->prio, 
													 frm->getAU()->dts, counter, last, 0, Rtp::Q_TS);
    (*li)->unsetPayload(); //payload will be freed when ACKd
    delete *li; // here, there is no payload to free any more...
    (*li)=NULL;
    slpl.pop_front();
    counter++;
  }
  
  IOFrameStat.numProcessedFrames++;
#ifdef _POSIX_PRIORITY_SCHEDULING
	if (sched_yield() != 0) { //this is necessary to give sendThread more realtime performance!
		perror("Rtp::writeFrame YIELD");
	}
#else
	// for windows pthread lib!
#if (defined WIN32) && (!defined WINCE)
	if (sched_yield() != 0) { //this is necessary to give sendThread more realtime performance!
		dprintf_full("Rtp::writeFrame YIELD");
	}
#endif
#endif
  return (int)result;
}


/***************************************************************/
bool Rtp::close(bool immediate) {
  if (state == CLOSED)
    return true;
  dprintf_full("Rtp::close: incoming state = %i\n",state);
  if  ((state == CLOSING) && (immediate)) { //multiple closes
    dprintf_full("Rtp::close: multiple close... ignoring\n");
    setState(FORCE_CLOSING);
    preQ.MUTEX_Q_DONE.release();  //signal that all data is sent
    sleep(2); //workaround to let datachannel::teardown wait!
    return true;
  }

  if((state == OPEN) || (state == PAUSED) || (state == MUTED)
     || (state == PREFETCHING) || (state == STREAMEOF)) {
    if (!immediate) {
      setState(CLOSING);
      dprintf_full("Rtp::close waiting for correct sending/displaying (NOT IMMEDIATE)\n");
    } else { //drop all frames, is IMMEDIATE
      dprintf_full("Rtp::close IMMEDIATE close requested\n");
      setState(FORCE_CLOSING);
    }
    if (immediate) { //drop all frames, is IMMEDIATE
      NetPacket_t *p;
			
      sendQupdate(-1);
      do {
				p = popPreQ(false);
				if ( (p != NULL) && (p->payload != NULL) ) {
					dprintf_full("freeing preQ Frame seq %i ts %i:%i\n", p->rtp_seq,p->rtp_ts,p->frag_no);
					delete p->payload;
					p->payload=NULL;
					delete p;
					p=NULL;
				}
      } while (p != NULL);
	  preQ.MUTEX_Q_DONE.tryLock();
    }
		
		
    dprintf_full("Rtp::close: waiting for data to be sent out...\n");
		
    preQ.MUTEX_Q.lock();
    do {
      if (state == OPEN) //happens after unpause or unmute
				setState(CLOSING);
      //if (preQ.Q == NULL) //make sure that other thread in popPreQ exits
      preQ.MUTEX_Q_DATA_AVAIL.release(); // simulate data
      preQ.MUTEX_Q.release();
			
      dprintf_full("Rtp::close: give the other thread a chance to work off Q (state %i)\n",state);
      //give the other thread a chance to work off Q
      sleep(1);
      preQ.MUTEX_Q.lock();
    } while ((state == CLOSING)||(state == PAUSED)||(state == MUTED));
    preQ.MUTEX_Q.release();

    if (state == OPEN) //happens on unpausing (hereby rewinding) a already closed inputIO
      return false;

    if (state != CLOSED) {
      //			if (preQ.Q == NULL) //make sure that other thread in popPreQ exits
      preQ.MUTEX_Q_DATA_AVAIL.release(); //simulate data
      dprintf_full("Rtp::close: waiting for MUTEX_Q_DONE\n");
      preQ.MUTEX_Q_DONE.lock();  //wait until all data is sent
      dprintf_full("Rtp::close: all frames sent, so we are done...\n");
      setState(CLOSED);

			
      //FIXME: this crashes the app!
      if(session!=NULL) {
				rtp_send_bye(session);
				rtp_done(session);
				session = NULL;
				portGen.closePortPair(localPort);
      }
      if (rtxEnabled && rtx->session) {
				rtp_send_bye(rtx->session);
				rtp_done(rtx->session);
				rtx->session = NULL;
				portGen.closePortPair(rtx->localPort);
      }
			
      // now delete mutex
      preQ.MUTEX_Q.destroy();
      preQ.MUTEX_Q_DATA_AVAIL.destroy();
      preQ.MUTEX_Q_FULL.destroy();
      preQ.MUTEX_Q_DONE.destroy();
      sendQ.MUTEX_Q.destroy();
    }
  }
  return true;
}

	/** returns a frame if one complete frame is available, otherwise null 
		is returned. This function is typically blocking. Don't use a NULL
		return value to conclude STREAMEOF, always check with getState()!
	*/
Frame *Rtp::getFrame() {
  struct AU *au=NULL;
  NetPacket_t *p=NULL;
  u8 *buf=NULL;
  long buf_siz=0;
  bool lastFrag;
  bool first_frag=true;
  bool old_ts=false;

  currentFrameNumber++;
  framesSeen++;
	
  dprintf_full("Rtp::getFrame %i (TS %i) of still buffered %i frames (total %i of %i TS %li)... expecting rtp_seq %i (state %i)\n",
							 IOFrameStat.numProcessedFrames, IOFrameStat.act_ts, preQ.num_full_frames,
							 currentFrameNumber,
							 es->getNumberOfMediaSamples(),(long)es->getDuration(), 
							 IOFrameStat.seqno+1,state);

  //PRE-buffer some frames
  struct timeval tv;
  double start_dtime, now_dtime=0,fetch_time=0;

  IO::State oldState;
	

  gettimeofday(&tv,0);
  start_dtime = now_dtime = tv.tv_sec + (double)tv.tv_usec / 1000000.0;

  if (!preQ.needsPrefetching && (stats->getBufAheadSec() <= RTP_BUF_PREFETCH)
      && ((state == OPEN) || (state == MUTED)) ) {
    dprintf_full("Rtp::getFrame: further PREFETCHING needed to fill empty buffers!\n");
    preQ.needsPrefetching=true;
    stats->setStillToPrefetchSecs( stats->getStillToPrefetchSecs() + RTP_PREFETCH_SECS);
     flushBuffer(0,-1); //clean up all packets, since this will lead to a DataChannel Pause()
  }

  // PREFETCHING ##########################################
  if (preQ.needsPrefetching) {	//initial or addtional prefetch is needed...
    bool done=false;
    oldState = state;
    setState(PREFETCHING);
    if (globalTimer)
      globalTimer->setPrefetching(true);
    fetch_time=0;
    do {
      //dprintf_full("...prefetching: left %f done %i needsPrefetch %i....state %i\n",
      //  stats->getStillToPrefetchSecs(), done,preQ.needsPrefetching,state);
      msleep(10);
      gettimeofday(&tv,0);
      now_dtime = tv.tv_sec + (double)tv.tv_usec / 1000000.0;
      if (state == STREAMERR) {
	dprintf_err("Rtp::getFrame() STREAMERR occured during prefetching, returning NULL\n");
	if (globalTimer)
	  globalTimer->setPrefetching(false);
	return NULL;
      }
      if ((state != PREFETCHING) && (state != PAUSED)) {
				dprintf_full("Rtp::getFrame: continued PREFETCHING requested! left %f  state %i\n",
										 stats->getStillToPrefetchSecs(),state);
				//preQ.needsPrefetching=false;
				done=false;
				setState(PREFETCHING);        // continue with prefetching...
      }
      if (firstFrame) {
		  double fpt = stats->getFirstPacketTime();
		  double gstps = stats->getStillToPrefetchSecs();
				if ( (fpt != 0) &&		//only take from real first frame
						 (now_dtime - fpt >= gstps) ) {
					fetch_time = now_dtime - fpt;
					done=true;
				}
      } else { //intermediate prefetching
				//minor FIXME: pausing during intermediate Prefetching will shorten prefetching
				if (now_dtime - start_dtime >= stats->getStillToPrefetchSecs()) {
					fetch_time = now_dtime - start_dtime;
					done=true;
				}
      }
    } while ( !done && ((state == PREFETCHING) || (state == OPEN)|| 
												(state == MUTED)|| (state == PAUSED)) );

    if (state == PREFETCHING)
      setState(oldState);
    stats->setPrefetchedSecs(stats->getPrefetchedSecs() + fetch_time); //fine tune spent time
    stats->setStillToPrefetchSecs(0);
    stats->setBufAheadSec(((double)(stats->getHighestPacketTS()-stats->getFirstPacketTS()) / (double)es->getMediaTimeScale()));
    preQ.needsPrefetching=false;
    dprintf_full("Rtp::getFrame: prefetching of %2.3f secs (%2.3f real data recv secs)"
		 " brought us %2.3f secs of %s! state %i, prefetched_secs: %f\n",
		 now_dtime - start_dtime, fetch_time,
		 stats->getBufAheadSec(), es->isVisualStream()?"Video":"Audio",state,
		 stats->getPrefetchedSecs());
    if (globalTimer)
      globalTimer->setPrefetching(false);
  }
	
  // MERGE FRAME FROM PACKETS ###############################
  do {
    p=popPreQ();
    if(p==NULL) //this is the end.... my friend...
      return NULL; 
    if (p==(Rtp::NetPacket_t *)-1)  //FIXME: after unpause, the Q was empty! this will force PREFETCHING
      return this->getFrame();

    dprintf_full("Rtp::getFrame popped rtp_seq %i, TS %i, last_frag %i  size %i\n",
								 p->rtp_seq,p->rtp_ts,p->last_frag,p->pl_size);

    if (IOFrameStat.seqno == -1)  // init the rtp sequence
      IOFrameStat.seqno = p->rtp_seq;
    else
      IOFrameStat.seqno++;
		
    if (first_frag) { //first AU fragment
      if ( IOFrameStat.act_ts > (int)p->rtp_ts) {
				dprintf_full("DELETING left overs from previous frames\n");
				old_ts = true; //left overs from previous frames
      } //else
      //IOFrameStat.act_ts = p->rtp_ts;
      first_frag = false;
    }

    if (old_ts || ((unsigned)IOFrameStat.seqno != p->rtp_seq)) { //old fragments or incomplete AU
      dprintf_err("Rtp::getFrame expected rtp_seq %i, I got %i\n",IOFrameStat.seqno,	p->rtp_seq);


      //drop all other following fragments of incomplete AU
      //ATTN: this will read one more pkt from the NEXT frame and has to push it back!
      while ((IOFrameStat.act_ts > (int)p->rtp_ts) 
						 || (!old_ts && (IOFrameStat.act_ts == (int)p->rtp_ts)) ) {
				dprintf_full("Rtp::getFrame popped for dropping: rtp_seq %i, TS %i, last_frag %i\n",
										 p->rtp_seq,p->rtp_ts,p->last_frag);
				delete p->payload;
				p->payload = NULL;
				delete p; 
				p = NULL;
				p=popPreQ();
				if(p==NULL)
					return NULL; //this is the end.... my friend...
				if (p==(Rtp::NetPacket_t *)-1)  //FIXME: after unpause, the Q was empty! this will force PREFETCHING
					return this->getFrame();
      }

      if ((unsigned)IOFrameStat.seqno != p->rtp_seq) {
				//pushback to preQ
				dprintf_err("LOSS: Not all fragments of TS %i where available\n",IOFrameStat.act_ts);
				addToPreQ(p,Q_TS,RTP_REINSERT);
				IOFrameStat.seqno = p->rtp_seq - 1; //restart seq_no from here
				if(buf) {
					delete[] buf;
					buf=NULL;
				}
				//try to get the next frame
				return this->getFrame();
      }
    }

    dprintf_full("Rtp::getFrame CONT'd with popped rtp_seq %i, TS %i, last_frag %i\n",
								 p->rtp_seq,p->rtp_ts,p->last_frag);
    IOFrameStat.act_ts = p->rtp_ts;
    lastFrag = p->last_frag;
    if(es->isAudioStream()) { //RFC2250 specific headers need to be removed
                              //the header info is not utilized currently
			p->pl_size -= 4;
			buf=(u8*)realloc(buf, buf_siz + p->pl_size);
			memcpy(buf+buf_siz, p->payload + 4, p->pl_size);
    }
    else {
			buf=(u8*)realloc(buf,buf_siz + p->pl_size);
			memcpy(buf+buf_siz,p->payload,p->pl_size);
    }
    buf_siz+=p->pl_size;
    delete p->payload;
		p->payload=NULL;
    delete p;
		p=NULL;
  } while (lastFrag == 0);

  dprintf_full("Rtp::getFrame  DONE TS %i\n", IOFrameStat.act_ts);


  //IF VIDEO ################################################3
  if (es->isVisualStream())
    if ((buf[3]==0xb0) && MP4Decoder::isValidDecoderConfig(buf)) {
      u8* tmp=buf+4;
      bool stop=false;
      // received a header!
      dprintf_full("Rtp::getFrame received header\n");
      // some servers, like the ITECMP4Server send header+payload in one packet
      // catch this case
      while(!stop && tmp < (buf+buf_siz) &&
						(tmp = (u8*)memchr(tmp, 0xb6, (buf+buf_siz)-tmp)) ) {
				if(tmp[-1] == 1 && tmp[-2] == 0 && tmp[-3] == 0) {
					stop=true;
					tmp-=3;
				} else
					tmp++; //avoid double findings
      }
      if (tmp)
				es->setDecoderConfig(buf,tmp-buf);
      else
				es->setDecoderConfig(buf,buf_siz);

			
      if(stop) {
				dprintf_full("Rtp::getFrame compound header  %x %x %x %x (%i bytes) + frame (%li bytes) %x %x %x %x!\n"
										 ,buf[0],buf[1],buf[2],buf[3], tmp-buf, buf_siz-(tmp-buf),
										 tmp[0],tmp[1],tmp[2],tmp[3]);
				//fix everything up for final CompressedVideoFrame() creation
				buf_siz -= tmp-buf;
				u8* new_buf = new u8[buf_siz];
				memcpy(new_buf,tmp,buf_siz);
				delete[] buf;
				buf = new_buf;
      }	else {
				delete[] buf; buf=NULL;
				return this->getFrame(); //get the real frame
      }
    }	else if ( !MP4Decoder::isValidVisualFrame(buf)) {
      dprintf_err("Rtp::getFrame LOSS: Not all frags of TS %i avail (broke header/missing first packet)\n",
									IOFrameStat.act_ts);
      delete[] buf;
			buf=NULL;
      //try to get the next frame
      return this->getFrame();
    }
	
  //IF AUDIO ################################################3
  if (es->isAudioStream()) 
    if (MP4audioDecoder::isValidAudioFrame(buf)) {
      dprintf_full("Rtp::getFrame received audio frame header\n");
      //Every audio frame consists of a 4 byte header + payload. These together comprise 
      //the payload of the access unit.
      if(buf_siz == MPEG_AUDIO_HEADER_SIZE) {
				es->setDecoderConfig(buf,buf_siz);
				delete[] buf; buf=NULL;
				return this->getFrame();
      }
      else {
				//FIXME: Decoder configuration may be required to be set for VBR audio for every frame!!!
				if (firstFrame)
					this->es->setDecoderConfig(buf,4);		//"4" is for mpeg audio only ..... may be diff for aac			
      }
    }	else {// not valid audio
      dprintf_err("Rtp::getFrame LOSS: Not all audio frags of TS %i avail (broke header/missing first packet)\n",
									IOFrameStat.act_ts);
      delete[] buf; buf=NULL;
      //try to get the next frame
      return this->getFrame();
    }

  //GENERAL FULL FRAME CREATION #######################
  IOFrameStat.numProcessedFrames++;
		
  //fill the merged access unit
  au = new AU();
  au->payload = buf;
  au->size = buf_siz;
  au->cts = au->dts = (u32) IOFrameStat.act_ts + (u32) ts_offset;
	Frame* f=NULL;
	if(es->isVisualStream())
		f = new CompressedVideoFrame(Frame::NN_VOP, ((VideoESInfo*)es)->getWidth(),
																 ((VideoESInfo*)es)->getHeight());
	else if(es->isAudioStream())
		f = new CompressedAudioFrame(Frame::NN_VOP);
	else
		f = new CompressedVideoFrame(Frame::NN_VOP, 0, 0);

  f->setAU(au); //FrameType is set here automatically
  f->setMediaTimeScale(es->getMediaTimeScale()); //used by GlobalTimer Adaptor

  dprintf_full("Rtp::getFrame got size %8i frameType %s  TS %i presTime: %2.3f\n",f->getAU()->size, 
							 Frame::VOPTypeToChar(f->getType()),IOFrameStat.act_ts,
							 (double)IOFrameStat.act_ts / es->getMediaTimeScale());
	
  if (firstFrame) {
    // not a valid ESHeader as first packet?
    // do we have one from the RTSP?
    if(this->es->getEncodedDecoderConfig()==NULL) {
      dprintf_err("FATAL: no header found!! was: 0x%x%x%x%x\n",
									buf[0],buf[1],buf[2],buf[3]);
      setState(STREAMERR);
      delete[] buf; buf=NULL;
			delete f; f=NULL;
			delete au; au=NULL;
      close(true); //immediate
      return NULL; //this is the end... my friend...
    }
    firstFrame = false;
  }
#ifdef _POSIX_PRIORITY_SCHEDULING
	if (sched_yield() != 0) { //this is necessary to give sendThread more realtime performance!
		perror("Rtp::writeFrame YIELD");
	}
#else
	// for windows pthread lib!
#if (defined WIN32) && (!defined WINCE)
	if (sched_yield() != 0) { //this is necessary to give sendThread more realtime performance!
		dprintf_full("Rtp::writeFrame YIELD");
	}
#endif
#endif

  return f;
}


/***************************************************************/
void Rtp::ucl_update(u32 ts, u32 timeout) {
  struct timeval tv; 

  dprintf_full("Rtp::ucl_update TS %i\n",ts);
  if (session) {
    rtp_update(session);
    rtp_send_ctrl(session, ts, NULL, RTCP_NORMAL);
    /* Receive control and data packets */
    tv.tv_sec  = 0;
    tv.tv_usec = timeout*1000;
    if (rtxEnabled)
      tv.tv_usec = tv.tv_usec >> 1; //wait half of the time on rtp
		assert(session);
		rtp_recv(session, &tv, ts);

    if (rtxEnabled) {
			//			dprintf_full("Rtp::ucl_update on RTX at TS %i\n",ts);
      rtp_send_ctrl(rtx->session, ts, NULL, RTCP_NORMAL);
      /* eventually receive NACK packets again */
      tv.tv_sec  = 0;
      tv.tv_usec = timeout*1000 >> 1; //wait other half of the time on rtcp
			assert(rtx->session);
      rtp_recv(rtx->session, &tv, ts);
      /* do maintenance */
      rtp_update(rtx->session);
    }
		bool ret;
		int rounds=0;
		do {
      tv.tv_sec  = 0;
      tv.tv_usec = 1; //wait nearly no time
			ret=rtp_recv(session, &tv, ts); //try to get as much data as possible
			if (rtxEnabled) {
				tv.tv_sec  = 0;
				tv.tv_usec = 1; //wait nearly no time
				ret=rtp_recv(rtx->session, &tv, ts);
			}
			rounds++;
		} while(session && ret && (rounds < 10)); //continue as long as  data or rtx avail, or 10 pks where read in a row
  }
}

/***************************************************************/
void Rtp::fb_print(struct rtp *session, uint32_t ssrc, rtcp_fb *fb) {
  int i;

  if (ssrc == rtp_my_ssrc(session))
    return;
  dprintf_full("RX_FB SSRC = 0x%08x  ", ssrc);
  switch (fb->subtype) {
  case RTCP_FB_FMT_ACK:
    dprintf_err("ACK  BLP from %6i: ",fb->fci.fb_ack.pid);
    if (fb->fci.fb_ack.r == 1) { //range
      dprintf_full(" to %6i",fb->fci.fb_ack.blp+fb->fci.fb_ack.pid);
    } else //bitmap
      for (i=0; i < 15; i++) {
				if ((fb->fci.fb_ack.blp & ((uint64_t)1 << i)) != 0) {
					dprintf_full("r");
				} else {
					dprintf_full(".");	
				}
      }
    break;
  case RTCP_FB_FMT_NACK:
    dprintf_full("NACK BLP from %6i: ",fb->fci.fb_ack.pid);
    for (i=0; i < 15; i++) {
      if ((fb->fci.fb_nack.blp & ((uint64_t)1 << i)) != 0) {
				dprintf_full("D");
      } else {
				dprintf_full(".");
      }	
    }
    break;
  default:
    dprintf_err("FATAL: unknown Feedback Format!");
    ::exit(1);
  }
  dprintf_full("\n");
}

/**************************************************************
 * called by: server-side ucl_update()
 */
void Rtp::fb_handle(struct rtp *session, uint32_t ssrc, rtcp_fb *fb) {
  int i;
  NetPacket_t *p=NULL;
  Rtp *caller=NULL;
  struct timeval now_tv;
  double now_dtime;
  int curr_rtt;
  int last_nack_seq=0;


  if (ssrc == rtp_my_ssrc(session))
    return;

  caller=(Rtp*)rtp_get_userdata(session);
  dprintf_full("RX_FB SSRC = 0x%08x  \n", ssrc);

  if (session == caller->session) { // feedback from DATA session
    switch (fb->subtype) {
    case RTCP_FB_FMT_ACK:
      dprintf_full("RTCP RX_FB: ACK  BLP from %6i: ",fb->fci.fb_ack.pid);
      p=caller->popNetPacket(fb->fci.fb_ack.pid);
      if (p) {
				p->acked=true;
				caller->pushNetPacket(p);
      }
      if (fb->fci.fb_ack.r == 1) { //range
				dprintf_full(" to %6i",fb->fci.fb_ack.blp+fb->fci.fb_ack.pid);
				for (i=1; i <= fb->fci.fb_ack.blp; i++) {
					p=caller->popNetPacket(fb->fci.fb_ack.pid+i);
					if (p) {
						p->acked=true;
						caller->pushNetPacket(p);
					}
				}
      } else //bitmap
				for (i=0; i < 15; i++) {
					if ((fb->fci.fb_ack.blp & ((uint64_t)1 << i)) != 0) {
						p=caller->popNetPacket(fb->fci.fb_ack.pid+i+1);
						if (p) {
							p->acked=true;
							caller->pushNetPacket(p);
						}
					}
				}
      break;
    case RTCP_FB_FMT_NACK:
      last_nack_seq = fb->fci.fb_nack.pid;
      dprintf_full("RTCP RX_FB: resending NACK BLP from %6i:\n",last_nack_seq);
      p=caller->popNetPacket(last_nack_seq);
      caller->reinsertToPreQ(p,last_nack_seq);

			
      //now add the bitmapped NACKs
      for (i=0; i < 15; i++) {
				if ((fb->fci.fb_nack.blp & ((uint64_t)1 << i)) != 0) {
					last_nack_seq = fb->fci.fb_nack.pid+i+1;
					dprintf_full("RTCP RX_FB: further resending NACK BLP from %6i:\n",last_nack_seq);
					p=caller->popNetPacket(last_nack_seq);
					caller->reinsertToPreQ(p,last_nack_seq);
				}
      }
      break;
    default:
      dprintf_err("FATAL: unknown Feedback Format!");
      ::exit(1);
    }
    dprintf_full("\n");
  } else /*************************************************/
    if (caller->rtx && (session == caller->rtx->session)) { //lost RTX packets are also reinserted!
      switch (fb->subtype) {
      case RTCP_FB_FMT_ACK:
				dprintf_full("RTCP RX_FB: RTX ACK  BLP from %6i: NOT SUPPORTED",fb->fci.fb_ack.pid);
				break;
      case RTCP_FB_FMT_NACK:
				last_nack_seq = fb->fci.fb_nack.pid;
				dprintf_full("RTCP RX_FB: resending RTX NACK BLP from %6i:\n",last_nack_seq);
				p=caller->popNetPacketByRTXseq(last_nack_seq);
				caller->reinsertToPreQ(p,last_nack_seq);
				
				//now add the bitmapped NACKs
				for (i=0; i < 15; i++) {
					if ((fb->fci.fb_nack.blp & ((uint64_t)1 << i)) != 0) {
						last_nack_seq = fb->fci.fb_nack.pid+i+1;

						dprintf_full("RTCP RX_FB: further resending RTX NACK BLP from %6i:\n",last_nack_seq);
						p=caller->popNetPacketByRTXseq(last_nack_seq);
						caller->reinsertToPreQ(p,last_nack_seq);
					}
				}
				break;
      default:
				dprintf_err("FATAL: unknown Feedback Format!");
				::exit(1);
      }
    } else { /*************************************************/
      dprintf_err("FATAL: unknown session 0x%08x is sending feedback!", ssrc);
      ::exit(1);
    }

  //use only LAST nack'd packet for RTT! 
  //pick out the NEXT packet, which was definitely recv'd by the client, this one
  //is the best candidate for correct RTT calculations!	
  if (caller->rtx && session == caller->rtx->session) //lost RTX packets are also reinserted!
    p=caller->popNetPacketByRTXseq(last_nack_seq+1);
  else
    p=caller->popNetPacket(last_nack_seq+1);

  if (p!=NULL) {
    gettimeofday(&now_tv,NULL);
    now_dtime = now_tv.tv_sec + now_tv.tv_usec/1000000.0;
    if (p->resent_dtime == 0) //fb for first try
      curr_rtt = (int)((now_dtime - p->sent_dtime)*1000);
    else //fb of already RTX'd packet
      curr_rtt = (int)((now_dtime - p->resent_dtime)*1000);
		
    caller->stats->setAvgRTT(curr_rtt); 
    dprintf_full("RTCP RX_FB: RTT curr: %i  avg: %i\n", curr_rtt, caller->stats->getAvgRTT());
  }
}

/**************************************************************
 * called by reader-side ucl_update()
 */
void Rtp::rtp_data_handle(struct rtp *session, uint32_t ssrc, rtp_packet *packet) {
  char *buf=NULL;
  int data_len;
  Rtp *caller=NULL;
  uint16_t rtp_seq=0;
  uint16_t rtx_rtp_seq=0;
  struct timeval now_tv;

  if (ssrc == rtp_my_ssrc(session))
    return;

  caller=(Rtp*)rtp_get_userdata(session);
  //  printf("RX_RTP DATA SSRC = 0x%08x \n", ssrc);

	if (caller->getState() == PAUSED) //ignore last packet, which are sent shortly before pause!
		return;

  gettimeofday(&now_tv,NULL);
  caller->interarrival_time = (int)(1000* (now_tv.tv_sec + now_tv.tv_usec/1000000.0 
																					 - caller->stats->getLastPacketTime()));
  caller->stats->setAvgInterarrivalTime(caller->interarrival_time);

  caller->stats->setLastPacketTime(now_tv.tv_sec + now_tv.tv_usec/1000000.0);
  if (caller->stats->getFirstPacketTime() == 0) {
    caller->stats->setFirstPacketTime(caller->stats->getLastPacketTime());
    caller->interarrival_time = 0; //reset to zero
    //if the server doesnt start at TS=0, but later, fix the offset....
    caller->ts_offset = packet->ph.ph_ts; //FIXME: problematic, if first packet comes out of order
    caller->stats->setFirstPacketTS(0);	// as packet->ph.ph_ts will be reduced by ts_offset below
    //FIXME: stats->avgInterarrivalTime cannot be reset...		
  }
  packet->ph.ph_ts-= caller->ts_offset; //always adjust TS to first detected offset

  int this_frame_sec = (int)floor((double)packet->ph.ph_ts / caller->es->getMediaTimeScale());
  int highest_ts_sec = (int)floor((double)caller->stats->getHighestPacketTS()
				  / caller->es->getMediaTimeScale());
  if (caller->stats->getHighestPacketTS() < packet->ph.ph_ts) {
    caller->stats->checkForNewPlayoutSec(this_frame_sec, highest_ts_sec);
		
    //dprintf_full("Rtp::rtp_data_handle: set new highest TS: %i\n",packet->ph.ph_ts);
    caller->stats->setHighestPacketTS(packet->ph.ph_ts);
  }

  if (this_frame_sec > (signed)highest_ts_sec - RTP_STAT_SECS) { //not out of range
    if (packet->ph.ph_m)
      caller->stats->incPlayoutSecNetFPS(this_frame_sec);
  }

  //make copy of rtp payload, since rtp pkt will be freed later
  if (caller->session == session) { // normal DATA packet
    caller->stats->incStreamoutSecDataBW(caller->stats->getStreamoutFullSec(),
																				 sizeof(rtp_packet_header) + packet->pd.rtp_pd_data_len);
    caller->stats->incStreamoutSecDataPkts(caller->stats->getStreamoutFullSec());

    if (this_frame_sec > (signed)highest_ts_sec - RTP_STAT_SECS) { //not out of range
      caller->stats->incPlayoutSecDataBW(this_frame_sec,
																				 sizeof(rtp_packet_header) + packet->pd.rtp_pd_data_len);
      caller->stats->incPlayoutSecDataPkts(this_frame_sec);
    }

    data_len = packet->pd.rtp_pd_data_len;
    buf=(char*)malloc(data_len);
    assert(buf != NULL);
    memcpy(buf,packet->pd.rtp_pd_data, data_len);
    rtp_seq = packet->ph.ph_seq;
    dprintf_full("Rtp::rtp_data_handle: received rtp packet seq no %i)\n",rtp_seq);

    if 	(caller->first_rtp_seq < 0) { //initialize
      caller->first_rtp_seq = rtp_seq;
      caller->last_rtp_seq = rtp_seq;
    } else
      if (++caller->last_rtp_seq != rtp_seq) { // FB only on NACK
				if (caller->feedbackEnabled) {
					// send immediate rtcp feedback packet
					dprintf_err("sending NACK for rtp seq %i up to seq %i\n", caller->last_rtp_seq, rtp_seq-1);
					rtp_send_ctrl(caller->session, packet->ph.ph_ts, NULL, RTCP_IMMEDIATE);
				}
				caller->last_rtp_seq = rtp_seq;
      }
  } else // RTX packet
    if ( (caller->rtxEnabled) && (caller->rtx->session == session)) {
      uint16_t* seq_p=NULL;

      caller->stats->incStreamoutSecRtxBW(caller->stats->getStreamoutFullSec(), 
																					sizeof(rtp_packet_header) + packet->pd.rtp_pd_data_len);
      caller->stats->incStreamoutSecRtxPkts(caller->stats->getStreamoutFullSec());

      if (this_frame_sec > (signed)highest_ts_sec - RTP_STAT_SECS) { //not out of range
				caller->stats->incPlayoutSecRtxPkts(this_frame_sec);
				caller->stats->incPlayoutSecRtxBW(this_frame_sec,
																					sizeof(rtp_packet_header) + packet->pd.rtp_pd_data_len);
      }

      rtx_rtp_seq = packet->ph.ph_seq;
      seq_p = (uint16_t*)packet->pd.rtp_pd_data;
      rtp_seq = ntohs(*seq_p);

      dprintf_full("RECV'd retransmitted RTX packet %i (RTX seq no %i) size %i\n",
									 rtp_seq,rtx_rtp_seq,packet->pd.rtp_pd_data_len);
      data_len = packet->pd.rtp_pd_data_len - sizeof(uint16_t);
      buf=(char*)malloc(data_len);
      assert(buf != NULL);
      memcpy(buf,packet->pd.rtp_pd_data + sizeof(uint16_t), data_len);
			
      if 	(caller->first_rtx_rtp_seq < 0) { //initialize
				caller->first_rtx_rtp_seq = rtx_rtp_seq;
				caller->last_rtx_rtp_seq = rtx_rtp_seq;
      } else
				if (++caller->last_rtx_rtp_seq != rtx_rtp_seq) { // FB only on NACK
					if (caller->feedbackEnabled) {
						// send immediate rtcp feedback packet
						dprintf_err("sending NACK for RTX rtp seq %i up to seq %i\n", 
												caller->last_rtx_rtp_seq,rtx_rtp_seq-1);
						rtp_send_ctrl(caller->rtx->session, packet->ph.ph_ts, NULL, RTCP_IMMEDIATE);
					}
					caller->last_rtx_rtp_seq = rtx_rtp_seq;
				}
    } else {
      dprintf_err("FATAL: recv'd data from unknown RTP session!");
      ::exit(1);
    }

	
  //add pkt sorted by TS (and frag_no==rtp_seq), insert with unknown VOP_type, vop_size and prio
  caller->insertToPreQ((u8*)buf, data_len, Frame::NN_VOP, -1, -1,
											 packet->ph.ph_ts, rtp_seq, packet->ph.ph_m, rtp_seq, Rtp::Q_TS);

}

/***************************************************************/
void Rtp::RtpCallback(struct rtp *session, rtp_event * e) {
  Rtp *caller = NULL;
  rtp_packet      *rtp_pkt=NULL;
  rtcp_fb         *fb=NULL;
  rtcp_rr         *rr=NULL;
  NetPacket_t *p=NULL;
  int ts;
  double this_rr_lost;

  if (e->ssrc == rtp_my_ssrc(session))
    return;

  switch (e->type) {
  case RX_RTP:
    //dprintf_full("RX_RTP SSRC = 0x%08x\n", e->ssrc);
    rtp_pkt = (rtp_packet*)e->data;
    caller=(Rtp*)rtp_get_userdata(session);
    if (caller->isInput()) { //only client handles incoming data!
      Rtp::rtp_data_handle(session,e->ssrc, rtp_pkt);
      if (caller->session == session)
				caller->sender_ssrc = e->ssrc; //set sender of normal RTP channel
    } else
      dprintf_err("Rtp::RtpCallback  incoming data on Server-side????  *wonders*\n");
    free(e->data);
    break;
  case RX_SDES:
    //dprintf_full("RX_SDES SSRC = 0x%08x\n", e->ssrc);
    break;
  case RX_BYE:
    dprintf_full("RX_BYE SSRC = 0x%08x\n", e->ssrc);
    break;
  case SOURCE_CREATED:
    //dprintf_full("New source created, SSRC = 0x%08x\n", e->ssrc);
    break;
  case SOURCE_DELETED:
    //		dprintf_full("Source deleted, SSRC = 0x%08x state=%i\n", e->ssrc,caller->getState());
    caller=(Rtp*)rtp_get_userdata(session);
    dprintf_full("Source deleted, SSRC = 0x%08x... state %i\n", e->ssrc,caller->state);
    if ( (caller->state != CLOSED) && (caller->state != CLOSING) 
				 && (caller->state != FORCE_CLOSING) ) {
      if (caller->isOutput())
				caller->setState(STREAMERR);
      else
				caller->setState(STREAMEOF);
      caller->preQ.MUTEX_Q_DATA_AVAIL.release(); //simulate data
    }
    break;
  case RX_SR:
    //dprintf_full("RX_SR SSRC = 0x%08x\n", e->ssrc);
    caller=(Rtp*)rtp_get_userdata(session);
    break;
  case RX_RR:
    rr = (rtcp_rr*)e->data;
    caller=(Rtp*)rtp_get_userdata(session);

    if (caller->session == session) { //only data channel, not RTX
      this_rr_lost = (double)(rr->fract_lost<<8)/(1<<16) * 100;
      if (caller->last_rr_sum_pkts_lost > -1) {
				int sum_iv=rr->last_seq - caller->last_rr_seq_no;
				int lost_iv=rr->total_lost - caller->last_rr_sum_pkts_lost;
				caller->last_rr_lost_fract = (double)lost_iv / (double)sum_iv * 100;
      } else //use total values for first lost_fract
				caller->last_rr_lost_fract = (double)rr->total_lost 
					/ (double)(rr->last_seq - caller->first_rtp_seq) * 100;
      caller->last_rr_sum_pkts_lost = rr->total_lost;
      caller->last_rr_seq_no = rr->last_seq;
			
      dprintf_full("RX_RR SSRC = 0x%08x Delay to last RR: %2.3f secs, "
									 "PacketLoss fract %3.3f%% myFract %3.3f%% total loss: %i pkts of %lli pkts\n",
									 e->ssrc,
									 caller->stats->getStreamoutFloatSec() - caller->last_rr_received,
									 this_rr_lost,
									 caller->last_rr_lost_fract, rr->total_lost, 
									 caller->last_rr_seq_no - caller->first_rtp_seq );

      caller->last_rr_received = caller->stats->getStreamoutFloatSec();

      //grab packet, get TS, push back...
      p = caller->popNetPacket(rr->last_seq);
      if (p) {
				ts = p->rtp_ts;
				caller->pushNetPacket(p);
				dprintf_full("RX_RR SSRC = 0x%08x RR from rtp seq %i (TS: %i)\n",
										 e->ssrc, rr->last_seq, ts);
      } else 
				dprintf_full("RX_RR SSRC = 0x%08x from rtp seq %i, which is allready off sendQ\n", 
										 e->ssrc,rr->last_seq);
    }
    break;
  case RX_RR_EMPTY:
    //dprintf_full("RX_RR_EMPTY SSRC = 0x%08x\n", e->ssrc);
    break;
  case RX_RTCP_START:
    //printf("RX_RTCP_START\n");
    break;
  case RX_RTCP_FINISH:
    //printf("RX_RTCP_FINISH\n");
    break;
  case RR_TIMEOUT:
    //dprintf_full("RR_TIMEOUT SSRC = 0x%08x\n", e->ssrc);
    break;
  case RX_APP:
    //dprintf_full("RX_APP SSRC = 0x%08x\n", e->ssrc);
    break;
  case RX_FB:
    //dprintf_full("RX_FB SSRC = 0x%08x\n", e->ssrc);
    fb = (rtcp_fb*)e->data;
    //Rtp::fb_print(session,e->ssrc, fb);
    Rtp::fb_handle(session,e->ssrc, fb);
    free(e->data);
    break;
  default:
    break;
  }
}

/***************************************************************/
bool Rtp::pushNetPacket(Rtp::NetPacket_t *p) {
  int slot;
	
  slot = p->rtp_seq % RTP_SENDQ_PRIME;

  sendQ.MUTEX_Q.lock();
  sendQ.size += p->pl_size;
  sendQ.num_elems++;
  p->next = sendQ.Q[slot];
  sendQ.Q[slot] = p;

  //DEBUG PRINTOUT QUEUE
  /*
    for (int i=0;i<RTP_SENDQ_PRIME;i++) {
    NetPacket_t *d = sendQ.Q[i];
    dprintf_full("sendQ slot %i\n",i);
    while (d != NULL) {
    dprintf_full("     RTP seq %i\n",d->rtp_seq);
    d = d->next;
    }
    }
  */
  sendQ.MUTEX_Q.release();
  return true;
}

/***************************************************************/
Rtp::NetPacket_t *Rtp::popNetPacket(uint16_t rtp_seq) {
  Rtp::NetPacket_t *p=NULL, *pprev=NULL;
  int slot;

  slot = rtp_seq % RTP_SENDQ_PRIME;
  sendQ.MUTEX_Q.lock();
  p = sendQ.Q[slot];
  pprev = NULL;
  while ( (p != NULL) && (p->rtp_seq != rtp_seq) ) {
    pprev = p;
    p = p->next;
  }
  
  if (p) {
    if (pprev)
      pprev->next = p->next;
    else
      sendQ.Q[slot] = p->next;
    p->next = NULL;
    sendQ.size -= p->pl_size;
    sendQ.num_elems--;
  }

  sendQ.MUTEX_Q.release();
  return p;
}

/***************************************************************/
Rtp::NetPacket_t *Rtp::popNetPacketByRTXseq(uint16_t rtx_rtp_seq) {
  Rtp::NetPacket_t *p=NULL, *pprev=NULL;
  int slot;

  sendQ.MUTEX_Q.lock();
  for (slot=0;((slot<RTP_SENDQ_PRIME) && (p==NULL));slot++) {
    p = sendQ.Q[slot];
    pprev = NULL;
    while ( (p != NULL) && (p->rtx_rtp_seq != rtx_rtp_seq) ) {
      pprev = p;
      p = p->next;
    }
			
    if (p) {
      if (pprev)
				pprev->next = p->next;
      else
				sendQ.Q[slot] = p->next;
      p->next = NULL;
      sendQ.size -= p->pl_size;
      sendQ.num_elems--;
    }
    pprev = NULL;
  }
  sendQ.MUTEX_Q.release();
  return p;
}

/***************************************************************/
void Rtp::sendQupdate(long streamout_sec) {
  Rtp::NetPacket_t *p=NULL, *pprev=NULL;
  int slot;
  long ssec_ticks, threshold;

  ssec_ticks = es->getMediaTimeScale() * streamout_sec;
  threshold = (long)(ssec_ticks - ((u64)ceil(stats->getPrefetchedSecs()) * es->getMediaTimeScale()));

  if ((streamout_sec >= 0) && (threshold < 0)) { //not "flushAll" and initial prefetching
		dprintf_full("Rtp::sendQupdate(ssec %li): ignoring since initial prefetching (threshold TS %li)\n", streamout_sec,threshold);
    return;
	}
	
  dprintf_full("Rtp::sendQupdate(ssec %li): drop all older than TS %li\n",streamout_sec, threshold);
	
  sendQ.MUTEX_Q.lock();
  for (slot=0; slot<RTP_SENDQ_PRIME;slot++) {
    p = sendQ.Q[slot];
    pprev = NULL;
    while (p != NULL)  {
      if ( (streamout_sec == -1) || //-1 == drop all
					 (p->rtp_ts + (u64)ceil(stats->getPrefetchedSecs()) * es->getMediaTimeScale() 
						< (unsigned)ssec_ticks) ||
					 (p->acked) ) {
				//dprintf_full("NOW TS: %li Frame seq %i ts %i:%i is outdated or acked... now dropped\n",
				//				 streamout_sec*es->getMediaTimeScale(),p->rtp_seq, p->rtp_ts,p->frag_no);
				sendQ.size -= p->pl_size;
				sendQ.num_elems--;
					
				if (pprev)
					pprev->next = p->next;
				else
					sendQ.Q[slot]=p->next;

				assert(p->payload != NULL);
				delete p->payload;
				p->payload=NULL;
				delete p;p=NULL;

				if (pprev)
					p = pprev; //to continue while
				else
					p = sendQ.Q[slot];
      }
				
      if (sendQ.Q[slot] == NULL)
				break;

      pprev = p;
      p = p->next;
    }
  }
  sendQ.MUTEX_Q.release();
}


/***************************************************************/
void Rtp::preQupdate() {
  Rtp::NetPacket_t *p=NULL;
  int act_seq = -1;
  int prev_seq = -1;
  int ret;

  if (!rtxEnabled) {
    dprintf_full("Rtp::preQupdate: no rtxEnabled... ignoring\n");
    return;
  }

  dprintf_full("Rtp::preQupdate\n");
  preQ.MUTEX_Q.lock();
  p = preQ.Q;
  while (p != NULL)  {
    act_seq=p->rtp_seq;
		//		dprintf_full("Rtp::preQupdate checks on %i ts %i (prev %i)\n",act_seq,p->rtp_ts,prev_seq);
    if (prev_seq != -1) {
      while (++prev_seq < act_seq) { //missing packet
				dprintf_full("Rtp::preQupdate: requesting %i (remaining: %i, final will be %i), sess %p ssrc %x\n",
										 prev_seq, act_seq-1-prev_seq, act_seq-1, session, sender_ssrc);
				ret=rtp_send_explicit_nack(session, sender_ssrc, prev_seq, stats->getHighestPacketTS(), NULL);
				assert(act_seq > prev_seq);
				//assert(act_seq-1 - prev_seq < 30 && "YOUR SYSTEM IS TOO SLOW! LOST TOO MANY PACKETS IN A ROW!!!");
				if (ret == -1) {
					dprintf_err("Rtp::preQupdate ERROR! rtp session wrong!!!!\n");
					break;
				}
      }
    }
    prev_seq = act_seq;
    p = p->next;
  }
  preQ.MUTEX_Q.release();
}




/***************************************************************/
bool Rtp::insertToPreQ(u8 *payload, long pl_size, Frame::FrameType type, long vop_size, int prio, 
											 u32 rtpTimestamp, u32 fragCounter, bool lastFrag, u32 rtp_seq, QOrder qo) {
  NetPacket_t *p=NULL;
  int diff;
  int rtp_cycle;

  p = new NetPacket_t;
  memset(p,0,sizeof(NetPacket_t));
  p->rtp_ts = rtpTimestamp;
  p->vop_type = type;
  p->frag_no = fragCounter;
  p->last_frag = lastFrag;
  p->payload = payload;
  p->pl_size = pl_size;
  p->vop_size = vop_size;
  p->drop_me = false;
  p->prio = prio;
  p->sent_dtime = p->resent_dtime = 0;
	
  preQ.MUTEX_Q.lock();
  /* to let this work, we assume no packets which are off more than 
   * 64K in seq_no space.... */
  rtp_cycle = (preQ.highest_rtp_seq / RTP_SEQ_MOD);
  if (preQ.highest_rtp_seq != 0) {
    diff=(int)(preQ.highest_rtp_seq % RTP_SEQ_MOD) - (int)(rtp_seq);
    if (diff > 32768) //new cycle
      rtp_cycle++;
    else
      if (diff < -32768) //previous cycle
				rtp_cycle--;
  }
  p->rtp_seq = rtp_seq + rtp_cycle*RTP_SEQ_MOD;
  if (isInput()) {  // clientSide fix for 32bit rtp_seq
    assert(rtp_seq  == fragCounter);
    p->frag_no = 	p->rtp_seq;
  }

  preQ.MUTEX_Q.release();
	
  //now put it onto the PreQ
  return addToPreQ(p,qo);
}


/***************************************************************
 * called by: server-side fb_handle()
 */
bool Rtp::reinsertToPreQ(NetPacket_t *p, uint rtp_seq) {

  if (p!=NULL) {
    int this_frame_sec = (int)floor((double)p->rtp_ts / es->getMediaTimeScale());
    int highest_ts_sec = (int)floor((double)stats->getHighestPacketTS() / es->getMediaTimeScale());

    // too late for this packet ....
    if (this_frame_sec < stats->getPlayoutSec()) {
      dprintf_full("Rtp::reinsertToPreQ too late for TS %i seq %i\n",p->rtp_ts,rtp_seq);
      if (p->resent_counter == 0) { //first RTX, adjust
				if (this_frame_sec > highest_ts_sec - RTP_STAT_SECS) { //in stats range
					if (p->last_frag)	// is a full frame
						stats->decPlayoutSecNetFPS(this_frame_sec);
					//general stats
					stats->decPlayoutSecDataPkts(this_frame_sec);
					stats->decPlayoutSecDataBW(this_frame_sec, p->pl_size + sizeof(rtp_packet_header));
				}
      }
      return false;
    }

    // when we recv this NACK, it might be for a previous streamout sec,
    // so fix up this previous streamout sec statistic with more nacked_bw
    int nack_sec = (int)floor(p->sent_dtime - stats->getFirstPacketTime());
    if (nack_sec >= MAX(0,(int)stats->getStreamoutFullSec() - RTP_STAT_SECS)) // within stat array
      stats->incStreamoutSecNackedBW(nack_sec, sizeof(rtp_packet_header) + p->pl_size);

    if (this_frame_sec > (signed)highest_ts_sec - RTP_STAT_SECS) //within stat array range
      stats->incPlayoutSecNackedBW(this_frame_sec, sizeof(rtp_packet_header) + p->pl_size);

    p->resent_counter++;
    dprintf_full("     adding to RESEND Q for %i%s time: seq %i ts %i:%i  prio %i\n",
								 p->resent_counter, p->resent_counter==1?"st":p->resent_counter==2?"nd":
								 p->resent_counter==3?"rd":"th",
								 p->rtp_seq, p->rtp_ts, p->frag_no,p->prio);
    return addToPreQ(p,Q_TS,RTP_REINSERT); //this will add to front, but keep RTX order
    //		return false; //DISABLE RTX
		
  }
  else
    dprintf_err("      resend of rtp_seq %i too late... already off sendQ!\n",rtp_seq);

  return false;
}


/***************************************************************/
/* @param qo: the QueueOrder 
 *     for the server: Q_TAIL for normal read from disk,
 *                    Q_HEAD for retransmissions, to put them to the immediate front;
 *    for the client: Q_SEQ in RtpHandler, if new data arrives, insert with RTP seqNo,
 *                    Q_SEQ in getFrame, if too many packets where read, for pushback;
 *  now, everything is done via Q_TS, which is more fair to RTX
 * @param reinsert: is used in Rtp::getFrame if there was popped too far in the Q...
 *                  or if a NACK'd packed is going to be added to the Q
 */
bool Rtp::addToPreQ(NetPacket_t *p, QOrder qo, bool reinsert) {
  NetPacket_t *pprev=NULL, *pold=NULL;

  preQ.MUTEX_Q.lock();
  p->next = NULL;
  if (reinsert)
    dprintf_full("Rtp::addtoPreQ reinsert!\n");
  dprintf_full("Rtp::addToPreQ (stored: %i full frames) packet seq %i ts %i:%i size %i (%s) type: %s prio: %i\n",
							 preQ.num_full_frames,p->rtp_seq,p->rtp_ts,
							 p->frag_no,p->pl_size,!p->last_frag?"intermediate":"last",
							 Frame::VOPTypeToChar(p->vop_type), p->prio);

  dprintf_full("          PreQ has %i packets and %li bytes (avg bw: %i Kbit/sec)\n",
							 preQ.num_elems,preQ.size,preQ.avg_bw*8/1024);
  dprintf_full("          now its %2.3f, PreQ is at %2.3f secs,  ClPlay sec is %2.3f, so it's %2.3f secs ahead\n",
							 stats->getStreamoutFloatSec(),(float)preQ.highest_ts/es->getMediaTimeScale(),
							 stats->getPlayoutSec(), stats->getBufAheadSec());
	
  if (!writeToNetwork && !reinsert)
    dprintf_full("          PreQ: last frame arrival@%6.3f secs, interarrival time: %i msecs\n",
								 stats->getLastPacketTime() - stats->getFirstPacketTime(), interarrival_time);

  //for convenience, resends are always enqueued, even if the preQ is full!
  //all others are bound to preQ MAX size
  if (!reinsert)
    while (preQ.size+p->pl_size > preQ.max_size) {
      preQ.MUTEX_Q_FULL.lock();
      preQ.MUTEX_Q.release();

      if (this->isInput())
				dprintf_full("Rtp::addToPreQ: preQ BUFFER FULL SO BLOCKED! ... we also might lose arriving packets!\n");
      //wait until run thread has sent some data
			msleep(200);
			//      preQ.MUTEX_Q_FULL.lock();
      preQ.MUTEX_Q_FULL.release();
      preQ.MUTEX_Q.lock();
    }

  preQ.isDirty=true;
  preQ.size += p->pl_size;
  preQ.num_elems++;
  /* add to tail of preQ */
  if (preQ.Q == NULL) {
    preQ.Q = preQ.Q_tail = p;
    //dprintf_full("preQ was NULL, now DATA AVAIL!\n");
    preQ.MUTEX_Q_DATA_AVAIL.release();
  }  else {
    if (qo == Q_HEAD) {
      p->next = preQ.Q;
      preQ.Q = p;
    } else 
      if (qo == Q_TAIL) { //add to tail
				preQ.Q_tail->next = p;
				preQ.Q_tail = p;
      } else
				if (qo == Q_TS) { //insert according to timestamp
					pprev = pold = preQ.Q;
				
					//also frag_no is a key for sorting!
					//SERVER: frag_no is set correctly, CLIENT: frag_no is the RTP_seq, also monotonic!
					while((pold!=NULL) && (pold->rtp_ts < p->rtp_ts)) {
						pprev = pold;
						pold = pold->next;
					}
					while ((pold!=NULL) && 
								 (pold->rtp_ts == p->rtp_ts) && 
								 (pold->frag_no < p->frag_no)) {
						pprev = pold;
						pold = pold->next;
					}

					if (pold!=NULL) {
						if (pold->rtp_ts > p->rtp_ts) {
							p->next = pold;
							if (pold == preQ.Q) {
								preQ.Q = p;
							} else
								pprev->next = p;
						} else						
							if (pold->frag_no == p->frag_no) {
								dprintf_err("MULTIPLE RECEPTION of rtp_seq %i TS %i:%i\n",p->rtp_seq,p->rtp_ts,p->frag_no);
								delete p->payload;p->payload=NULL;
								delete p;p=NULL;
								preQ.MUTEX_Q.release();
								return false;
							}	else 
								if (pold->frag_no > p->frag_no) {
									p->next = pold;
									if (pold == preQ.Q)
										preQ.Q = p;
									else
										pprev->next = p;
								} else {
									if (pold == preQ.Q_tail) {
										preQ.Q_tail->next = p;
										preQ.Q_tail = p;
									} else {
										if (pprev == pold) { //first elem
											preQ.Q=p;
											p->next = pold;
										} else {
											p->next=pprev->next;
											pprev->next=p;
										}
									}
								}
					} else {//add to tail
						//preQ.Q_tail->next = p;  //redundant
						pprev->next = p;
						preQ.Q_tail = p;
					}
				} else
					if (qo == Q_SEQ) { //insert according to RTP sequence number
						pprev = pold = preQ.Q;
						while((pold!=NULL) && (pold->rtp_seq < p->rtp_seq)) {
							pprev = pold;
							pold = pold->next;
						}
						if (pold!=NULL) { //insert in between
							if (pold->rtp_seq == p->rtp_seq) {
								dprintf_err("MULTIPLE RECEPTION of rtp_seq %i TS %i:%i\n",p->rtp_seq,p->rtp_ts,p->frag_no);
								delete p->payload;p->payload=NULL;
								delete p;p=NULL;
								preQ.MUTEX_Q.release();
								return false;
							}	else {
								p->next = pold;
								if (preQ.Q == pold)
									preQ.Q = p;
								else
									pprev->next = p;
							}
						} else { //add to tail
							preQ.Q_tail->next = p;
							preQ.Q_tail = p;
						}
					}
  }

  /*  VITOOKI_DEBUG PRINT OUTPUT */
	/*
    NetPacket_t *dummy=p;
    NetPacket_t *pr;
    int64_t monotonic_seq;
    dprintf_full("preQ (%li bytes, %i frms, %i KBps) elem: ",
    preQ.size,preQ.num_elems,(int)(preQ.avg_bw*8/1024.0));
    //  printf(" %i:%i#%i\n",p->rtp_seq,p->rtp_ts,p->frag_no);
    p=pr=preQ.Q;
    monotonic_seq=-1;
    while (p != NULL) {
    printf(" %i:%i#%i(%i)",p->rtp_seq, p->rtp_ts, p->frag_no, p->pl_size);
    fflush(stdout);
    //		if (!writeToNetwork)
    //			assert(monotonic_seq < p->rtp_seq);
    monotonic_seq=p->rtp_seq;
    pr=p;
    p=p->next;
    }
    printf("\n");
    p=dummy;
    assert(preQ.Q_tail == pr);
    assert(pr->next == NULL);
	*/

  if (!reinsert && (p->resent_counter == 0)) {
    if (preQ.highest_ts < p->rtp_ts) {
      if (preQ.highest_ts == 0) { //initial jump size (eg. 3000==30fps or 3600=25fps)
				//FIXME: dirty hack to determine FPS even if one frame lost
				if(allowVopTimeDetection) {
					if (p->rtp_ts == 7200)
						es->setVOPTimeIncrement(3600);
					else if (p->rtp_ts == 6000)
						es->setVOPTimeIncrement(3000);
					else
						es->setVOPTimeIncrement(p->rtp_ts);
						//					es->setVOPTimeIncrement(90000); //MP3
				}
			}
      preQ.highest_ts = p->rtp_ts;
    }
    if (preQ.highest_rtp_seq < p->rtp_seq) 
      preQ.highest_rtp_seq = p->rtp_seq;

    if (p->last_frag) {
      preQ.num_full_frames++;
      if (preQ.num_full_frames > 0)
				preQ.avg_bw = (int)(preQ.size/preQ.num_full_frames * 
														((double)es->getMediaTimeScale())/((double)es->getVOPTimeIncrement()));
      else
				preQ.avg_bw = 0;
    }
  }

  preQ.MUTEX_Q.release();
  //dprintf_full("insertQ unlocked\n");

  return true;
}

/***************************************************************/
int Rtp::getNextPreQ_TS() {
  int ts;

  preQ.MUTEX_Q.lock();
  if (preQ.Q_tail != NULL)
    ts = preQ.Q_tail->rtp_ts;
  else
    ts = -1;
  preQ.MUTEX_Q.release();
	
  return ts;
}

/***************************************************************/
Rtp::NetPacket_t *Rtp::popPreQ(bool blocking) {
  NetPacket_t *p=NULL;
  int counter = 0;

	//dprintf_full("Rtp::popPreQ %s\n",blocking?"blocking":"non-blocking");

  //FIXME: WHY !firstframe??  	if (!firstFrame && blocking && (state != STREAMEOF))
  if (blocking && (state != STREAMEOF))
    preQ.MUTEX_Q_DATA_AVAIL.lock(); //wait for data
	//dprintf_full("Rtp::popPreQ data avail\n");

  //	preQ.MUTEX_Q.lock();
  if ((state == FORCE_CLOSING) || (state == CLOSED)) {
    //	preQ.MUTEX_Q.release();
    if (writeToNetwork)
      preQ.MUTEX_Q_DONE.release();
    dprintf_full("Rtp::popPreQ is quitting because of %s\n",(state==CLOSED?"CLOSED":"FORCE_CLOSING"));
    return NULL;
  }
  //	preQ.MUTEX_Q.release();


  preQ.MUTEX_Q.lock();
  if ((state!=OPEN) && blocking && (preQ.Q == NULL)) { // if other thread is closing...
    dprintf_full("Rtp::popPreQ was blocking, but bailing out! empty Q and state %i!\n",state);
    preQ.MUTEX_Q.release();
    return NULL;
  }
  if (preQ.Q == NULL) {
    dprintf_full("erroneously recv'd preQ.MUTEX_Q unlock.... no data after unpausing and state %i\n",state);
    preQ.MUTEX_Q.release();
    return (Rtp::NetPacket_t *)-1; //FIXME: THIS IS A HACK!!!!
  }

  while (preQ.Q == NULL) {
    dprintf_full("Rtp::popPreQ has empty Q! (%ith try)!\n",counter);
    msleep(100);
    if (counter++ > 10) {
      preQ.MUTEX_Q.release();
      dprintf_full("Rtp::popPreQ is quitting because of empty Q and state %i!\n",state);
      if (firstFrame)     //this is needed for a special error case 
				firstFrame=false; //during prefetching!
      return NULL;
    }
  }
  p=preQ.Q;
  preQ.Q = p->next;
  p->next = NULL;
  preQ.size -= p->pl_size;
  preQ.num_elems--;
  if ( (p->last_frag) &&  (p->resent_counter == 0) )
    preQ.num_full_frames--;
  if (preQ.num_elems == 0) {
    //			assert(preQ.num_full_frames == 0);
    preQ.Q_tail = NULL;
    preQ.avg_bw = 0;
  } else
    preQ.avg_bw = (int)(preQ.size/preQ.num_elems * ((double)es->getMediaTimeScale())/((double)es->getVOPTimeIncrement()));
			
  preQ.MUTEX_Q_FULL.release();
  if (preQ.Q != NULL) { //still data available
    preQ.MUTEX_Q_DATA_AVAIL.release();
  }
  preQ.MUTEX_Q.release();
  return p;
}

/***************************************************************/
void Rtp::run() {
  dprintf_full("Rtp::run  (ports %i:%i) is starting the according RTP thread\n",localPort,remotePort);

#ifndef QT_THREAD_SUPPORT
	int ret = setRoundRobinScheduling(2);
	if (ret == 0) {
		dprintf_full("Rtp::sendThread switched to RT-RR SCHEDULING!\n");
	} else {
		dprintf_full("Rtp::sendThread couldnt change to RT-RR SCHEDULING! (got root?) : err %i\n",ret);
	}
#endif

  if (writeToNetwork) //Sender
    sendThread();
  else
    readThread();
}


/***************************************************************/
void Rtp::readThread() {
  struct timeval now_tv;
  double now_dtime;
  double paused_dtime=0.0;
  IO::State myState;
  bool new_streamout_sec;

  dprintf_full("Rtp::ReadThread started\n");
  dprintf_full("RTP_PREFETCH_SECS: %i  Avg BW: %i kbit/sec  BufferSize: %li KByte\n",
							 RTP_PREFETCH_SECS, es->getAvgBandwidth()/1024, preQ.max_size/1024);
  //is enough buffer avail for selected prefetching?
  if (preQ.max_size < RTP_PREFETCH_SECS * es->getAvgBandwidth()/8) {
    dprintf_err("WARNING: maybe not enough buffer to fit client prefetching!!!!\n");
    //::exit(1);
  }

  while(1) {
    myState = state; // disable race conds without large locking...

    if ((myState == CLOSING) || (myState == FORCE_CLOSING)) {
      dprintf_full("Rtp::readThread DONE!\n");
      preQ.MUTEX_Q_DONE.release(); 
      return;
    }

    gettimeofday(&now_tv,NULL);
    now_dtime = now_tv.tv_sec + now_tv.tv_usec/1000000.0;
    dprintf_full("Rtp::readThread() loop now_dtime %lf\n",now_dtime);
    if (!threadWasPaused && (myState == PAUSED)) {
      dprintf_full("Rtp::readThread set to PAUSED! waiting for PLAY again...\n");
      threadWasPaused=true;
      paused_dtime = now_dtime;
    }
    if ( (threadWasPaused) && (myState != PAUSED)) {
      threadWasPaused=false;
      dprintf_full("Rtp::readThread returns from paused to OPEN after %6.3f secs! adjusting first_packet_dtime!\n",
									 now_dtime - paused_dtime);
      if (stats->getFirstPacketTime() != 0.0) { // PAUSED at very beginning
				//set sendout base timing right
				stats->setLastPacketTime(now_dtime);
				stats->setFirstPacketTime(stats->getFirstPacketTime() + (now_dtime - paused_dtime));
      }
    }

    if (myState != PAUSED) { //normal code in OPEN or MUTED or PREFETCHING
      new_streamout_sec = stats->checkForNewStreamoutSec(now_dtime);

      if (new_streamout_sec) {
				dprintf_full("Rtp::readThread: NEW streamout %2.3f "
										 " ClPlaySec %2.3f ClBuf ahead %2.3f\n",
										 stats->getStreamoutFloatSec(),
										 stats->getPlayoutSec(), stats->getBufAheadSec());

				stats->writeStreamoutSecClientStats(preQ.size);

				//this does (maybe) senseless requests for B-VOPs too!!
				preQupdate(); //re-request still missing packets
      }

      if ((!firstFrame) && (now_dtime - stats->getLastPacketTime() > 1.0)) // TIMEOUT
				//ignore, if state is PAUSED (dont use myState, since the real state might have changed!)
				if ((state == OPEN) || (state == MUTED) || (state == PREFETCHING)) {
					dprintf_err("Rtp::readThread: now its %2.3f no more data arriving! EOF? state=%i buf ahead %2.3f secs\n",stats->getStreamoutFloatSec(),myState,stats->getBufAheadSec());
					setState(STREAMEOF);
					preQ.MUTEX_Q_DATA_AVAIL.release(); //simulate data
				}
    } // !PAUSED

    //always try to get packets, even if PAUSED...
    if ((state != CLOSED) && (state != CLOSING) && (state != FORCE_CLOSING))
      ucl_update(stats->getHighestPacketTS(), UCL_TIMEOUT); //blocking for 20 msecs
  }//while
}


/***************************************************************/
int Rtp::doSend(NetPacket_t *p) {
  int ret=0;
  struct timeval now_tv;
  double now_dtime;
  if(!p)
    return -1;

  gettimeofday(&now_tv,NULL);
  interarrival_time = (int)(1000* (now_tv.tv_sec + now_tv.tv_usec/1000000.0
																	 - stats->getLastPacketTime()));
  stats->setAvgInterarrivalTime(interarrival_time);
  now_dtime = now_tv.tv_sec + now_tv.tv_usec/1000000.0;

  if (interarrival_time > 0) //not first frame...
    dprintf_full("Rtp::doSend: last sent: %3.3f secs, now %3.3f secs, interarrival time: %i msecs\n",
								 stats->getLastPacketTime() - stats->getFirstPacketTime(), 
								 now_dtime - stats->getFirstPacketTime(), 
								 interarrival_time);
 
  if (stats->getFirstPacketTime() == 0.0) {
    stats->setFirstPacketTime(now_dtime);
    stats->setFirstPacketTS(p->rtp_ts);
  }
  stats->setLastPacketTime(now_dtime);
  if (globalTimer) //adjust not to TS but to packet sendout time
    globalTimer->adjustToTS((int)((now_dtime - stats->getFirstPacketTime()) * es->getMediaTimeScale()),
			    es->getMediaTimeScale());

  int this_frame_sec = (int)floor((double)p->rtp_ts / es->getMediaTimeScale());
  int highest_ts_sec = (int)floor((double)stats->getHighestPacketTS() / es->getMediaTimeScale());
  if (stats->getHighestPacketTS() < p->rtp_ts) { //find highest TS ever sent...
    dprintf_full("Rtp::doSend: set new highest TS: %i\n",p->rtp_ts);
    stats->setHighestPacketTS(p->rtp_ts);

    stats->checkForNewPlayoutSec(this_frame_sec,highest_ts_sec);
  }

  //put packet into sendQ, only if not a retransmission
  //FIRST TIME SEND ##############
  if (p->resent_counter == 0) {
    stats->incStreamoutSecDataBW(stats->getStreamoutSec(), sizeof(rtp_packet_header) + p->pl_size);
    stats->incStreamoutSecDataPkts(stats->getStreamoutSec());

    if (!p->last_frag) //only if there are other packets of this frame
      markFullFramePartially(false,true,p->rtp_ts);
    p->rtp_seq = rtp_get_act_seqno(session);
    p->sent_dtime = now_dtime;
    dprintf_full("data available for sending! (%6i bytes, rtp_seq %6i cts %6i:%i prio %i %s)\n",
								 p->pl_size,p->rtp_seq,p->rtp_ts,p->frag_no,p->prio,Frame::VOPTypeToChar(p->vop_type));

    if (this_frame_sec > highest_ts_sec - RTP_STAT_SECS) { //in stats range
      if (p->last_frag && (p->vop_type != Frame::HEADER_VOP))	// is a full frame
				stats->incPlayoutSecNetFPS(this_frame_sec);
      // general stats on this packet
      stats->incPlayoutSecDataPkts(this_frame_sec);
      stats->incPlayoutSecDataBW(this_frame_sec,p->pl_size + sizeof(rtp_packet_header));
    }

    if 	(first_rtp_seq < 0) //initialize
      first_rtp_seq = p->rtp_seq;
    last_rtp_seq = p->rtp_seq;

    //ret holds the number of sent bytes or -1 on error
    ret=rtp_send_data(session,p->rtp_ts, RTP_PAYLOAD_TYPE, p->last_frag, 0, NULL,
											(uint8_t *) p->payload,p->pl_size, NULL, 0, 0);
  } else { // RETRANSMISSION ###############
    if (rtxEnabled) {
      uint8_t *buf=NULL;
      uint16_t *seq_p=NULL;

      stats->incStreamoutSecRtxBW(stats->getStreamoutSec(), sizeof(rtp_packet_header) + p->pl_size);
      stats->incStreamoutSecRtxPkts(stats->getStreamoutSec());
      stats->incPlayoutSecRtxPkts(this_frame_sec);
      stats->incPlayoutSecRtxBW(this_frame_sec,p->pl_size + sizeof(rtp_packet_header));

      p->rtx_rtp_seq = rtp_get_act_seqno(rtx->session);
      p->resent_dtime = now_dtime;
      dprintf_full("RESENT packet seq %i rtx_seq %i ts %i:%i size %6i bytes\n",
									 p->rtp_seq,p->rtx_rtp_seq, p->rtp_ts,p->frag_no,p->pl_size);
      // add original RTP seq_no to front of payload
      // according to draft-ietf-avt-rtp-retransmission-04
      buf=(uint8_t*)malloc(p->pl_size + sizeof(uint16_t));
      assert(buf!=NULL);
      seq_p = (uint16_t*)buf;
      *seq_p = htons(p->rtp_seq); //store real rtp seqNo
      assert(p);
      assert(p->payload);
      memcpy(buf+sizeof(uint16_t),(uint8_t *) p->payload,p->pl_size);

      if 	(first_rtx_rtp_seq < 0) //initialize
				first_rtx_rtp_seq = p->rtp_seq;
      last_rtx_rtp_seq = p->rtp_seq;

      //ret holds the number of sent bytes or -1 on error
      ret=rtp_send_data(rtx->session,p->rtp_ts, RTP_PAYLOAD_TYPE, p->last_frag, 0, NULL,
												buf,p->pl_size+sizeof(uint16_t), NULL, 0, 0);
      free(buf);
    } else
      dprintf_err("no RTX initialized... NO resend of packet seq %i ts %i:%i\n",
									p->rtp_seq,p->rtp_ts,p->frag_no);
  }
	
  if (ret == -1) // problem with sending data
    close(true);
  return ret;
}

/***************************************************************/
void Rtp::sendThread() {
  NetPacket_t *p=NULL;
  struct timeval timeout;
  struct timeval now_tv;
  double now_dtime=0, now2_dtime=0;
  double byte_msec_fact;
  bool new_streamout_sec;
  int act_wait=0;
  int net_bw;
	int inc_in_a_row = 0;
	int dec_in_a_row = 0;
  int bke=15; // estimation of needed time for bookkeeping
  int awe=10; // estimation of needed time for avg. wait time
  IO::State myState;
	int sent_pkts = 0;

#ifdef NETBENCH
	int NETBENCH_BASE_streamBW = stats->getStreamBW();
	int NETBENCH_lastSec = 0;
	int NETBENCH_highSec = 0;
	bool NETBENCH_go_down=true;
#endif

  //this is for gdb'd client or QuickTime to come up receiving!
  sleep(RTP_SENDER_PRESLEEP);
	//stream_bw is set in setESInfo()!!!
	//	setESInfo(es);

  net_bw_before_superdrain = -1;
  //	net_bw = (int)(stats->getStreamBW() * (1+RTP_BUF_FILL_FACT)); //initially assume max. avail BW (optimum)
  net_bw = stats->getStreamBW();

  byte_msec_fact = 1000.0 / net_bw;
  dprintf_full("Rtp::sendThread started with avgBW %i streamBW %i netBW %i KBps\n",
							 es->getAvgBandwidth() / 1024, stats->getStreamBW()/128, net_bw/128);

  while (1) {
    new_streamout_sec = false;
    myState = state; //prevent race conds and avoid large locking

    gettimeofday(&now_tv,NULL);
    now_dtime = now_tv.tv_sec + now_tv.tv_usec/1000000.0;

    if (!threadWasPaused && (myState == PAUSED)) {
      dprintf_full("Rtp::sendThread set to PAUSED! waiting for PLAY again...\n");
      threadWasPaused=true;
    }
    if (threadWasPaused)
      if (myState != PAUSED) {
				threadWasPaused=false;
				dprintf_full("Rtp::sendThread returns from paused to OPEN! adjusting first_packet_dtime!\n");
				if (stats->getFirstPacketTime() != 0) { // PAUSED not at very beginning
					//set sendout base timing right
					stats->setFirstPacketTime(stats->getFirstPacketTime() + now_dtime - stats->getLastPacketTime());
					stats->setLastPacketTime(now_dtime);
				}
      }	else { //keep alive
				stats->incStreamoutSecNumBKeepChecks(stats->getStreamoutSec());
				// does all rtp, rtcp updates & callbacks, blocking 20msec
				ucl_update(stats->getHighestPacketTS(), UCL_TIMEOUT);
      }

    if (myState != PAUSED) { //normal code in OPEN or MUTED, but also CLOSE et al.

			
      new_streamout_sec = stats->checkForNewStreamoutSec(now_dtime);

      if (new_streamout_sec) {
				float pkt_loss;
				int oldSec = stats->getStreamoutFullSec()-2;

				//SUPERHACK TO FORCE ADAPTATION
#ifndef NETBENCH
				if (stats->getStreamoutFullSec() >= 2) 
#endif
					{
					//ADJUST AVAIL BW
					if  (stats->getStreamoutFullSec() % 2) { //every two secs we see the results of adjustment
						if (feedbackEnabled) {
							//adjust bw to network state of 2 secs ago
							pkt_loss = stats->getStreamoutSecLossPercent(oldSec);
							//dprintf_full("Rtp::sendThread calcing pkt_loss from rtp feedback: data %i rtx %i nack %i == %3.3f pkt loss\n",stats->getStreamoutSecDataBW(oldSec),stats->getStreamoutSecRtxBW(oldSec),stats->getStreamoutSecNackedBW(oldSec),pkt_loss );
						} else {
							pkt_loss = last_rr_lost_fract; //use RTCP Receiver report
							last_rr_lost_fract = 0; //reset, 0 means later increase!
							dprintf_full("Rtp::sendThread calcing pkt_loss from RTCP RX_RR: %3.3f pkt loss\n",pkt_loss );
						}

#ifndef NO_ADJUST

						if (pkt_loss > RTP_BW_DEC_LOSS) { //high packet loss!
							dec_in_a_row++;
#ifdef NETBENCH
							net_bw = (int)((double)net_bw / 100 * (100 - pkt_loss)); //greedy adjustment
#else
							net_bw = (int)((float)net_bw * RTP_BW_DEC_FACT);

							//TCP-friendly sawtooth: multiplicative decrease
							//net_bw = (int)((float)net_bw - (dec_in_a_row * 7 *  pLayer->getMaxPacketPayloadSize()));
#endif
							if (net_bw < RTP_BW_MINIMUM)
								net_bw = RTP_BW_MINIMUM;
							inc_in_a_row = 0; //reset bw inc counter
						} else 
							dec_in_a_row = 0;
							//dont be greedy: only allow a little more netBW than streamBW...
							//at least enough to fill up empty buffers....
							if (pkt_loss < RTP_BW_INC_LOSS) { // virtually no packet loss!
								if ( ((stats->getStreamoutSecClientBufFill(stats->getStreamoutFullSec()-1)*100)
											/ stats->getClientPreQMaxSize() ) > RTP_BUF_SUPERHIGHWATER) { 
									//buffer fill exceeds SUPERHIGHWATER: this shouldnt happen!
									dprintf_full("Rtp::sendThread: superdrain mode: ACTIVELY reducing net BW"
															 " to drain overfull buffers (%i %%)!\n",
															 (stats->getStreamoutSecClientBufFill(stats->getStreamoutFullSec()
																																		-1)*100) 
															 / stats->getClientPreQMaxSize() );
									if (net_bw_before_superdrain == -1)
										net_bw_before_superdrain=net_bw;
									net_bw = RTP_BW_MINIMUM; //  _REALLY_ reduce BW
								} else {
									if ( ( ((stats->getStreamoutSecClientBufFill(stats->getStreamoutFullSec()-1)*100)
													/ stats->getClientPreQMaxSize() ) < RTP_BUF_SUPERHIGHWATER_LOW) && 
											 (net_bw_before_superdrain != -1) ) { 
										//buffer falls back from SUPER to normal HIGHWATER... reset old net_bw
										net_bw = net_bw_before_superdrain; //reset old state
										net_bw_before_superdrain = -1;
									} else { //slowly upscale

										dprintf_full("Rtp::sendThread: STEP UP: old netBW: %i getbaseBW %i\n",net_bw/128, 
																 stats->getBaseStreamBW() );

										inc_in_a_row++;
										int new_bw = (int)((float)net_bw  // * RTP_BW_INC_FACT);
																			 //+ (RTP_BW_INC_PACKETS * pLayer->getMaxPacketPayloadSize() ));
																			 // + (RTP_BW_INC_PACKETS * pLayer->getMaxPacketPayloadSize() / ((inc_in_a_row % 5)+1) ) );
										 + (inc_in_a_row  *  pLayer->getMaxPacketPayloadSize()  ) );

										net_bw = new_bw;
										//force to MAX BaseStreamBW, wich is not affected by switching
										if (new_bw < stats->getBaseStreamBW() )// * (1+RTP_BUF_FILL_FACT))//a little more...
											net_bw = new_bw;
										else //exactly find top level
											net_bw = stats->getBaseStreamBW();										
										
									}
								}
							} else { //plateau: keep netBW steady
							}
#endif

						//SUPERHACK TO FORCE ADAPTATION to eg. 30%
#ifdef NETBENCH
						//		net_bw = (int)(stats->getStreamBW() * 0.85);

						//implements oscillating bandwidth adjustments
						//stepwise go down and then up again
#define NETBENCH_MIN 0.4
						if (NETBENCH_go_down)
							net_bw = (int)( NETBENCH_BASE_streamBW -            //degrade BW step by step 
															NETBENCH_lastSec * (pLayer->getMaxPacketPayloadSize()*2)); 
						else
							net_bw = (int)( NETBENCH_BASE_streamBW -            //upgrade BW step by step 
															(NETBENCH_highSec - NETBENCH_lastSec) * (pLayer->getMaxPacketPayloadSize()*2));
						NETBENCH_lastSec++;

						if (net_bw > NETBENCH_BASE_streamBW  && !NETBENCH_go_down) { //top reached
							dprintf_full("net_bw > NETBENCH_BASE_stream:  top reached\n");
							NETBENCH_go_down = true;
							NETBENCH_lastSec = 0;
							NETBENCH_highSec = 0;
						} else
							if (net_bw < NETBENCH_BASE_streamBW * NETBENCH_MIN  && NETBENCH_go_down) { //bottom reached
								dprintf_full("net_bw < NETBENCH_BASE_streamBW * NETBENCH_MIN:  bottom reached\n");
								NETBENCH_go_down = false;
								NETBENCH_lastSec = 1;
							}
							else {
								if (NETBENCH_go_down)
									NETBENCH_highSec++;
							}
#endif

						byte_msec_fact = 1000.0 / net_bw;

						calcAdaptPreQValues(net_bw);
						
						
						dprintf_full("Rtp::sendThread: NEW streamout %2.3f ADJUST BW: really measured netBW %i"
												 " (after %2.2f %% pkt loss)  assumedBW %i, adapt %i %% to send %2.2f secs"
												 " (ClBuffer %i KBits of %i (%i %%): %s)\n",
												 stats->getStreamoutFloatSec(),
												 (stats->getStreamoutSecDataBW(oldSec)
													+ stats->getStreamoutSecRtxBW(oldSec)
													- stats->getStreamoutSecNackedBW(oldSec)) /128,
												 pkt_loss, 
												 net_bw*8/1024,
												 stats->getAdaptRate(), 
												 stats->getAdaptSecs(),
												 stats->getStreamoutSecClientBufFill(stats->getStreamoutFullSec()-1)/128,
												 stats->getClientPreQMaxSize()/128,
												 (stats->getStreamoutSecClientBufFill(stats->getStreamoutFullSec()-1)*100)
												 /stats->getClientPreQMaxSize(),
												 stats->getAdaptSecs()>1.0?"FILL":stats->getAdaptSecs()==1.0?"STABLE":"DRAIN");
						
					}

					stats->writeStreamoutSecServerStats();

					//delete outdated packets (no RTX feasible any more)
					//still, keep 2 outdated secs, to get better playsec_stat
					sendQupdate(stats->getStreamoutFullSec()-2);
				} // streamout_sec >= 2
				
				stats->setStreamoutSecNetBW(stats->getStreamoutSec(), net_bw);
				dprintf_full("Rtp::sendThread: NEW streamout %2.3f (est.BW %i KBps fact %f)"
										 " ClPlaySec %2.3f ClBuf ahead %2.3f\n",
										 stats->getStreamoutFloatSec(),net_bw*8/1024, byte_msec_fact,
										 stats->getPlayoutSec(), stats->getBufAheadSec());
				
      } //new streamout sec

      adaptPreQ(net_bw); //this will only happen if really necessary!

      //each popPreQ will block until data is avail
      while ( (p = popPreQ()) ) { //actually execute adaptation
				if (p==(Rtp::NetPacket_t *)-1)
					continue; //FIXME: after unpause, the Q was empty! this will force PREFETCHING
				if (p->drop_me || p->partially_dropped || (!rtxEnabled && p->resent_counter > 0)) {
					dprintf_full("Rtp::sendThread ADAPTING (dropping) prio %i VOP_type %s rtpSeq %i TS %i:%i\n",
											 p->prio,Frame::VOPTypeToChar(p->vop_type),p->rtp_seq,
											 p->rtp_ts,p->frag_no);
					if (!p->partially_dropped) //dont search again, work already done
						markFullFramePartially(true,false,p->rtp_ts); //partially_dropped
					int this_frame_sec = (int)floor((double)p->rtp_ts / es->getMediaTimeScale());
					int highest_ts_sec = (int)floor((double)stats->getHighestPacketTS() / es->getMediaTimeScale());
					if (this_frame_sec > highest_ts_sec - RTP_STAT_SECS) { //in stats range
						if (p->resent_counter >= 1) //RTX
							if (p->last_frag)	// is a full frame
								stats->decPlayoutSecNetFPS(this_frame_sec);
					}
					//free everything!
					assert(p->payload != NULL);
					delete p->payload;
					p->payload=NULL;
					delete p;p=NULL;
				} else
					break;
      }
			
      /* what happens, if this is blocked and we get an immediate close?
       * mutexes will be deleted, popPreQ will return and p==NULL,
       * so just return... */
      if ((p==NULL) || (myState == STREAMERR)) {
				myState = state; //get new state after blocking...
				//if ((myState == CLOSED) || (myState == CLOSING) ||(myState == FORCE_CLOSING)) {
				dprintf_full("Rtp::writeThread DONE!\n");
				setState(STREAMEOF);
				preQ.MUTEX_Q_DONE.release(); 
			
				return;
      }

      doSend(p); // also modifies p->rtp_seq!
      pushNetPacket(p);	// put the sent packet on the sendQ
      if (p->last_frag) {
				currentFrameNumber++;
				framesSeen++;
      }
			sent_pkts++;

			/*
			//force fast sending of always TWO packets in a row
			if (sent_pkts % 2) {
				ucl_update(stats->getHighestPacketTS(), 0);
				//busy wait
					double timeout_dtime;
					//this hopefully prevents  UDP pkt loss by the kernel
					// minimum with msleep(), nanosleep() and pthread_cond_timedwait() is 20 msecs !!!
					//minimum with select() is still 10 msecs  :((((				
					gettimeofday(&now_tv,NULL);
					now2_dtime = timeout_dtime = now_tv.tv_sec + now_tv.tv_usec/1000000.0;
					timeout_dtime += ((float)(RTP_MIN_INTERARRIVAL_TIME) / 1000.0);
					while (now2_dtime < timeout_dtime) { // BUSY WAITING LOOP
						gettimeofday(&now_tv,NULL);
						now2_dtime = now_tv.tv_sec + now_tv.tv_usec/1000000.0;
					}

				continue; //jump back to big while(1)
			}
			*/

      gettimeofday(&now_tv,NULL);
      now_dtime = now_tv.tv_sec + now_tv.tv_usec/1000000.0;
			
			
      if ((stats->getStreamoutSecDataBW(stats->getStreamoutSec())
					 +stats->getStreamoutSecRtxBW(stats->getStreamoutSec()) ) 
					> (signed)(net_bw - pLayer->getMaxPacketPayloadSize())) {
				act_wait = (int)(((float)stats->getStreamoutFullSec() + 1.0 
													- (now_dtime - stats->getFirstPacketTime()))*1000) + 2;
				dprintf_full("Rtp::sendThread: ALL DONE FOR THIS SEC: act_wait %i\n",act_wait);
      } else
				act_wait=(int) (stats->getStreamoutFullSec()*1000.0
												+ (int)( (stats->getStreamoutSecDataBW(stats->getStreamoutSec())
																	+stats->getStreamoutSecRtxBW(stats->getStreamoutSec()) )
																 * byte_msec_fact)
												- (int)((now_dtime - stats->getFirstPacketTime())*1000.0));

      if (now_dtime - stats->getLastPacketTime() + MAX(0,act_wait)/1000.0
					< RTP_MIN_INTERARRIVAL_TIME/1000.0) {
				dprintf_full("Rtp::sendThread: act_wait %i (intrtime %3.3f) would break "
										 "MIN_INTERARRIVAL_TIME. increasing...\n",act_wait, 
										 now_dtime - stats->getLastPacketTime() + MAX(0,act_wait)/1000.0);
				act_wait = (int)((now_dtime - stats->getLastPacketTime() 
													+ RTP_MIN_INTERARRIVAL_TIME/1000.0)*1000);
      }
    } // normal code in OPEN or MUTED

    // BOOKKEEPING
    // either a new streamout sec began or we have time left...
    // PAUSE is specially handled above!
    // use this time for useful things like receiving NACKs for RTX requests 
    // and other rtp maintenace
    if ( ((new_streamout_sec) || (act_wait >= bke)) && !threadWasPaused) {
      stats->incStreamoutSecNumBKeepChecks(stats->getStreamoutSec());
      // do all rtp, rtcp updates & callbacks, non-blocking!
      ucl_update(stats->getHighestPacketTS(), 0);
      gettimeofday(&now_tv,NULL);
      now2_dtime = now_tv.tv_sec + now_tv.tv_usec/1000000.0;
      bke = (int) (bke * 0.7 + ((now2_dtime - now_dtime)*1000 * 0.3));
      dprintf_full("I spent %i msecs for bookkeeping... new est. bke %i msecs\n",
									 (int) ((now2_dtime - now_dtime)*1000), bke);
    }

    int real_wait;
    int now_wait;
    while(1) {
      gettimeofday(&now_tv,NULL);
      now2_dtime = now_tv.tv_sec + now_tv.tv_usec/1000000.0;
      real_wait = (int)(((now2_dtime - now_dtime)*1000000.0)/1000.0);
      if (real_wait >= act_wait)
				break; //great, the time is over...
      now_wait =  MAX(0,act_wait) - real_wait;
      dprintf_full("Rtp::sendThread entering act_wait %i loop: now_wait: %i\n",act_wait,now_wait);
      /******* SLEEP UNTIL NEXT ROUND  ********/	
      if (now_wait > RTP_BUSY_WAIT_MS) { //if we want to have bad sleep granularity
				//if (1) { //always do select
				//we only wait if we will have to wait longer than the avg
				//by this, we prefer pushing out netPackets than being late...
				timeout.tv_sec  = 0;
				timeout.tv_usec = (now_wait>0)?now_wait * 1000:1000; //at least wait 1 msec
				select(0,NULL,NULL,NULL,&timeout);
				
      } else // do busy waiting
				{
					double timeout_dtime;
					//this hopefully prevents  UDP pkt loss by the kernel
					// minimum with msleep(), nanosleep() and pthread_cond_timedwait() is 20 msecs !!!
					//minimum with select() is still 10 msecs  :((((				
					gettimeofday(&now_tv,NULL);
					now2_dtime = timeout_dtime = now_tv.tv_sec + now_tv.tv_usec/1000000.0;
					timeout_dtime += ((float)RTP_BUSY_WAIT_MS / 1000.0);
					while (now2_dtime < timeout_dtime) { // BUSY WAITING LOOP
						gettimeofday(&now_tv,NULL);
						now2_dtime = now_tv.tv_sec + now_tv.tv_usec/1000000.0;
					}
				}
    }

    if (act_wait > 0) // if we had to wait, get avg wait time...
      //awe = (int)(awe * 0.7 + ((now2_dtime - now_dtime)*1000000.0)/1000.0 * 0.3);
      awe = (int)(awe * 0.7 + act_wait * 0.3);

    gettimeofday(&now_tv,NULL);
    now2_dtime = now_tv.tv_sec + now_tv.tv_usec/1000000.0;
    real_wait = (int)(((now2_dtime - now_dtime)*1000000.0)/1000.0);
    dprintf_full("time %2.3f -> %2i:%03i(should be %03i)  'act_wait'ed for %i msec"
								 " (really was %i, awe %i), %6i KBit of %6i\n",
								 stats->getStreamoutFloatSec(),
								 stats->getStreamoutFullSec(),
								 (int)(((now2_dtime - stats->getFirstPacketTime())
												- stats->getStreamoutFullSec())*1000.0),  //now offset
								 (int)((stats->getStreamoutSecDataBW(stats->getStreamoutSec())
												+stats->getStreamoutSecRtxBW(stats->getStreamoutSec()))
											 * byte_msec_fact), //should be
								 act_wait,
								 real_wait,
								 awe,
								 (stats->getStreamoutSecDataBW(stats->getStreamoutSec())
									+stats->getStreamoutSecRtxBW(stats->getStreamoutSec()))/128, 
								 net_bw/128);
  } //while(1)
}


/***************************************************************
 * @param net_bw: what is the actual network capability? normally less :(
 */
void Rtp::calcAdaptPreQValues(int net_bw) {
  CompareNetPackets cmpFnc;
  priority_queue<NetPacket_t*,std::vector<NetPacket_t*>,CompareNetPackets> pq(cmpFnc);


	if (stats->getStreamoutSec() < RTP_PREFETCH_SECS)
		return;
  stats->setAdaptSecs(1.0); //normal adaptation
  if (stats->getBufAheadSec() <  RTP_BUF_LOWWATER) {
    stats->setAdaptSecs(stats->getAdaptSecs() + (float)RTP_BUF_FILL_FACT);  //fill buf  (eg. 1.10)
	}

  if (net_bw_before_superdrain != -1) { //we are in super draining phase
    dprintf_full("Rtp::calcAdaptPreQValues turning off adaptation, totally use buffer, we have sooo much of it...\n");
    real_bw = net_bw; //turn off adaptation, totally use buffer, we have sooo much of it...
  } else {
    real_bw = (int)((float)stats->getStreamBW() * stats->getAdaptSecs());
    if (real_bw < net_bw) //if buf_drain more than net_bw
      real_bw = net_bw; //no need to spoil net bw! :)
  }
  //if (stats->getAdaptRate(0)) //smoother adapt changes
  stats->setAdaptRate(net_bw*100 / real_bw);
  //else
  //	stats->setAdaptRate((int)((float)stats->getAdaptRate()*(1 - 0.3) + (net_bw*100 / real_bw)*0.3) + 1);

  if (stats->getBufAheadSec() >  RTP_BUF_HIGHWATER) {
		if (net_bw <= stats->getStreamBW()) { //so adaptSecs is < 1.0
			stats->setAdaptSecs((float)net_bw / (float)stats->getStreamBW());
			real_bw = (int)((float)stats->getStreamBW() * stats->getAdaptSecs());
			stats->setAdaptRate(net_bw*100 / real_bw);
			dprintf_full("Rtp::calcAdaptPreQValues: superdrain mode: TURNING OFF ADAPTATION since we have so much buffer!\n");
		}
	}

	if	((stats->getBufAheadSec() >  RTP_BUF_LOWWATER) && (stats->getAdaptRate() < (float)RTP_MAX_ADAPT/100.0)) {
    //we have enough buffers to reduce necessary adaptation to B-frames only
    //stats->setAdaptSecs((float)net_bw / ((float)RTP_MAX_ADAPT/100.0 * (float)stats->getStreamBW()));
   stats->setAdaptSecs(stats->getAdaptSecs() - (float)RTP_BUF_DRAIN_FACT);  //drain buf  (eg. 0.70)

    real_bw = (int)((float)stats->getStreamBW() * stats->getAdaptSecs());
    stats->setAdaptRate(net_bw*100 / real_bw);
    dprintf_full("Rtp::calcAdaptPreQValues: superdrain mode: ACTIVELY DRAINING BUFFER to preserve quality!\n");
  }

  dprintf_full("Rtp:calcAdaptPreQValues: keeping %2i%% reach netBW %i KBits of "
							 "streamBW %i KBits for %2.3f secs == realBW %i (ClBuf: %2.3f)\n",
							 stats->getAdaptRate(), net_bw*8/1024, stats->getStreamBW()*8/1024,
							 stats->getAdaptSecs(),real_bw/128, stats->getBufAheadSec());

}



/***************************************************************/
void Rtp::markFullFramePartially(bool dropped, bool sent, u32 ts) {
  NetPacket_t *p=NULL;

  dprintf_full("Rtp::markFullFramePartially%s TS %i\n",dropped?"DROPPED":"SENT",ts);
  preQ.MUTEX_Q.lock();

  p=preQ.Q;
  while (p != NULL) {
    if (p->rtp_ts == ts) {
      assert ( (!p->partially_dropped && !p->partially_sent) || 
							 (p->partially_dropped && !p->partially_sent) || 
							 (!p->partially_dropped && p->partially_sent));

      p->partially_dropped = dropped;
      p->partially_sent = sent;
			//dprintf_full("Rtp::markFullFramePartially%s TS %i:%i\n",dropped?"DROPPED":"SENT",ts,p->frag_no);
    }
    p=p->next;
  }
 
  preQ.MUTEX_Q.release();
  return;
}


/***************************************************************
 * @param net_bw: what is the actual network capability? normally less :(
 */
void Rtp::adaptPreQ(int net_bw) {
  CompareNetPackets cmpFnc;
  priority_queue<NetPacket_t*,std::vector<NetPacket_t*>,CompareNetPackets> pq(cmpFnc);
  int size=0;
  NetPacket_t *p=NULL, *pp=NULL;

  dprintf_full("Rtp:adaptPreQ: keeping %2i%% reach netBW %i KBits of streamBW"
							 " %i KBits for %2.3f secs == realBW %i (ClBuf: %2.3f)\n",
							 stats->getAdaptRate(), net_bw*8/1024, stats->getStreamBW()*8/1024,
							 stats->getAdaptSecs(),real_bw/128, stats->getBufAheadSec());

  //	dprintf_full("Rtp::adaptPreQ waiting for preQ mutex\n");
  preQ.MUTEX_Q.lock();
  //dprintf_full("Rtp::adaptPreQ got preQ mutex\n");
  p=preQ.Q;

  // read in full streamout block and mark it to non-drop
  // this includes all partially dropped or sent frame-packets, I_VOPs, P_VOPS.... and RTX
  // because all of them shall be streamed in the next sec.
  while (p && (size+p->pl_size <= (unsigned)real_bw)) {
    pq.push(p);
    size += p->pl_size;
    p->drop_me = false;
    p=p->next;
  }

  // chop off the packets to meet the net_bw
  // we have to ignore partially sent frames and RTX, we force dropping of partially dropped ones
  while (!pq.empty() && (size > net_bw)) {
    p=pq.top();
    pq.pop();

		//    if (p->prio == IO_NETWORK_HIGHEST_PRIORITY) {
    if (p->prio <= RTP_MIN_PRIO) { //keep at least 10 fps
      //FIXME: what now? switching?
      dprintf_full("Rtp::adaptPreQ  BAILING OUT because below RTP_MIN_PRIO!! (prio %i VOP_type %s TS %i:%i)\n"
									 ,p->prio,Frame::VOPTypeToChar(p->vop_type),p->rtp_ts,p->rtp_seq);
      break;
    }

    if ( ((p->resent_counter == 0) && (!p->partially_sent)) || (p->partially_dropped) ) { 
      //BVOP for first send, so drop it!
      size -= p->pl_size; //dont reduce size for RTX, so we will adapt more to fit RTX in network
      p->drop_me = true;
      dprintf_full("Rtp::adaptPreQ  marking prio %i VOP_type %s TS %i:%i for dropping... partially_dropped %i\n",p->prio,Frame::VOPTypeToChar(p->vop_type),p->rtp_ts,p->rtp_seq,p->partially_dropped);
    } else
      dprintf_full("Rtp::adaptPreQ  NOT marking since its a RTX or partially sent (prio %i VOP_type %s TS %i:%i)\n"
									 ,p->prio,Frame::VOPTypeToChar(p->vop_type),p->rtp_ts,p->rtp_seq);
  }

  // DELETE the rest of FULL FRAME
  // only if B-VOP for first send, 
  if (p && !pq.empty() && (p->prio > RTP_MIN_PRIO)
      && (p->resent_counter == 0) && (!p->partially_sent) ) {
    do {
      pp = pq.top();
      pq.pop();
      if (pp->rtp_ts == p->rtp_ts) { //no checking on RTX is needed, since list is sorted by TS:seqNo 
				pp->drop_me = true;
				dprintf_full("Rtp::adaptPreQ  rest-marking prio %i VOP_type %s TS %i:%i for dropping...\n",pp->prio,Frame::VOPTypeToChar(p->vop_type),pp->rtp_ts,p->rtp_seq);
      } else
				break;
    } while (!pq.empty() );
  }

  preQ.isDirty = false;
  preQ.MUTEX_Q.release();
}



/***************************************************************
 * sets a shallow copy of @param rtx at the ESInfo object */
void Rtp::setRtxInfo(const rtx_info* rt) {
  assert(rt);
  dprintf_full("Rtp::setRtxInfo  \n");
  if(rtx) {
    delete rtx; 
  }
  rtx= new rtx_info(rt);
};
