 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	Rtp.hpp                                                       *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#ifndef _MK_RTP_HPP_
#define _MK_RTP_HPP_
#ifdef WIN32
#pragma warning(disable:4786)
#endif

#include "VMutex.hpp"

#include "IO.hpp"
#include "ESInfo.hpp"
#include "adaptors/GlobalTimer.hpp"
#include "net/PortGenerator.hpp"
#ifdef WINCE
#include "3rdparty/commonucl/rtp.h" // UCL library
#else
#include "commonucl/rtp.h" // UCL library
#endif

class Frame;
class PacketizationLayer;
class SyncLayer;
class TerminalCapabilities;
class Statistics;

extern PortGenerator portGen;

/* how many secs worth of data should be prefetched on startup or if buffers dry out? 
 * Type: SECONDS */
#ifdef LARGE_PREFETCH
#define RTP_PREFETCH_SECS 20
#else
#define RTP_PREFETCH_SECS 8
#endif

/* this is for gdb'd client or QuickTime to come up receiving!
 * Type: SECONDS */
#define RTP_SENDER_PRESLEEP 1


/* General preQ Buffer Settings */
/* preQ is set to max 60 secs of fitting stream_BW of chosen video (eg. 60*500KBit/sec
 * Type: SECONDS */
#define RTP_CLIENT_PreQ_MAX_SECS  60

/* if client buffer has less than eg. 5 secs of video we force higher adaptation to fill buf
 * Type: SECONDS */
#define RTP_BUF_LOWWATER  5

/* if client buffer has more than eg. 30 secs of video we allow lower adaptation and drain buf
 * Type: SECONDS */
#define RTP_BUF_HIGHWATER  25


/**
 *
 * <Class Description goes here>
 *
 *@author Michael Kropfberger and Peter Schojer
 *
 *@version $Id: Rtp.hpp,v 1.30 2006/01/26 11:42:00 mkropfbe Exp $
 *
 *@short <short description>
 *
 */
class Rtp:public IO {
public:
	
  /** creates an RTP object
      @param url: the source/dest for this IO class, deep-copied
      @param remotePrt: the remote port where data is sent to
      @param address: the ip address of the remote host, deep-copied
			@param localPrt: the local port, where we are binding and receiving
      @param es: the ElementaryStream sent, shallow-copied
      @param packetization: specifies the transport mode: can be RFC3016,
                            multiSL or flexmux
      @param writeonly specifies if the RTP connection is for sending
      (writeonly is set to true), of for reading
			@param tc: the TerminalCapabilities, shallow-copied (eg. RTX on/off)
			@param allowVopTimeDetection: if set to true (default), allows rtp to detect and CHANGE the VOPTIMEincrement
			for the ESInfo object
  */
  Rtp(const char *url, int remotePrt, const char *address, int localPrt,
			ESInfo * es, PacketizationLayer * packetization, bool writeOnly,
			const TerminalCapabilities *tc,  GlobalTimer * globalTimer, Statistics *statistics,bool allowVopTimeDetection=true);
  virtual ~ Rtp();

	Statistics *getStatistics() { return stats; };

	void setESInfo(ESInfo * new_es);
	ESInfo* getESInfo() { return es;};
	/** returns a frame if one complete frame is available, otherwise null 
		is returned. This function is typically blocking. Don't use a NULL
		return value to conclude STREAMEOF, always check with getState()!
	*/
  Frame *getFrame();
  
  /** just adds a frame to the send buffer. sending per se is handled in
      the run thread. If the send buffer is full, writeFrame will block.
			@param frm
      @returns the number of packets delivered. Returns 0 on error */
  int writeFrame(Frame * frm, ESInfo *out_es=NULL);

	/** if this is called, the next sent frame will also include a decoder header
	 * this could be used for multicast sessions 
	 */
	void resendHeader() { resendFrameHeader=true;} ;


  /** this thread works on the read buffer, using the estimated avail. 
      bandwidth for send out rate  */
  void run();

  /** opens the IO connection. State is set to OPENING. Depending on the
      underlying QIODevice, a network connection or a file connection is 
      established. When the connection is ready for use, State is OPEN and
      the run thread is started.  */
  bool open();

	State play(double prefetchTime);

	State pause();

	State mute();

  bool close(bool immediate=false);
  
  int getLocalPort() const { return localPort; };

  /** returns preQ buffer fill in percent */
  int getBufferFillLevelInPercent() const { return (int)(preQ.size/((float)preQ.max_size/100.0)); };
  int getBufferFillLevel() const { return getBufferFillLevelInPercent(); };

  /** returns preQ buffer fill in Bytes */
  int getBufferFillLevelInBytes() const { return preQ.size; };

	/** flushes the preQ buffer from starting TimeStamp to end TS */
	int flushBuffer(long from_ts = 0, long to_ts = -1);

	/* is this Rtp object for reading from network? (eg. client side) */
	bool isInput() {return !writeToNetwork; };
	/* is this Rtp object for writing to network? (eg. server side) */
	bool isOutput() {return writeToNetwork; };

  /** you can't delete a remote url :-) will always return true */
  bool destroy() { return true; } 
  const char *getURL() const { return url; };
  /** returns NULL if the session was not opened, Otherwise,
   * a char* containing a %08x hexValue */
  char *getSSRC();

	/** sets a shallow copy of @param rtx at the ESInfo object */
	void setRtxInfo(const rtx_info* rtx);

	const rtx_info* getRtxInfo() const { return rtx;};

	/* flushes the buffer before the given @param frameNumber
	 * be aware that this currentFrameNumber will be reset if the input is changed with setESInfo()
	 * and will be increased with every streamed out frame later on
	 */
	bool setToFrameNumber(u32 frameNumber);

	/* flushes the buffer after the given @param frameNumber
	 * be aware that this endFrameNumber will be reset if the input is changed with setESInfo()
	 */
	bool setEndFrameNumber(u32 frameNumber);


/*******************************************************/
protected:

	struct NetPacket_t; //fwd decl

  static void RtpCallback(struct rtp *session, rtp_event * e);
  static void fb_print(struct rtp *session, uint32_t ssrc, rtcp_fb *fb);
  static void fb_handle(struct rtp *session, uint32_t ssrc, rtcp_fb *fb);
	static void rtp_data_handle(struct rtp *session, uint32_t ssrc, rtp_packet *packet);

	typedef enum {Q_HEAD, Q_TAIL, Q_TS, Q_SEQ} QOrder;
	/**
	 * inserts an SL packet to the preQ in the given order
	 */
  bool insertToPreQ(u8 *payload, long pl_size,  Frame::FrameType type, long vop_size, int prio,
										u32 rtpTimestamp, u32 fragCounter, bool lastFrag, u32 rtp_seq,
										QOrder qo);

	/* used to ignore preQ fill level and for pushback */
#define RTP_REINSERT true
	/**
	 * reinserts a SL packet for retransmission to the head of the preQ
	 */
  bool reinsertToPreQ(NetPacket_t *p, uint rtp_seq);

	/**
	 * adds a NetPacket to the head or tail of the preQ
	 * (this is used from insertToPreQ or for retransmissions)
	 * or sorted by a frames' timestamp (used for receiving RTP) */
	bool addToPreQ(NetPacket_t *p, QOrder qo, bool pushback=false);

	/**
	 * block or get the first packet off the preQ
	 * this does neither free the packet, nor the packet payload!
	 * @param blocking if false, popPreQ might return with a NULL NetPacket */
	NetPacket_t *popPreQ(bool blocking=true);


	/**
	 * it decides, how many stream seconds should be brought thru in this second
	 *    (this is needed to fill up the buffers or send a better quality even if the network
	 *     does not allow it --> only if client buffers are full!)
	 * it also sets the percentage of adaptation
	 * @param net_bw sets the available bandwidth of the network for the next streamout second
 	 */
	void calcAdaptPreQValues(int net_bw);


	/**
	 * searches all packets with the same TS (so they belong to one single frame)
	 * and marks them all as partially_dropped or partially_sent
	 * this is used for better adaptPreQ() results (it has to adapt full frames, not packets)
	 */
	void markFullFramePartially(bool dropped, bool sent, u32 ts);

	/** 
	 * marks unimportant packets for the next streamout sec, according to their priority
	 * marks packets for the next streamout second starting from the preQ head.
	 * if new data or RTX makes the preQ dirty, a new marking is done, now starting
	 * at exactly from the new preQ head. RTX is never marked for adaptation.
	 * it has to adapt full frames, not packets. to do so, it looks at partially_sent and
	 * partially_dropped markers of packets
	 * @param net_bw sets the available bandwidth of the network for the next streamout second
 	 */ 
	void adaptPreQ(int net_bw);

	/**
	 * @returns the timestamp of the latest packet in the preQ (the one that would be popped next)
	 * or -1 if no packets where in queue */
	int getNextPreQ_TS();

	// SendQ related methods
  bool pushNetPacket(NetPacket_t *p);
  NetPacket_t *popNetPacket(uint16_t rtp_seq);
  NetPacket_t *popNetPacketByRTXseq(uint16_t rtx_rtp_seq);

	/**
	 * scans the SendQ for already ack'd or out-dated frames and releases them.
   * @param streamout_sec determines the actual time, or if set to -1, every frame 
   * in the SendQ will be dropped	*/
	void sendQupdate(long streamout_sec);

	/**
	 * scans the preQ for missing packets and requests their retransmission
	 * since in the client, not frame type is known, EVERY missing packet is reqeusted...
	 */  
	void preQupdate();

	/* @param NetPacket_t takes a NetPacket and sends it either as new data or as RTX 
	 * ATTN: it sets the p->rtp_seq and, if applicable, p->rtx_rtp_seq and timestamps
	 * @returns the number of sent bytes or -1 on error */
	int doSend(NetPacket_t *p);

	void sendThread();
	void readThread();

  /** calls UCL lib for RTP,RTCP updates and callback functions (recv data)
	 * @param ts is the current timestamp
	 * @param timeout: if zero, non-blocking, else in msecs
	 */
  void ucl_update(u32 ts, u32 timeout);


	/************************************************
  * Variables *************************************/

  typedef struct NetPacket_t {
    uint32_t    rtp_seq;         // rtp sequence number

		uint16_t    rtx_rtp_seq;     // seqNo of this resent packet on the RTX channel
		int         resent_counter;  // how often was this pkt resent?
		bool        acked;           // already acked?

    u32         rtp_ts;          // rtp composition timestamp
		double      sent_dtime;      // time passed since thread_start_dtime
		double      resent_dtime;

    Frame::FrameType   vop_type;        // I,P,B, NN... this packet might be just a fragment!
		u32         vop_size;        // NOT the actual packet size, but the full frame size!
		int         prio;            // adaptation priority   0..high (definitely keep) (eg. I,P)

		bool        partially_sent;    // an arbitrary fragment of this TS is already on the net!
		bool        partially_dropped; // an arbitrary fragment of this TS is already dropped!
		bool        drop_me;         // set if this packet was selected for adaptation

    u32         frag_no;         // 0..first packet, 1..n following fragments
		                             // on clientside: rtp_seq, which also gives a monotonic order 
                                 //                within TS!
    bool        last_frag;       // flag, if this is the last fragment of frame
    u8          *payload;
    u32         pl_size;         // payload bytes
    NetPacket_t *next;
  } NetPacket_t;

	struct CompareNetPackets
	{
		bool operator()(NetPacket_t*& np1,
										NetPacket_t*& np2) const
		{
      return (np1->prio < np2->prio);
		}
	};
  
  struct PreQ_t {
    NetPacket_t *Q; /* will be taken and sent. 
			    will then go into sendQ with the new rtp_seq */
    NetPacket_t *Q_tail; /* points to the last elem */

    VMutex MUTEX_Q;
    VMutex MUTEX_Q_DATA_AVAIL;
    VMutex MUTEX_Q_FULL;
    VMutex MUTEX_Q_DONE;
		bool isDirty;
		bool needsPrefetching;

		unsigned long max_size;
    unsigned long size;
    uint num_elems;
		uint num_full_frames;
    uint avg_bw;   // avg BW/sec of frames in PreQ, not of whole video!

		uint32_t highest_rtp_seq;
		uint32_t highest_ts;
  } preQ;


#define RTP_SENDQ_PRIME 37
	struct SendQ_t {
		NetPacket_t *Q[RTP_SENDQ_PRIME];
		long size;
		uint num_elems;
		uint avg_bw;  // avg BW/sec of frames in SendQ, not of whole video!
		VMutex MUTEX_Q;
	} sendQ;


	Statistics *stats;
  struct rtp *session;
	uint32_t sender_ssrc;
  int remotePort;
  int localPort;
  char *remoteAddress;
#ifndef WIN32
  static const int TTL = 16;
  static const int MAX_PAYLOAD_SIZE = 65000;
  static const int RTP_PAYLOAD_TYPE = 96;
#endif
  bool writeToNetwork;
  bool firstFrame;

	//global state used by getFrame
	struct {
		int numProcessedFrames;
		int seqno;
		int act_ts;
	} IOFrameStat;

	GlobalTimer * globalTimer;
  PacketizationLayer *pLayer;
  ESInfo *es;
  char *url;
	rtx_info* rtx;
	bool feedbackEnabled;
	bool rtxEnabled;

	int64_t last_rr_seq_no;
	int64_t last_rr_sum_pkts_lost;
	double last_rr_lost_fract;
	double last_rr_received;

	int64_t first_rtp_seq;
  uint16_t last_rtp_seq;
	int ts_offset;   //if the server doesnt start at TS=0, but later, fix the offset....

	int64_t  first_rtx_rtp_seq;
  uint16_t last_rtx_rtp_seq;
	int interarrival_time;       // msecs
	int net_bw_before_superdrain; // when super-draining is necessary, we keep the old (good) net_bw
	int real_bw;       // stream_bw, adjusted by 1 second +- DRAIN/FILL fact

  bool threadWasPaused;
	bool allowVopTimeDetection;
};

#endif
