 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:    SDLRenderer.cpp                                            *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *  $Id: SDLRenderer.cpp,v 1.36 2006/06/22 10:02:26 mkropfbe Exp $
 *                                                                      *
 ***********************************************************************/

#ifndef WINCE
#include <sys/types.h>
#include <signal.h>
#endif
#include "global.hpp"
#include "IO.hpp"
#include "Frame.hpp"
#include "SDLRenderer.hpp"
#include "SDLFrameBuffer.hpp"
#include "VideoRenderer.hpp"


#ifdef ENABLE_X11
#include <X11/Xatom.h>
#include <unistd.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#include <X11/extensions/Xrandr.h>
#endif


SDLRenderer::SDLRenderer(const VideoESInfo *_es, UncompressedVideoFrame::ColorSpace colorSpace, 
												 int width, int height, bool fullScreen, bool pollEvents,
												 bool isLabelResizable, bool fixAspectRatio, bool allowUpscaleX,
												 bool rotateXDisplay)
	: VideoRenderer(_es, colorSpace, width, height, isLabelResizable, fixAspectRatio, allowUpscaleX) {
	dprintf_full("SDLRenderer Construct %s ES %p\n",this->url, this->es);
	strcpy(this->url,"SDLRenderer");
	this->fullScreen = fullScreen;
	this->pollEvents = pollEvents;
	this->rotateXDisplay = rotateXDisplay;

	overlay=NULL;
	surface=NULL;
	video_info=NULL;
}


SDLRenderer::~SDLRenderer() {

	if (rotateXDisplay)
		doRotateXDisplay(0);

	if (overlay)
		SDL_FreeYUVOverlay(overlay);
	overlay=NULL;
	if (surface)
		SDL_FreeSurface(surface);
	surface=NULL;

	//free video_info not needed, is read-only from SDL!

	SDL_ShowCursor(SDL_ENABLE);
	SDL_Quit();
}

void SDLRenderer::setESInfo(ESInfo * es) {
	this->es = (VideoESInfo*)es;
}


bool SDLRenderer::updateDisplayRegion() {
	
  if (!es || !es->isVisualStream()) {
    dprintf_err("SDLRenderer::updateDisplayRegion(): invalid members: es = %p\n",es);
		::exit(1);
    return false;
  }

	return VideoRenderer::updateDisplayRegion();
}


void SDLRenderer::doRotateXDisplay(int orientation) {

#ifdef BUILD_ARCH_ARM
#ifdef ENABLE_X11
	int status;
	char *hostname = new char[MAX_STR_LEN];
	char *envhost = getenv("DISPLAY");


	if (!envhost)
		strcpy(hostname,"localhost:0");
	else
		strcpy(hostname,envhost);
	
	Display *display = XOpenDisplay(hostname);
	assert(display);
	
	XRRScreenConfiguration *sc;
	sc = XRRGetScreenInfo (display, DefaultRootWindow(display));
	status = XRRSetScreenConfigAndRate(display,sc,DefaultRootWindow (display),0,1 << orientation,0,0);
	dprintf_full("SDLRenderer::open  rotating X11 screen, status: %i\n",status);
	
	XRRFreeScreenConfigInfo(sc);
	delete [] hostname;
#endif
#endif
}


bool SDLRenderer::open() {
	if (state == OPEN) {
		dprintf_full("SDLRenderer::open  state was already OPEN, ignoring opening request.\n");
		return true;
	}
	if (state != CLOSED) {
		dprintf_err("SDLRenderer::open  state was not CLOSED but state=%i!\n",state);
		return false;
	}

  if (!updateDisplayRegion()) {
    state = STREAMERR;
    return false;
  }

	char *buf = new char[MAX_STR_LEN];
	//on iPAQ, dont forget: export DISPLAY=:0
	//if without X11, maybe this helps: export SDL_NOMOUSE=1
	if (SDL_Init(SDL_INIT_VIDEO) < 0 || !SDL_VideoDriverName(buf, 1)) {
		dprintf_err("Could not init SDL video: %s\r\n", SDL_GetError());
		delete [] buf;
		::exit(-1);
	}
	atexit(SDL_Quit);
	video_info = SDL_GetVideoInfo();
	uint flags = SDL_SWSURFACE;
	if (fullScreen)
		flags |= SDL_FULLSCREEN;

#ifndef WINCE
	surface = SDL_SetVideoMode(out_label_info.width, out_label_info.height, 0, flags);
#else
#ifndef DISABLE_CE_RESIZE
	int myWidth = GetSystemMetrics(SM_CXSCREEN);
	int myHeight = GetSystemMetrics(SM_CYSCREEN);
#else
	int myWidth = out_label_info.width;
	int myHeight = out_label_info.height;
#endif
	surface = SDL_SetVideoMode(myWidth, myHeight, 0, flags);
#endif
	
	if (surface == NULL) {
		dprintf_err("Couldn't initialize window with dimension %ix%i: %s\n", out_label_info.width, out_label_info.height, SDL_GetError());
		delete [] buf;
		::exit(1);	
	}

	if (rotateXDisplay)
		doRotateXDisplay(1);

	SDL_ShowCursor(SDL_DISABLE);
	if  ( (surface->format->BitsPerPixel == UncompressedVideoFrame::mapColorSpaceToBitDepth(colorSpace) ) &&
				(out_label_info.width ==  frame_info.width) && (out_label_info.height ==  frame_info.height) &&
				(display_region.left == 0) &&  (display_region.top == 0) )
		directBitBlt=true;  //optimize via rgbSurface bitBlitting
	else
		directBitBlt=false; //use slow pixel-per-pixel painting and scaling

	
	dprintf_full("SDLRenderer::open() set SDL video mode to (%dx%d - %dx%d) bpp %i, using %s.\n",
							 out_label_info.left, out_label_info.width, out_label_info.top, out_label_info.height, 
							 surface->format->BitsPerPixel, directBitBlt?"fastBitBlit":"slow pixel painting");
		
	//set 8, 16 or both  24/32  to  24
	setOutputDepth(surface->format->BitsPerPixel>=24?24:surface->format->BitsPerPixel); 

	SDL_WM_SetCaption("ViTooKi SDLRenderer", "");
	
	if (overlay) 
		SDL_FreeYUVOverlay(overlay);
	overlay=NULL;

  /* initialize frameBuffer */
  if (frameBuffer) {
    delete frameBuffer;
    frameBuffer = NULL;
  }
	bool gray=false;
  switch (colorSpace) {
  case UncompressedVideoFrame::ColorSpaceGRAY8:
		dprintf_full("SDLRenderer::open()  ColorSpaceGRAY8\n");
    frame_info.size = frame_info.width * frame_info.height;
		gray=true;
    break;
  case UncompressedVideoFrame::ColorSpaceRGB565:
		dprintf_full("SDLRenderer::open()  ColorSpaceRGB565 with 16bit depth\n");
    frame_info.size = frame_info.width * frame_info.height * 2;
    break;
  case UncompressedVideoFrame::ColorSpaceRGB24:
		dprintf_full("SDLRenderer::open()  ColorSpaceRGB24\n");
    frame_info.size = frame_info.width * frame_info.height * 3;
    break;
  case UncompressedVideoFrame::ColorSpaceRGB32:
		dprintf_full("SDLRenderer::open()  ColorSpaceRGB32\n");
    frame_info.size = frame_info.width * frame_info.height * 4;
    break;
	case UncompressedVideoFrame::ColorSpaceYV12:
		dprintf_full("SDLRenderer::open()  ColorSpaceYV12\n");
		//overlay is scalable, so frame_info is used, and then up/down-scaled by hardware with SDL_YV12
		overlay = SDL_CreateYUVOverlay(frame_info.width, frame_info.height, SDL_YV12_OVERLAY, surface);
    frame_info.ySize = frame_info.width * frame_info.height;
    frame_info.uvSize = frame_info.width * frame_info.height / 4;
		frame_info.size = frame_info.ySize + 2 * frame_info.uvSize;
		dstrect.x = out_label_info.left;
		dstrect.y = out_label_info.top;
		dstrect.w = out_label_info.width;
		dstrect.h = out_label_info.height;		
		break;
	case UncompressedVideoFrame::ColorSpaceI420:
		dprintf_full("SDLRenderer::open()  ColorSpaceI420\n");
		//overlay is scalable, so frame_info is used, and then up/down-scaled by hardware with SDL_IYUV
		overlay = SDL_CreateYUVOverlay(frame_info.width, frame_info.height, SDL_IYUV_OVERLAY, surface);
    frame_info.ySize = frame_info.width * frame_info.height;
    frame_info.uvSize = frame_info.width * frame_info.height / 4;
		frame_info.size = frame_info.ySize + 2 * frame_info.uvSize;
		dstrect.x = out_label_info.left;
		dstrect.y = out_label_info.top;
		dstrect.w = out_label_info.width;
		dstrect.h = out_label_info.height;		
		break;
  default:
    dprintf_err("SDLRenderer::open(): invalid color space: %d\n", colorSpace);
    state = STREAMERR;
	delete [] buf;
    return false;
  }

#ifdef WINCE
#ifndef DISABLE_CE_RESIZE
  dstrect.w = GetSystemMetrics(SM_CXSCREEN);
  dstrect.h = GetSystemMetrics(SM_CYSCREEN);
#else
  dstrect.w = out_label_info.width;
  dstrect.h = out_label_info.height;
#endif
#endif

	if (!overlay)
	frameBuffer = new SDLFrameBuffer(surface, UncompressedVideoFrame::mapColorSpaceToBitDepth(colorSpace),
																		 gray, out_label_info.width, out_label_info.height );

  if (!frameBuffer && !overlay) {
    dprintf_err("SDLRenderer::open(): SDLFrameBuffer creation FAILED! (w = %d, h = %d)\n",
								out_label_info.width, out_label_info.height );
    state = STREAMERR;
	delete [] buf;
    return false;
  }

  state = OPEN;
  framesSeen = 0;
  currentFrameNumber = 0;
  dprintf_small("SDLRenderer::open() (state=%i): initialized frame and display regions:\n"
								"    frame: width = %d, height = %d, size = %d, scaleX = %5.3f, scaleY = %5.3f\n"
								"    display: x(%d:%d), y(%d:%d), width = %d, height = %d\n",
								state, frame_info.width, frame_info.height, frame_info.size,
								FIXED2DOUBLE(frame_info.scaleX), FIXED2DOUBLE(frame_info.scaleY),
								display_region.left, display_region.right,
								display_region.top, display_region.bottom,
								display_region.width, display_region.height);
  delete [] buf;
  return true;
}


int SDLRenderer::writeFrame(Frame * frm, ESInfo *out_es) {
	if (pollEvents && SDL_PollEvent( &event )) {
			SDL_KeyboardEvent *key = &event.key;
			char *buf = SDL_GetKeyName( key->keysym.sym );
			char c;
			if (buf[1]=='\0') //only single character (a-zA-Z...)
				c = toupper(buf[0]);
			switch( event.type ) {
			case SDL_KEYUP:
				PrintKeyInfo( &event.key );
				if ( ((key->keysym.mod & KMOD_CTRL)  && (c == 'C')) ||
						 (strstr(buf, "escape") != NULL)) {
					dprintf_err("SDLRenderer::writeFrame   STREAMERR because of CTRL-C/ESC!\n");
					state = STREAMERR;
					//kill(0,SIGQUIT); //kill the app incl. other eg. audio stream
					return 0;
				}
				break;
				
			case SDL_KEYDOWN:
				PrintKeyInfo( &event.key );
				break;
				
				/* SDL_QUIT event (window close) */
			case SDL_MOUSEBUTTONDOWN:
			case SDL_QUIT:
				dprintf_err("SDLRenderer::writeFrame   STREAMERR because of window close!\n");		
				state = STREAMERR;
				return 0;
				break;
				
			default:
				break;
			}
	}

	//SDL-YUV will not need a frameBuffer so don't supercall!
	if ((colorSpace == UncompressedVideoFrame::ColorSpaceI420) || (colorSpace == UncompressedVideoFrame::ColorSpaceYV12)) {
		if (state != OPEN)
			return 1;
		framesSeen++;
		/* we don't set currentFrameNumber, as we have no way to determine it! */

#ifndef WINCE
		if (SDL_MUSTLOCK(surface)) {
			if (SDL_LockSurface(surface) < 0) {
				::exit(2);
			}
		}
		
		if(SDL_LockYUVOverlay(overlay) !=0) {
			return -1;
		}
#endif

#ifndef WINCE
		memcpy(overlay->pixels[0], frm->getAU()->payload, frame_info.ySize); // Y part

		if (colorSpace == UncompressedVideoFrame::ColorSpaceI420) {
			memcpy(overlay->pixels[1], frm->getAU()->payload + frame_info.ySize, frame_info.uvSize); // U part
			memcpy(overlay->pixels[2], frm->getAU()->payload + frame_info.ySize + frame_info.uvSize, frame_info.uvSize); // V Part
		} else { // YV12, so exchange U and V
			memcpy(overlay->pixels[2], frm->getAU()->payload + frame_info.ySize, frame_info.uvSize); // U part
			memcpy(overlay->pixels[1], frm->getAU()->payload + frame_info.ySize + frame_info.uvSize, frame_info.uvSize); // V Part
		}
#else	
		int i,j,k;
		int edgedWidth = (frame_info.width + 15) / 16;
		int pitch = 16 * edgedWidth + 2 * EDGE_SIZE;


		u8 *y = frm->getAU()->y;
		u8 *u = frm->getAU()->u;
		u8 *v = frm->getAU()->v;

		u8 *pixels1 = overlay->pixels[0];
		u8 *pixels2 = overlay->pixels[1];
		u8 *pixels3 = overlay->pixels[2];

		int halfWidth = frame_info.width >> 1;
		int halfHeight = frame_info.height >> 1;
		int halfPitch = pitch >> 1;

		for (i=0; i < frame_info.height; i++) {
			memcpy(pixels1, y, frame_info.width);
			if ( i < halfHeight) {
				memcpy(pixels2, u, halfWidth);
				memcpy(pixels3, v, halfWidth);

				u += halfPitch;
				v += halfPitch;
				pixels2 += halfWidth;
				pixels3 += halfWidth;
			}
			y += pitch;
			pixels1 += frame_info.width;
			
		}

#endif


		SDL_DisplayYUVOverlay(overlay, &dstrect);
		dprintf_full("SDLRenderer::writeFrame -> Displaying YUV Frame (%i bytes) to x:%i y:%i -- w:%i h:%i\n",
								 frm->getAU()->size,dstrect.x,dstrect.y,dstrect.w,dstrect.h);
#ifndef WINCE
		SDL_UnlockYUVOverlay(overlay);
		
		if (SDL_MUSTLOCK(surface)) {
			SDL_UnlockSurface(surface);
		}
#endif
		

#ifdef _POSIX_PRIORITY_SCHEDULING
		sched_yield(); //this is necessary to give parallel getFrames a chance
#endif

		return 1;

	} else //RGB scale/write+yield is done in superClass
		return VideoRenderer::writeFrame(frm);

}


/* Print all information about a key event */
void SDLRenderer::PrintKeyInfo( SDL_KeyboardEvent *key ) {
	/* Is it a release or a press? */
	if( key->type == SDL_KEYUP )
		printf( "SDLRenderer::PrintKeyInfo Release:- " );
	else
		printf( "SDLRenderer::PrintKeyInfo Press:- " );
	
	/* Print the hardware scancode first */
	printf( "Scancode: 0x%02X", key->keysym.scancode );
	/* Print the name of the key */
	printf( ", Name: %s", SDL_GetKeyName( key->keysym.sym ) );
	/* We want to print the unicode info, but we need to make */
	/* sure its a press event first (remember, release events */
	/* don't have unicode info                                */
	if( key->type == SDL_KEYDOWN ){
		/* If the Unicode value is less than 0x80 then the    */
		/* unicode value can be used to get a printable       */
		/* representation of the key, using (char)unicode.    */
		printf(", Unicode: " );
		if( key->keysym.unicode < 0x80 && key->keysym.unicode > 0 ){
			printf( "%c (0x%04X)", (char)key->keysym.unicode,
							key->keysym.unicode );
		}
		else{
			printf( "? (0x%04X)", key->keysym.unicode );
		}
	}
	printf( "\n" );
	/* Print modifier info */
	PrintModifiers( key->keysym.mod );
}


/* Print modifier info */
void SDLRenderer::PrintModifiers( SDLMod mod ) {
	printf( "Modifers: " );
	
	/* If there are none then say so and return */
	if( mod == KMOD_NONE ){
		printf( "None\n" );
		return;
	}
	
	/* Check for the presence of each SDLMod value */
	/* This looks messy, but there really isn't    */
	/* a clearer way.                              */
	if( mod & KMOD_NUM ) printf( "NUMLOCK " );
	if( mod & KMOD_CAPS ) printf( "CAPSLOCK " );
	if( mod & KMOD_LCTRL ) printf( "LCTRL " );
	if( mod & KMOD_RCTRL ) printf( "RCTRL " );
	if( mod & KMOD_RSHIFT ) printf( "RSHIFT " );
	if( mod & KMOD_LSHIFT ) printf( "LSHIFT " );
	if( mod & KMOD_RALT ) printf( "RALT " );
	if( mod & KMOD_LALT ) printf( "LALT " );
	if( mod & KMOD_CTRL ) printf( "CTRL " );
	if( mod & KMOD_SHIFT ) printf( "SHIFT " );
	if( mod & KMOD_ALT ) printf( "ALT " );
	printf( "\n" );
}
