 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 *  title:   SDLRenderer.hpp                                            *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *  $Id: SDLRenderer.hpp,v 1.11 2006/01/20 15:37:18 mkropfbe Exp $
 *                                                                      *
 ***********************************************************************/

#ifndef _SDLRENDERER_HPP
#define _SDLRENDERER_HPP

#include "IO.hpp"
#include "VideoESInfo.hpp"
#include "adaptors/MP4Decoder.hpp"

#include <SDL.h>
#include "VideoRenderer.hpp"

class SDLFrameBuffer;


/**
 * @short This I/O class renders a gray-scale or RGB24/32 or YUV I420/YV12 stream  
 * the used SDL surface is created within this function, 
 * incl. CTRL-C/ESC signal/event catchers
 *
 * It is intended to be used as a DataSink connected to a DataChannel.
 * <p>
 * Video frames will be (down or up) scaled to fit into the given size by
 * preserving the aspect ratio given in VideoESInfo.</p>
 * <p>
 * The expected format of the gray-scale stream is: 1 byte per pixel.</p>
 * <p>
 * The expected format of the RGB24 stream is: 3 bytes per pixel,
 * in the order RGB (red, green, blue).</p>
 * <p>
 * The expected format of the RGB32 stream is: 4 bytes per pixel,
 * in the order BGRA (blue, green, red, alpha). This conforms to the
 * PIX_FMT_RGBA32 format defined in ffmpeg/libavcodec/avcodec.h on
 * little endian architectures.</p>
 *
 *@author Michael Kropfberger
 *@version $Id: SDLRenderer.hpp,v 1.11 2006/01/20 15:37:18 mkropfbe Exp $
 */
class SDLRenderer : public VideoRenderer
{
public:

  /** Constructor. If the parameters are invalid, the #open() method will fail.
   * @param es        elementary stream information about the video stream to
   *                  be rendered.
   * @param colorSpace the color space of the video stream to be rendered;
   *                   must be one of: VideoFrame::::ColorSpaceGRAY8,
   *                   UncompressedVideoFrame::ColorSpaceRGB24, UncompressedVideoFrame::ColorSpaceRGB32,
	 *                   UncompressedVideoFrame::ColorSpaceRGB565,
   *                   UncompressedVideoFrame::ColorSpaceI420, UncompressedVideoFrame::ColorSpaceYV12
	 * @param width     the width of the SDL window
	 * @param height    the height of the SDL window
	 * @param fullScreen  opens the SDL window in FullScreen Mode
	 * @param pollEvents  if <code>true</code>, the SDL event handler is checked for
	 *                  keypresses like CTRL-C or ESC, and a "SDL window close" is detected
	 *                  if such an event occurs, the state of the stream is set to <code>STREAMERR</code>
	 *                  beware, that this polling might grab away ALL events, so implement on
	 *                  on your own, if you want to support eg. joysticks or more keycodes!
	 * @param isLabelResizable if <code>true</code>, the given width/height is 
	 *                      assumed to be flexible, so it will be tried to keep
	 *                      the original video size (plus adhering #fixAspectRatio)
	 *                      setting this to true, width&height may be zero, 
	 *                      only the video is used to determine the resolution
   * @param fixAspectRatio if <code>true</code>, video frames may be resized 
	 *                      according to their ESInfo aspectRatio, which means
	 *                      upscaling in the y-dimension
   * @param allowUpscaleX if <code>true</code>, video frames may be upscaled
   *                      in x-dimension to fit the label, which may
   *                      result in significant quality loss, since a very
   *                      simple upscaling algorithm is used (assumes isLabelResizable==false).
	 * @param rotateXDisplay uses XRandr X11 extension to rotate the screen, play SDL video, then rotate back
	 *
   */
  SDLRenderer(const VideoESInfo *es, UncompressedVideoFrame::ColorSpace colorSpace = UncompressedVideoFrame::ColorSpaceI420,
							int width=0, int height=0, bool fullScreen = false,
							bool pollEvents = true,
							bool isLabelResizable = true, bool fixAspectRatio = true, 
							bool allowUpscaleX = false, bool rotateXDisplay = false);
	
  /** Destructor. */
  virtual ~SDLRenderer();

  /** Renders a frame to the destination YUV_Overlay/RGB_Surface
   * @return 1 on success, 0 on error.
   */
  virtual int writeFrame(Frame * frm, ESInfo *out_es=NULL);

  /** Creates internal frame buffers and checks whether the stream to be
   * rendered matches the widget dimensions.
   * @return <code>true</code> on success, <code>false</code> on error.
   * @see IO::open()
   */
  virtual bool open();


  void setESInfo(ESInfo * es);



protected:
  /** Update {@link #frame_info} (except <code>size</code>)
   * and {@link #display_region}.
   * This function is not thread-safe, and is internally
   * called by the {@link #open()} method.
  */
  bool updateDisplayRegion();


	/** rotate the X11 display using the XRandr extension
	 * @param orientation 0 for normal, 1 for 90 degrees
	 */
	void doRotateXDisplay(int orientation);


private:

	const SDL_VideoInfo *video_info;
	SDL_Rect dstrect;
	SDL_Overlay *overlay;
	SDL_Surface *surface;
	SDL_Event event;
	bool pollEvents;
	bool fullScreen;
	bool rotateXDisplay;

	void PrintKeyInfo( SDL_KeyboardEvent *key );
	void PrintModifiers( SDLMod mod );
};

#endif /* _SDLRENDERER_HPP */
