 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title: SDLaudioIO.hpp                                                *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

// SDLaudioIO.hpp: Class for Audio output.
//
//////////////////////////////////////////////////////////////////////

#ifdef ISOMP4
#include "ISOMovies.h" //XXX for data types definition
#endif

#ifndef _SDLAUDIOIO_HPP_
#define _SDLAUDIOIO_HPP_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "AudioIO.hpp"
#include "adaptors/GlobalTimer.hpp"
#include "SDL.h"
#include <SDL_thread.h>
#include "VMutex.hpp"
#include "VCondition.hpp"

#include "AudioESInfo.hpp"
#include "Frame.hpp"
#include "io/Rtp.hpp"

#define SDL_MAX_AUDIOQ_SIZE (2 * 4 * 48000)	// 2 media seconds @48 kHz, 16-bit stereo
                                        // this may be only 2 jumbo packets!
/* SDL audio buffer size, in samples. Should be small to have precise
 * A/V sync as SDL does not have hardware buffer fullness info. */
#define SDL_AUDIO_BUFFER_SAMPLES 1024

class Frame;
class AudioESInfo;

typedef struct APacket {
    u64 pts; /* presentation time stamp in stream units */
    u8 *data;
    int   size;
} APacket;

typedef struct APacketList {
    APacket pkt;
    struct APacketList *next;
} APacketList;


typedef struct PacketQueue {
    APacketList *first_pkt, *last_pkt;
    int nb_packets;
    int size;
    int abort_request;
#ifndef WINCE
    SDL_mutex *mutex;
    SDL_cond *cond;
#else
	VMutex *mutex;
	VCondition *cond;
#endif
} PacketQueue;

class SDLaudioIO; //forward declaration

typedef struct AudioState {
	SDLaudioIO *selfClass;
        SDL_AudioSpec spec;
	int abort_request;
	PacketQueue audioq;
	APacket audio_pkt;	// current packet to be processed by sdl_audio_callback()
        int audio_pkt_index;	// index into audio_pkt.data
	u32 audio_pkt_time_increment; // in ms
        u32 ticks_per_second;
        u32 queue_full_delay;	// in ms
} AudioState;



/**
 *
 * provides raw-audio output via the SDL library methods
 *
 * @author Mithlesh Kumar
 *
 * @version $Id: SDLaudioIO.hpp,v 1.13 2006/01/20 15:37:18 mkropfbe Exp $
 *
 * @short <short description>
 *
 */
class SDLaudioIO : public AudioIO {
public:
	/** Constructor.
	 * @param ess       the ESSynchronizer object attached to the <em>video</em>
	 *                  DataChannel; needed for A/V synchronization.
	 */
	SDLaudioIO(AudioESInfo *es, ESSynchronizer *ess = NULL);
	
	virtual ~SDLaudioIO();
	
	int initialize();
	
	// IO Methods
	Frame * getFrame() ;

	/** returns the number of packets sent. Returns 0 on error */
	int writeFrame(Frame * frm, ESInfo *out_es=NULL);

	/** opens the IO connection. State is set to OPENING.
			When the connection is ready for use, State is OPEN */
	bool open();

	bool close(bool immediate=false);

	IO::State play(double prefetchTime=0.0);

	IO::State pause();

	IO::State mute();

	int packet_queue_get(PacketQueue *q, APacket *pkt, int block);

	void resume();
	//static void sdl_audio_callback(void *opaque, Uint8 *stream, int len);

        /** clear packet buffer */
        void clearBuffer();

	int getPacketsInQueue();

        friend void sdl_audio_callback(void *opaque, Uint8 *stream, int length);

private:
	bool initialized;
	bool requestUnpauseESS;
	AudioState *is;
	void packet_queue_init();
	void packet_queue_end();
	int packet_queue_put(APacket *pkt);
	void packet_queue_abort();
	void packet_queue_clear();
	
	/** For debugging output.
	 * @return a pointer to a static string.
	 */
	const char *get_sdl_audio_state() const;
	
	/** Re-start SDL audio thread without clearing packet buffers. 
	 * @return 0 on success, -1 on failure.
	 */
	int restartSDLThread();
};

#endif
