#ifndef WINCE

#ifndef _PS_SIMPLERTP_HPP_
#define _PS_SIMPLERTP_HPP_

#include "IO.hpp"
#include "ESInfo.hpp"
#include "net/SyncLayer.hpp"
#include "net/PacketizationLayer.hpp"
#include "net/PortGenerator.hpp"

#ifndef WIN32
#include "uclconf.h"		//for endian
#endif
#ifdef WINCE
#include "3rdparty/commonucl/rtp.h"	// UCL library
#else
#include "commonucl/rtp.h"	// UCL library
#endif

class Frame;
class SimpleStatistics;
#include "FastPointerHashTable.hpp"
extern PortGenerator portGen;

class SimpleRtp:public IO {

public:
	enum Delay { NO_DELAY,CTS_BASED,BITRATE_BASED,MIXED_CTS_BITRATE,AGGRESSIVE_MIXED_CTS_BITRATE};

	/** creates an RTP object
			@param url: the source/dest for this IO class, deep-copied
			@param remotePort: the clientPort
			@param localPort: the serverPort
			@param clientaddress: the ip address of the client, deep-copied
			@param es: the ElementaryStream sent, shallow-copied
			@param statsFileName: textfile where statistic information is saved, for no statistic leave at NULL
			@param overwriteExistingStatFile: will append to an existing filename if set to false!
			@param writeonly specifies if the RTP connection is for sending
			(writeonly is set to true), of for reading
			@param packetization specifies the transport mode: can be RFC3016,
			multisl or flexmux
	*/
    SimpleRtp(const char *url, int remotePort, int localPort, const char *clientaddress, ESInfo * es,
			PacketizationLayer * packetization, bool writeOnly,
			const char* statsFileName=NULL, bool overwriteExistingStatFile=true,
			Delay delay=MIXED_CTS_BITRATE,
			bool updateVopTimeIncrement=true, bool enableSwitching=true);
     virtual ~ SimpleRtp();
	/** returns a frame  */
    Frame *getFrame();
	/** returns the number of packets sent. Returns 0 on error */
    int writeFrame(Frame * frm, ESInfo *out_es=NULL);
	/** opens the IO connection. State is set to OPENING. Depending on the
	    underlying QIODevice, a network connection is stablished or a file
	    connection. When the connection is ready for use, State is OPEN */
    bool open();
    bool close(bool immediate=false);

    State getState() const {
	return state;
    };
    int getLocalPort() const {
	return localPort;
    };
	/** returns -1, when at end of File, returns 0 when file is closed
	  * returns 50 otherwise */
    int getBufferFillLevel() const {
	return 50;
    };
	/** you can't delete a remote url :-) will always return true */
  bool destroy() {	return true;}
	/* is this Rtp object for reading from network? (eg. client side) */
	bool isInput() { return !writeOnly; };
	/* is this Rtp object for writing to network? (eg. server side) */
	bool isOutput() { return writeOnly; };
	const char *getURL() const { return url; };

	/** returns NULL if the session was not opened, Otherwise,
	  * a new char* is created and returned containing a %08x hexValue */
  char *getSSRC();

	/** sends control packets */
  void update();

	void setESInfo(ESInfo * es) { this->es=es;};

	virtual ESInfo* getESInfo() { return es;};

	IO::State play(double prefetchTime=0.0);

	IO::State pause();
	/** blocking method which will block for @param millisecs milliseconds.
	  * When called with 0 extracts all pending packets from the rtp_buffer
		* and returns immediately.
		*/
	void preBufferPackets(long millisecs);
	/** empties buffer */
	bool setToFrameNumber(u32 frameNumber);
	const SimpleStatistics* getSimpleStatistics() const { return stats;};

	u32 getctsForFirstPlay()
	{
		return this->ctsForFirstPlay;
	}

protected:
	u32 frameToStart;

	static void RtpCallback(struct rtp *session, rtp_event * e);

	/** blocks until we are in sync with the timing information */
	void syncToCts(u32 cts);

	/** if a config header is in front followed by a frame, the header will
	  * be removed and the frm will be updated
		*/
	void stripHeaderFromFrame(Frame* frm);
	long calcSleepTimeInMs(u32 cts);
	long calcSleepTimeCTSBased(u32 cts);
	/** @param speedScaleFactor is in Percent! valid ranges are 1 to 400. If an invalid range is detected
	  * it is set to the default 115
	  */
	long calcSleepTimeBitRateBased(u32 cts, u32 speedScaleFactor=115);
	long calcSleepTimeMixedBased(u32 cts,u32 clientPreBufferTime);
	long getElapsedTimeInMSSinceLastPlay();

	bool sendPacket(u8 * data, u32 size, u32 rtpTimestamp, bool last);

	Frame* extractCompleteFrame(u32 rtpSeqNr);

	void insertPacket(AU* p, u32 seqNr);
	void removeIncompleteFrame(u32 seqNr);
	struct rtp *session;
	int clientPort;
	int localPort;
	char *clientAddress;

#ifndef WIN32
    static const int TTL = 16;
    static const int MAX_PAYLOAD_SIZE = 65000;
    static const int RTP_PAYLOAD_TYPE = 96;
	static const int UPDATE_INTERVAL_MS=20;
	static const int MINI_PREBUFFER_INTERVALL=500;
	static const unsigned int MAX_NULL_FRAMES=10;
#endif

	bool writeOnly;
	bool firstFrame;
	bool packetInsertedDuringLastCallback;
	bool firstPacket;
	PacketizationLayer *pLayer;
	ESInfo *es;
	char *url;
	/** the timepoint of the last play */
	struct timeval lastTimeStamp;
	/** the cts value of the first payload(!) frame after the play */
	u32 ctsForFirstPlay;
	/** the cts value of the last frame that was returned to the caller */
	u32 lastCTSReturned;
	/** the first cached packet in the packet buffer */
	u32 rtpSeqNrStart;
	/** the last cached packet in the packet buffer */
	u32 rtpSeqNrEnd;
	/** how often was an overflow in the 16bit rtp_seq_nr detected */
	u32 overflow;
	/** how many packets were sent */
	u32 sendPacketCounter;
	/** if set to true it will cause lastTimeStamp and ctsForFirstPlay ro reinitialize */
	bool resetPlayTime;
	/** packet buffer for storing packets */
	FastPointerHashTable<AU*>* packetBuffer;
	/** must be locked prior accessing packetBuffer */
	VMutex lockPacketBuffer;
	/** initial prefetch time */
	long prefetchTimeInMs;
	/** if set to true we will correct the voptimeincrement according to the
	  * first two frames */
	bool updateESInfo;
	/** if set to false header packets in the stream will be ignored */
	bool enableSwitching;
	/** how often did we return a NULL frame consecutively */
	u32 nullFrameCounter;
	/** bytes Transferred since last play */
	u64 bytesTransferred;
	int intermediateMiniSleep;
	Delay delayType;
	SimpleStatistics *stats;
};

#endif


#endif //wince
