 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title: VideoRenderer.cpp                                             *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *  $Id: VideoRenderer.cpp,v 1.19 2006/02/23 15:42:29 tkuglits Exp $
 *                                                                      *
 ***********************************************************************/

#include "global.hpp"
#include "IO.hpp"
#include "Frame.hpp"
#include "VideoRenderer.hpp"

VideoRenderer::VideoRenderer(const VideoESInfo *_es,  UncompressedVideoFrame::ColorSpace colorSpace, 
														 int width, int height, 
														 bool isLabelResizable, bool fixAspectRatio,
														 bool allowUpscaleX)
  : IO() {
	strcpy(this->url,"VideoRenderer");
  this->es = _es;
  this->colorSpace = colorSpace;	
	this->outputDepth = 24; //default value assumes modern display
	this->directBitBlt = false;
	frameBuffer = NULL;

	dprintf_full("VideoRenderer Construct (colorSpace %s, depth %i bit,  resolution %ix%i, allowUpscaleX: %s)\n",
							 colorSpace==UncompressedVideoFrame::ColorSpaceGRAY8?"RGB8gray":
							 colorSpace==UncompressedVideoFrame::ColorSpaceRGB24?"RGB24":
							 colorSpace==UncompressedVideoFrame::ColorSpaceRGB32?"RGB32":
							 colorSpace==UncompressedVideoFrame::ColorSpaceRGB565?"RGB565":
							 colorSpace==UncompressedVideoFrame::ColorSpaceI420?"I420":
							 colorSpace==UncompressedVideoFrame::ColorSpaceYV12?"YV12":
							 "UNKNOWN",
							 UncompressedVideoFrame::mapColorSpaceToBitDepth(colorSpace),
							 width, height, allowUpscaleX?"yes":"no");
  frame_info.width = 0;
  frame_info.height = 0;
  frame_info.size = 0;
  frame_info.scaleX = INT2FIXED(1);
  frame_info.scaleY = INT2FIXED(1);
  frame_info.fixAspectRatio = fixAspectRatio;
  frame_info.allowUpscaleX = allowUpscaleX;
  out_label_info.isLabelResizable = isLabelResizable;
	out_label_info.left=0;
	out_label_info.top=0;

	if (isLabelResizable) {
		//get first hint from ES, aspectRatio may be calc'd later
		out_label_info.width=es->getWidth();
		out_label_info.height=es->getHeight();
	} else {
		out_label_info.width=width;
		out_label_info.height=height;
	}
  state = CLOSED;
}


VideoRenderer::~VideoRenderer() {
}


void VideoRenderer::setOutputDepth(int depth) {
	assert (depth == 16 || depth == 24 /* || depth == 8 */);
	dprintf_full("VideoRenderer::setOutputDepth to %i\n",depth);
	outputDepth = depth;
}


void VideoRenderer::setVideoFrameBuffer( VideoFrameBuffer *vfb ) {
	assert(frameBuffer == NULL);
	frameBuffer = vfb;
}


bool VideoRenderer::updateDisplayRegion() {
  int width, height;
  double scalex, scaley;
  double aspect;

  if (!es || !es->isVisualStream()) {
    dprintf_err("VideoRenderer::updateDisplayRegion(): invalid member: es = %p\n", es);
    return false;
  }
  /* update frame_info */
  frame_info.width = (int) es->getWidth();
  frame_info.height = (int) es->getHeight();
  /* compute frame display region with correct aspect ratio */
  display_region.top = 0;
  display_region.left = 0;
  display_region.width = frame_info.width;
  display_region.height = frame_info.height;
  
  aspect = (double) frame_info.width / (double) frame_info.height;
  if (es->getAspectRatio() > 0.0 && es->getAspectRatio() < 10.0)
    aspect = es->getAspectRatio();
  width=display_region.width;
  height = (int) ((double) display_region.width / aspect);
  if (height < display_region.height) { //always scale up, not down!
    height= display_region.height;
    width = (int) ((double) display_region.height * aspect);		
  }
  
  // adjust to even value
  height += height % 2;
	if (height == 242 || height == 642) //quick hack, but works
		height -= 2;
  display_region.height = height;
  display_region.width = width;
  dprintf_full("VideoRenderer::updateDisplayRegion() aspect-correct frame size (%d x %d)\n"
	       "    aspect ratio = %5.3f (using: %5.3f)\n",
	       display_region.width, display_region.height,
	       es->getAspectRatio(), aspect);

  /* if label/widget is resizeable, adjust out_label_info to display_region */
  if (out_label_info.isLabelResizable) {
    out_label_info.width = display_region.width;
    out_label_info.height = display_region.height;
    scalex = scaley = 1.0;
  } else {
    /* otherwise, fit display_region into label */
    width = out_label_info.width;
    height = out_label_info.height;
    if (display_region.width < width && !frame_info.allowUpscaleX) {
      width = display_region.width;
    }
    scalex = (double) width / (double) display_region.width;
    scaley = (double) height / (double) display_region.height;
    if ((scalex >= scaley && scaley >= 1.0)
	|| (scaley <= scalex && scaley < 1.0)) {
      /* scale to height */
      if (height == display_region.height)
	width = display_region.width;
      else
	width = (int) (display_region.width * scaley);
    } else if ((scaley >= scalex && scalex >= 1.0)
	       || (scalex <= scaley && scalex < 1.0)) {
      /* scale to width */
      if (width == display_region.width)
	height = display_region.height;
      else
	height = (int) (display_region.height * scalex);
    }
    // adjust width and height to even values
    width += width % 2;
    height += height % 2;
    display_region.width = width;
    display_region.height = height;
  }

  dprintf_full("VideoRenderer::updateDisplayRegion() set display_region to (%d x %d)\n"
	       "    scalex = %5.3f, scaley = %5.3f, label/widget (%d x %d)\n",
	       display_region.width, display_region.height,
	       scalex, scaley, out_label_info.width, out_label_info.height);

  if (display_region.width > out_label_info.width
      || display_region.height > out_label_info.height
      || display_region.width <= 0
      || display_region.height <= 0) {
    dprintf_err("VideoRenderer::updateDisplayRegion(): invalid size:\n"
		"    video (%d x %d), qlabel (%d x %d)\n",
		display_region.width, display_region.height,
		out_label_info.width, out_label_info.height);
    return false;
  }
  /* center displayed region within label/widget */
  display_region.left += (out_label_info.width - display_region.width) / 2;
  display_region.top += (out_label_info.height - display_region.height) / 2;
  /* update frame_info.scale{X,Y} */
  frame_info.scaleX = DIV_FIXED(frame_info.width, display_region.width);
  frame_info.scaleY = DIV_FIXED(frame_info.height, display_region.height);

	display_region.right = display_region.left + display_region.width;
	display_region.bottom = display_region.top + display_region.height;
  return true;
}


bool VideoRenderer::open() {
	/* implement in subclass! */
  return true;
}


int VideoRenderer::writeFrame(Frame * frm, ESInfo *out_es) {
  struct AU *au;
  register int x, x0, y, y0;
  register u8 *pix, *pixline;
  register u32 yframe;
  register u32 one = INT2FIXED(1);

  if (!frm || !frm->getAU()) {
    dprintf_err("VideoRenderer::writeFrame(): invalid parameter: frm = %p, au = %p\n",
		frm, frm ? (void *) frm->getAU() : NULL);
    state = STREAMERR;
    return 0;
  }
  if (state != OPEN)
    return 1;
  framesSeen++;
  /* we don't set currentFrameNumber, as we have no way to determine it! */

  au = frm->getAU();
  if (au->size != (u32) frame_info.size) {
    dprintf_err("VideoRenderer::writeFrame(): frame size mismatch: expected %d bytes, "
		"got %u bytes\n", frame_info.size, au->size);
    state = STREAMERR;
    return 0;
  }

	/* convert frame to frameBuffer,
	 * scaling the frame according to frame_info.scaleX and frame_info.scaleY 
	 */
	frameBuffer->lock();
	switch (colorSpace) {
	case UncompressedVideoFrame::ColorSpaceGRAY8:
		CHOOSE_SCALE_FRAME(1, *pix)
			break;
	case UncompressedVideoFrame::ColorSpaceRGB24:
	  if (outputDepth >= 24) {
	    //note: it is BGR!
	    CHOOSE_SCALE_FRAME(3, (pix[0] << 16) | (pix[1] << 8) | pix[2])
	  } else {
	    CHOOSE_SCALE_FRAME(3,  ((pix[0] >> 3) << 11) + ((pix[1] >> 2) << 5) + (pix[2] >> 3))
	    // CHOOSE_SCALE_FRAME(3,  ((pix[2] >> 11) << 11) + ((pix[1] >> 10) << 5) + (pix[0] >> 11))
	  }
	  break;
	case UncompressedVideoFrame::ColorSpaceRGB32:
		/* Qt expected input format: BGRA */
		//    CHOOSE_SCALE_FRAME(4, qRgba(pix[2], pix[1], pix[0], pix[3]))
		/* SDL expected input format: ARGB */
		//				CHOOSE_SCALE_FRAME(4,  (pix[3] << 24) | (pix[2] << 16) | (pix[1] << 8) | pix[0])
		/* X11 expected input format: ARGB */


		//note: it is RGB!
		if (outputDepth >= 24)
			CHOOSE_SCALE_FRAME(4,  (pix[3] << 24) | (pix[2] << 16) | (pix[1] << 8) | pix[0])
		else
			CHOOSE_SCALE_FRAME(4,  ((pix[2] >> 3) << 11) + ((pix[1] >> 2) << 5) + (pix[0] >> 3))
		
			break;
	case UncompressedVideoFrame::ColorSpaceRGB565:
		/* How Do I Convert RGB5XX into a 16-bit number?
		 * #define RGB565(r, g, b) ((r >> 3) << 11)| ((g >> 2) << 5)| ((b >> 3) << 0)
		 * #define RGB555(r, g, b) ((r >> 3) << 10)| ((g >> 3) << 5)| ((b >> 3) << 0)
		 */

	  if (outputDepth >= 24) {
			//FIXME: the RGB green-part is fucked up, and I don't find the problem :(
			
	    //scale up from 16 to 24 bit
				//				CHOOSE_SCALE_FRAME(2,    ((pix[0] << 3) <<0)  |  (pix[0] >>5 |  pix[1] >>5 <<3)<<8 | ((pix[1] << 3)>>3)<<16  )
			CHOOSE_SCALE_FRAME(2,    ((pix[0] << 3)) /*blue*/ | ((pix[1] << 3)>>3<<16) /*red*/
												 | pix[0] >>5<<10  |  pix[1] >>5 <<13 /*green*/   )

				} else {
	    CHOOSE_SCALE_FRAME(2,  (pix[1]<<8) | pix[0])
	  }
	  break;
	default:
		/* make compiler happy */
		break;
	}
	frameBuffer->unlock();
	dprintf_full("VideoRenderer::writeFrame() posting paint event to frame buffer\n");
	frameBuffer->postPaintEvent();
	


#ifdef _POSIX_PRIORITY_SCHEDULING
  sched_yield(); //this is necessary to give parallel getFrames a chance
#endif

  return 1;
}

