 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 *  title:   VideoRenderer.hpp                                          *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *  $Id: VideoRenderer.hpp,v 1.15 2006/01/26 11:42:00 mkropfbe Exp $
 *                                                                      *
 ***********************************************************************/

#ifndef _VIDEORENDERER_HPP
#define _VIDEORENDERER_HPP

#include "IO.hpp"
#include "VideoESInfo.hpp"
#include "adaptors/MP4Decoder.hpp"
#include "VideoFrameBuffer.hpp"

class VideoFrameBuffer;


/**
 * @short This I/O class renders a gray-scale or RGB stream  
 * (via a QFrameBuffer) onto a QLabel widget. Don't forget to set the appropriate
 * VideoFrameBuffer!
 *
 * It is intended to be used as a DataSink connected to a DataChannel.
 * <p>
 * Video frames will be (down or up) scaled to fit into the QLabel widget by
 * preserving the aspect ratio given in VideoESInfo.</p>
 * <p>
 * The expected format of the gray-scale stream is: 1 byte per pixel.</p>
 * <p>
 * The expected format of the RGB24 stream is: 3 bytes per pixel,
 * in the order RGB (red, green, blue).</p>
 * <p>
 * The expected format of the RGB32 stream is: 4 bytes per pixel,
 * in the order BGRA (blue, green, red, alpha). This conforms to the
 * PIX_FMT_RGBA32 format defined in ffmpeg/libavcodec/avcodec.h on
 * little endian architectures.</p>
 *
 *@author Mario Taschwer and Michael Kropfberger
 *@version $Id: VideoRenderer.hpp,v 1.15 2006/01/26 11:42:00 mkropfbe Exp $
 */
class VideoRenderer : public IO
{
public:
	/* struct used for various video rectangles */
	typedef struct {
    int width;
    int height;
    int size;			// in bytes
		int ySize;    // if ColorSpace == YUV, then size of Y-part
		int uvSize;   // if ColorSpace == YUV, then size of UV-parts
		int top;
		int bottom;
		int left;
		int right;
    u32 scaleX;
    u32 scaleY;
		bool isLabelResizable;
    bool allowUpscaleX;
    bool fixAspectRatio;
	} VideoRect;



  /** Constructor. If the parameters are invalid, the #open() method will fail.
   * @param es        elementary stream information about the video stream to
   *                  be rendered.
   * @param colorSpace the color space of the video stream to be rendered;
   *                   must be one of: MP4Decoder::ColorSpaceGRAY8,
   *                   MP4Decoder::ColorSpaceRGB24,
   *                   MP4Decoder::ColorSpaceRGB32.
	 * @param width the width of the output label/window
	 * @param height the height of the output label/window
	 * @param isLabelResizable if <code>true</code>, the given width/height is 
	 *                      assumed to be flexible, so it will be tried to keep
	 *                      the original video size (plus adhering #fixAspectRatio)
	 *                      setting this to true, width&height may be zero, 
	 *                      only the video is used to determine the resolution
   * @param fixAspectRatio if <code>true</code>, video frames may be resized 
	 *                      according to their ESInfo aspectRatio, which means
	 *                      upscaling in the y-dimension
   * @param allowUpscaleX if <code>true</code>, video frames may be upscaled
   *                      in x-dimension to fit the label, which may
   *                      result in significant quality loss, since a very
   *                      simple upscaling algorithm is used (assumes isLabelResizable==false).
   */
  VideoRenderer(const VideoESInfo *es, UncompressedVideoFrame::ColorSpace colorSpace, 
								int width, int height, bool isLabelResizable = true, bool fixAspectRatio = true, 
									 bool allowUpscaleX = false);

  /** Destructor. */
  virtual ~VideoRenderer();

  /** Returns NULL, as this class is intended to be an output class only. */
  virtual Frame *getFrame() {
#ifdef _POSIX_PRIORITY_SCHEDULING
    sched_yield(); //this is necessary to give parallel getFrames a chance
#endif
    return NULL;
  };

	/* necessary for eg. X11, where the output may be 16bit, but the decoded video is RGB24/RGB32
	 * @param outputDepth may be either 8, 16, 24, which means the output RGB coding (defaults to 24)
	 */
	void setOutputDepth(int depth);
	
	/* if <code>direct</code> is set, calls the overwritten and optimized @bitBlt method
	 * of the subclass, else the slow pixel-per-pixel painting is used
	 */
	void setDirectBitBlt(bool direct);
	
	/* sets the VideoFrameBuffer, where the pixels are drawn to */
	void setVideoFrameBuffer( VideoFrameBuffer *vfb );

 /** Renders a frame to the destination QLabel widget.
   * @return 1 on success, 0 on error.
   */
  virtual int writeFrame(Frame * frm, ESInfo *out_es=NULL);


  /** Creates internal frame buffers and checks whether the stream to be
   * rendered matches the widget dimensions.
   * @return <code>true</code> on success, <code>false</code> on error.
   * @see IO::open()
   */
  virtual bool open();

  /** Does nothing but setting the object state to CLOSED.
   * @param immediate ignored.
   * @return true
   * @see IO::close(bool)
   */
  virtual bool close(bool immediate=false) {
    state=CLOSED;
    return true;
  };

  /** Not applicable to this class.
   * @return true
   * @see IO::destroy()
   */
  virtual bool destroy() {
    return true;
  };
 
  /** Not applicable to this class.
   * @return 50
   * @see IO::getBufferFillLevel()
   */
  virtual int getBufferFillLevel() const {
    return 50;
  };
  
  /** Returns the fixed string <code>VideoRenderer:</code> */
  virtual const char *getURL() const {
    return url;
  };

  /** Not implemented. */
  virtual void setESInfo(ESInfo * es) {};

  /** Not implemented.
   * @return NULL
   */
  virtual ESInfo* getESInfo() { return NULL;};


protected:
  /** Update {@link #frame_info} (except <code>size</code>)
   * and {@link #display_region}.
   * This function is not thread-safe, and is internally
   * called by the {@link #open()} method.
  */
  bool updateDisplayRegion();

	//private:
  char url[MAX_STR_LEN];

  /** Elementary video stream information. */
  const VideoESInfo *es;

  /** The frame buffer. */
  VideoFrameBuffer *frameBuffer;

  /** The color space of the frames to be rendered. */
  UncompressedVideoFrame::ColorSpace colorSpace;

  /** Dimension, storage size (in bytes), and scaling factors of a 
   * decoded video frame. The scaling factors are defined as<br>
   * <code>sx = frame_info.width / display_region.width</code><br>
   * <code>sy = frame_info.height / display_region.height</code>
   * <p>
   * We use fixed point arithmetic with FIXED_DIGITS binary precision,
   * so<br>
   * <code>scaleX = (int) (sx * 2^FIXED_DIGITS)</code>
   * </p><p>
   * The product <code>x * sx</code> for an integer <code>x</code>
   * is then computed as the scaled integer product<br>
   * <code>(x * scaleX) >> FIXED_DIGITS</code><br>
   * which should be considerably faster than floating point
   * multiplication.
   * </p>
   */
	VideoRect frame_info;
	/* exact size of output label (this might be larger than frame or display) */
	VideoRect out_label_info;

  /** apect-ratio adjusted region of qlabel where frames will be displayed. 
   * Frame dimensions may be smaller than display_region and out_label_info dimensions.
   */
  VideoRect display_region;

	int outputDepth;

	bool directBitBlt;



};







/* The number of binary precision digits for fixed point arithmetic.
 * As the product of a pixel coordinate x and frame_info.scaleX
 * must fit into an u32 data type, the condition
 * log2(MAX(frame_info.width, frame_info.height)) + FIXED_DIGITS <= 32
 * must hold. So a value of 20 for FIXED_DIGITS seems to be reasonable.
 */
#define FIXED_DIGITS 20

/* Convert an unsigned integer to a fixed point number of type u32.
 * The integer must be less than 2^(32 - FIXED_DIGITS).
 */
#define INT2FIXED(i) ( (u32) ( (u32)(i) << FIXED_DIGITS ) )

/* Convert a fixed point number of type u32 to floating point. */
#define FIXED2DOUBLE(f) ( (double)(f) / pow(2.0, (double)FIXED_DIGITS) )

/* Multiply an integer by a fixed point number, resulting in an integer
 * of type u32.
 */
#define MULT_INT_FIXED(i, f) ( (u32) ( ( (u32)(i)*(u32)(f) ) >> FIXED_DIGITS ) )

/* Divide two integers, resulting in a fixed point number of type u32. */
#define DIV_FIXED(i, k) ( (u32) ( ( (u32)(i) << FIXED_DIGITS ) / (u32)(k) ) )






/* Macro used in writeFrame() to compute a scaled frame in frameBuffer.
 * Parameters: depth - pixel depth in bytes.
 *             rgb   - expression computing an RGB value from the pointer pix
 *                     to a pixel in the decoded frame.
 */
#define SCALE_FRAME(depth, rgb)                                                            \
    for (y = display_region.top, y0 = 0; y < display_region.bottom; y++, y0++) { \
      yframe = (frame_info.scaleY == one) ? y0 : MULT_INT_FIXED(y0, frame_info.scaleY);	\
      pixline = au->payload + yframe * frame_info.width * (depth);			\
      for (x = display_region.left, x0 = 0; x < display_region.right; x++, x0++) { \
				pix = pixline + MULT_INT_FIXED(x0, frame_info.scaleX) * (depth); \
				frameBuffer->setPixel(x, y, (uint)(rgb));												\
      }																																	\
    }

/* Macro used in writeFrame() to compute a scaled frame in frameBuffer.
 * Optimized version for scaling in y-direction only.
 * Parameters: depth - pixel depth in bytes.
 *             rgb   - expression computing an RGB value from the pointer pix
 *                     to a pixel in the decoded frame.
 */
#define SCALE_FRAME_Y(depth, rgb)                                                          \
    for (y = display_region.top, y0 = 0; y < display_region.bottom; y++, y0++) {  \
      yframe = (frame_info.scaleY == one) ? y0 : MULT_INT_FIXED(y0, frame_info.scaleY);    \
      pix = au->payload + yframe * frame_info.width * (depth);					\
      for (x = display_region.left; x < display_region.right; x++) { \
				frameBuffer->setPixel(x, y, (uint)(rgb));												\
        pix += (depth);																									\
      }																																	\
    }

/* Macro used in writeFrame() to copy a frame to frameBuffer.
 * Optimized version for no scaling.
 * if directBitBlt is set, a even more sophisticated subclass method is calledmuh
 * Parameters: depth - pixel depth in bytes.
 *             rgb   - expression computing an RGB value from the pointer pix
 *                     to a pixel in the decoded frame.
 */
#define COPY_FRAME(depth, rgb)																					\
	if (directBitBlt) {																										\
		frameBuffer->bitBlt(au);																						\
	} else {																															\
  dprintf_full("CHOOSE_SCALE slow COPY_FRAME depth %i\n",depth);				\
		pix = au->payload;																									\
		for (y = display_region.top; y < display_region.bottom; y++) {			\
			for (x = display_region.left; x < display_region.right; x++) {		\
				frameBuffer->setPixel(x, y, (uint)(rgb));												\
				pix += (depth);																									\
			}																																	\
		}																																		\
	}




/* Macro used in writeFrame() to choose an optimized scaling routine
 * based on the values of frame_info.scaleX and frame_info.scaleY.
 * Parameters: depth - pixel depth in bytes.
 *             rgb   - expression computing an RGB value from the pointer pix
 *                     to a pixel in the decoded frame.
 */
#define CHOOSE_SCALE_FRAME(depth, rgb)                          \
	if (frame_info.scaleX == one && frame_info.scaleY == one) {		\
		COPY_FRAME(depth, rgb);																			\
			} else if (frame_info.scaleX == one) {										\
		SCALE_FRAME_Y(depth, rgb)																		\
			} else {																									\
		SCALE_FRAME(depth, rgb)																			\
			}



#endif /* _VIDEORENDERER_HPP */
