 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:    X11Renderer.cpp                                            *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *  $Id: X11Renderer.cpp,v 1.12 2006/01/20 15:37:18 mkropfbe Exp $
 *                                                                      *
 ***********************************************************************/

#ifndef WIN32

#include "global.hpp"
#include "IO.hpp"
#include "Frame.hpp"
#include "X11Renderer.hpp"
#include "X11FrameBuffer.hpp"
#include "VideoRenderer.hpp"


X11Renderer::X11Renderer(const VideoESInfo *_es, UncompressedVideoFrame::ColorSpace colorSpace, 
																	 int width, int height, bool pollEvents,
																	 bool isLabelResizable, bool fixAspectRatio, bool allowUpscaleX)
	: VideoRenderer(_es, colorSpace, width, height, isLabelResizable, fixAspectRatio, allowUpscaleX)
{
	dprintf_full("X11Renderer Construct %s ES %p\n",this->url, this->es);
	strcpy(this->url,"X11Renderer");
	this->pollEvents = pollEvents;
}

X11Renderer::~X11Renderer()
{
	showCursor(display,xwin);

	XDestroyWindow(display,xwin);
  XCloseDisplay(display);
}

bool X11Renderer::updateDisplayRegion() {
	
  if (!es || !es->isVisualStream()) {
    dprintf_err("X11Renderer::updateDisplayRegion(): invalid members: es = %p\n",es);
		::exit(1);
    return false;
  }

	return VideoRenderer::updateDisplayRegion();
}


void X11Renderer::hideCursor(Display * disp, Window win)
{
    Cursor no_ptr;
    Pixmap bm_no;
    XColor black, dummy;
    Colormap colormap;
    static char bm_no_data[] = { 0, 0, 0, 0, 0, 0, 0, 0 };


    colormap = DefaultColormap(disp, DefaultScreen(disp));
    XAllocNamedColor(disp, colormap, "black", &black, &dummy);
    bm_no = XCreateBitmapFromData(disp, win, bm_no_data, 8, 8);
    no_ptr = XCreatePixmapCursor(disp, bm_no, bm_no, &black, &black, 0, 0);
    XDefineCursor(disp, win, no_ptr);
    XFreeCursor(disp, no_ptr);
    if (bm_no != None)
        XFreePixmap(disp, bm_no);
    XFreeColors(disp,colormap,&black.pixel,1,0);
}

void X11Renderer::showCursor(Display * disp, Window win)
{
    XDefineCursor(disp, win, 0);
}


bool X11Renderer::open()
{
	if (state != CLOSED) {
		dprintf_err("X11Renderer::open  state was not CLOSED but state=%i!\n",state);
		return false;
	}

  if (!updateDisplayRegion()) {
    state = STREAMERR;
    return false;
  }

	char hostname[MAX_STR_LEN];
	char *envhost = getenv("DISPLAY");
	if (!envhost)
		strcpy(hostname,"localhost:0");
	else
		strcpy(hostname,envhost);

	display = XOpenDisplay(hostname);
	assert(display);
	screennr = XDefaultScreen(display);
	screen = XDefaultScreenOfDisplay(display);
	depth = XDefaultDepth(display,screennr);
	winAttrib.background_pixel =  BlackPixel(display, screennr);
	winAttrib.border_pixel = BlackPixel(display, screennr);
	winAttrib.backing_store = Always;
	winAttrib.bit_gravity = StaticGravity;
	XVisualInfo visual_info;
	Visual *def_visual = XDefaultVisual(display,screennr);
	unsigned int attr_val = CWBackingStore | CWBackPixel | CWBorderPixel; 
	int visualClass = -1;

	setOutputDepth(depth);

	/* FIXME: only tested for DirectColor on 16/24 bit X11 display! */
	if (!XMatchVisualInfo(display, screennr, depth, PseudoColor, &visual_info)) {
      if (!XMatchVisualInfo(display, screennr, depth, DirectColor, &visual_info)) {
         /* No PseudoColor visual available at default_depth;
           * some applications might try for a GrayScale
           * visual here if they can use gray to advantage,
           * before giving up and using black and white */
				dprintf_err("X11Renderer::open() could not find PseudoColor or DirectColor visual!\n");
				::exit(0);
      } else
				visualClass = DirectColor;
   } else
		visualClass = PseudoColor;

	dprintf_full("X11Renderer:open   outputDepth %i, defaultVisual %p, pseudo/directColor visualID 0x%x\n",
							 depth, def_visual, (int)visual_info.visualid);
	
	
	
	
	if (visualClass != DirectColor)
		XCreateColormap(display, RootWindowOfScreen(screen), visual_info.visual, AllocNone);
	else {				
		XColor cols[256];
		unsigned int cm_size, red_mask, green_mask, blue_mask;
		unsigned int k, r, g, b, ru, gu, bu, m, rv, gv, bv, rvu, gvu, bvu;
		cm_size = visual_info.colormap_size;
    red_mask = visual_info.red_mask;
    green_mask = visual_info.green_mask;
    blue_mask = visual_info.blue_mask;
    ru = (red_mask & (red_mask - 1)) ^ red_mask;
    gu = (green_mask & (green_mask - 1)) ^ green_mask;
    bu = (blue_mask & (blue_mask - 1)) ^ blue_mask;
    rvu = 65536ull * ru / (red_mask + ru);
    gvu = 65536ull * gu / (green_mask + gu);
    bvu = 65536ull * bu / (blue_mask + bu);
    r = g = b = 0;
    rv = gv = bv = 0;
    m = DoRed | DoGreen | DoBlue;
    for (k = 0; k < cm_size; k++) {
        unsigned int t;

        cols[k].pixel = r | g | b;
        cols[k].red = rv;
        cols[k].green = gv;
        cols[k].blue = bv;
        cols[k].flags = m;
        t = (r + ru) & red_mask;
        if (t < r)
            m &= ~DoRed;
        r = t;
        t = (g + gu) & green_mask;
        if (t < g)
            m &= ~DoGreen;
        g = t;
        t = (b + bu) & blue_mask;
        if (t < b)
            m &= ~DoBlue;
        b = t;
        rv += rvu;
        gv += gvu;
        bv += bvu;
    }
    winAttrib.colormap = XCreateColormap(display, RootWindowOfScreen(screen), visual_info.visual, AllocAll);
    XStoreColors(display, winAttrib.colormap, cols, cm_size);
	}

		XInstallColormap(display, winAttrib.colormap);	// put cm into hardware!

	attr_val |= CWColormap;

	xwin=XCreateWindow(display, XRootWindow(display,screennr)
											,0, 0, out_label_info.width, out_label_info.height, 1
											,depth, InputOutput,
										 visual_info.visual 
										 //def_visual 
										 , attr_val, &winAttrib);
	
	XStoreName(display, xwin, "ViTooKi X11Renderer");
	XSetIconName(display, xwin, "ViTooKi X11Renderer");
 	XSetWindowBackground(display, xwin, 0xFFFFFF);
	XMapWindow(display, xwin);
	XSync(display, 0);

	hideCursor(display,xwin);
	
  /* initialize frameBuffer */
  if (frameBuffer) {
    delete frameBuffer;
    frameBuffer = NULL;
  }

	bool gray = false;
  switch (colorSpace) {
  case UncompressedVideoFrame::ColorSpaceGRAY8:
		dprintf_full("X11Renderer::open()  ColorSpaceGRAY8\n");
		gray=true;
    frame_info.size = frame_info.width * frame_info.height;
    break;
  case UncompressedVideoFrame::ColorSpaceRGB24:
		dprintf_full("X11Renderer::open()  ColorSpaceRGB24\n");
    frame_info.size = frame_info.width * frame_info.height * 3;
    break;
  case UncompressedVideoFrame::ColorSpaceRGB32:
		dprintf_full("X11Renderer::open()  ColorSpaceRGB32\n");
    frame_info.size = frame_info.width * frame_info.height * 4;
    break;
  case UncompressedVideoFrame::ColorSpaceRGB565:
		dprintf_full("X11Renderer::open()  ColorSpaceRGB565\n");
    frame_info.size = frame_info.width * frame_info.height * 2;
    break;
	case UncompressedVideoFrame::ColorSpaceYV12:
	case UncompressedVideoFrame::ColorSpaceI420:
		dprintf_err("X11Renderer::open()  ColorSpaceYV12 or I420    NOT SUPPORTED on X11!!!\n");
		::exit(1);
		break;
  default:
    dprintf_err("X11Renderer::open(): invalid color space: %d\n", colorSpace);
    state = STREAMERR;
    return false;
  }

	if  ( (depth == 16) && //FIXME: sorry, but 24 bitBlt does not seem to work...  32bit is untested
			 (depth == UncompressedVideoFrame::mapColorSpaceToBitDepth(colorSpace) ) &&
				(out_label_info.width ==  frame_info.width) && (out_label_info.height ==  frame_info.height) &&
				(display_region.left == 0) &&  (display_region.top == 0) )
		directBitBlt=true;  //optimize via rgbSurface bitBlitting
	else
		directBitBlt=false; //use slow pixel-per-pixel painting and scaling


	frameBuffer = new X11FrameBuffer(display, &xwin, depth, gray, display_region.width, display_region.height );

  if (!frameBuffer) {
    dprintf_err("X11Renderer::open(): X11FrameBuffer creation FAILED! (w = %d, h = %d)\n",
								display_region.width, display_region.height );
    state = STREAMERR;
    return false;
  }

  state = OPEN;
  framesSeen = 0;
  currentFrameNumber = 0;
  dprintf_small("X11Renderer::open(): initialized frame and display regions:\n"
								"    frame: width = %d, height = %d, size = %d, scaleX = %5.3f, scaleY = %5.3f\n"
								"    display: x(%d:%d), y(%d:%d), width = %d, height = %d\n",
								frame_info.width, frame_info.height, frame_info.size,
								FIXED2DOUBLE(frame_info.scaleX), FIXED2DOUBLE(frame_info.scaleY),
								display_region.left, display_region.left + display_region.width,
								display_region.top, display_region.top + display_region.height,
								display_region.width, display_region.height);
  return true;
}


int X11Renderer::writeFrame(Frame * frm, ESInfo *out_es) {
	return VideoRenderer::writeFrame(frm, out_es);

}

#endif /* WIN32 */

