 /***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 *  title:   X11Renderer.hpp                                            *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

 /***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *  $Id: X11Renderer.hpp,v 1.10 2006/05/12 11:32:50 mkropfbe Exp $
 *                                                                      *
 ***********************************************************************/

#ifndef WIN32

#ifndef _X11RENDERER_HPP
#define _X11RENDERER_HPP

#include <X11/Xatom.h>
#include <unistd.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>

#include "IO.hpp"
#include "VideoESInfo.hpp"
#include "adaptors/MP4Decoder.hpp"

#include "VideoRenderer.hpp"


class X11FrameBuffer;


/**
 * @short This I/O class renders a gray-scale or RGB24/32 stream  
 * the used X11-Window is created within this function, 
 * incl. CTRL-C/ESC signal/event catchers
 *
 * It is intended to be used as a DataSink connected to a DataChannel.
 * <p>
 * Video frames will be (down or up) scaled to fit into the given size by
 * preserving the aspect ratio given in VideoESInfo.</p>
 * <p>
 * The expected format of the gray-scale stream is: 1 byte per pixel.</p>
 * <p>
 * The expected format of the RGB24 stream is: 3 bytes per pixel,
 * in the order RGB (red, green, blue).</p>
 * <p>
 * The expected format of the RGB32 stream is: 4 bytes per pixel,
 * in the order BGRA (blue, green, red, alpha). This conforms to the
 * PIX_FMT_RGBA32 format defined in ffmpeg/libavcodec/avcodec.h on
 * little endian architectures.</p>
 *
 *@author Michael Kropfberger
 *@version $Id: X11Renderer.hpp,v 1.10 2006/05/12 11:32:50 mkropfbe Exp $
 */
class X11Renderer : public VideoRenderer
{
public:

  /** Constructor. If the parameters are invalid, the #open() method will fail.
   * @param es        elementary stream information about the video stream to
   *                  be rendered.
   * @param colorSpace the color space of the video stream to be rendered;
   *                   must be one of: UncompressedVideoFrame::ColorSpaceGRAY8, UncompressedVideoFrame::ColorSpaceRGB565,
   *                   UncompressedVideoFrame::ColorSpaceRGB24, UncompressedVideoFrame::ColorSpaceRGB32,
   *                   UncompressedVideoFrame::ColorSpaceI420, UncompressedVideoFrame::ColorSpaceYV12
	 * @param width     the width of the SDL window
	 * @param height    the height of the SDL window
	 * @param pollEvents  if <code>true</code>, the SDL event handler is checked for
	 *                  keypresses like CTRL-C or ESC, and a "SDL window close" is detected
	 *                  if such an event occurs, the state of the stream is set to <code>STREAMERR</code>
	 *                  beware, that this polling might grab away ALL events, so implement on
	 *                  on your own, if you want to support eg. joysticks or more keycodes!
	 * @param isLabelResizable if <code>true</code>, the given width/height is 
	 *                      assumed to be flexible, so it will be tried to keep
	 *                      the original video size (plus adhering #fixAspectRatio)
	 *                      setting this to true, width&height may be zero, 
	 *                      only the video is used to determine the resolution
   * @param fixAspectRatio if <code>true</code>, video frames may be resized 
	 *                      according to their ESInfo aspectRatio, which means
	 *                      upscaling in the y-dimension
   * @param allowUpscaleX if <code>true</code>, video frames may be upscaled
   *                      in x-dimension to fit the label, which may
   *                      result in significant quality loss, since a very
   *                      simple upscaling algorithm is used (assumes isLabelResizable==false).
   */
  X11Renderer(const VideoESInfo *es, UncompressedVideoFrame::ColorSpace colorSpace,
									 int width, int height,
									 bool pollEvents = true,
									 bool isLabelResizable = true, bool fixAspectRatio = true, 
									 bool allowUpscaleX = false);

  /** Destructor. */
  virtual ~X11Renderer();

  /** Renders a frame to the destination YUV_Overlay/RGB_Surface
   * @return 1 on success, 0 on error.
   */
  virtual int writeFrame(Frame * frm, ESInfo *out_es=NULL);

  /** Creates internal frame buffers and checks whether the stream to be
   * rendered matches the widget dimensions.
   * @return <code>true</code> on success, <code>false</code> on error.
   * @see IO::open()
   */
  virtual bool open();

protected:
  /** Update {@link #frame_info} (except <code>size</code>)
   * and {@link #display_region}.
   * This function is not thread-safe, and is internally
   * called by the {@link #open()} method.
  */
  bool updateDisplayRegion();


private:

	bool     pollEvents;
  Display *display;
  Screen  *screen;
  int      screennr, depth;
  Window   xwin;
  XSetWindowAttributes winAttrib;

	void hideCursor(Display * disp, Window win);


  void showCursor(Display * disp, Window win);


};


#endif /* _X11RENDERER_HPP */

#endif /* WIN32 */
