/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	MP7File.cpp                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

/***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#include "MP7File.hpp"
#include "ContainerInfo.hpp"
#include "VideoESInfo.hpp"
#include "AdaptorChain.hpp"

#include "../container/ISOMP4ContainerFile.hpp"
#include "../container/ContainerDemux.hpp"

#include "../adaptors/Forwarder.hpp"
#include "../adaptors/SpatialReductionAdaptor.hpp"
#include "../net/Url.hpp"

const char *const
MP7File::NL = "\r\n";
const char *const
MP7File::content = "Content";
const char *const
MP7File::name = "Name";
const char *const
MP7File::fileFormat = "FileFormat";
const char *const
MP7File::variationSet = "VariationSet";
const char *const
MP7File::fileSize = "FileSize";
const char *const
MP7File::mediaProfile = "MediaProfile";
const char *const
MP7File::componentMediaProfile = "ComponentMediaProfile";
const char *const
MP7File::bitRate = "BitRate";
const char *const
MP7File::mediaLocator = "MediaLocator";
const char *const
MP7File::preHeader =
"<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>%s<Mpeg7 xmlns=\"urn:mpeg:mpeg7:schema:2001\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns:mpeg7=\"urn:mpeg:mpeg7:schema:2001\" xsi:schemaLocation=\"urn:mpeg:mpeg7:schema:2001 definitions2.xsd\">%s\t<Description xsi:type=\"VariationDescriptionType\">%s";
const char *const
MP7File::closeHeader = "\t</Description>%s</Mpeg7>%s";

/** generates an MPEG-7 description for a locally stored ContainerInfo object.
 * This method will apply different Adaptors to the video 
 * and use this information to generate the mp7file. 
 * Note: this method can take very long.
 * In case of an error NULL is returned, otherwise the name of the generated MPEG-7 file
 */
char *MP7File::generateMP7Description(ContainerInfo * mp4Info,
																			char *mp7File, bool showVisualResult)
{

	FILE *fp = MP7File::openMP7File(mp7File);
	if (!fp)
		return NULL;

	/** demultiplex to determine size of each ES */
	ContainerDemux::demultiplexAndUpdateIO(mp4Info,".");

	ContainerInfo *clone = mp4Info->clone();
	VideoESInfo *esi = clone->getFirstVisualES();
	unsigned int w = esi->getWidth() / 2;
	unsigned int h = esi->getHeight() / 2;
	Adaptor **adaptors = new Adaptor *[3];
	adaptors[0] = new Forwarder();
	adaptors[1] = new TemporalAdaptor();
	adaptors[2] = new StrongTemporalAdaptor();
	MP7File::writeVariationSet(fp, adaptors, 3, clone, showVisualResult);
	delete clone;
	delete adaptors[0];
	delete adaptors[1];
	delete adaptors[2];
	delete adaptors;

	clone = mp4Info->clone();
	esi = clone->getFirstVisualES();
	adaptors = new Adaptor *[1];
	adaptors[0] =
		new SpatialReductionAdaptor(esi, &w, &h,
																esi->getAvgBandwidth() / 4);
	MP7File::writeVariationSet(fp, adaptors, 1, clone, showVisualResult);
	delete clone;
	delete adaptors[0];
	delete adaptors;


	MP7File::closeMP7File(fp);

	return mp7File;
};


FILE *MP7File::openMP7File(const char *mp7File)
{
	FILE *fp = fopen(mp7File, "wb");
	if (!fp)
		return NULL;
	fprintf(fp, preHeader, NL, NL, NL);
	return fp;
};

/** closes an mpeg-7 file, writes the closing header, returns false if fp is NULL */
bool MP7File::closeMP7File(FILE * fp)
{
	if (!fp)
		return false;
	fprintf(fp, closeHeader, NL, NL);
	fclose(fp);
	return true;
}

bool MP7File::writeVariationSet(FILE * fp, Adaptor ** adaptors,
																int numAdaptors, ContainerInfo * mp4Info,
																bool showVisualResult)
{
	MP7File::writeEntry(fp, 2, variationSet, false);
	MP7File::writeEntry(fp, 3, "Source xsi:type=\"VideoType\"", false);
	MP7File::writeEntry(fp, 4, "Video id=\"SourceVideo\"", false);
	MP7File::writeMediaInformation(fp, 5,mp4Info);
	MP7File::writeEntry(fp, 4, "Video", true);
	MP7File::writeEntry(fp, 3, "Source", true);
	char *adaptorName = new char[MAX_STR_LEN];
	dprintf_full("MP7File::writeVariationSet Doing %i adaptations\r\n",numAdaptors);
	// description of the video is finished
	// now do the Variation description
	float fidelity = 1.0;	// FIXME: calculate fidelity
	for (int i = 0; i < numAdaptors; i++) {
		strcpy(adaptorName,adaptors[i]->getName());
		dprintf_full("MP7File::writeVariationSet Doing %s\r\n", adaptors[i]->getName());
		if (showVisualResult) {
	    ContainerDemux::demultiplexAndUpdateIO(mp4Info, ".", adaptors[i], false,
																						 false, true);
		} else
	    ContainerDemux::demultiplexAndUpdateIO(mp4Info, ".", adaptors[i], false,
																						 false, false);

		fidelity -= (float) (1.0 / (numAdaptors + 1));	// FIXME
		MP7File::writeAdaptation(fp, 3, mp4Info, adaptorName, i + 1, i + 1,
														 fidelity);
	}

	MP7File::writeEntry(fp, 2, variationSet, true);

	delete [] adaptorName;
	return true;
};

u32 MP7File::writeMediaInformation(FILE * fp, u32 indent,
																	 ContainerInfo * mp4Info)
{
	MP7File::writeEntry(fp, indent, "MediaInformation", false);
	MP7File::writeEntry(fp, indent + 1, mediaProfile, false);

	// write for each ES a ComponentMediaProfile entry
	list < ESInfo * >*l = mp4Info->getESList();
	list < ESInfo * >::iterator li = l->begin();
	u64 totalSize = 0;
	u32 maxDurInMs = 1;
	while (li != l->end()) {
		totalSize += MP7File::writeCMP(fp, indent + 2, (*li));
		if (maxDurInMs < (*li)->getDurationInMs())
			maxDurInMs = (*li)->getDurationInMs();
		++li;
	}
	// now write the info for the complete mp4 stream
	MP7File::writeEntry(fp, indent + 2, "MediaFormat", false);

	MP7File::writeEntry(fp, indent + 3, "Content href=\"MPEG7ContentCS\"",
											false);
	MP7File::writeEntry(fp, indent + 4, "Name>audiovisual</Name", false);
	MP7File::writeEntry(fp, indent + 3, "Content", true);

	MP7File::writeEntry(fp, indent + 3,
											"FileFormat href=\"urn:mpeg:mpeg7:cs:FileFormatCS:2001:5\"",
											false);
	MP7File::writeEntry(fp, indent + 4, "Name xml:lang=\"en\">mp4</Name",
											false);
	MP7File::writeEntry(fp, indent + 3, fileFormat, true);

	// now write the filesize & bitrate of the complete stream
	MP7File::writeVideoDescriptionEntry(fp, indent + 3, fileSize,
																			totalSize);
	MP7File::writeVideoDescriptionEntry(fp, indent + 3, bitRate,
																			((8000 * totalSize) / maxDurInMs));
	// debug info
	//      MP7File::writeVideoDescriptionEntry(fp,indent+3,"DEBUG durationInMs", maxDurInMs);

	MP7File::writeEntry(fp, indent + 2, "MediaFormat", true);

	// now write MediaInstance
	MP7File::writeEntry(fp, indent + 2, "MediaInstance", false);
	MP7File::writeEntry(fp, indent + 3,
											"InstanceIdentifier organization=\"MPEG\" type=\"MPEG7ContentSetOnLineId\"> mpeg7 </InstanceIdentifier",
											false);
	MP7File::writeEntry(fp, indent + 3, mediaLocator, false);
	const char* u=mp4Info->getLocalFile();
	MP7File::writeVideoDescriptionEntry(fp, indent + 4, "MediaUri",
																			u);
	MP7File::writeEntry(fp, indent + 3, mediaLocator, true);
	MP7File::writeEntry(fp, indent + 2, "MediaInstance", true);

	MP7File::writeEntry(fp, indent + 1, mediaProfile, true);
	MP7File::writeEntry(fp, indent, "MediaInformation", true);
	return totalSize;
};

/** sets the MetaItems (incl. Adaptors) in the QualityInfo object of the ContainerInfo object
 * by parsing the mp7File. The .mp7 file must be in ASCII format, BIM is not supported.
 * If sth. went wrong false is returned, otherwise true.
 */
bool MP7File::setAdaptors(ContainerInfo * mp4Info, char *mp7File)
{
	return false;
};

/** convert the Meta-Info stored in the mp4Info object to an MPEG-7 description.
 * Already applied adaptors will be excluded from the MPEG-7 description.
 * Not yet implemented!
 */
bool MP7File::convertMetaInfoToMP7(const ContainerInfo * mp4Info,
																	 char *mp7File)
{
	return false;
};

/** writes the following MPEG-7 entries: <key/> */
void MP7File::writeSimpleEntry(FILE * fp, u32 indent,
															 const char *const key)
{
	for (u32 i = 0; i < indent; i++) {
		fputc((int) '\t', fp);
	}
	fprintf(fp, "<%s/>%s", key, NL);
}

void MP7File::writeEntry(FILE * fp, u32 indent, const char *const key,
												 bool close)
{
	for (u32 i = 0; i < indent; i++) {
		fputc((int) '\t', fp);
	}
	if (close)
		fprintf(fp, "</%s>%s", key, NL);
	else
		fprintf(fp, "<%s>%s", key, NL);
};

u32 MP7File::writeCMP(FILE * fp, u32 indent, ESInfo * es)
{
	char *temp = new char[4096];
	temp[0] = 0;
	sprintf(temp, "ComponentMediaProfile id=\"ES%llu\"", es->getStreamId());
	MP7File::writeEntry(fp, indent, temp, false);
	MP7File::writeEntry(fp, indent + 1, "MediaFormat", false);

	// write content
	MP7File::writeEntry(fp, indent + 2, "Content href=\"MPEG7ContentCS\"",
											false);

	temp[0] = 0;
	if (es->isVisualStream()) {
		MP7File::writeEntry(fp, indent + 3, "Name>visual</Name", false);
		MP7File::writeEntry(fp, indent + 2, "Content", true);
		//write fileformat
		//<VisualCoding>
		//      <Format href="urn:mpeg:mpeg7:cs:VisualCodingFormatCS:2001:1" colorDomain="color"/>
		//      <Frame height="240" rate="" width="320"/>
		//</VisualCoding>
		MP7File::writeEntry(fp, indent + 2, "VisualCoding", false);
		MP7File::writeSimpleEntry(fp, indent + 3,
															"Format href=\"urn:mpeg:mpeg7:cs:VisualCodingFormatCS:2001:3\"");
		MP7File::writeEntry(fp, indent + 3,
												"Name xml:lang=\"en\">MPEG-4 Visual</Name",
												false);
		sprintf(temp, "Frame height=\"%i\" width=\"%i\" rate=\"%.2f\"",
						((VideoESInfo*)es)->getHeight(), ((VideoESInfo*)es)->getWidth(), ((VideoESInfo*)es)->getFPS());
		MP7File::writeSimpleEntry(fp, indent + 3, temp);
		MP7File::writeEntry(fp, indent + 2, "VisualCoding", true);

	} else if (es->isAudioStream()) {
		MP7File::writeEntry(fp, indent + 3, "Name>audio</Name", false);
		MP7File::writeEntry(fp, indent + 2, "Content", true);
		// FIXME: no possibility to detect which audio we have: hard-code zu MP3
		//write fileformat
		MP7File::writeEntry(fp, indent + 2, "AudioCoding", false);
		MP7File::writeSimpleEntry(fp, indent + 3,
															"Format href=\"urn:mpeg:mpeg7:cs:AudioCodingFormatCS:2001:3.3\"");
		MP7File::writeEntry(fp, indent + 3,
												"Name xml:lang=\"en\">MPEG-1 Audio Layer III</Name",
												false);
		// close fileformat
		MP7File::writeEntry(fp, indent + 2, "AudioCoding", true);
	} else {
		MP7File::writeEntry(fp, indent + 3, "Name>generic</Name", false);
		MP7File::writeEntry(fp, indent + 2, "Content", true);
		// FIXME: no possibility to detect if BIFS or ODstream
		// set to MPEG-4 Scene Description Simple 2D Profile
		//write fileformat
		MP7File::writeEntry(fp, indent + 2, "SceneCodingFormat", false);
		MP7File::writeSimpleEntry(fp, indent + 3,
															"Format href=\"urn:mpeg:mpeg7:cs:SceneCodingFormatCS:2001:1.1\"");
		MP7File::writeEntry(fp, indent + 3,
												"Name xml:lang=\"en\">MPEG-4 Scene Description</Name",
												false);
		// close fileformat
		MP7File::writeEntry(fp, indent + 2, "SceneCodingFormat", true);
	}

	u64 size = 0;
	const char *inp = es->getInput();
	dprintf_full(inp);
	FILE *f;
	if ((f=fopen(inp,"r"))) {
		fseek(f,0L, SEEK_END);
		size=ftell(f);
		fclose(f);
	}

	// write size: <FileSize>10000000</FileSize>
	// Note: size contains header too 
	//      MP7File::writeVideoDescriptionEntry(fp,indent+2,fileSize,size);

	/*	// now calc the bitrate -> 
			u32 bitrate=0;
			if(es->isVisualStream() || es->isAudioStream()) {
			bitrate=(u32)((8000*size)/es->getDurationInMs());
			}	
			else { //generic streams are soo small that the produce irregulary high bitrate values	
			if(es->getDurationInMs()<1000) {
			bitrate=size*8;
			}
			else {	
			bitrate=(u32)((8000*size)/es->getDurationInMs());
			}
			}	*/
	//      MP7File::writeVideoDescriptionEntry(fp,indent+2,bitRate,bitrate);
	MP7File::writeEntry(fp, indent + 1, "MediaFormat", true);
	MP7File::writeEntry(fp, indent, componentMediaProfile, true);

	delete [] temp;
	return size;
};

void MP7File::writeVideoDescriptionEntry(FILE * fp, u32 indent,
																				 const char *const key,
																				 u32 intValue)
{
	for (u32 i = 0; i < indent; i++) {
		fputc((int) '\t', fp);
	}
	fprintf(fp, "<%s>%u</%s>%s", key, intValue, key, NL);
};

void MP7File::writeVideoDescriptionEntry(FILE * fp, u32 indent,
																				 const char *const key,
																				 const char *stringVal)
{
	for (u32 i = 0; i < indent; i++) {
		fputc((int) '\t', fp);
	}
	fprintf(fp, "<%s>%s</%s>%s", key, stringVal, key, NL);
};

void MP7File::writeAdaptation(FILE * fp, u32 indent, ContainerInfo * mp4,
															const char * adaptName, u32 variationId,
															u32 priority, float fidelity)
{
	//              <Variation id="VARIATION1" fidelity="0.85" priority="1">
	//                              <!-- <SourceRef idref="ES1"/> -->
	//                              <Content xsi:type="VideoType">
	//                                      <Video id="var1">
	//                                              <MediaInformation>
	char *a = new char[128];
	sprintf(a,
					"Variation id=\"VARIATION%i\" fidelity=\"%f\" priority=\"%i\"",
					variationId, fidelity, priority);
	MP7File::writeEntry(fp, indent, a, false);
	MP7File::writeEntry(fp, indent + 1, "Content xsi:type=\"VideoType\"",
											false);
	sprintf(a, "Video id=\"var%i\"", variationId);
	MP7File::writeEntry(fp, indent + 2, a, false);
	MP7File::writeMediaInformation(fp, indent + 3, mp4);
	//                                      </Video>
	//                              </Content>
	//                              <VariationRelationship>qualityReduction</VariationRelationship>
	//                      </Variation>    
	MP7File::writeEntry(fp, indent + 2, "Video", true);
	MP7File::writeEntry(fp, indent + 1, "Content", true);
	MP7File::writeVideoDescriptionEntry(fp, indent + 1,
																			"VariationRelationship",
																			adaptName);
	MP7File::writeEntry(fp, indent, "Variation", true);

	delete [] a;
};
