 /***********************************************************************
 *                                                                      *
 * ViTooKi Video Annotation Tool                                        *
 *                                                                      *
 * title: MP7Time.hpp                                                   *
 *                                                                      *
 * $Id: MP7Time.hpp,v 1.4 2005/01/14 12:01:19 mkropfbe Exp $
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

#ifndef _MP7TIME_HPP_
#define _MP7TIME_HPP_

#include <global.hpp>

/** @file
 * Declaration of the MP7Time class.
 */

/** maximal length of MPEG-7 date string (part of TimePoint datatype) */
#define MP7TIME_DATE_LEN      15

/** maximal length of MPEG-7 time zone string (part of TimePoint datatype) */
#define MP7TIME_TZ_LEN        7

/** This class represents an MPEG-7 time value. 
 * It may denote a time point or a time interval.
 * Note that this class may also represent media time.
 * @author Mario Taschwer
 * @version $Id: MP7Time.hpp,v 1.4 2005/01/14 12:01:19 mkropfbe Exp $
 */
class MP7Time
{
public:
  /** Default constructor. Initializes to time point 0 with duration 0. */
  MP7Time();

  /** Set time point from a tick value. 
   * @return <code>true</code> on success.
   * @see #timePointIsValid()
   */
  bool setTimePoint(uint value, uint ticksPerSecond = 1);

  /** Set time point from the MPEG-7 representation. 
   * @return <code>true</code> on success.
   */
  bool setTimePoint(const char *mp7Value);

  /** Copy time point from other MP7Time value.
   * The duration value is not modified.
   * @return <code>true</code> on success.
   */
  bool setTimePoint(const MP7Time &otherTime);

  /** Move time point by the given number of ticks.
   * If the resulting time point would be negative,
   * this time point is reset to 0 and an error status
   * is returned.
   * The date part and time zone of this object are ignored.
   * @return <code>true</code> on success.
   */
  bool moveTimePoint(int value, uint ticksPerSecond);
  
  /** Set duration from a tick value. 
   * @return <code>true</code> on success.
   */
  bool setDuration(uint value, uint ticks);

  /** Set duration from the MPEG-7 representation. 
   * @return <code>true</code> on success.
   */
  bool setDuration(const char *mp7Value);

  /** Set duration from another MP7Time instance.
   * The time point value remains unchanged.
   * @return <code>true</code> on success.
   */
  bool setDuration(const MP7Time &duration);

  /** Set the duration according to this object's time point
   * and the given end time point.
   * <p><b>Note:</b>
   * This function currently ignores the date part and time zone
   * of the operands.</p>
   * @return <code>true</code> on success.
   */
  bool setDurationFromEnd(const MP7Time &end);

  /** Add the duration of <code>otherTime</code>.
   * This function supports non-negative durations only.
   * If one of the operands has a negative duration,
   * an error status is returned.
   * @return <code>true</code> on success.
   */
  bool addDuration(const MP7Time &otherTime);

  /** Pretty-print time point.
   * A fresh object returns <code>00:00:00</code>.<br>
   * Output format: <code>YYYY-MM-DD hh:mm:ss.cc[+|-]hh:mm</code>
   * @param forceSubSec   if <code>true</code>, print sub-seconds digits also if they are 0.
   * @return a pointer to a static buffer, which will be overwritten by
   *         each invocation of this method.
   */
  const char *timePointPretty(bool forceSubSec = false) const;

  /** Pretty-print duration. 
   * @return a pointer to a static buffer, which will be overwritten by
   *         each invocation of this method.
   */
  const char *durationPretty() const;

  /** Return end time point (time point + duration) of this interval. 
   * <p>
   * <b>Note:</b> This method currently ignores the date and time-zone parts
   * of an MP7Time instance.
   * </p>
   */
  MP7Time endTimePoint() const;

  /** Pretty-print end time point (time point + duration).
   * <p>
   * <b>Note:</b> This method currently ignores the date and time-zone parts
   * of an MP7Time instance.
   * </p>
   * @return a pointer to a static buffer, which will be overwritten by
   *         each invocation of this method.
   * @see #endtimePoint()
   * @see #timePointPretty(bool)
   */
  const char * endTimePointPretty(bool forceSubSec = false) const;

  /** Pretty-print time and duration.
   * @return a pointer to a static buffer, which will be overwritten by
   *         each invocation of this method.
   */
  const char *toString() const;

  /** Return MPEG-7 representation of time point.
   * @return a pointer to a static buffer, which will be overwritten by
   *         each invocation of this method.
   */
  const char *timePointMP7() const;

  /** Return MPEG-7 representation of duration. 
   * If the duration is an IncrDuration type, only the
   * timeUnit will be returned.
   * @return a pointer to a static buffer, which will be overwritten by
   *         each invocation of this method.
   */
  const char *durationMP7() const;

  /** Return time point in seconds. 
   * The date and time zone parts of this instance are ignored.
   */
  double timeToSec() const;

  /** Return time point in milliseconds. 
   * The date and time zone parts of this instance are ignored.
   */
  int timeToMSec() const;

  /** Return duration in milliseconds. */
  int durationToMSec() const;

  /** Return duration in seconds. */
  double durationToSec() const;

  /** Return <code>true</code> if the time point is valid,
   * i.e. it has been set before.
   */
  bool timePointIsValid() const;

  /** Return <code>true</code> if the duration is 0. */
  bool durationIsNull() const;

  /** Compare this time point to another one.
   * An invalid time point is considered to be earlier than any valid
   * time point. Two invalid time points are considered to be equal.
   * <p>
   * <b>Note:</b> The date part and time zone of the operands are
   * currently ignored.
   * </p>
   * @return -1 if this time point is earlier than <code>other</code>;<br>
   *         0  if they are equal;<br>
   *         1  if this time point is later than <code>other</code>.
   * @see #timePointIsValid()
   */
  int compareTimePoint(const MP7Time &other) const;

  /** Return <code>true</code> if this interval contains the given
   * time point or interval.
   */
  bool contains(const MP7Time &time) const;

private:
  /** Reset to default values. */
  void reset();

  /** Reset time point to default value. */
  void resetTimePoint();

  /** Reset duration to default value. */
  void resetDuration();

  /** Time point (start of interval). */
  struct timepoint {
    char date[MP7TIME_DATE_LEN+1]; // date part, in MPEG-7 format, sign included
    bool hasTime;		   // false if both sec and num fields are invalid
    uint sec;			   // time in seconds
    uint num;			   // numerator of second fraction
    uint denom;			   // denominator of second fraction
    char tz[MP7TIME_TZ_LEN+1];	   // time zone
  } tpoint;

  /** Duration (interval length). */
  struct duration {
    int sign;			// +1 or -1
    uint days;
    uint sec;			// seconds
    uint num;			// numerator of second fraction
    uint denom;			// denominator of second fraction
    char tz[MP7TIME_TZ_LEN+1];	// time zone
    int units;			// number of time units (for IncrDuration type)
  } duration;

  /** Find a common multiple of the operands. */
  static uint commonMultiple(uint op1, uint op2);
};

#endif /* _MP7TIME_HPP_ */
