/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	SDP.cpp                                                       *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

/***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/

#include <time.h>
#ifdef WIN32
#ifndef WINCE
#include <sys/timeb.h>
#endif
#else
#include <sys/time.h>
#endif

/* included from common-UCL RTP for IOD parsing */
#ifdef WINCE
#include "3rdparty/commonucl/base64.h"
#else
#include "base64.h"
#endif

#include "SDP.hpp"
#include "ContainerInfo.hpp"
#include "VideoESInfo.hpp"
#include "adaptors/MP4Encoder.hpp"
#include "../metadata/TerminalCapabilities.hpp"


SDP::~SDP()
{
  clear();
}

void SDP::clear()
{
  list<struct session *>::iterator it = sessionList.begin();
  while (it != sessionList.end()) {
    delete *it;
    it = sessionList.erase(it);
  }
}

#define SDP_BUF_SIZE    4096

bool SDP::parse(const char *sdp)
{
  char *buf = new char[SDP_BUF_SIZE];
  int bufsize = strlen(sdp)+1;
  if (bufsize > SDP_BUF_SIZE) {
	delete [] buf;
    return false;
  }
  strcpy(buf, sdp);
  struct session *sp = NULL;
  bool err = false;
  bool valid_version = false;
  char *bufend = buf + strlen(buf);	// points to terminating '\0'
  char *p = buf;
  char *field, *value;
  do {
    /* ignore empty lines */
    while (p < bufend && (*p == '\n' || *p == '\r'))
      p++;
    if (p >= bufend)
      break;
    field = p;
    if (++p >= bufend || *p != '=') {
      err = true;
      break;
    }
    /* value is anything up to the end of line or buffer */
    value = ++p;
    while (p < bufend && *p != '\n' && *p != '\r')
      p++;
    *p++ = '\0';
    switch (*field) {
    case 'v':
      if (strcmp(value, "0") == 0)
	valid_version = true;
      else
	err = true;
      break;
    case 's':
      sp = new struct session;
      sp->name.assign(value);
      sessionList.push_back(sp);
      break;
    case 'a':
      if (sp) {
	if (strncmp(value, "control:", 8) == 0) {
	  sp->url.assign(value + 8);
	} else if (strncmp(value, "range:", 6) == 0) {
	  char *p1 = strstr(value + 6, "npt=");
	  if (p1) {
	    p1 += 4;
	    char *p2 = strchr(p1, '-');
	    if (p2) {
	      *p2++ = '\0';
	      if (strlen(p2) > 0 && sscanf(p2, "%lf", &sp->nptEnd) != 1)
		err = true;
	    }
	    if (strlen(p1) > 0 && sscanf(p1, "%lf", &sp->nptStart) != 1)
	      err = true;
	  }
	}
      }
      break;
    } // switch
  } while (p < bufend);
  if (err || !valid_version) {
    clear();
	delete [] buf;
    return false;
  }
#if VITOOKI_DEBUG_LEVEL >= VITOOKI_DEBUG_FULL
  dprintf_full("SDP::parse() parsed %d session description(s):\n", sessionList.size());
  for (list<struct session *>::iterator it = sessionList.begin();
       it != sessionList.end(); it++) {
    dprintf_full("    name = %s, url = %s, nptStart = %lf, nptEnd = %lf\n",
		 (*it)->name.c_str(), (*it)->url.c_str(),
		 (*it)->nptStart, (*it)->nptEnd);
  }
#endif
  delete [] buf;
  return true;
}

const struct SDP::session *SDP::firstSession() const
{
  return sessionList.empty() ? NULL : sessionList.front();
}

char *SDP::generate(ContainerInfo* stream, 	bool perfectHit,
										const char *filename, const char *localAddr,
										int SL_PAYLOAD_TYPE, int sizeLength,
										int dtsDeltaLength, const TerminalCapabilities* tc, 
										bool includeRtxInfo,list<rtx_group*>* groups)
{

  // mslc.SizeLength  + mslc.DTSDeltaLength replaced by
  // sizeLength + dtsDeltaLength
	assert(stream!=NULL);
  char *buffer = new char[MSG_BUFFER_SIZE];
  buffer[0] = 0;
  char *next = buffer;
  int cnt = 0;
	//disable rtx when rtxconfig is missing
	if(groups==NULL)
		includeRtxInfo = false;

#if defined WIN32 && !(defined WINCE)
  struct timeb tv;
  ftime(&tv);
  cnt =
		sprintf(next,	"v=0\r\no=%s %i %i IN IP4 %s\r\ns=%s\r\nu=http:///\r\ne=admin@\r\n",SERVER_ID,
						(int) tv.time, (int) tv.millitm, localAddr, filename);
#else
  timeval tv;
  gettimeofday(&tv, 0);
  cnt =
		sprintf(next, "v=0\r\no=%s %i %i IN IP4 %s\r\ns=%s\r\nu=http:///\r\ne=admin@\r\n",SERVER_ID,
						(int) tv.tv_sec, (int) tv.tv_usec, localAddr, filename);
#endif

  next += cnt;
	list <ESInfo*> * es = stream->getESList();
  list < ESInfo * >::const_iterator li = es->begin();
  float duration = 0;
  u32 bitRateInKilobits = 0;
	
  while (li != es->end()) {
		if (duration < ((float) (*li)->getDurationInMs() / 1000.0)) {
			duration = ((float) (*li)->getDurationInMs() / 1000.0);
		}
		if ((*li)->isAudioStream() || (*li)->isVisualStream()) {
			
			bitRateInKilobits += ((*li)->getAvgBandwidth() / 1024);
		}
		++li;
	}
	if(tc) {
		u32 bw = tc->getNetworkCapacityInByte()*8;
		if(bw==0)
			bw = tc->getMaxDecoderBitRateInBit();
		else if(tc->getMaxDecoderBitRateInBit() && 
						tc->getMaxDecoderBitRateInBit() < bw) {
			bw = tc->getMaxDecoderBitRateInBit();
		}
		if((bw/1024)<bitRateInKilobits)
			bitRateInKilobits=bw;
	}
  /*
		Connection Data

		c=<network type> <address type> <connection address>

		The "c=" field contains connection data.

		A session announcement must contain one "c=" field in each media
		description (see below) or a "c=" field at the session-level.  
	*/
  cnt =	sprintf(next, "c=IN IP4 %s\r\nb=AS:%i\r\nt=0 0\r\n", localAddr,
								bitRateInKilobits);
	//      cnt = sprintf(next, "c=IN IP4 0.0.0.0\r\nb=AS:%i\r\nt=0 0\r\n",bitRateInKilobits);
  next += cnt;

  char *iodstr = new char[MAX_IOD_SIZE];
  iodstr[0] = 0;
	u32 iodHandleSize=0;

#ifndef WINCE
	int iodenclen = base64encode(stream->getIODHandle(iodHandleSize), iodHandleSize, 
															 (unsigned char*)iodstr, MAX_IOD_SIZE);
#else
	int iodenclen = 0;
#endif

	if(iodenclen<0 || iodenclen>=MAX_IOD_SIZE)
		iodenclen=0;
  iodstr[iodenclen] = 0;
  int streamcnt = es->size();	// #streams is without iod
	
  li = es->begin();		//a=control:*\r\n
  cnt =	sprintf(next,	"a=control:*\r\na=mpeg4-iod:\"data:application/mpeg4-iod;base64,%s\"\r\n",
								iodstr);
  next += cnt;
	
	//
	cnt =	sprintf(next,		"a=isma-compliance:1,1.0,1\r\na=range:npt=0-%5.5f\r\n",
								duration);
	next += cnt;
	
	if(includeRtxInfo && !es->empty()) {
		// write the grouping infos
		list < ESInfo*>::const_iterator ei=es->begin();
		u32 lastMid=1;
		while(ei!=es->end()) {
			if((*ei)->isAudioStream() ||
				 (*ei)->isVisualStream() ) {
				rtx_group* group=stream->getRtxGroup( (*ei)->getStreamId(),groups);
				if(!group) {
					// no entry found, generate one
					group=new rtx_group(NACK,3000);
					group->midESInfo=lastMid++;
					group->rtx->mId=lastMid++;
					group->es=(*ei);
					group->rtx->ticks=(*ei)->getMediaTimeScale();
					groups->push_back(group);
				}
				if(!group->rtx) {
					group->rtx=new rtx_info(NACK,3000);
					group->midESInfo=lastMid++;
					group->rtx->mId=lastMid++;
					group->es=(*ei);
					group->rtx->ticks=(*ei)->getMediaTimeScale();
				}
				cnt=sprintf(next,"a=group:FID %u %u\r\n",
										group->midESInfo,
										group->rtx->mId);
				next+=cnt;
			}
			++ei;
		}

	}
  int payloadType = SL_PAYLOAD_TYPE;
	u32 bw2=0;

  for (int i = 1; i <= streamcnt; i++) {

		u32 tid = (*li)->getStreamId();
		u32 handlerType = (*li)->getHandlerType();
		//const u8 *decoderConfig = (*li)->getEncodedDecoderConfig();
		//              u32 objectType=(*li)->getObjectType();
		//	u32 streamType = (*li)->getStreamType();
		//        u32 bufferSize=(*li)->getBufferSize();
		u32 sze = 0;
		u8 volHeader[4096];
		u8* newHeader=NULL;
		u8* src=NULL;
		uint width = 0, height = 0;uint bitrate=0;

		if ('/' == *filename)
			filename++;
		rtx_group* currentRtx=NULL;
		if(includeRtxInfo)
			currentRtx=stream->getRtxGroup(tid,groups);

		switch (handlerType) {
		case MP4VisualHandlerType:
			width = ((VideoESInfo*)(*li))->getWidth();
			height = ((VideoESInfo*)(*li))->getHeight();
			bitrate=((VideoESInfo*)(*li))->getAvgBandwidth();
			if(tc && !perfectHit) {
				// check if tc say sth different
				if(tc->getDisplayHeight()!=0 && 
					 tc->getDisplayHeight() < height)
					height=tc->getDisplayHeight();
				if(tc->getDisplayWidth()!=0 && 
					 tc->getDisplayWidth() < width)
					width=tc->getDisplayWidth();
				bw2=tc->getNetworkCapacityInByte()*8;
				if(bw2==0)
					bw2=tc->getMaxDecoderBitRateInBit();
				else if(tc->getMaxDecoderBitRateInBit() && 
								tc->getMaxDecoderBitRateInBit() < bw2) {
					bw2=tc->getMaxDecoderBitRateInBit();
				}
				u32 new_bitrate = width * height * 6;
				if (new_bitrate > bw2)
					new_bitrate = bw2; //reduce to given termCaps
				if (new_bitrate < bitrate)
					bitrate = new_bitrate;
				newHeader = MP4Encoder::createNewHeader((VideoESInfo*)(*li), new_bitrate,
																							width, height, 300, (int)((VideoESInfo*)(*li))->getFPS(), &sze);
			}
			if(!newHeader) {
				sze = (*li)->getHeaders(&src);
				SDP::encodeDecoderConfig(volHeader, src, sze);
			}
			else {
				SDP::encodeDecoderConfig(volHeader,newHeader,sze);
				delete newHeader;newHeader=NULL;
			}
			cnt = sprintf(next,	"m=video 0 RTP/AVP %i\r\nb=AS:%i\r\n" 
										"a=rtpmap:%i MP4V-ES/%i\r\n" 
										"a=control:trackID=%i\r\n" 
										"a=cliprect:0,0,%i,%i\r\n" 
										"a=fmtp:%i profile-level-id=1;config=%s\r\n"	// StreamType=%i; "SizeLength=%i; " 
										//      "DTSDeltaLength=%i; config=%s\r\n"
										"a=mpeg4-esid:%i\r\n"
										"a=aspect-ratio=%6.6f\r\n"
										"a=frame-pattern=%s\r\n"
										"a=bframesamount=%f\r\n",
										payloadType, (bitrate / 1024), 
										payloadType, (*li)->getMediaTimeScale(), 
										tid, 
										width, height, 
										SL_PAYLOAD_TYPE, volHeader, // streamType, //sizeLength,
										// dtsDeltaLength, volHeader,
										tid,	// "mpeg4-generic" 
										((VideoESInfo*)(*li))->getAspectRatio(),
										(((VideoESInfo*)(*li))->hasStaticFramePattern()?"static":"dynamic"),
										((VideoESInfo*)(*li))->getAvgBFrameSize());
			next += cnt;
			// add bframes and gop size for static frame pattern
			if(((VideoESInfo*)(*li))->hasStaticFramePattern()) {
				cnt = sprintf(next, "a=num-bframes %i\r\n"
											"a=gop-size %i\r\n",
											((VideoESInfo*)(*li))->getNum_B_frames(),
											((VideoESInfo*)(*li))->getGOP_size());
				next+=cnt;
			}
			payloadType++;

			if(	includeRtxInfo && currentRtx &&
					currentRtx->rtx && currentRtx->rtx->state!=NONE) {
				u32 maxRtxDelay=3000;
				if(	currentRtx->rtx &&
						currentRtx->rtx->rtxTimeInMs>0)
					maxRtxDelay=currentRtx->rtx->rtxTimeInMs;
				else {
					currentRtx->rtx=new rtx_info(NACK,maxRtxDelay); // impossible code
				}
				
				
				cnt = sprintf(next, 
											"a=rtcp-fb:%i %s\r\n"
											"a=mid:%u\r\n" // write the mid for the media
											"m=video 0 RTP/AVPF %i\r\n" 
											"a=rtpmap:%i rtx/%u\r\n" 
											"a=fmtp:%i apt=%i;rtx-time=%u\r\n"
											"a=mid:%u\r\n", 
											payloadType-1, sRTXState[currentRtx->rtx->state],
											currentRtx->midESInfo,
											payloadType, 
											payloadType, currentRtx->rtx->ticks, 
											payloadType,(payloadType-1),maxRtxDelay,
											currentRtx->rtx->mId);
				next += cnt;
				currentRtx->rtx->payloadType=payloadType;
				payloadType++;
			}
	    break;
		case MP4AudioHandlerType:
			// only send audio when TermCaps indicate that the client can handle audio
			if(!tc || (tc && tc->getNumAudioChannels()>0)) {
				// audio config still not correct
				u8 decoderConfig[4096];	
				src = NULL;
				payloadType=14;
				sze = (*li)->getHeaders(&src);
				SDP::encodeDecoderConfig(decoderConfig, src, sze);
				cnt = sprintf(next, "m=audio 0 RTP/AVP %i\r\nb=AS:%i\r\n" // payload
											"a=control:trackID=%i\r\n" // trackid
											"a=mpeg4-esid:%i\r\n"			 // esid
					                                                "a=fmtp:%i DTSDeltaLength=%i; config=%s\r\n" 
											"a=rtpmap:%i MPA/%i\r\n", 
											payloadType,((*li)->getAvgBandwidth() / 1024), tid, tid,
					                                                payloadType, dtsDeltaLength, decoderConfig,
											payloadType, (*li)->getMediaTimeScale());
				next += cnt;
				payloadType++;

				if(includeRtxInfo) {
					u32 maxRtxDelay=3000;
					if(	currentRtx->rtx &&
							currentRtx->rtx->rtxTimeInMs>0)
						maxRtxDelay=currentRtx->rtx->rtxTimeInMs;
					else {
						currentRtx->rtx=new rtx_info(NACK,maxRtxDelay);
					}
					cnt = sprintf(next, 
												"a=mid:%ui\r\n" // write the mid for the media
												"m=audio 0 RTP/AVPF %i\r\n" 
												"a=rtpmap:%i rtx/%i\r\n" 
												"a=fmtp:%i apt=%i;rtx-time=%ui\r\n"
												"a=mid:%i\r\n", 
												currentRtx->midESInfo,
												payloadType, 
												payloadType, currentRtx->rtx->ticks, 
												payloadType,(payloadType-1),maxRtxDelay,
												currentRtx->rtx->mId);
					next += cnt;
					currentRtx->rtx->payloadType=payloadType;
					payloadType++;
				}
			} // if !tc...
	    break; 
		default:	//application is not recognized by many players!
			/*	    cnt = sprintf(next, "m=application 0 RTP/AVP %i\r\n" "a=control:trackID=%i\r\n" "a=mpeg4-esid:%i\r\n" "a=fmtp:%i "	// StreamType=%i; "// SizeLength=%i; " 
			//      "config=%s\r\n" //      "DTSDeltaLength=%i; config=%s\r\n"
			"\r\na=rtpmap:%i %s\r\n", payloadType, tid, tid, payloadType,	// streamType, //sizeLength,
			//dtsDeltaLength, decoderConfig,
			//			  payloadType, "X-MP4-ES/1000");	//mpeg4-generic
	    next += cnt;
	    payloadType++; */
	    ;
		}
		++li;			// increase list iterator
		if(src)
			delete src;
	}

	*next = 0;
	char *result = new char[strlen(buffer) + 1];
	strcpy(result, buffer);

	delete [] iodstr;
	delete [] buffer;
	return result;
};



void SDP::encodeDecoderConfig(u8 * dest, const u8 * source, u32 size)
{
	//encode
	size=size*2;
	for (u32 j = 0; j < size;j += 2) {
		u8 help = (*source) >> 4;
		dest[j] = help > 9 ? (char) help - 10 + 'A' : (char) help + '0';
		help = (*source) & 0xf;
		dest[j + 1] =
	    help > 9 ? (char) help - 10 + 'A' : (char) help + '0';
		source++;
	}
	dest[size] = 0;
};

void SDP::decodeDecoderConfig(u8 * dest, const u8 * source, u32 srcSize)
{
	//decode: during encode 1 Byte -> 2 Bytes (higher 4 bits were first, second byte is lower 4 bits
	for (u32 j = 0; j < srcSize; j += 2) {
		u8 help = (*source);
		dest[j / 2] =
	    (help >=
			 'A') ? ((char) help - 'A' + 10) << 4 : ((char) help - '0') << 4;
		source++;
		help = (*source);
		dest[j / 2] +=
	    (help >= 'A') ? ((char) help - 'A' + 10) : ((char) help - '0');
		source++;
	}
};

