/***********************************************************************
 *                                                                      *
 * ViTooKi                                                              *
 *                                                                      *
 * title:	ServerSession.cpp                                             *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 * ITEC institute of the University of Klagenfurt (Austria)             *
 * http://www.itec.uni-klu.ac.at                                        *
 *                                                                      *
 *                                                                      *
 * For more information visit the ViTooKi homepage:                     *
 * http://ViTooKi.sourceforge.net                                       *
 * vitooki-user@lists.sourceforge.net                                   *
 * vitooki-devel@lists.sourceforge.net                                  *
 *                                                                      *
 * This file is part of ViTooKi, a free video toolkit.                  *
 * ViTooKi is free software; you can redistribute it and/or             *
 * modify it under the terms of the GNU General Public License          *
 * as published by the Free Software Foundation; either version 2       *
 * of the License, or (at your option) any later version.               *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,               *
 * MA  02111-1307, USA.                                                 *
 *                                                                      *
 ***********************************************************************/

/***********************************************************************
 *                                                                      *
 *  REVISION HISTORY:                                                   *
 *                                                                      *
 *                                                                      *
 *                                                                      *
 ***********************************************************************/
#ifndef WIN32
#include <sys/select.h>
#include <arpa/inet.h>
#endif
#include "ServerSession.hpp"
#include "RTSP.hpp"
#include "DataChannel.hpp"
#include "DataSink.hpp"
#include "IO.hpp"
#include "RFC3016.hpp"
#include "RFC2250.hpp"
#include "BitField.hpp"
#include "Statistics.hpp"

#include "cache/MetaObject.hpp"
#include "cache/AdmissionControl.hpp"
#include "io/Rtp.hpp"
#include "io/SimpleRtp.hpp"
#include "io/IOCreator.hpp"
#include "io/FilteredIO.hpp"
#include "io/MPGStreamIO.hpp"
#include "metadata/MP21.hpp"
#include "adaptors/TemporalAdaptor.hpp"
#include "adaptors/SharedAdaptor.hpp"
#include "container/Video4LinuxContainerFile.hpp"


/***********************************************************************/
ServerSession::ServerSession(int socket, const struct sockaddr_in *clint,
									CacheManager * c, AdmissionControl* am,
									 bool enableRTX,SessionCounter* counter,TerminalCapabilities* tc):Session(),enableRtx(enableRTX)
{
	assert(c);
	assert(am);
	this->counter = counter;
	termCap=tc;
	admControl=am;
  state = SESSION_CLOSED;
  cm = c;
  if (clint) {
		client = new sockaddr_in();
		memcpy(client, clint, sizeof(sockaddr_in));
		clientId=(u32)client->sin_addr.s_addr;
  } else {
		client = NULL;
		clientId=0;
	}

  sessionControlChannel=socket;
	sockaddr_in localaddr;
	canAdmitRequest=true;

#ifdef WIN32
  int slen = sizeof(localaddr);
#else	/*  */
	socklen_t slen = sizeof(localaddr);
#endif	/*  */

	memset(&localaddr, 0, sizeof(localaddr));
  getsockname(socket, (sockaddr *) & localaddr, &slen);
	const char* tmp=inet_ntoa(localaddr.sin_addr);
	localAddress=new char[strlen(tmp)+1];
	strcpy(localAddress,tmp);
  prot = NULL;
  url = NULL;

	userPref=NULL;
	suggestedAdaptor=NULL;
	//	clientObj=NULL;
	mp4Stream=NULL;
	playerId=NULL;
	serverID=NULL;

	gtRefCount = new ReferenceCounter();
	globalTimer = new GlobalTimer();
	//this will prevent deletion on DataChannel closes:
	gtSharedAdapt = new SharedAdaptor(gtRefCount, (Adaptor*&)globalTimer);
	resUsage.cpu=resUsage.disk=resUsage.network=0.0;
}

/***********************************************************************/
ServerSession::~ServerSession()
{
	admControl->freeRequest(&resUsage);
#ifdef WIN32
	closesocket(sessionControlChannel);
#else
	::close(sessionControlChannel);
#endif
	if(mp4Stream)
		delete mp4Stream;

	if (prot) {
		delete prot;
		prot = NULL;
  }
	if (client) {
		delete client;
		client = NULL;
	}
	if (url) {
		delete url;
		url = NULL;
	}
	if(localAddress)
		delete localAddress;

	if(playerId)
		delete playerId;
	if(serverID)
		delete serverID;
	list < DataChannel * >::iterator i;
	while(!channels.empty()) {
		i=channels.begin();
		(*i)->teardown(TEARDOWN_ALL,true);
		delete (*i);
		channels.pop_front();
	}
	list < rtx_group* >::iterator j;
	while(!rtxInfo.empty()) {
		j=rtxInfo.begin();
		delete *j;
		rtxInfo.pop_front();
	}
	if(termCap)
		delete termCap;
	if(userPref)
		delete userPref;

	delete gtSharedAdapt; //this will also del gtRefCount and globalTimer

}

/***********************************************************************/
bool ServerSession::options(const Url* fileName, const char* remaining)
{
  prot->incSeqNr();
  return true;
}


/***********************************************************************/
bool ServerSession::connect(const Url* fileName, const char* remaining)
{
	assert(fileName && remaining);
	// keep existing termcap when invalid ones are found
	TerminalCapabilities* dummyTermCap=MP21::createTerminalCapability(remaining);
	if(dummyTermCap) {
		// overwrite the default with the new ones
		dprintf_full("ServerSession::connect Found TermCaps\n");
		if(termCap)
			delete termCap;
		// if termcaps are specified, they take precedence over default UserPrefs!
		if(userPref)
			delete userPref;
		userPref=NULL;
	}
	if(dummyTermCap)
		termCap=dummyTermCap;

	if(!termCap) {
		// maybe UserPreferences?
		// keep existing UserPref
		UserPreferences* upDummy=MP21::createUserPreferences(remaining);
		dprintf_full("ServerSession::connect Found UserPreferences %p in XX%sXX\n",upDummy,remaining);

		if(userPref && upDummy)
			delete userPref;
		if(upDummy)	{
			userPref=upDummy;
		}
	}
	setUrl(fileName); //find perfect fitting mp4stream
	bool ret=prot->parseCMDDescribe(this, fileName, remaining,&playerId);
	if(!playerId) {
		// HACK: assume it's the stupid Real player
		playerId=new char[strlen(REAL_PLAYER_ID)+1];
		strcpy(playerId,REAL_PLAYER_ID);
	}
	dprintf_full("ServerSession::connect Playerid \"%s\"\n",playerId);
	

	/*if(resUsage.network < 0.000001) {
			canAdmitRequest=false;
			state=SESSION_ERR;
	}
	*/
	
	if (this->getContainerInfo() == NULL)
		printf("ServerSession::connect : vontinfo NULL!!!\n");
	if(canAdmitRequest)
		ret=prot->generateCMDDescribeReply(this,fileName,localAddress,playerId);
	else {// only way to indicate a server overload is with notenoughbandwith
		dprintf_full("ServerSession::connect NOT ENOUGH BANDWIDTH!!!\n");
		prot->generateNotEnoughBandwidth();
		ret=true;
	}
	

	const char* buf=prot->getBuffer();
	dprintf_full("ServerSession::connect %s\n",buf);
	strcpy(buffer,buf);
	prot->incSeqNr();
	return ret;
};


/***********************************************************************/
bool ServerSession::setup(const Url* fileName, const char* remaining)
{
	assert(fileName && remaining);
	dprintf_full("ServerSession::setup %s\n",url->toString());
	bool hasSessionKey=false,hasRtxPort=false;
	char* protocol=NULL;
	char* unicast=NULL;
  char* clientAddress=NULL;

	int sesKey=0,esId=0,rPort=0,rPort2=0,lPort=0;
	u32 clientRtxPort=0, serverRtxPort=0;

	bool err=!prot->parseCMDSetup(this,fileName,remaining,&hasSessionKey,&sesKey,&protocol,
																&unicast,&esId,&rPort,&rPort2,&hasRtxPort,&clientRtxPort,
                                &clientAddress);
	hasRtxPort&=enableRtx; //disable rtx if it is switched off
	if(!enableRtx)
		clientRtxPort=0;

	if(!err) {
		if(!prot->isSupportedTransport(protocol)) {
			prot->generateIllegalTransport();
			err=true;
		}
	}
	else {
		prot->generateBadCMD();
	}

	if(err) {
		dprintf_err("ServerSession::setup error in parsing setup request\n");
		const char* buf=prot->getBuffer();
		strcpy(buffer,buf);
		if(protocol)
			delete protocol;
		if(unicast)
			delete unicast;
		prot->incSeqNr();
		return false;
	}
	if(hasSessionKey==false) {
		sesKey=prot->generateSessionKey();
	}
	lPort=portGen.getNextFreePortPair();
	if(hasRtxPort && clientRtxPort>0) {
		serverRtxPort=portGen.getNextFreePortPair();
	}

	// now that we have all the info, do the following:
	// create an RTP with RFC3016 output channel
	// create DataChannel
	// open RTP session

  //char *clientAddress = inet_ntoa(client->sin_addr);	// points to a static array in inet..
  if (clientAddress == NULL) clientAddress = inet_ntoa(client->sin_addr);

	ESInfo* es=NULL;
	IO* input=NULL;

	if(mp4Stream)
		es=mp4Stream->getES(esId);
	if(es) {
		input=IOCreator::createInput(es, false);
	}
  if (!(this->mp4Stream!=NULL && es!=NULL) || input==NULL) {
		prot->generateFileNotFound();
		portGen.closePortPair(lPort);
		if(hasRtxPort && clientRtxPort>0) {
			portGen.closePortPair(serverRtxPort);
		}
		prot->incSeqNr();

		return false;
  }

  PacketizationLayer * p;
  if (es->isAudioStream())
	  p = new RFC2250();
  else
	  p = new RFC3016();

	// no probs with inet_ntoa: rtp deep-copies the clientaddress
	IO* output=NULL;
	Statistics *outStats=NULL;
	if(suggestedAdaptor){
		suggestedAdaptor->setESInfo(es);
		suggestedAdaptor->initialize();
	}

	// only use Rtp when retransmission is requested, otherwise SimpleRtp is cheaper
//	if(hasRtxPort && clientRtxPort>0) {
	if(enableRtx) {
		outStats = new Statistics(es,true);
		dprintf_full("ServerSession::setup: created RTP with retransmission\n");
		output = new Rtp(fileName->getPath(), rPort, clientAddress, lPort,
												 es, p, true, termCap, globalTimer, outStats, false);
	}
	else {
#ifndef WINCE
		dprintf_full("ServerSession::setup: created SimpleRtp without retransmission\n");
		char logName[MAX_STR_LEN];
		struct timeval t;
		gettimeofday(&t,NULL);
		sprintf(logName,"log%10i.%3i_server.out",(int)t.tv_sec,(int)(t.tv_usec/1000));
		output=new SimpleRtp(fileName->getPath(), rPort, lPort,
		clientAddress, es, p, true,logName,true,SimpleRtp::MIXED_CTS_BITRATE);
#else
		dprintf_err("\nServersession::setup() RTX MUST BE ENABLED !!!");
#endif
	}
/*
	if(	playerId && suggestedAdaptor==NULL && es && es->isVisualStream() &&
			strstr(playerId, QT_PLAYER_ID) &&
			strstr(playerId, OLD_BFRAMES_DISABLED_QT_PLAYER) ) {
		// adaptor changes esInfo!!! no prob here, we already have made a clone in setUrl
		suggestedAdaptor=new TemporalAdaptor((VideoESInfo*)es);
	}
*/
	DataSink* clientObj = new DataSink((u32)client->sin_addr.s_addr,output, clientAddress,
																		 ntohl(client->sin_port),
																		 this->suggestedAdaptor);

	suggestedAdaptor=NULL; //unlink Adaptor, will be freed in DataSink
	this->prot->generateSetupReply(this,fileName,sesKey,protocol,unicast,clientAddress,
																 esId,rPort,rPort2,lPort,clientRtxPort,serverRtxPort);
	const char* buf=prot->getBuffer();
	strcpy(buffer,buf);

	rtx_group* group=getRtxGroup(esId);
	if(group && group->rtx) {
		group->rtx->remotePort=clientRtxPort;
		group->rtx->localPort=serverRtxPort;
		output->setRtxInfo(group->rtx);
	}
	else {
		dprintf_err("ServerSession::setup: Session has stored %i groups \n",rtxInfo.size());
		dprintf_err("ServerSession::setup: Failed to get rtx_port for es %i\n", esId);
	}

	// now create the DataChannel and set the outstats of the DataSink IO (RTP)
	DataChannel * d = new DataChannel(input, es);
	d->setOutputStatistics(outStats);

  d->setSessionId(sesKey);
  d->insert(clientObj); // add the client to the DC
	// set the DataChannel at the Session
	setDataChannel(d);
	if(protocol)
		delete protocol;
	if(unicast)
		delete unicast;

	state = SESSION_INITIALIZED;
	prot->incSeqNr();
	return true;
}

/***********************************************************************/
bool ServerSession::startDCs(int sessionId, double startTime, double endTime,
														 double prefetch, bool backwardsClosestIFrame) {
	u32 startFrame=0, endFrame=0;
	bool res = false;
	long ivop_align = -9999;

	if(endTime <= 0.00001) {
		endTime = mp4Stream->getFirstVisualES()->getDurationInMs() / 1000.0;
	}

	const ESInfo* es=NULL;
	for (list < DataChannel * >::iterator i = this->channels.begin(); i != channels.end(); i++) {
		dprintf_full("ServerSession::startDCs: inspecting ES with session %u\n", (*i)->getSessionId());
		if ((sessionId < 0) || ((*i)->getSessionId() == (u32)sessionId)) {
			es=(*i)->getESInfo();

			if(es) {
				startFrame=es->getFrameNumber(startTime);
				endFrame=es->getFrameNumber(endTime);
				dprintf_full("ServerSession::startDCs startFrame %u (@sec %6.3f) endFrame %u (@sec %6.3f)\n",
										 startFrame,startTime,endFrame,endTime);

				IO* input=(*i)->getInput();
				//input->setToFrameNumber(startFrame);
				dprintf_full("ServerSession::startDCs find real I-VOP\n");
				if( //input->getState()==IO::OPEN &&
					 ( -1 == (ivop_align = input->setToClosestIFrame(startFrame, backwardsClosestIFrame))) ) {
					dprintf_err("ServerSession::startDCs Failed to set input to frame %i (state %i)\r\n",
											startFrame,input->getState());
				}
				dprintf_full("ServerSession::startDCs real I-VOP startFrame %li \n",ivop_align);
				input->setEndFrameNumber(endFrame);
				DataSink* clientOut=(*i)->getDataSink(clientId);
				IO* clientOutput=NULL;
				if(clientOut && (clientOutput=clientOut->getOutput())) {
					// flush the Rtp output buffer !!!
					clientOutput->setToFrameNumber(ivop_align); //new startFrame
				}
				(*i)->play(clientId,prefetch);


				// if not yet started -> start it
				if (!(*i)->running() || (*i)->finished() ) {
					dprintf_full("ServerSession::startDCs starting inputIO thread\n");
					(*i)->start();
				}
				res=true;
			}
			else {
				startFrame=0;
				dprintf_err("ServerSession::startDCs: RTSP::handleCMDPlay Missing ES for request? PANIC!\r\n");
			}
		}
	}
	return res;
}

/***********************************************************************/
bool ServerSession::play(const Url* fileName, const char* remaining)
{
	int sessionId;

	assert(fileName && remaining);

	double startTime=0.0, endTime=0.0, prefetch=0.0;
	if(!prot->parseCMDPlay(this,fileName,remaining,&sessionId,&startTime,&endTime,&prefetch)) {
		dprintf_err("ServerSession::play RTSP::parseCmdPlay failed\n");
		prot->generateSessionNotFound();
		const char* buf=prot->getBuffer();
		strcpy(buffer,buf);
		prot->incSeqNr();
		return false;
	}

	//start up all DCs with their inputs and outputs, backwardsClosestIFrame
	dprintf_err("ServerSession::play START OK\n");
	bool res = startDCs(sessionId, startTime, endTime, prefetch, true );
	if(res)
		prot->generateOk();
	else {
		dprintf_full("ServerSession::play didn't find any ES for the sessionId\n");
		prot->generateSessionNotFound();
	}
	const char* buf=prot->getBuffer();
	strcpy(buffer,buf);

	state = SESSION_ACTIVE;
	prot->incSeqNr();
  return res;
};


/***********************************************************************/
bool ServerSession::pause(const Url* fileName, const char* remaining)
{
	bool found=false;
	assert(fileName && remaining);
	int sessionId=prot->extractSessionKeyFromCMD(remaining);

  if (sessionId > 0 && (unsigned)sessionId!=INVALID_SESSIONID) {
    // search for the session
    // start it
    for (list < DataChannel * >::iterator i = this->channels.begin();
				 i != this->channels.end(); i++) {
      if ((*i)->getSessionId() == (u32)sessionId) {
				found=true;
				(*i)->pause(clientId);
      }
    }
  }
	else
		prot->generateSessionNotFound();

	if(found)
		prot->generateOk();
	const char* buf=prot->getBuffer();
	strcpy(buffer,buf);

	state = SESSION_PAUSED;
	prot->incSeqNr();
  return found;
}


/***********************************************************************/
bool ServerSession::tearDown(int sessionId,bool immediate, const Url* fileName, const char* remaining)
{
	dprintf_full("ServerSession::tearDown(%i, %i)\r\n",sessionId,immediate);
	if(fileName && remaining) {
		sessionId=RTSP::extractSessionId(remaining);
	}
	if((u32)sessionId==TEARDOWN_ALL) {
		// teardown all
		state = SESSION_CLOSED;
		list < DataChannel * >::iterator dcI;
		while (!channels.empty()) {
			dcI=channels.begin();
			(*dcI)->teardown(clientId,immediate);
			// check if we have to delete it, might be shared with other sessions
			if((*dcI)->getNumberOfActiveDataSinks()==0) {
				(*dcI)->wait(); // wait for thread to finish
				delete (*dcI);
			}
			channels.pop_front();
		}
		if(prot) { // can happen that no prot is set when teardown called first
			prot->generateOk();
			const char* buf=prot->getBuffer();
			strcpy(buffer,buf);
			prot->incSeqNr();
			return true;
		}
		else
			return false;
	}
	else {
		// teardown one session
		bool sesFound=false;
		list < DataChannel * >::iterator dcI;
		if (!channels.empty()) {
			dcI = channels.begin();
			while (dcI != channels.end()) {
				if ((*dcI)->getSessionId()== (unsigned)sessionId) {
					sesFound=true;
					(*dcI)->teardown(clientId,immediate);
					// check if we have to delete it, might be shared with other sessions
					if((*dcI)->getNumberOfActiveDataSinks()==0) {
						(*dcI)->wait(); // wait for thread to finish
						delete (*dcI);
					}
					dcI = channels.erase(dcI);
					if(channels.empty()) {
						state = SESSION_CLOSED;
						prot->generateOk();
						const char* buf=prot->getBuffer();
						strcpy(buffer,buf);
						prot->incSeqNr();
						return true;
					}
				} else
				  ++dcI;
			}
	  }
		if(sesFound) {
			prot->generateOk();
		}
		else {
			prot->generateSessionNotFound();
		}
		const char* buf=prot->getBuffer();
		strcpy(buffer,buf);
		prot->incSeqNr();
		return sesFound;
	}
	prot->generateSessionNotFound();
	const char* buf=prot->getBuffer();
	strcpy(buffer,buf);
	prot->incSeqNr();
  return false;
};

//is teardown on streams or sessions ???


/***********************************************************************/
void ServerSession::run()
{
	fd_set rfds;
	struct timeval tv;
	if(this->counter != NULL)
		this->counter->increment();
	int retval;
	int lastSec = 1;
	int lastSwitchSec = -99;

	// while not end of session
	//check for control command
	// do control command
	state = SESSION_NEW;
	/* the buffer for incoming rtsp/http commands */
	char *inbuffer = new char[MSG_BUFFER_SIZE + 1];
	dprintf_full("ServerSession::run() SESSION START\n");
	dprintf_full("ServerSession::run: Listening for data on ControlChannel\r\n");
	while (state!=SESSION_ERR && state!=SESSION_CLOSED) {
		inbuffer[0] = 0;
    FD_ZERO(&rfds);
    FD_SET(sessionControlChannel, &rfds);
    tv.tv_sec = 0;
    tv.tv_usec = 40 * 1000; //microsecs
    retval = select(sessionControlChannel+1, &rfds, NULL, NULL, &tv);

    if (retval == -1) {
			dprintf_err("select()");
			::exit(1);
    }else if (retval) {
			int bytesRead = readRequest(inbuffer, MSG_BUFFER_SIZE);

			dprintf_full("ServerSession::run: readBlock ended. read %i bytes\r\n",bytesRead);
			if (bytesRead == 0) {
				//no data read
				// the other side has closed the connection, teardown all
				dprintf_full("ServerSession::run: No data read from ControlChannel\r\n");
				tearDown(TEARDOWN_ALL,true,NULL,NULL);
			} else if (bytesRead < 0) {
				// error:
				dprintf_err("ServerSession::run Error reading from controlchannel");
				tearDown(TEARDOWN_ALL,true,NULL,NULL);
				state = SESSION_ERR;
			} else {		// we got a command
				dprintf_full("ServerSession::run: got command\r\n");
				// check for protocol if not yet set
				if (!prot) {
					if (determineProtocol(inbuffer) == PROTO_UNKNOWN) {
						dprintf_err("ServerSession::run: Unknown protocol: %s\n",
												inbuffer);
					}
				}
				if (prot) {
					switch (prot->getID()) {
					case PROTO_RTSP:
					case PROTO_HTTP:
						if (bytesRead >= MSG_BUFFER_SIZE) {
							dprintf_err("received command exceeded buffer size");
							break;
						}
						break;
					default:
						dprintf_err("ServerSession::run: Unknown protocol: %s\n",inbuffer);
						break;
					}

					if (!prot->applyReqToSession(inbuffer, bytesRead, this)) {
						dprintf_err("ServerSession::run: Failed to parse request >%s<\n",	inbuffer);
					}
					else {
						// check if we have options
						if(strncmp("OPTIONS", inbuffer,strlen("OPTIONS"))==0) {
							const char* buf=prot->getBuffer();
							strcpy(buffer,buf);
						}
					}
					sendResponse(buffer,strlen(buffer));
				}
			}
		} else if ((url != NULL && (strcmp(url->getFileName(), "v4l")) !=0))
			if((state == SESSION_ACTIVE) && (cm->findMetaObject(url)->getNumberOfVariations() > 1)) {
			//check for SWITCHING only if PLAYING and there is a choice for other streams
			
			int nowSec = (int)ceil(globalTimer->getActualSec());
			if (nowSec > lastSec) { //new streamout sec
				lastSec = nowSec;
				int sumNetBW = 0;
				bool doSwitch=false;
				for (list < DataChannel * >::iterator dc = channels.begin();
						 dc != channels.end(); dc++) {
					//FIXME: completely ignores SessionIDs, resets all DCs
					//FIXME: outStats is only from first DataChannel, but sync'd by globalTimer
					if (!(*dc)->tearDownFlagSet()) {
						Statistics *outStats = (*dc)->getOutputStatistics();
						if (outStats && (3 * (*dc)->getESInfo()->getMediaTimeScale() + outStats->getHighestPacketTS()
														 < (*dc)->getESInfo()->getDuration() ) ) { //FIXME: global nowSec!
							// stream will continue to run for at least 3 more seconds!
							double realBWfact = (double)outStats->getBWfromStreamoutSec(nowSec-2)/(double)outStats->getStreamBW();
							dprintf_full("ServerSession::run now %6.3f sec ES-Id %llu Pkt Loss %4.2f %%  netBW %li kbps streamBW %i kbps realBWfact %2.2f VSec %6.3f (of %6.3f) adaptRate %i %%\n",
													 globalTimer->getActualSec(), (*dc)->getESInfo()->getStreamId(),
													 outStats->getStreamoutSecLossPercent(nowSec-2),
													 outStats->getBWfromStreamoutSec(nowSec-2)*8 / 1024,
													 outStats->getStreamBW()*8 / 1024,
													 realBWfact,
													 (double)outStats->getHighestPacketTS() / (*dc)->getESInfo()->getMediaTimeScale(),
													 (double)(*dc)->getESInfo()->getDuration() / (double)(*dc)->getESInfo()->getMediaTimeScale(),
													 outStats->getAdaptRate());

							// too much adaptation, switch down
							if ((outStats->getAdaptRate() < 70) && (outStats->getAdaptSecs() >= 1.0)) {
								doSwitch = true;
								sumNetBW += (int)((double)outStats->getStreamBW() * 0.7);
								//								sumNetBW += (int)((double)outStats->getBWfromStreamoutSec(nowSec-2) * 1.15); //add adapt range
								dprintf_full("ServerSession::run DECIDED ON STREAM SWITCHING :    DOWN for adapt\n");
							} else

							//buffer low, so force even lower quality stream!
							if ((nowSec - lastSwitchSec > 5) &&  (outStats->getAdaptSecs() > 1.0))  {
								doSwitch = true;
								sumNetBW += (int)((double)outStats->getStreamBW() * 0.7);
								dprintf_full("ServerSession::run DECIDED ON STREAM SWITCHING :    DOWN to fill buf\n");
							} else


							//wait at least 5 secs, then if no adaptation and super-high BW:  SWITCH UP!
							if ((nowSec - lastSwitchSec > 15) && (outStats->getAdaptRate() >= 100)
									&& (outStats->getAdaptSecs() <= 1.0) //force high quality if buffers high
									&& ((realBWfact > 1.25) || (outStats->getBufferFillLevelInSecs() > RTP_BUF_HIGHWATER) ) ) {
								doSwitch = true;
								sumNetBW += (int)((double)outStats->getStreamBW() * 1.4); //accept frame dropping for higher quality stream!
								dprintf_full("ServerSession::run DECIDED ON STREAM SWITCHING :    UP!!\n");
							}
						}
					}
				}

				if (doSwitch) {
					dprintf_full("ServerSession::run DECIDED ON STREAM SWITCHING out of %i variations! new overall NetworkCapacity: %i kbps\n", cm->findMetaObject(url)->getNumberOfVariations(), sumNetBW*8/1024);
					//sumNetBW in termcaps is needed for multi-stream containers: with Video+Audio!
					//termCap->setNetworkCapacityInByte((u32)(sumNetBW * 1.15)); //add a faked adaptation range
					termCap->setNetworkCapacityInByte((u32)(sumNetBW));

					lastSwitchSec = nowSec;

					//exchange and set to the new input
					assert(mp4Stream);
					ContainerInfo *oldmp4 = mp4Stream;
					setUrl(new Url(url->toString()), false); //re-issue optimum stream finding process, forbid transcoding
					if (0 == strcmp(mp4Stream->getFirstVisualES()->getInput(),oldmp4->getFirstVisualES()->getInput())) {
						// I chose the same stream, so its useless
						dprintf_err("ServerSession::run: Sorry, there was no fitting stream... I took the same, so ignoring!\n");
						delete mp4Stream;
						mp4Stream = oldmp4;
					} else {
						//set everything paused (input, output, DCs)
						//FIXME: completely ignores SessionIDs, resets all DCs
						for (list < DataChannel * >::iterator dc = channels.begin();
								 dc != channels.end(); dc++) {
							dprintf_full("DC ESid %llu  newID %llu\n",(*dc)->getESInfo()->getStreamId(),
													 mp4Stream->getFirstVisualES()->getStreamId());
							ESInfo *new_es = mp4Stream->getES((*dc)->getESInfo()->getStreamId());
							assert(new_es != NULL);
							if ( (*dc)->getESInfo() !=  new_es ) {
								long ivop_align;
								double vsec_time;
								list <u32> ids;
								list <u32>::iterator id;

								//pause clients and RTP to get stable HighestPacketTS
								ids = (*dc)->getListOfActiveClientIds();
								for(id = ids.begin(); id != ids.end(); ++id) {
									(*dc)->pause((*id));
								}

								//FIXME: only first DataSink
								DataSink* clientOut=(*dc)->getDataSink(clientId);
								IO* clientOutput=clientOut->getOutput();
								IO *input = IOCreator::createInput(new_es, false);
								assert(input);
								input->setEndFrameNumber((*dc)->getInput()->getEndFrameNumber());
								vsec_time=(double)(*dc)->getOutputStatistics()->getHighestPacketTS()
									/ (double)(*dc)->getESInfo()->getMediaTimeScale();
								u32 startFrame = new_es->getFrameNumber(vsec_time) + 1;
								ivop_align = input->setToClosestIFrame(startFrame, false);
								assert (ivop_align > 0);

								//force DC to continue until switchPoint is reached
								// (normally already far after that.... but still...)
								(*dc)->setFinalFrameNumber(ivop_align - 1);

								//ATTENTION: RTP preQ might (will) be _far_ ahead to switchPoint
							//so chop of unneeded data
								clientOutput->setEndFrameNumber(ivop_align); //new startFrame

								//restart clients and RTP to flush out data until switchPoint
								// we have to wait until all data is really sent out to the network!
								ids = (*dc)->getListOfPausedClientIds();
								for(id = ids.begin(); id != ids.end(); ++id) {
									(*dc)->play((*id), 0);
								}

								//wait until DC-thread reaches switchPoint and stops
								(*dc)->initiateThreadStop(false);
								//reset DC to stream all data which is further arriving with the new_input
								(*dc)->setFinalFrameNumber(-1);

								long actFrame;
								do {//wait until RTP is done
									actFrame = (*dc)->getOutputStatistics()->getHighestPacketTS()
										/ (*dc)->getESInfo()->getVOPTimeIncrement();
									dprintf_full("ServerSession::run waiting for RTP to reach switchPoint %li (now at frame %li)\n",
															 ivop_align-1, actFrame);
									msleep(40);
								} while (actFrame < ivop_align-1);

								//pause all clients and change to new ESs and input
								ids = (*dc)->getListOfActiveClientIds();
								for(id = ids.begin(); id != ids.end(); ++id) {
									(*dc)->pause((*id));
									//change RTP-ES, automatically also changes Statistics-ES and adaptPreQ
									(*dc)->getDataSink((*id))->getOutput()->setESInfo(new_es);
								}
								(*dc)->setESInfo(new_es);

								//switch to new input with all clients paused
								(*dc)->setInput(input);

								dprintf_full("ServerSession::run STREAM SWITCHING DONE! now restarting on new stream !\n");

								//restart with no prefetch and continue with the new stream
								ids = (*dc)->getListOfPausedClientIds();
								for(id = ids.begin(); id != ids.end(); ++id) {
									(*dc)->play((*id), 0);
								}
								//DC was stopped in setInput, so restart it
								(*dc)->start();
							}
						}
						delete oldmp4; //also deletes ESes!
					}//got the same stream
				}//SWITCH NOW!!!
			}

		}//enough variations?
	}				// while valid
	if(this->counter != NULL)
		this->counter->decrement();
	dprintf_full("ServerSession::run() end!\n");

	delete [] inbuffer;
 	this->exit();

}



/***********************************************************************/
void ServerSession::setUrl(const Url *uri, bool makeExactMatch)
{
	bool insertResUsage=false;
  	if (cm && uri) {
		if (url) {
			delete url;
		}
		dprintf_full("ServerSession::setUrl: %s \r\n", uri->toString());
		//if (url->getFileName() == NULL)
		//	dprintf_full("ServerSession::setUrl: file NULL!!!\n");
		
		dprintf_full("ServerSession::setUrl: %s filename: %s\r\n", uri->toString(), uri->getFileName());
		url = new Url(uri->toString());
		bool exactMatch;
		MetaObject *meta = cm->findMetaObject(url);
		if (meta != NULL) {
			dprintf_full("ServerSession::setUrl metadata found\n");
		} else {
			dprintf_full("ServerSession::setUrl metadata NOT found\n");		
		}

		dprintf_full("ServerSession::setUrl termcap: %p, userpref %p meta: %p\n", termCap, userPref, meta);	

		if( (!termCap && !userPref)) {
			// just send ANY version
			dprintf_full("ServerSession::setUrl NO metaobject/metadata found!\n");
			mp4Stream = cm->getVideo(url);
			if (mp4Stream == NULL) {
				fprintf(stderr,"\nERROR: demuxed stream not found or maybe you are starting the server from a different computer (hostname/IP) now? DELETE demux directory!\n");
				//::exit(1);
				return;
			}
			
			mp4Stream=mp4Stream->clone(); // CM returns the orig video, we don't want it in server
			// calc resUsage
			VideoESInfo* vis=mp4Stream->getFirstVisualES();
			if(vis) {
				
				dprintf_full("ServerSession::setUrl VIS found\n");
				Feature src(vis);
				src.isCached=true;
				ResourceUsage* tmpRU=CostFunction::calcResourceUsage(&src,&src,admControl->getResourceLimit());
				if(tmpRU) {
					resUsage.network=tmpRU->network;
					resUsage.disk=tmpRU->disk;
					resUsage.cpu=tmpRU->cpu;
					delete tmpRU;
					insertResUsage=true;
				}
			}
		}
		
		else if(meta && userPref) {
			// FIXME: hardcode delay values to 100ms
			dprintf_full("ServerSession::setUrl meta && userPref\n");
			suggestedAdaptor=admControl->makeBestMatch(userPref,100,meta,&mp4Stream,&resUsage);

			if(mp4Stream)
				mp4Stream=mp4Stream->clone();
		}
		else if(meta && termCap )
		{ //use MetaObject for best match
			// bandwidth should be the minimum of network and decodeBandwidth
			// if one of them is zero, ignore the zero param
			dprintf_full("ServerSession::setUrl meta && termCap\n");
			u32 bandwidth=termCap->getNetworkCapacityInByte()*8;
			if(bandwidth==0)
				bandwidth=termCap->getMaxDecoderBitRateInBit();
			else if(termCap->getMaxDecoderBitRateInBit() &&
							termCap->getMaxDecoderBitRateInBit() < bandwidth) {
				bandwidth=termCap->getMaxDecoderBitRateInBit();
			}
			dprintf_full("ServerSession::setUrl: termCap->getColorDisplay()=%u\r\n",termCap->getColorDisplay());
			mp4Stream=meta->findMp4Stream(0,termCap->getDisplayWidth(), 0,termCap->getDisplayHeight(), 0,bandwidth, (termCap->getColorDisplay()==0),
			(float)termCap->getDisplayRefreshRate(), &exactMatch);
			if(mp4Stream)
				mp4Stream=mp4Stream->clone();
			if(makeExactMatch && !exactMatch) {
				suggestedAdaptor=admControl->makeExactMatch(termCap->getDisplayWidth(),
																							termCap->getDisplayHeight(),
																							bandwidth,
																							(termCap->getColorDisplay()==0),
																							(float)termCap->getDisplayRefreshRate(),
																							mp4Stream,&resUsage);
			}
			else 
			{
				// calc the resource usage for the non transcoding case
				if(mp4Stream) 
				{
					VideoESInfo* vis=mp4Stream->getFirstVisualES();
					if(vis) 
					{
						Feature src(vis);
						src.isCached=true;
						ResourceUsage* tmpRU=CostFunction::calcResourceUsage(&src,&src,admControl->getResourceLimit());
						if(tmpRU) 
						{
							resUsage.network=tmpRU->network;
							resUsage.disk=tmpRU->disk;
							resUsage.cpu=tmpRU->cpu;
							delete tmpRU;
							insertResUsage=true;
						}
					}
				}


			}

		}
#ifndef WINCE
		else if(termCap && ((strcmp(uri->getFileName(), "v4l")) ==0))
		{
			dprintf_full("ServerSession::setUrl  v4l && termCap &&\n");
			u32 bandwidth=termCap->getNetworkCapacityInByte()*8;
			if(bandwidth==0)
				bandwidth=termCap->getMaxDecoderBitRateInBit();
			else if(termCap->getMaxDecoderBitRateInBit() && termCap->getMaxDecoderBitRateInBit() < bandwidth) 
			{
				bandwidth=termCap->getMaxDecoderBitRateInBit();
			}
			dprintf_full("ServerSession::setUrl: v4l termCap->getColorDisplay()=%u\r\n",
									 termCap->getColorDisplay());
			
			mp4Stream = Video4LinuxContainerFile::loadContainerInfo();
			if(mp4Stream)
			{
				dprintf_full("ServerSession::setUrl: v4l mp4stream ok\n");
				mp4Stream=mp4Stream->clone();
			}
			else
				dprintf_full("ServerSession::setUrl: v4l mp4stream NULL!! \n");
			if(makeExactMatch && !exactMatch) 
			{
				suggestedAdaptor=admControl->makeExactMatch(termCap->getDisplayWidth(), termCap->getDisplayHeight(),bandwidth, (termCap->getColorDisplay()==0), (float)termCap->getDisplayRefreshRate(),mp4Stream,&resUsage);
			}		
			dprintf_full("ServerSession::setUrl: v4l OK\r\n");

		}
#endif
	}		

	// now insert the resource usage for those other cases
	if(insertResUsage)
	{
		if(!userPref && !admControl->insertRequest(&resUsage)) 
		{
			resUsage.cpu=resUsage.disk=resUsage.network=0.0;
			if(mp4Stream)
				delete mp4Stream;
			mp4Stream=NULL;
			if(suggestedAdaptor)
				delete suggestedAdaptor;
			suggestedAdaptor=NULL;
		}
	}
};


/***********************************************************************/
bool ServerSession::setOptions(const Url* fileName, const char* remaining)
{

	assert(fileName && remaining);
	u32 session=0;
	std::map<char*, char*, struct stringCompare>* param=new std::map<char*, char*, stringCompare>();
	this->prot->parseSetParameter(remaining,&session,param);
	if(param->size()>0) {
		// search the DC for the session
		dprintf_full("ServerSession::setOptions: detected %i params\n",param->size());
		DataChannel* dc=NULL;
		list < DataChannel * >::iterator dcI;
		if (!channels.empty()) {
			dcI = channels.begin();
			while (dcI != channels.end()) {
				if ((*dcI)->getSessionId()== (u32)session) {
					dc=(*dcI);
					dcI=channels.end();
				}
			}
		}
		if(dc) {
			dprintf_full("ServerSession::setOptions: found DataChannel\n");
			// analyze params: we currently only support FRAME_FILTER
			char** invalid=new char*[param->size()];
			int invalidCnt=0;
			std::map<char*, char*, stringCompare>::iterator mapI;
			for(mapI=param->begin();mapI!=param->end();++mapI) {
				if( !strstr((*mapI).first,"FRAME_FILTER")) {
					invalid[invalidCnt]=(*mapI).first;
					invalidCnt++;
				}
				else if(strstr((*mapI).first,"FRAME_FILTER")) {
					// set it at input
					dprintf_full("ServerSession::setOptions: detected FRAME_FILTER\n");
					IO* io=dc->getInput();
#ifdef WINCE
					FilteredIO * fio = (FilteredIO*) io;
#else
					FilteredIO* fio=dynamic_cast<FilteredIO*>(io);
#endif
					if(fio) {
					//if( typeid(io) == typeid(FilteredIO)) {
						dprintf_full("ServerSession.setOptions found %s\n",(*mapI).second);
						BitField* b=BitField::fromHexString( (*mapI).second);

						if(b) {
							// we must invert the bitfield, because a 1 there tells us what frames the proxy
							// already has, we need to send the 0 --> invert
							b->invert();
							((FilteredIO*) io)->setBitField(b);
							// prot->generateOk();
						}
						else {
							dprintf_err("ServerSession::setOptions: failed to create BitField for String\n");
							prot->generateInvalidParameter(0,NULL);
							delete invalid;
							strcpy(buffer,prot->getBuffer());
							prot->incSeqNr();
							RTSP::freeMapMemory(param);
							delete param;
							delete[] invalid;
							return true;
						}
					}
					else { // IO doesn't support it
						invalid[0]="FRAME_FILTER";
						dprintf_full("ServerSession::setOptions: FRAME_FILTERing not supported by IO\n");
						prot->generateInvalidParameter(1,invalid);
						delete invalid;
						strcpy(buffer,prot->getBuffer());
						prot->incSeqNr();
						RTSP::freeMapMemory(param);
						delete param;
						return true;
					}
				}
			} // for
			if(invalidCnt==0)
				prot->generateOk();
			else
				prot->generateInvalidParameter(invalidCnt,invalid);
			delete[] invalid;
		}
		else { // Session not found
			prot->generateSessionNotFound();
		}
	}
	else { //set without params
		prot->generateOk();
	}
	prot->incSeqNr();
	strcpy(buffer,prot->getBuffer());
	RTSP::freeMapMemory(param);
	delete param;
	return true;
};



/***********************************************************************/
bool ServerSession::getOptions(const Url* fileName, const char* remaining)
{
	assert(fileName && remaining);
	u32 session=0;
	int numparams=0;
	char** params=NULL;
	this->prot->parseGetParameter(remaining,&session,&numparams,&params);
	// FIXME: implement

	//handle parameters. currently, only 'position' is supported!
	for (int i = 0; i < numparams; i++) {

		if (strcmp(params[i], "position") == 0) {
			for (list < DataChannel * >::iterator dc = channels.begin();
					dc != channels.end(); dc++) {
				//calculate time in msecs from requested session
				if ((*dc)->getSessionId() == session) {
					if ((*dc)->getOutputStatistics()) {
						u32 rtptimems = (u32)((*dc)->getOutputStatistics()->getPlayoutSec() * 1000);

						//add result into map and generate response msg
						std::map<char*, char*, struct stringCompare>* param=new std::map<char*, char*, stringCompare>();
						char rtptimesStr[11];
						sprintf(rtptimesStr,"%d", rtptimems);
						std::pair<char*, char*> pos("position", rtptimesStr);
						param->insert(param->begin(), pos);

						prot->generateResponseGetParameter(param);
						const char* buf=prot->getBuffer();
						strcpy(buffer,buf);
					} else {
						dprintf_err("ServerSession::setup() \nCannot get output statistics!\n");
					}
				}
			}
		}
	}

	prot->incSeqNr();
	return true;
};

